// Test the GhostNode distributed simulation
// George F. Riley, Georgia Tech, Fall 2003

#include <iostream>

#include "simulator.h"
#include "node.h"
#include "linkp2p.h"
#include "tcp-tahoe.h"
#include "application-webserver.h"
#include "application-webbrowser.h"
#include "rng.h"
#include "ratetimeparse.h"
#include "validation.h"

using namespace std;

int main(int argc, char** argv)
{
  Validation::Init(argc, argv);
  if (argc < 3)
    {
      cout << "Usage: testghost myid number-lps" << endl;
      exit(1);
    }
  SystemId_t  myId = atol(argv[1]);
  SystemId_t  nProcs = atol(argv[2]);
  IPAddr_t    baseIP = IPAddr("192.168.1.1");
  IPAddr_t    adderIP = IPAddr("0.0.1.0");
  
  IPAddrVec_t servers; // IP Addresses of possible servers

  Simulator s(myId);
  Linkp2p   myLink(Rate("10Mb"), Time("10ms"));

  Trace* tf = Trace::Instance();
  TCPTahoe tcp;
  TCP::LogFlagsText(true);    // Log flags in text mode
  
  if (myId == 0)
    { // Debug..get a trace file for lp 0 only
      tf->Open("testghost.txt");
      tf->IPDotted(true);
      // Enable all l3 tracing
      IPV4::Instance()->SetTrace(Trace::ENABLED);
      // Enable tcp tracing
      tcp.SetTrace(Trace::ENABLED);
    }
  
  // Simple topology in a straight line, one node per lp
  Node* prior = nil;
  for (SystemId_t i = 0; i < nProcs; ++i)
    {
      Node* n = new Node(i);
      cout << "Node " << i << " is " << n 
           << " id " << n->Id()
           << " IsReal() " << n->IsReal()
           << endl;
      
      if (prior) prior->AddDuplexLink(n, myLink,
                                      baseIP, Mask(24),
                                      baseIP + 1, Mask(24));
      baseIP += adderIP;
      // Add a web server to each
      n->AddApplication(WebServer(tcp));
      prior = n;
    }
  // Make a list of servers for each browser (excluding self)
  // and create the browsers on the "Real" nodes
  for (SystemId_t i = 0; i < nProcs; ++i)
    {
      servers.clear();
      Node* n = Node::GetNode(i);
      cout << "Node " << n->Id() 
           << " ipa is " << (string)IPAddr(n->GetIPAddr()) << endl;
      for (SystemId_t j = 0; j < nProcs; ++j)
        {
          if (i != j)
            {
              Node* nj = Node::GetNode(j);
              servers.push_back(nj->GetIPAddr());
            }
        }
      WebBrowser* wb =
          (WebBrowser*)n->AddApplication(
              WebBrowser(servers, Uniform(0, servers.size())));
      if (wb)
        { // Application added, not a ghost
          wb->ThinkTimeBound(600.0);
          wb->GetTCP()->SetTrace(Trace::ENABLED); // Enable l4 tracing
          wb->Start(0);
          cout << " Added web browser on node " << n->Id()
               << " server size " << servers.size() << endl;
          for (Count_t k = 0; k < servers.size(); ++k)
            {
              cout << "  Server ip is " << (string)IPAddr(servers[k]) << endl;
            }
        }
    }
  s.Progress(100);
  s.StopAt(100000);
  s.Run();
  if (myId == 0)
    {
      tf->Close();
    }
  
  cout << "Simulation complete, lp " << myId << endl;
  cout << "Total Web Browser connections " 
       << WebBrowser::completedConnections << endl;
  
#ifdef JUST_FOR_DEBUGGING
  cout << "next nodeid is " << Node::nextId << endl;
  
  for (SystemId_t i = 0; i < nProcs; ++i)
    {
      Node* n = Node::GetNode(i);
      NodeWeightVec_t neighbors;
      n->Neighbors(neighbors);
      cout << "Printing neighbors for node " << n->Id() << endl;
      for (NodeWeightVec_t::size_type j = 0; j < neighbors.size(); ++j)
        {
          NodeIfWeight& nw = neighbors[j];
          cout << "  node " << nw.node
               << " id " << nw.node->Id() << endl;
        }
    }
#endif
}

  
