/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

* $Id: application-aodv.h,v 1.1.1.1 2006/12/13 13:09:35 dnguyen Exp $
*/

// Georgia Tech Network Simulator - AODV Routing Protcol Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the AODV routing protocol

#ifndef __aodv_application_h__
#define __aodv_application_h__

#include <list>
#include "assert.h"

#include "common-defs.h"
#include "application.h"
#include "udp.h"
#include "routing-aodv.h"
#include "aodv-queue.h" //for queue.

#include "timer.h" //for timers

//typedef the packet type for AODV
#define AODV_REQUEST  100
#define AODV_REPLY    101
#define AODV_HELLO    102
#define AODV_ERROR    103

typedef unsigned char   Byte; // define the byte. 
							  // Don't understand why common-defs.h use short?
#define MY_ROUTE_TIMEOUT        10     // 10 s?
#define ACTIVE_ROUTE_TIMEOUT    10     // 10 s
#define REV_ROUTE_LIFE          6      // 6 seconds
#define BCAST_ID_SAVE           6      // 6 seconds
#define HELLO_INTERVAL          1      // 1000 ms
#define ALLOWED_HELLO_LOSS      3      // 3 packets
#define MinHelloInterval        (0.75 * HELLO_INTERVAL)
#define MaxHelloInterval        (1.25 * HELLO_INTERVAL)
#define RREQ_RETRIES            3      //Number of times to do network-wide search before timeout
#define MAX_RREQ_TIMEOUT        10.0   //sec. timeout after doing network-wide search.

#define TTL_START               5      //various constants used for expanding ring search
#define TTL_THRESHOLD           7
#define TTL_INCREMENT           2

// This should be somewhat related to arp timeout
 #define NODE_TRAVERSAL_TIME     0.03   // 30 ms
 #define LOCAL_REPAIR_WAIT_TIME  0.15   //sec
 

#define NETWORK_DIAMETER        30     //should be set by user using best guess
#define AODV_MAX_ERRORS         100

//Timer events
class AODVEvent : public TimerEvent {
	public:
		// the types of the AODV timer events
		typedef enum {BROADCAST_TIMEOUT, HELLO_TIMEOUT, NEIGHBOR_TIMEOUT
				, ROUTECACHE_TIMEOUT, LOCALREPAIR_TIMEOUT}AODVTimeout_t;
	public:
		AODVEvent(Event_t ev) : TimerEvent(ev) { };
		virtual ~AODVEvent();
};


class BroadcastID {
    public: 
	     BroadcastID(IPAddr_t i, Long_t b) { src = i; id = b;  } 
    public: 
	     IPAddr_t     src;
	     Long_t       id;
	     double       expire;         // now + BCAST_ID_SAVE s
};

typedef std::list<BroadcastID *> BroadcastIDList_t;

struct AODVRequestHdr {
        Byte            Type;	    // Packet Type
        Byte            Reserved[2];
        Byte            HopCount;   // Hop Count
        Long_t          BcastId;    // Broadcast ID

        IPAddr_t        Dst;         // Destination IP Address
        Seq_t       	DstSeqNo;   // Destination Sequence Number
        IPAddr_t        Src;         // Source IP Address
        Seq_t           SrcSeqNo;   // Source Sequence Number

        double          TimeStamp;   // when REQUEST sent;
					// used to compute route discovery latency

  inline int size() { 
  int sz = 0;
  /*
  	sz = sizeof(Byte)		// rq_type
	     + 2*sizeof(Byte) 	// reserved
	     + sizeof(Byte)		// rq_hop_count
	     + sizeof(double)	// rq_timestamp
	     + sizeof(Long_t)	// rq_bcast_id
	     + sizeof(IPAddr_t)	// rq_dst
	     + sizeof(Seq_t)	// rq_dst_seqno
	     + sizeof(IPAddr_t)	// rq_src
	     + sizeof(Seq_t);	// rq_src_seqno
  */
  	sz = 7*sizeof(Long_t);
  	assert (sz >= 0);
	return sz;
  }
};

struct AODVReplyHdr {
        Byte            Type;       // Packet Type
        Byte            Reserved[2];
        Byte            HopCount;   // Hop Count
        IPAddr_t        Dst;        // Destination IP Address
        Seq_t           DstSeqNo;   // Destination Sequence Number
        IPAddr_t        Src;        // Source IP Address
        double	        LifeTime;   // Lifetime

        double          TimeStamp;  // when corresponding REQ sent;
						// used to compute route discovery latency
						
  inline int size() { 
  int sz = 0;
  /*
  	sz = sizeof(Byte)		// rp_type
	     + 2*sizeof(Byte) 	// rp_flags + reserved
	     + sizeof(Byte)		// rp_hop_count
	     + sizeof(double)	// rp_timestamp
	     + sizeof(IPAddr_t)	// rp_dst
	     + sizeof(Seq_t)	// rp_dst_seqno
	     + sizeof(IPAddr_t)	// rp_src
	     + sizeof(Long_t);	// rp_lifetime
  */
  	sz = 6*sizeof(Long_t);
  	assert (sz >= 0);
	return sz;
  }

};

struct AODVErrorHdr {
        Byte        Type;            // Type
        Byte        Reserved[2];     // Reserved
        Byte        DestCount;       // DestCount
        // List of Unreachable destination IP addresses and sequence numbers
        IPAddr_t     UnreachableDst[AODV_MAX_ERRORS];   
        Seq_t        UnreachableDstSeqNo[AODV_MAX_ERRORS];   

  inline int size() { 
  int sz = 0;
  /*
  	sz = sizeof(Byte)		// type
	     + 2*sizeof(Byte) 	// reserved
	     + sizeof(Byte)		// length
	     + length*sizeof(IPAddr_t); // unreachable destinations
  */
  	sz = (DestCount*2 + 1)*sizeof(Long_t);
	assert(sz);
        return sz;
  }

};

struct AODVRrep_ackHdr {
	Byte	rpack_type;
	Byte	reserved;
};


#define AODV_PORT 564

static void RouteResolveCallbackAODV(Packet *pPacket, void *data);

class AODVApplication : public Application, public TimerHandler {
public:
  AODVApplication(Node*);
  AODVApplication(const AODVApplication&); // Copy constructor
  //void Handle(Event*, Time_t); // Uncomment if specific event handling needed
  // Upcalls from L4 protocol
  virtual void Receive(Packet*,L4Protocol*, Seq_t = 0);   // Data received
  // Called from application base class
  virtual void StartApp();    // Called at time specified by Start
  virtual void StopApp();     // Called at time specified by Stop
  virtual Application* Copy() const;// Make a copy of the application
public:
  UDP*          l4Proto;      // Points to the specified Layer 4 protocol
  // Must certainly need more state here
public:
  // Timer handle methods
  virtual void Timeout(TimerEvent*); // Called when timers expire
  AODVQueue		rQueue; //used to buffer data which has no route.
  
  void RouteResolve(Packet *pPacket); //call back from low layer.
  
public:
  // Timers
  Timer          timer;           // Timer for various AODV timer events
  AODVEvent*     BroadcastTimeout;  // Broadcast timer timeout event
  AODVEvent*     HelloTimeout;      // Hello timer timeout
  AODVEvent*     NeighborTimeout;   // Neighbor timer timeout
  AODVEvent*     RouteCacheTimeout; // Route cache timer timeout 
  AODVEvent*     LocalRepairTimeout;  // LocalRepair timer Timeout
  bool 			 useTimerBuckets;

  // Timer management
  void ScheduleTimer(Event_t, AODVEvent*&, Time_t);
  void CancelTimer(AODVEvent*&, bool delTimer = false); 
  void CancelAllTimers();
  
protected:
  Seq_t          SeqNo;				//Sequence number.??
  int            BroadcastId;       //Really should add this??
  BroadcastIDList_t BroadcastIdList;//TODO:
  AODVNeighborList_t  NeighborList;
  RoutingAODV	 *pAODVRouting;     //used to refer routing AODV object.

  double  PerHopTime(AODVRoutingEntry *pRt);


  //utilites to maintain.
  //Broadcast ID Management
  void BroadcastIDPurge(void);
  bool BroadcastIDLookup(IPAddr_t id, Long_t bid);
  void BroadcastIDInsert(IPAddr_t id, Long_t bid);
  
  //TODO: implement the following 3 routines in .cc file.

  //when lower layer finds link down call the following
  //TODO: about its parameter.
  void RouteLinkLayerFailed(IPAddr_t);
  void HandleLinkFailure(IPAddr_t);
		  
  void RoutePurge();
  void RouteUpdate(AODVRoutingEntry *rt, Seq_t seqnum,Word_t metric
				  ,IPAddr_t nexthop, double expire_time);
  void RouteDown(AODVRoutingEntry *rt);

  //Neighbor management.
  void NeighborPurge();
  AODVNeighbor * NeighborLookup(IPAddr_t id); 
  void NeighborInsert(IPAddr_t id);
  void NeighborDelete(IPAddr_t id);

  //Packet TX Routines
  void Forward(AODVRoutingEntry *rt, Packet *p, double delay);
  void SendHello(void);
  void SendRequest(IPAddr_t dst);

  void SendReply(IPAddr_t ipdst, Long_t hop_count,
                 IPAddr_t rpdst, Seq_t rpseq, 
				 Long_t lifetime, double timestamp);
  void SendError(struct AODVErrorHdr *p, bool jitter = true);

  // Packet RX Routines
  void RecvHello(Packet *p);
  void RecvRequest(Packet *p);
  void RecvReply(Packet *p); 
  void RecvError(Packet *p);
  
  
};

#endif

