/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Implementation of the Gnutella Peer-to-Peer Protocol
// George F. Riley, Georgia Tech, Spring 2003

#include <stdio.h>

#ifdef HAVE_QT
#include <qnamespace.h>
#endif

#include "application-gnutella.h"
#include "tcp.h"
#include "rng.h"

using namespace std;

// Static members
IPAddrSet_t Gnutella::gcacheHosts;        // IPAddress of list of GCache's
PortId_t    Gnutella::defaultPort = 6346; // Default port for incoming report
Count_t     Gnutella::outPeerCount = 4;   // Limit on desired out connections
Count_t     Gnutella::inPeerCount = 4;    // Limit on desired in connections
Count_t     Gnutella::transferCount = 4;  // Limit on simulataneous tranfers

// Gnutella specific methods

Gnutella::Gnutella() 
  : node(nil),
    gcacheTCPh(nil),
    gcacheTCPu(nil),
    listenTCP((TCP*)TCP::Default().Copy()),
    gcacheActive(false),
    gcacheIP(IPADDR_NONE),
    gnuPort(defaultPort),
    outConnectionLimit(outPeerCount), inConnectionLimit(inPeerCount),
    transferLimit(transferCount),
    goodOutConnections(0), failedOutConnections(0),
    goodInConnections(0), rejectedInConnections(0),
    gcacheRequests(0), gcacheFailed(0)
{
  CopyOnConnect(false); // Do not copy this app on incomming connections
  listenTCP->AttachApplication(this);
}

Gnutella::Gnutella(const TCP& t) 
  : node(nil),
    gcacheTCPh(nil),
    gcacheTCPu(nil),
    listenTCP((TCP*)t.Copy()),
    gcacheActive(false),
    gcacheIP(IPADDR_NONE),
    gnuPort(defaultPort),
    outConnectionLimit(outPeerCount), inConnectionLimit(inPeerCount),
    transferLimit(transferCount),
    goodOutConnections(0), failedOutConnections(0),
    goodInConnections(0), rejectedInConnections(0),
    gcacheRequests(0), gcacheFailed(0)
{
  CopyOnConnect(false); // Do not copy this app on incomming connections
  listenTCP->AttachApplication(this);
}

Gnutella::Gnutella(const Gnutella& c)
  : Application(c),
    node(nil),
    gcacheTCPh(nil),
    gcacheTCPu(nil),
    listenTCP((TCP*)c.listenTCP->Copy()),
    gcacheActive(false),
    gcacheIP(c.gcacheIP),
    gnuPort(c.gnuPort),
    outConnectionLimit(c.outConnectionLimit),
    inConnectionLimit(c.inConnectionLimit),
    transferLimit(c.transferLimit),
    outConnections(c.outConnections),
    inConnections(c.inConnections),
    transferConnections(c.transferConnections),
    goodOutConnections(0), failedOutConnections(0),
    goodInConnections(0), rejectedInConnections(0),
    gcacheRequests(0), gcacheFailed(0)
{
  listenTCP->AttachApplication(this);
}

Gnutella::~Gnutella()
{
  //delete listenTCP;
}

void Gnutella::Receive(Packet* p, L4Protocol* l4,Seq_t)
{   // Data received
  if (l4 == gcacheTCPh)
    {
      cout << "Gnutella " << this << " got response from gcache" << endl;
      Data* d = (Data*)p->PeekPDU();
      if (d->Contents())
        {
          string r(d->Contents());
          cout << "Response is " << r <<endl;
          gcacheRequests++;
          while(r.length())
            {
              string::size_type j = r.find('\n');
              if (j > 0)
                { // Found a non-empty line
                  string t = r.substr(0, j);
                  string::size_type k = t.find(':');
                  string ipa = t.substr(0, k);
                  string port = t.substr(k+1, string::npos);
                  AddPossiblePeer(IPPort_t(
                                           IPAddr(ipa.c_str()),
                                           atol(port.c_str())));
                }
              r = r.substr(j+1, string::npos);  // remove this one
            }
          ConnectToPeer(); // Try a new connection(s)
        }
    }
  else if (l4 == gcacheTCPu)
    { // ?? Do we get a response to the url update?
    }
  else
    {
      cout << "Gnutella " << this << " got message from peer" << endl;
    }
  delete p;
}

void Gnutella::Sent(Count_t, L4Protocol*)
{     // Data has been sent
}

void Gnutella::CloseRequest(L4Protocol* l4)
{      // Close request from peer
  l4->Close(); // Just close in response
  if (l4 == gcacheTCPh)
    {
      // Time to post the update
      // Make a copy of the listening tcp endpoint.
      gcacheTCPu = (TCP*)listenTCP->Copy();
      gcacheTCPu->AttachApplication(this);
      gcacheTCPu->Attach(node);
      gcacheTCPu->DeleteOnTWait();        // Delete this one when done
      gcacheTCPu->Bind();                 // Get a local port number
#ifdef HAVE_QT
      gcacheTCPu->SetColor(Qt::green);    // Use green for updates
#endif
      gcacheTCPu->Connect(gcacheIP, 80);  // Establish another connection
      gcacheActive = false;
      if (outConnections.size() < outConnectionLimit)
        { // Need more hosts to choose from
          GCacheConnect();
        }
    }
}

void Gnutella::Closed(L4Protocol*)
{ // Connection has closed
}

void Gnutella::ConnectionComplete(L4Protocol* l4) // Connection rq success
{
  if (l4 == gcacheTCPh)
    {
      cout << "Gnutella::ConnectionComplete sending hostfile" << endl;
      string hostfile("gcache.php?hostfile=1\n");
      Data d(hostfile);
      l4->Send(d);
      return;
    }
  if (l4 == gcacheTCPu)
    {
      cout << "Gnutella::ConnectionComplete sending update" << endl;
      string ipUpdate("gcache.php?ip=");
      char work[255];
      sprintf(work, "%ld", gnuPort);
      ipUpdate = ipUpdate + 
        (string)IPAddr(gcacheTCPu->localNode->GetIPAddr()) +
        string(":") +
        string(work);
            
      Data d1(ipUpdate);
      cout << "Gnutella update sending " << d1.Contents() << endl;
      gcacheTCPu->Send(d1);
      return;
    }
  // Check for peer connection
  PeerList_t::iterator i = FindPeerInfo(outConnections, (TCP*)l4);
  if (i != outConnections.end())
    { // Successful connection
      i->connected = true;
      cout << "Gnutella " << (string)IPAddr(node->GetIPAddr())
           << " successful connection to peer " 
           << (string)IPAddr(i->ipPort.first)
           << " port " << i->ipPort.second << endl;
      goodOutConnections++;
    }
}

void Gnutella::ConnectionFailed(L4Protocol* l4)   // Connection request failed
{
  if (l4 == gcacheTCPh)
    {
      cout << "Gnutella::ConnectionFailed gcache, retrying" << endl;
      gcacheActive = false;
      gcacheFailed++;
      GCacheConnect();
      return;
    }
  // Check for peer connection
  PeerList_t::iterator i = FindPeerInfo(outConnections, (TCP*)l4);
  if (i != outConnections.end())
    { // Out connection failed
      cout << "Gnutella " << (string)IPAddr(node->GetIPAddr())
           << " failed connection to " << (string)IPAddr(i->ipPort.first)
           << " port " << i->ipPort.second << endl;
      failedOutConnections++;
      outConnections.erase(i); // Remove from list
      ConnectToPeer(); // Try another
      return;
    }
  // Check for file transfer connection
}

bool Gnutella::ConnectionFromPeer(L4Protocol* l4, IPAddr_t i, PortId_t p)
{// Listener has cn rq from peer with specified ip/port
  // ?? Is this right?
  if (inConnections.size() >= inConnectionLimit)
    {
      rejectedInConnections++;
      return false; // Cannot accept
    }
  inConnections.push_back(
      GnuPeerInfo((TCP*)l4,
                  IPPort_t(i,p)));             // Add a new connection
#ifdef HAVE_QT
  l4->SetColor(Qt::blue);                      // Use blue for p2p connections
#endif
  GnuPeerInfo& gpi = outConnections.back();    // Get the one just added
  gpi.connected = true;                       // Note connected
  cout << "Gnutella " << (string)IPAddr(node->GetIPAddr())
       << " accepted connection from " << (string)IPAddr(i)
       << " port " << p << endl;
  goodInConnections++;
  return true;
}

void Gnutella::StartApp()
{ // Called at time specified by Start
  GCacheConnect(); // Initiate a GCache connection
}
void Gnutella::StopApp()
{     // Called at time specified by Stop
}

void Gnutella::AttachNode(Node* n)
{
  node = n;
  listenTCP->Attach(n);
  if (!listenTCP->Bind(gnuPort))
    {
      cout << "Gnutella::AttachNode, can't bind port " << gnuPort << endl;
    }
  listenTCP->Listen();
#ifdef HAVE_QT
  //listenTCP->SetColor(Qt::blue);
#endif

}

Application* Gnutella::Copy() const
{    // Make a copy of the application
  return new Gnutella(*this);
}

void Gnutella::DBDump()
{ // Print out debug info
  cout << "Debug info from Gnutella " << (string)IPAddr(node->GetIPAddr())
       << " port " << gnuPort << endl
       << "    Active outconnections " << outConnections.size()
       << " goodOutConnections " << goodOutConnections
       << " failedOutConnections " << failedOutConnections << endl
       << "    inConnections " << inConnections.size()
       << " goodInConnections " << goodInConnections
       << " rejectedInConnections " << rejectedInConnections << endl
       << "    gcacheRequests " << gcacheRequests
       << " gcacheFailed " << gcacheFailed << endl;
}

// Private members
void Gnutella::GCacheConnect()      // Initiate a gcache connection
{
  if (gcacheActive) return;         // Already connecting
  if (gcacheHosts.size() == 0)
    {
      cout << "GnutellaStartApp, no gcache's to choose from" << endl;
      return;
    }
  Uniform r(0, gcacheHosts.size());
  Count_t gc = r.IntValue(); // Choose a random gcache
  IPAddrSet_t::iterator i = gcacheHosts.begin();
  for (Count_t j = 0; j < gc; ++j) i++; // Advance interator random times
  gcacheIP = *i;                        // Get the IPAddress of the gcache
  gcacheTCPh = (TCP*)listenTCP->Copy(); // Get a tcp to use
  gcacheTCPh->AttachApplication(this);
  gcacheTCPh->Attach(node);             // Attach to this node
  gcacheTCPh->DeleteOnTWait();          // Delete this agent when done
  gcacheTCPh->Bind();                   // Get transient port
#ifdef HAVE_QT
  gcacheTCPh->SetColor(Qt::red);        // Use green for host file requests
#endif
  gcacheTCPh->Connect(gcacheIP, 80);    // Connect to the gcache
  gcacheActive = true;
  cout << "Gnutella::GCacheConnect " << this
       << " connecting to gcache " << (string)IPAddr(gcacheIP) << endl;
}

void Gnutella::AddPossiblePeer(const IPPort_t& ipport)
{
  cout << "Adding peer " << (string)IPAddr(ipport.first)
       << " port " << ipport.second << endl;
  if (node->LocalIP(ipport.first)) return; // Don't add self
  // First see if duplicate
  deque<IPPort_t>::iterator i;
  for (i = pendingHosts.begin(); i != pendingHosts.end(); ++i)
    {
      if (ipport == *i) break;    // Found duplicate
    }
  if (i == pendingHosts.end())
    { // Not duplicate, insert this one
      cout << "Not duplicate, adding" << endl;
      pendingHosts.push_back(ipport); // Add to list of potential hosts
    }
  
}

void Gnutella::ConnectToPeer()
{
  cout << "CTP oc.size() " << outConnections.size()
       << " oclimit " << outConnectionLimit << endl;
  while (outConnections.size() < outConnectionLimit)
    { // Need more
      if (pendingHosts.empty()) break;         // No more known peers
      IPPort_t ipport = pendingHosts.front();  // Get the next
      pendingHosts.pop_front();                // And remove
      outConnections.push_back(
          GnuPeerInfo((TCP*)listenTCP->Copy(),
                      ipport));                // Add a new connection
      GnuPeerInfo& gpi = outConnections.back();// Get the one just added
      // Initialize the TCP endpoint and start the connection
#ifdef HAVE_QT
      // Use blue for peer connections
      gpi.tcp->SetColor(Qt::blue);
#endif
      gpi.tcp->AttachApplication(this);
      gpi.tcp->Attach(node);
      gpi.tcp->DeleteOnTWait();        // Delete this one when done
      gpi.tcp->Bind();                 // Get a local port number
      // Establish another connection
      cout << "Gnutella " << (string)IPAddr(node->GetIPAddr())
           << " connecting to " << (string)IPAddr(gpi.ipPort.first)
           << " port " << gpi.ipPort.second << endl;
      gpi.tcp->Connect(gpi.ipPort.first, gpi.ipPort.second);
    }
  if (outConnections.size() < outConnectionLimit)
    { // Need more hosts to choose from
      GCacheConnect();
    }
}

PeerList_t::iterator Gnutella::FindPeerInfo(PeerList_t& peers , TCP* t)
{
  PeerList_t::iterator i;
  for (i = peers.begin(); i != peers.end(); ++i)
    {
      if (i->tcp == t) return i; // Found it
    }
  return i; // Not found
}


// Static members
void Gnutella::AddGCache(IPAddr_t ip)
{
  gcacheHosts.insert(ip); // Add to the list of known gcaches
}
