/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Implementation of the Gnutella Peer-to-Peer Protocol
// George F. Riley, Georgia Tech, Spring 2003

#ifndef __gnutella_h__

#include <set>
#include <list>

#include "common-defs.h"
#include "ipaddr.h"
#include "application.h"
#include "tcp.h"

// Define type for IPAddr_t set
// Used for defining the IPAddresses of the well known GCache list

typedef std::set<IPAddr_t> IPAddrSet_t;

//Doc:ClassXRef
class GnuPeerInfo { // Information about active peers
public:
  GnuPeerInfo() : tcp(nil), ipPort(0,0), connected(false) { }
  GnuPeerInfo(TCP* t, IPPort_t i)
    : tcp(t), ipPort(i), connected(false) { }
public:
  TCP*     tcp;       // TCP Endpoint for this peer
  IPPort_t ipPort;    // IP and Port
  bool     connected; // True if connected, false if pending
};

typedef std::list<GnuPeerInfo> PeerList_t; // List of peers

//Doc:ClassXRef
class Gnutella : public Application {
  //Doc:Class Class {\tt Gnutella} provides a model of the behavior 
  //Doc:Class of {\em Gnutella} peer--to--peer applications.  When started,
  //Doc:Class a {\tt Gnutella} application will randomly select a
  //Doc:Class {\em GCache} server (from the list of servers specified
  //Doc:Class with the static {\tt AddGCache} method) and requests 
  //Doc:Class a list of potential peers from the GCache.  The application
  //Doc:Class also posts it's own \IPA\ to the GCache.  At the same
  //Doc:Class time, each {\tt Gnutella} application opens a listening
  //Doc:Class {\tt TCP} endpoint to accept connections from peers.
  //Doc:Class Once a list of potential clients is know, the application
  //Doc:Class selects one of the potential  peers randomly and tries
  //Doc:Class to connect.  It will continue  querying the GCache servers
  //Doc:Class and attempting peer connections until it has successfully
  //Doc:Class connected with the number of desired peers (this limit
  //Doc:Class is specified by static variable {\tt outPeerCount}.
  //Doc:Class The applications also enforce a limit on the number of
  //Doc:Class allowable accepted connections, specified by static
  //Doc:Class variable {\tt inPeerCount}.

  //Doc:Class Once a connection is established, the {\em Ping} and
  //Doc:Class {\em Query} messages are sent (Note: With this release
  //Doc:Class this functionality is not implemented.

public:
  //Doc:Method
  Gnutella();
    //Doc:Desc Constructor for {\tt Gnutella} objects.  Uses
    //Doc:Desc the default {\tt TCP} type for all connections.

  //Doc:Method
  Gnutella(const TCP&);
    //Doc:Desc Constructor, specifying  a reference to the type of {\tt TCP}
    //Doc:Desc protocol to use.
    //Doc:Arg1 Reference to any subclass of {\tt TCP}.

  //Doc:Method
  Gnutella(const Gnutella&);
    //Doc:Arg1     //Doc:Desc Copy constructor.
    //Doc:Arg1 {\tt Gnutella} object to copy.

  virtual ~Gnutella();
public:
  // Connect to a peer, IP, Port, and timeout
  void Connect(IPAddr_t, PortId_t, Time_t);
  // Send Ping Request
  void Ping();
  // Send Query
  void Query();
  // Set port number
  //Doc:Method
  void SetPort(PortId_t p) { gnuPort = p;}
    //Doc:Desc Set the port number to bind to for incoming connections.
    //Doc:Arg1 Port number for incoming connections

  // Upcalls
  // Overridden calls from application base class
  virtual void Receive(Packet*,L4Protocol*,Seq_t = 0);   // Data received
  virtual void Sent(Count_t, L4Protocol*);     // Data has been sent
  virtual void CloseRequest(L4Protocol*);      // Close request from peer
  virtual void Closed(L4Protocol*);            // Connection has closed
  virtual void ConnectionComplete(L4Protocol*);// Connection request succeeded
  virtual void ConnectionFailed(L4Protocol*);  // Connection request failed
  // Listener has cn rq from peer 
  virtual bool ConnectionFromPeer(L4Protocol*, IPAddr_t, PortId_t);
  virtual void StartApp();           // Called at time specified by Start
  virtual void StopApp();            // Called at time specified by Stop
  virtual void AttachNode(Node*);    // Note which node attached to
  virtual Application* Copy() const; // Make a copy of the application
  void    DBDump();                  // Print debug info
  // Private Methods
  //Doc:Method
  virtual void GCacheConnect();      // Initiate a gcache connection
    //Doc:Desc Initiate a connection to a randomly selected GCache server.

  //Doc:Method
  virtual void AddPossiblePeer(const IPPort_t&);
    //Doc:Desc Add a peer (\IPA\ and port) to list of possible peers.
    //Doc:Arg1 Reference to a pair with \IPA\ and port of possible peer.

  //Doc:Method
  virtual void ConnectToPeer();      // Try another connection
    //Doc:Desc Attempt a connection to a randomly selected peer.

  // Find a peer in a peer list
  PeerList_t::iterator FindPeerInfo(PeerList_t&, TCP*);
  // Static Methods

  //Doc:Method
  static void DefaultPort(PortId_t); // Set the default port
    //Doc:Desc Specify the default port (incoming connections) for all
    //Doc:Desc future {\tt Gnutella} applications.
    //Doc:Arg1 Port to bind to.

  // Members
  Node*    node;                     // Node we are attached to
  TCP*     gcacheTCPh;               // TCP gcache hostfile requests
  TCP*     gcacheTCPu;               // TCP gcache ip update requests
  TCP*     listenTCP;                // TCP to use for incomming connections
  bool     gcacheActive;             // GCache request is active
  IPAddr_t gcacheIP;                 // IP Address of current GCache
  PortId_t gnuPort;                  // Port to use for incoming connections
  Count_t  outConnectionLimit;       // Number of desired output connections
  Count_t  inConnectionLimit;        // Number of allowed input connections
  Count_t  transferLimit;            // Number of allowed transfer connections
  PeerList_t outConnections;         // Connections we initiated
  PeerList_t inConnections;          // Connections we accepted
  PeerList_t transferConnections;    // Connections for file transfers
  Count_t    goodOutConnections;     // Number of out connections succeeded
  Count_t    failedOutConnections;   // Out connections that failed
  Count_t    goodInConnections;      // In connections we accepted
  Count_t    rejectedInConnections;  // In connections we rejected
  Count_t    gcacheRequests;         // Number successful gcache requests
  Count_t    gcacheFailed;           // Number of gcache failures
  std::deque<IPPort_t> pendingHosts; // List of hosts we can try to connect to

  // Static methods
  //Doc:Method
  static  void AddGCache(IPAddr_t);  // Add a gcache to the list
    //Doc:Desc Specify the \IPA\ of a potential GCache to query and post.
    //Doc:Arg1 \IPA\ to add.

  // Static members
  //Doc:Member
  static IPAddrSet_t gcacheHosts;    // IPAddress of list of GCache's
    //Doc:Desc List of all potential GCache servers (\IPA{s})

  //Doc:Member
  static PortId_t    defaultPort;    // Port number for incoming connections
    //Doc:Desc Default port number to bind for incoming connections.

  //Doc:Member
  static Count_t     outPeerCount;   // Limit on desired out connections
    //Doc:Desc Limit on the number of connections that this {\tt Gnutella}
    //Doc:Desc application maintains to other {\tt Gnutella} applications.

  //Doc:Member
  static Count_t     inPeerCount;    // Limit on desired in connections
    //Doc:Desc Limit on the number of accepted incoming connections.

  //Doc:Member
  static Count_t     transferCount;  // Limit on simulataneous tranfers
    //Doc:Desc Limit on the number of simultaneous file transfers.
    //Doc:Desc (Note: Not presently implemented)
};

#endif
