#include "debug.h"
#include "application-kneighlev.h"
#include "udp.h"
#include "node.h"
#include "simulator.h"
#include "interface.h"
#include "debug.h"

#include <assert.h>
#include <iostream>

using namespace std;

#ifdef USE_CISCO_AIRONET350
  // Cisco Aironet 350 Discrete Power Levels in mW 
  double KNeighLevApplication::txPower[] = {0.001, 0.005, 0.020, 
					    0.030, 0.050, 0.100};
  double KNeighLevApplication::radioRange[] = {24.0, 55.0, 109.0, 
					       134.0, 173.0, 244.0};
  double KNeighLevApplication::MAX_POWER = 0.100;
  // TL is the timeout between stepping power levels
  double KNeighLevApplication::TL[] = {0.1, 0.1, 0.1, 0.1, 0.1, 0.1};
#endif

int KNeighLevApplication::K_CONST = 0; // must be modified from simulation code

double KNeighLevApplication::GetTxPowerFromRange(Meters_t range)
{
  int i = 0;
  do
  {
    if (radioRange[i] == range)
      return txPower[i];
    else
      i++;
  } while (radioRange[i] != MAX_POWER);
  return -1.0;
}

KNeighLevApplication::KNeighLevApplication(Node *n, Time_t start)
  : node(n), iw(nil), ip(IPADDR_NONE), l4Proto(nil),
    pendingEvent(0), myTxPowerLevel(0), startTime(start),
    changedPower(false)
{
  
}

KNeighLevApplication::KNeighLevApplication(const KNeighLevApplication &c)
  : l4Proto(nil)
{
}

Application* KNeighLevApplication::Copy() const
{
  return new KNeighLevApplication(*this);
}

void KNeighLevApplication::Handle(Event* e, Time_t t)
{
  //cout << "In handle..." << endl;;
  AppKNeighLevEvent *ev = (AppKNeighLevEvent*)e;
  //cout << "\n" << IPAddr::ToDotted(ip) << "\n" 
  //     << "Nsymmetric.size(): " << Nsymmetric.size() << "." << endl;
  switch (ev->event) {
  case AppKNeighLevEvent::START:
  {
    DEBUG0((cout << "we got a start event" << endl));
    // initialize this node's power level
    SetTxPowerLevel(0);
    // send beacon message
    SendMessage(K_BEACON);
    
    //Time_t nextTime = t + TL[myTxPowerLevel]; 
    Time_t nextTime = TL[myTxPowerLevel]; 
    AppKNeighLevEvent* ev = new AppKNeighLevEvent(AppKNeighLevEvent::TIMEOUT);
    pendingEvent = ev;
    Simulator::instance->Schedule(ev, nextTime, this); // Schedule again
    break;
  }
  case AppKNeighLevEvent::TIMEOUT:
  {
    //cout << "Got a TIMEOUT event." << endl;
    // if node's symmetric neighbors list has at least k nodes
    if (Nsymmetric.size() >= K_CONST)
    {
      //  stop
      //cout << IPAddr::ToDotted(ip) 
      //     << ": We've reached k at time " << t << " seconds." << endl;
      pendingEvent = NULL;
      break;
    }
    // else if current node's power level is maximum
    else if (KNeighLevApplication::txPower[myTxPowerLevel] >= MAX_POWER)
    {
      //  stop
      //cout << IPAddr::ToDotted(ip) 
      //     << ": We've reached maximum power." << endl;
      pendingEvent = NULL;
      break;
    }
    // else 
    else
    {
      //cout << "We need to increment power levels and try again." << endl;
      // increment power level
      if (!changedPower)
	SetTxPowerLevel(myTxPowerLevel + 1);
      changedPower = false;
      //  send help message
      SendMessage(K_HELP);
    }

    //Time_t nextTime = t + TL[myTxPowerLevel]; // Time till next packet
    Time_t nextTime = TL[myTxPowerLevel]; 
    if (pendingEvent) delete pendingEvent;
    ev = new AppKNeighLevEvent(AppKNeighLevEvent::TIMEOUT);
    pendingEvent = ev;
    Simulator::instance->Schedule(ev, nextTime, this); // Schedule again
    //cout << "returning from handle..." << endl;
    return;
  }
  }
  Application::Handle(e, t);  // May be application event
  //cout << "leaving handle..." << endl;
}

// Application Methods
void KNeighLevApplication::StartApp()    // Called at time specified by Start
{
  // Copied from original constructor by GFR
  l4Proto = new UDP();
  //cout << "in KNeighLev Application constructor...\n";
  IFVec_t ifs = node->Interfaces();
  
  iw = NULL;
  for(IFVec_t::iterator i = ifs.begin(); i != ifs.end(); ++i)
    {
      DEBUG0((cout << "Node " << (string)IPAddr(node->GetIPAddr())
              << " checking iface " << (*i) << endl));
      if ((*i)->IsWireless())
	{
	  iw = (InterfaceWireless*)(*i);
          DEBUG0((cout << "Found a wireless interface" << endl));
	  break;
	}
    }
  
  if (iw == NULL)
    {
      cout << "KNeighLev: node does not have wireless interface." << endl;
      return;
    }
  
  //cout << "found the wireless interface...\n";
  l4Proto->Attach(node);
  //node->Bind(l4Proto->Proto(), K_PORT, l4Proto);
  if (l4Proto->Bind(K_PORT) == NO_PORT)   
    // Bind the local l4 protocol to available port
    {
      cout << "KNeighLev Application Local Protoocl Bind Failed!" << endl;
      return;
    }
  if (node->Bind(l4Proto->Proto(), K_PORT, l4Proto))
    {
      cout << "KNeighLev Application Local Protoocl Bind Failed!" << endl;
      return;
    }
  l4Proto->AttachApplication(this); // Attached application
  //cout << "Attached application...done with constructor.\n";

  // End of copied from constructor
  StopApp();                               // Insure no pending event
  l4Proto->Connect(IPAddrBroadcast, K_PORT);  

  AppKNeighLevEvent* ev = new AppKNeighLevEvent(AppKNeighLevEvent::START);
  pendingEvent = ev;
  Simulator::instance->Schedule(ev, startTime, this); // Schedule again
}

void KNeighLevApplication::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending transmit event
      Simulator::instance->Cancel(pendingEvent);
      pendingEvent = NULL;
    }
}

void KNeighLevApplication::Receive(Packet* packet, L4Protocol *, Seq_t)
{
  Data *d = (Data *)packet->PeekPDU();
  struct KNeighLevHdr *hdr = (struct KNeighLevHdr *)(d->data);
  int level = hdr->txPowerLevel;
  IPAddr_t v = hdr->ip;
  // if message is a beacon
  if (hdr->type == K_BEACON)
  {
    //cout << IPAddr::ToDotted(ip); 
    //cout << ": Received a BEACON from " << IPAddr::ToDotted(v) 
    // << " with power level: " << KNeighLevApplication::txPower[level] 
    // << "." << endl;
    //  if v (source node) is not in the incoming neighbor list
    if (Nincoming.find(v) == Nincoming.end())
    {
      //   add v to the incoming neighbor list
      Nincoming.insert(TxPowerDataMap_t::value_type(v, level));
      //   if my power level is greater than or equal to v's
      if (myTxPowerLevel >= level)
	//    add v to the symmetric neighbor list
	Nsymmetric.insert(TxPowerDataMap_t::value_type(v, level));
    }
  }
  // if message is a help
  else if (hdr->type == K_HELP)
  {
    //cout << IPAddr::ToDotted(ip); 
    //cout << ": Received a HELP msg from " << IPAddr::ToDotted(v)
    // << " with power level: " << KNeighLevApplication::txPower[level] 
    // << "." << endl;
    //  if v (source node) is in the incoming neighbor list
    //   and not in the symmetric neighbor list
    if (Nincoming.find(v) != Nincoming.end() &&
	Nsymmetric.find(v) == Nsymmetric.end())
    {
      //    step-wise increase to v's power level
      StepWiseIncrease(myTxPowerLevel + 1, level);
    }
    //  else if v is not in the incoming neighbor list
    else if (Nincoming.find(v) == Nincoming.end())
    {
      //   add v to the incoming neighbor list
      Nincoming.insert(TxPowerDataMap_t::value_type(v, level));
      //   if my power level is less than v's
      if (myTxPowerLevel < level)
      {
	//    step-wise increase to v's power level
	StepWiseIncrease(myTxPowerLevel + 1, level);
      }
      //   else
      else
      {
	//    add v to the symmetric neighbor list
	Nsymmetric.insert(TxPowerDataMap_t::value_type(v, level));
      }
    }
  }
  else
    cout << IPAddr::ToDotted(ip) 
	 << ": Received an invalid message." << endl;

  delete packet;
}

void KNeighLevApplication::SendMessage(KMsgType_t type)
{
#ifdef CHANGED_BY_GFR
  struct KNeighLevHdr *hdr = new struct KNeighLevHdr;
  memset((void *)hdr, 0, sizeof(struct KNeighLevHdr));
  hdr->type = type;
  hdr->txPowerLevel = myTxPowerLevel;
  hdr->ip = ip;
  //cout << IPAddr::ToDotted(ip) 
  //     << ((type == K_BEACON) ? ": BEACON sent." : ": HELP sent.") << endl;
  l4Proto->SendTo((char*)hdr, hdr->size(), IPAddrBroadcast, K_PORT);
#endif
  struct KNeighLevHdr hdr;
  memset((void*)&hdr, 0, sizeof(hdr));
  hdr.type = type;
  hdr.txPowerLevel = myTxPowerLevel;
  hdr.ip = ip;
  l4Proto->SendTo((char*)&hdr, sizeof(hdr), IPAddrBroadcast, K_PORT);
}

void KNeighLevApplication::SetTxPowerLevel(int level)
{
  myTxPowerLevel = level;
  assert(iw);
  iw->setTxPower(KNeighLevApplication::txPower[myTxPowerLevel]);
  node->SetRadioRange(KNeighLevApplication::radioRange[myTxPowerLevel]);
  //cout << IPAddr::ToDotted(ip) 
  //     << ": Power set to " << iw->getTxPower() << " mW. "
  //     << "Range set to " << node->GetRadioRange() << " m." << endl;
  changedPower = true;
}

void KNeighLevApplication::StepWiseIncrease(int levI, int levF)
{
  //cout << IPAddr::ToDotted(ip) 
  //     << ": in step-wise increase..." << endl;
  //cout << "levI: " << levI << ", levF: " << levF << endl;
  for (int h = levI; h <= levF; h++)
  {
    SetTxPowerLevel(h);
    SendMessage(K_BEACON);

    // if there exists a node i in the incoming neighbor list
    //  with the same power level
    //   add it to the symmetric neighbor list
    TxPowerDataMap_t::iterator i;
    for (i = Nincoming.begin(); i != Nincoming.end(); i++)
      if (i->second == h)
	Nsymmetric.insert(*i);
  }
  //cout << "leaving step-wise increase..." << endl;
}
