/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - On/Off Data Source Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define an on/off data source application class

#ifndef __cbr_application_h__
#define __cbr_application_h__

#include "common-defs.h"
#include "application.h"
#include "rng.h"
#include "udp.h"

// Define the start/stop events
//Doc:ClassXRef
class AppOOEvent : public Event {
public:
  typedef enum { SEND_PKT = 100, START_GEN, STOP_GEN } AppOOEvent_t;
  AppOOEvent() { }
  AppOOEvent(Event_t ev) : Event(ev) { }
};

//Doc:ClassXRef
class OnOffApplication : public Application {
  //Doc:Class Defines an application that uses the on--off data generator
  //Doc:Class model.

public:
  // Define a number of constructors for different needs
  // Endpoints, on/off rng, l4 protocol
  //Doc:Method
  OnOffApplication(IPAddr_t, PortId_t,
                   const Random&, const L4Protocol& = UDP(),
                   Rate_t r = defaultRate,
                   Size_t s = defaultSize);
    //Doc:Desc Constructor specifying destination IP/Port, a single random
    //Doc:Desc number generator for both the {\em On} and {\em Off} 
    //Doc:Desc distributions, and a layer 4 protocol object.  Optionally
    //Doc:Desc specifies bit rate (when {\em On} (defaults to a globally
    //Doc:Desc specified default rate), and optionally a packet size
    //Doc:Desc (defaults to a globally specified default size).  See
    //Doc:Desc {\tt SetDefaultRate} and {\tt SetDefaultSize} below.
    //Doc:Arg1 IPAddress of remote  endpoint (destination)
    //Doc:Arg2 Port number for remote endpoing (destination)
    //Doc:Arg3 Random number generator to use for both {\em On} and {\em Off}
    //Doc:Arg3 time periods.
    //Doc:Arg4 A layer 4 protocol object to copy.
    //Doc:Arg5 Data rate to generate when {\em On}.
    //Doc:Arg6 Packet size.

  // Endpoints, on rng, off rng, l4 protocol
  //Doc:Method
  OnOffApplication(IPAddr_t, PortId_t,
                   const Random&, const Random&,
                   const L4Protocol& = UDP(),
                   Rate_t r = defaultRate,
                   Size_t s = defaultSize);
    //Doc:Desc Constructor specifying remote IP/Port, a random
    //Doc:Desc number generator for the {\em On} and a second
    //Doc:Desc random number generator for the {\em Off} 
    //Doc:Desc distribution, and a layer 4 protocol object.  Optionally
    //Doc:Desc specifies bit rate (when {\em On} (defaults to a globally
    //Doc:Desc specified default rate), and optionally a packet size
    //Doc:Desc (defaults to a globally specified default size).  See
    //Doc:Desc {\tt SetDefaultRate} and {\tt SetDefaultSize} below.
    //Doc:Arg1 IPAddress of remote  endpoint (destination)
    //Doc:Arg2 Port number for remote endpoing (destination)
    //Doc:Arg3 Random number generator to use for {\em On} period.
    //Doc:Arg4 Random number generator to use for {\em Off} period.
    //Doc:Arg5 A layer 4 protocol object to copy.
    //Doc:Arg6 Data rate to generate when {\em On}.
    //Doc:Arg7 Packet size.

  OnOffApplication(const OnOffApplication&); // Copy constructor
  void Handle(Event*, Time_t);
  virtual void StartApp();    // Called at time specified by Start
  virtual void StopApp();     // Called at time specified by Stop
  virtual void AttachNode(Node*); // Note which node attached to
  virtual Application* Copy() const;// Make a copy of the application
  virtual L4Protocol*  GetL4() const { return l4Proto;}
  virtual L4Protocol*  GetLocalProto() { return l4Proto;}

  //Doc:Method
  virtual void MaxBytes(Count_t m) { maxBytes = m;}
    //Doc:Desc Specify a maximum number of bytes to send, after which the
    //Doc:Desc application shuts down.
    //Doc:Arg1 Maximum number of bytes to send.

public: // Static methods
  //Doc:Method
  static void DefaultRate(Rate_t r) { defaultRate = r;}
    //Doc:Desc Specify a default data rate to use for all On/Off applications
    //Doc:Desc by default
    //Doc:Arg1 The default data rate (when the application is {\em On}.

  //Doc:Method
  static void DefaultSize(Size_t s) { defaultSize = s;}
    //Doc:Desc Specifies a default packet size to use for all
    //Doc:Desc On/Off applcations
    //Doc:Desc by default.
    //Doc:Arg1 The default packet size.

public:
  L4Protocol*   l4Proto;      // Points to the specified Layer 4 protocol
  IPAddr_t      peerIP;       // PeerIP
  PortId_t      peerPort;     // Peer port
  bool          connected;    // True if connected
  Random*       onTime;       // rng for On Time
  Random*       offTime;      // rng for Off Time
  Rate_t        cbrRate;      // Rate that data is generated
  Size_t        pktSize;      // Size of packets
  AppOOEvent*   pendingEvent; // Pending SendPkt event
  AppOOEvent*   pendingOO;    // Pending on or off event
  Size_t        residualBits; // Number of generated, but not sent, bits
  Time_t        lastStartTime;// Time last packet sent
  Count_t       maxBytes;     // Limit total number of bytes sent
  Count_t       totBytes;     // Total bytes sent so far
public:
  static Rate_t defaultRate;
  static Size_t defaultSize;
private:
  // void ConstructorHelper( IPAddr_t, PortId_t);
  void ScheduleNextTx();
};

#endif

