/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// UDPStorm DDOS Attack application for GTNetS
// George F. Riley, Georgia Tech, Summer 2003

#include <iostream>

#ifdef HAVE_QT
#include <qnamespace.h>
#endif

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "application-udpstorm.h"
#include "udp.h"
#include "packet.h"
#include "datapdu.h"
#include "ipv4.h"

using namespace std;

Count_t     UDPStorm::defaultPktPerSec = 10;
Count_t     UDPStorm::defaultPktLength = 1000;

UDPStorm::UDPStorm() : pktsPerSec(defaultPktPerSec),
                       pktLength(defaultPktLength),
                       useSpoofing(true),
                       timeoutEvent(nil), node(nil),
                       rngV(nil), rngS(nil), rngP(nil)
{
}

UDPStorm::UDPStorm(const UDPStorm& r)
  : pktsPerSec(r.pktsPerSec),
    pktLength(r.pktLength),
    useSpoofing(r.useSpoofing),
    timeoutEvent(nil), node(nil),
    rngV(nil), rngS(nil), rngP(nil)
{
}

void UDPStorm::Timeout(TimerEvent* ev)
{
  if (!node)
    {
      cout << "UDPStorm::Timeout with no attached node" << endl;
      return;
    }
  DEBUG0((cout << "UDPStorm::Timeout " << Simulator::Now() << endl));
  // Create the packet and TCP header
#ifdef HAVE_QT
  // Use red for storm packets
  ColoredPacket* p = new ColoredPacket();
  p->SetColor(Qt::red);
#else
  Packet* p = new Packet();
#endif
  // Create the data pdu
  Data* d = new Data(pktLength);
  p->PushPDU(d);

  // Create the UDP pdu
  UDPHeader* h = new UDPHeader();
  if (!rngP) rngP = new Uniform(10000, 50000); // Range for random ports
  h->sourcePort = rngP->IntValue();
  h->destPort = rngP->IntValue(); // Victim's port
  h->messageLength = pktLength;   // Length of storm packets
  h->checksum= 0;
  h->fid = 0;
  // Add this header to the packet
  p->PushPDU(h); 

  // Create the layer 3 request info
  if (!rngV) rngV = new Uniform(0, victims.size());
  IPAddr_t victimIP = victims[rngV->IntValue()]; // Get random victim
  IPAddr_t spoofIP;
  if (!useSpoofing)
    { // Not spoofing, use node's ip address
      spoofIP = node->GetIPAddr();
    }
  else
    {
      if (!spoofs.empty())
        { // Use a spoof'ed ip from the list
          if (!rngS) rngS = new Uniform(0, spoofs.size());
          spoofIP = spoofs[rngS->IntValue()]; // Get random spoof
        }
      else
        {
          if (!rngS) rngS = new Uniform(0xa0000000, 0xcfffffff); // IP Range
          spoofIP = rngS->IntValue(); // Get spoofed IP address
          DEBUG0((cout << "UDPStorm using spoofed IP "
                  << (string)IPAddr(spoofIP) << endl));
        }
    }

  IPV4ReqInfo r(victimIP, spoofIP, 64, 6); // Info for l3
  // Pass to layer 3 for further processing
  IPV4::Instance()->DataRequest(node, p, &r);

  // Schedule the next pkt
  StartApp();
}

void UDPStorm::StartApp()
{ // Schedule the timer for the next pkt-pkt
  if (!timeoutEvent) timeoutEvent = new TimerEvent();
  Time_t future = 1.0 / (double)pktsPerSec;
  timer.Schedule(timeoutEvent, future, this);
}

void UDPStorm::StopApp()
{ // Cancel the timer if active
  if (timeoutEvent) timer.Cancel(timeoutEvent);
}

void UDPStorm::AttachNode(Node* n)
{
  node = n;
}


Application* UDPStorm::Copy() const
{
  return new UDPStorm(*this);
}

void UDPStorm::Victim(IPAddr_t ip)
{ // Add victim ip
  victims.push_back(ip);
  delete rngV; // Any existing victim RNG is not valid
  rngV = nil;
}

void UDPStorm::Spoof(IPAddr_t ip)
{ // Add spoofed IP address
  spoofs.push_back(ip);
  delete rngS; // Any existing victim RNG is not valid
  rngS = nil;
}

// Static Methods
void UDPStorm::DefaultPktPerSec(Count_t sps)
{
  defaultPktPerSec = sps;
}

void UDPStorm::DefaultPktLength(Count_t l)
{
  defaultPktLength = l;
}
