/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Web Browser Application Class
// George F. Riley.  Georgia Tech, Fall 2002

// Define a Web Browser Application class
// Based on the Web Browsing models by Bruce Mah, Infocom 1997

// Uncomment below to enable debug 
//#define DEBUG_MASK 0x02
#include "debug.h"
#include "application-webbrowser.h"
#include "tcp.h"
#include "simulator.h"
#include "globalstats.h"

#include "http-distributions.h"

using namespace std;

// Number of simulataneous connections
#define DEFAULT_CONNECTIONS 2

// Static memeber variables
HttpPrimaryRequest*   WebBrowser::primaryRequest;
HttpSecondaryRequest* WebBrowser::secondaryRequest;
HttpPrimaryReply*     WebBrowser::primaryReply;
HttpSecondaryReply*   WebBrowser::secondaryReply;
HttpFilesPerPage*     WebBrowser::filesPerPage;
HttpConsecutivePages* WebBrowser::consecutivePages;
HttpThinkTime*        WebBrowser::thinkTime;
Count_t               WebBrowser::defaultConcurrentConn = DEFAULT_CONNECTIONS;
Count_t               WebBrowser::activeBrowsers;
Count_t               WebBrowser::activeConnections;
Count_t               WebBrowser::completedConnections;
Count_t               WebBrowser::failedConnections;
Count_t               WebBrowser::startedThisPeriod;
Count_t               WebBrowser::completedThisPeriod;
Time_t                WebBrowser::totRespTime;
DCount_t              WebBrowser::totRespSize;
Count_t               WebBrowser::totObjectsPerPage;
Count_t               WebBrowser::totBrowserSessions;
Time_t                WebBrowser::largestRespTime;
Count_t               WebBrowser::largestRespSize;
Time_t                WebBrowser::totWaitTime;
Time_t                WebBrowser::largestWaitTime;
Count_t               WebBrowser::waitCount;
DCount_t              WebBrowser::totReqBytes;   // Total size of requests
DCount_t              WebBrowser::totRespBytes;  // Total size of responses
Count_t               WebBrowser::cleanConnections;
Count_t               WebBrowser::dirtyConnections;
Count_t               WebBrowser::clientTimeouts;
Count_t               WebBrowser::serverTimeouts;

// Browser Connection methods
BrowserConnection::BrowserConnection() 
    : t(nil), start(0),
      rqSize(0), rsSize(0), rxSize(0), cnCount(0),
      active(false), logged(false), closed(false),
      first(true)
{
}

BrowserConnection::BrowserConnection(const TCP& tcp) 
  : t((TCP*)tcp.Copy()), start(0),
    rqSize(0), rsSize(0), rxSize(0), cnCount(0),
    active(false), logged(false), closed(false)
{
}


// Constructor
WebBrowser::WebBrowser(const IPAddrVec_t& s, // List of servers
                       const Random& ss,     // Random to select server
                       const TCP& t)         // TCP Variant to use
    : pendingEvent(nil), node(nil), tcp((TCP*)(t.Copy())),
      stats(nil), dbstats(nil),
      servers(s),
      selectServer(ss.Copy()),
      concurrentConn(defaultConcurrentConn),
      serverIP(IPADDR_NONE), consecutive(0),
      nObjects(0), nObjectsRq(0), nObjectsRx(0),
      thinkTimeBound(INFINITE_TIME), idleTime(INFINITE_TIME),
      verbose(false)
{
#ifdef MOVED_TO_START
  for (Connections_t::size_type i = 0; i < connections.size(); ++i)
    {
      connections[i].t = (TCP*)t.Copy(); // Load the tcp endpoints 
      connections[i].t->AttachApplication(this); // Notify of application
      DEBUG(4,(cout << "WebBrowser tcp " << connections[i].t << endl));
    }
#endif
}

WebBrowser::WebBrowser(const WebBrowser& c)
    : Application(c), pendingEvent(nil), node(c.node),
      tcp((TCP*)(c.tcp->Copy())),
      stats(c.stats),
      dbstats(c.dbstats),
      servers(c.servers),
      selectServer(c.selectServer->Copy()),
      concurrentConn(c.concurrentConn),
      serverIP(c.serverIP), consecutive(c.consecutive),
      nObjects(c.nObjects),
      nObjectsRq(c.nObjectsRq), nObjectsRx(c.nObjectsRx),
      thinkTimeBound(c.thinkTimeBound), idleTime(c.idleTime),
      verbose(c.verbose)
{ // Copy constructor
#ifdef MOVED_TO_START
  for (Connections_t::size_type i = 0; i < connections.size(); ++i)
    { // Load the tcp endpoints 
      if (connections[i].t)
        connections[i].t = (TCP*)connections[i].t->Copy();
    }
#endif
}

WebBrowser::~WebBrowser()
{ // Destructor
  // Delete the select server RNG
  if (selectServer) delete selectServer;
  if (tcp) delete tcp;
  // Delete the tcp connections
  for (ConnectionVec_t::size_type i = 0; i < connections.size(); ++i)
    { // Load the tcp endpoints 
      if (connections[i].t)
        delete connections[i].t;
    }
}

// Handler Methods
void WebBrowser::Handle(Event* e, Time_t t)
{
  AppWebBrowserEvent* ev = (AppWebBrowserEvent*)e;
  switch (ev->event) {
    case AppWebBrowserEvent::REQUEST_PAGE :
      DEBUG0((cout << "Request page Event" << endl));
      if (Simulator::Now() < idleTime)
        {
          RequestWebPage();  // Request the next web page
        }
      pendingEvent = nil;
      delete e;
      return;
    default :
      break;
  }
  Application::Handle(e, t);  // May be application event
}

  // Upcalls from L4 protocol
void WebBrowser::Receive(Packet* p,L4Protocol* pr,Seq_t s)
{ // Data received, see if we have all
  BrowserConnection* c = FindConnection(pr);
  if (!c)
    {
      cout << "HuH?  WebBrowser::Receive, can't find connection"
           << endl;
      return;
    }
  Data* d = (Data*)p->PopPDU();  // Get the data pdu
  if (c->rxSize != s) 
    { // debug testing
      cout << "HuH?  TCP " << c->t
           << " seq/rx mismatch, seq " << s
           << " rxSize " << c->rxSize << endl;
      exit(1);
    }
  c->rxSize += d->Size();        // Count bytes received
  DEBUG(2,(cout << "AppWB, RX " << d->Size() 
          << " tot " << c->rxSize
          << " exp " << c->rsSize << endl));
  delete p; // Delete the packet
}

void WebBrowser::CloseRequest(L4Protocol* pr)
{ // Close request from peer
  BrowserConnection* c = FindConnection(pr);
  DEBUG0((cout << "WebBrowser::CloseRequest() " << this << endl));
  pr->Close(); // Close in response
  Time_t now = Simulator::Now();
  if (stats)
    { // Log response time in statistics object
      stats->Record(now - c->start);
      totRespSize += c->rxSize;
      totRespTime += (now - c->start);
      if (c->rxSize > largestRespSize)
        largestRespSize=c->rxSize;
      if ((now - c->start) > largestRespTime)
        largestRespTime = now - c->start;
      DEBUG(2,(cout << "Size " << c->rxSize << " resp " << now - c->start 
               << " toc " << c->t->TimeoutCount()
               << " tod " << c->t->TimeoutDelay()));
      TCP* pp = c->t->Peer();
      if (pp)
        {
          DEBUG(2,(cout << " ptoc " << pp->TimeoutCount()
                   << " ptod " << pp->TimeoutDelay()));
        }
      DEBUG(2,(cout << endl));
    }
  if (verbose)
    {
      cout << "rqSize " << c->rqSize
           << " rxSize " << c->rxSize
           << " rxTime " << now - c->start
           << " ";
      Statistics* st = c->t->GetRTTEstimator()->GetPrivateStatistics();
      if (st) st->Log(cout);
    }
  c->logged = true;
  DEBUG(1,(cout << "AppWB, CloseReq, Time " << now
       << " logging resp time " << now - c->start
       << " objsize " << c->rsSize
       << " rxSize " << c->rxSize
           << endl));
  // More debug
  TCP* tc = c->t->Peer();
  if (tc)
    {
      if (tc->TimeoutCount() == 0 && tc->RetransmitCount() == 0)
        {
          cleanConnections++;
        }
      else
        {
          dirtyConnections++;
        }
      serverTimeouts += tc->TimeoutCount();
    }
  clientTimeouts += c->t->TimeoutCount();
}

void WebBrowser::Closed(L4Protocol* pr)
{ // Connection has closed
  DEBUG0((cout << "WebBrowser::Closed() " << this << endl));
  BrowserConnection* c = FindConnection(pr);
  // debug
  if (dbstats)
    {
      TCP* tc = c->t->Peer();
      if (tc) dbstats->Record(tc->TimeoutCount());
    }
  if (!c->logged)
    {
      Time_t now = Simulator::Now();
      if (stats)
        { // Log response time in histogram
          stats->Record(now - c->start);
          totRespSize += c->rxSize;
          totRespTime += (now - c->start);
          if (c->rxSize > largestRespSize)
            largestRespSize = c->rxSize;
          if ((now - c->start) > largestRespTime)
            largestRespTime = now - c->start;
          DEBUG(2,(cout << "Size " << c->rxSize << " resp " << now - c->start 
                   << " toc " << c->t->TimeoutCount()
                   << " tod " << c->t->TimeoutDelay()));
          TCP* pp = c->t->Peer();
          if (pp)
            {
              DEBUG(2,(cout << " ptoc " << pp->TimeoutCount()
                       << " ptod " << pp->TimeoutDelay()));
            }
          DEBUG(2,(cout << endl));
        }
      if (verbose)
        {
          cout << "rqSize " << c->rqSize
               << " rxSize " << c->rxSize
               << " rxTime " << now - c->start
               << " ";
          Statistics* st = c->t->GetRTTEstimator()->GetPrivateStatistics();
          if (st) st->Log(cout);
        }
      c->logged = true;
      // More debug
      TCP* tc = c->t->Peer();
      if (tc)
        {
          if (tc->TimeoutCount() == 0 && tc->RetransmitCount() == 0)
            {
              cleanConnections++;
            }
          else
            {
              dirtyConnections++;
              serverTimeouts += tc->TimeoutCount();
            }
        }
      clientTimeouts += c->t->TimeoutCount();
      // end debug
    }
  if (c->active)
    { // Finshed, see if we need another
      DEBUG(2,(cout << "AppWB, Closed "
               << " tot " << c->rxSize
               << "  exp " << c->rsSize << endl));
      if (c->rxSize != c->rsSize)
        { // ! HuH?
          cout << "WebBrowser " << this << " debug exit" << endl;
          cout << "AppWB, " << this << " Closed, time " << Simulator::Now()
               << " TCP " << c->t
               << " tot " << c->rxSize
               << " exp " << c->rsSize << endl;
          cout << "MyIP " << (string)IPAddr(c->t->localNode->GetIPAddr())
               << " myport " << c->t->localPort << endl;
        }
      ObjectComplete(c);
    }
}

void WebBrowser::ConnectionComplete(L4Protocol* pr)
{ // Connection request succeeded
  BrowserConnection* c = FindConnection(pr);
  if (!c)
    {
      cout << "HuH?  WebBrowser::ConnectionComplete, can't find connection"
           << endl;
      return;
    }
  DEBUG(1,(cout << "WebBrowser::ConnectionComplete, "
           << " TCP " << c->t
           << " req " << c->rqSize
           << " resp " << c->rsSize << endl));
  Data d(c->rqSize);           // Request size
  d.msgSize = c->rqSize;       // Request size
  d.responseSize = c->rsSize;  // Response size
  c->t->Send(d);               // Send the request
}

void WebBrowser::ConnectionFailed(L4Protocol* pr)
{ // Connection request failed
  BrowserConnection* c = FindConnection(pr);
  failedConnections++;
  if (!c)
    {
      cout << "HuH?  WebBrowser::ConnectionFailed, can't find connection"
           << endl;
      return;
    }
  if (c->cnCount < CONNECT_RETRY)
    { // Try again
      c->cnCount++;
      cout << "WebBrowser::ConnectionFailed, TCP " << c->t 
               << " retrying try " << c->cnCount
               << " of " << CONNECT_RETRY 
               << " time " << Simulator::Now()
               << endl;
      c->t->Reset(); // Reset the connection
      c->t->Connect(serverIP, HTTP_PORT);
    }
  else
    {
      cout << "WebBrowser::ConnectionFailed, TCP " << c->t 
               << " giving up" 
               << " time " << Simulator::Now()
               << endl;
      c->active = false;
      ObjectComplete(c);
    }
}

// Application Methods
void WebBrowser::StartApp()    // Called at time specified by Start
{
  DEBUG0((cout << "Starting browser " << this 
          << " time " << Simulator::Now() << endl));
  StopApp();                   // Insure no pending event
  if (connections.size() == 0)
    { // Need to create the connections list
      for (ConnectionVec_t::size_type i = 0; i < concurrentConn; ++i)
        {
          connections.push_back(BrowserConnection(*tcp));
          BrowserConnection& c = connections.back();
          c.t->AttachApplication(this);
        }
    }
  RequestWebPage();            // Request the next web page    
}

void WebBrowser::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending start event
      Simulator::instance->Cancel(pendingEvent);
      delete pendingEvent;
      pendingEvent = nil;
    }
}

void WebBrowser::AttachNode(Node* n)
{ // Note attached node
  node = n;
  tcp->Attach(n);
}

Application* WebBrowser::Copy() const
{ // Make a copy of this application
  return new WebBrowser(*this);
}

void WebBrowser::ConcurrentConnections(Count_t c)
{ // Set number of concurrent connections
  concurrentConn = c;
}

void WebBrowser::DBDump()
{
  for (ConnectionVec_t::size_type i = 0; i < connections.size(); ++i)
    {
      if (connections[i].active)
        {
          connections[i].t->DBDump(); // Dump active tcp connection
          cout << "       rsSize " << connections[i].rsSize
               << " rxSize " << connections[i].rxSize << endl;
        }
    }
}

// Private Methods

BrowserConnection* WebBrowser::FindConnection(L4Protocol* l4)
{
  for (ConnectionVec_t::size_type i = 0; i < connections.size(); ++i)
    {
      if (connections[i].t == l4) return &connections[i]; // Found it
    }
  return nil; // Not found
}

void WebBrowser::RequestWebPage()
{ // Request the next web page
  if (servers.size() == 0)
    {
      cout << "WebBrowser::RequestWebPage, no servers!" << endl;
      return;
    }
  activeBrowsers++;
  if (!consecutive)
    { // Time to select a new server
      if (!consecutivePages) InitializeRandom();
      consecutive = consecutivePages->IntValue();
      IRandom_t ss = selectServer->IntValue(); // Choose a server
      if (ss >= servers.size())
        { 
          cout << "WebBrowser::RequestWebPage, bad server selector" << endl;
          ss = 0;
        }
      serverIP = servers[ss];
    }
  nObjects = filesPerPage->IntValue();
  if (!nObjects) nObjects++; // Zero not allowed
  totObjectsPerPage += nObjects;
  totBrowserSessions++;
  nObjectsRq = 0;
  nObjectsRx = 0;
  DEBUG0((cout << "WebBrowser::RWP, nObj " << nObjects
          << "conn.size() " << connections.size()
          << endl));
  for (ConnectionVec_t::size_type i = 0; i < connections.size(); ++i)
    {
      RequestOneFile(&connections[i]);
      if (nObjectsRq == nObjects) break; // Requested all
    }
}

void WebBrowser::RequestOneFile(BrowserConnection* c)
{
  c->t->Reset(); // Reset the tcp endpoint
  c->start = Simulator::Now(); // Note the start time
  //if (!nObjectsRq)
  if (nObjectsRq < 2) // The Jeffay experiment had a bug, primary twice
    { // Primary
      c->rqSize = primaryRequest->IntValue();  // Request size
      c->rsSize = primaryReply->IntValue();    // Response size
    }
  else
    { // Secondary
      c->rqSize = secondaryRequest->IntValue();// Request size
      c->rsSize = secondaryReply->IntValue();  // Response size
    }
  DEBUG0((cout << "Browser " << this << " rqSize " << c->rqSize
          << " rsSize " << c->rsSize << endl));
  if (!c->rqSize) c->rqSize++;
  if (!c->rsSize) c->rsSize++;        // Insure non-zero for sizes
  c->rxSize = 0;                      // Clear received count   
  nObjectsRq++;                       // Note this request
  c->t->Bind();                       // Bind to available local port
  c->t->FlowId(++TCP::nextFlowId);    // Unique id for debugging
  DEBUG(1,(cout << "AppWB Connecting to server IP " << (string)IPAddr(serverIP)
          << " time " << Simulator::Now()
          << endl));
  
  c->t->Connect(serverIP, HTTP_PORT); // Connect to the server
  c->active = true;
  c->logged = true;
  c->cnCount = 0;                     // Connection Count
  activeConnections++;
  startedThisPeriod++;
  totReqBytes  += c->rqSize;
  totRespBytes += c->rsSize;
  DEBUG(2,(cout << "AppWB, RequestSize " << c->rqSize
           << " ReplySize " << c->rsSize << endl));
  DEBUG(2,(cout << "AppWB, " << this << " Rq " << c->rqSize
           << " Rs " << c->rsSize 
           << " ip " << (string)IPAddr(c->t->localNode->GetIPAddr())
           << " port " << c->t->localPort << endl));
  DEBUG0((cout << "WebBrowser::ROF, obj " << nObjectsRq <<
          " of " << nObjects << endl));
}

void WebBrowser::ObjectComplete(BrowserConnection* c)
{
  // Remove binding
  c->t->Unbind(c->t->ProtocolNumber(), c->t->Port());
  c->active = false;
  nObjectsRx++; // Count received objects
  activeConnections--;
  completedConnections++;
  completedThisPeriod++;
  if (nObjectsRq < nObjects)
    { // Still pending objects
      RequestOneFile(c);
    }
  if (nObjectsRx == nObjects)
    { // All done, get the sleep time and schedule wakeup event
      activeBrowsers--;
      consecutive--;
      if (Simulator::Now() < idleTime)
        { // Not time to idle
          if (!consecutive)
            { // Try to match a bug in the UNC jeffay experiment
              // They had a bug that did not sleep when choosing
              // a new server
              RequestWebPage();
            }
          else
            {
              Time_t st = thinkTime->Value();
              DEBUG0((cout << "Browser " << this << "tt " << st << endl));
              if (st > thinkTimeBound)
                {
                  DEBUG0((cout << "Bounding TT " << st 
                          << " to " << thinkTimeBound << endl));
                  st = thinkTimeBound;
                }
              pendingEvent = 
                new AppWebBrowserEvent(AppWebBrowserEvent::REQUEST_PAGE);
              // Schedule again    
              DEBUG(1,(cout << "Sleeping for " << st << " secs" << endl));
              Simulator::instance->Schedule(pendingEvent, st, this);
              // Stats for debugging
              totWaitTime += st;
              if (st > largestWaitTime) largestWaitTime = st;
              waitCount++;
            }
        }
    }
}

// Static Methods
void WebBrowser::InitializeRandom(Seed_t s)
{
  if (primaryRequest)   delete primaryRequest;
  if (secondaryRequest) delete secondaryRequest;
  if (primaryReply)     delete primaryReply;
  if (secondaryReply)   delete secondaryReply;
  if (filesPerPage)     delete filesPerPage;
  if (consecutivePages) delete consecutivePages;
  if (thinkTime)        delete thinkTime;

  primaryRequest   = new HttpPrimaryRequest(s);
  secondaryRequest = new HttpSecondaryRequest(s);
  primaryReply     = new HttpPrimaryReply(s);
  secondaryReply   = new HttpSecondaryReply(s);
  filesPerPage     = new HttpFilesPerPage(s);
  consecutivePages = new HttpConsecutivePages(s);
  thinkTime        = new HttpThinkTime(s);
}

