/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Web Browser Application Class
// George F. Riley.  Georgia Tech, Fall 2002

// Define a Web Browser Applicatoinn class

#ifndef __tcp_application_browser_h__
#define __tcp_application_browser_h__

#include "common-defs.h"
#include "application.h"
#include "tcp.h"
#include "rng.h"
#include "statistics.h"
#include "http-distributions.h"

#define HTTP_PORT 80
#define CONNECT_RETRY 3

typedef std::vector<IPAddr_t> IPAddrVec_t;     // List of ip addresses

// Define a class with info for a single connection
//Doc:ClassXRef
class BrowserConnection {
public:
  BrowserConnection();
  BrowserConnection(const TCP& tcp);
public:
  TCP*    t;       // TCP Connection
  Time_t  start;   // Time request initialized
  Count_t rqSize;  // Request size
  Count_t rsSize;  // Response size
  Count_t rxSize;  // Bytes received so far
  Count_t cnCount; // Connection re-try Count
  bool    active;  // True if active
  bool    logged;  // True if already logged
  bool    closed;  // True if already closed
  bool    first;   // True if first connection
};

typedef std::vector<BrowserConnection> ConnectionVec_t;

// Define the start/stop events
//Doc:ClassXRef
class AppWebBrowserEvent : public Event {
public:
  typedef enum { REQUEST_PAGE = 100 } AppWebBrowserEvent_t;
  AppWebBrowserEvent() { }
  AppWebBrowserEvent(Event_t ev) : Event(ev) { }
};

//Doc:ClassXRef
class WebBrowser : public Application {
  //Doc:Class Simulation model of a web browser.  The browser behavior uses
  //Doc:Class empirical distributions (based on a study by Bruce Mah
  //Doc:Class published in Infocom 1997) to define web browsing actions.
  //Doc:Class Each browser performs the following actions.  First a
  //Doc:Class server is selected randomly from the server list passed
  //Doc:Class in the constructor, and a connection with that server
  //Doc:Class is established.  Then the number of objects (individual
  //Doc:Class connections) needed to get the web page is sampled from
  //Doc:Class the empirical distributions.  For each object needed,
  //Doc:Class a request size and a response size are sampled from the
  //Doc:Class empirical distribution, and the request is sent to the server.
  //Doc:Class If more than one object is needed, a second simultaneous
  //Doc:Class connection is opened and a second request is sent.
  //Doc:Class When the entire request has been received by a web server,
  //Doc:Class the number of bytes from the response size are sent to the
  //Doc:Class browser, and the connection is closed.  When the
  //Doc:Class browser receives all the bytes of the requested object
  //Doc:Class and detects the connection closing, another object is
  //Doc:Class requested (if any remain for this page).  After all
  //Doc:Class of the objects for a given page are received, the browser
  //Doc:Class "sleeps" for a period of time (sampled from the empirical
  //Doc:Class distribution).  After the sleep time is over, another
  //Doc:Class web page (again with multiple objects) is requested,
  //Doc:Class either from the same server or a new one, depending
  //Doc:Class on the empirical random distribution of {\em Consecutive
  //Doc:Class Pages}, which indicates how many times the same server
  //Doc:Class is accessed consecutively.


public:
  // Constructor
  //Doc:Method
  WebBrowser(const IPAddrVec_t&,           // List of servers
             const Random&,                // Random to select server
             const TCP& = TCP::Default()); // TCP Variant to us
    //Doc:Desc Constructor for Web Browser objects.
    //Doc:Arg1 A vector of \IPA{s} to sample for selecting a web server.
    //Doc:Arg2 A random number generator to sample for selecting a web server.
    //Doc:Arg3 A reference to a {\tt TCP} variant to use for all connections.

  //Doc:Method
  WebBrowser(const WebBrowser& c);             // Copy constructor
    //Doc:Desc Copy Constructor
    //Doc:Arg1 Web Browser object to copy.

  virtual ~WebBrowser();
  void Handle(Event*, Time_t);
  // Upcalls from L4 protocol
  virtual void Receive(Packet*,L4Protocol*,Seq_t);   // Data received
  virtual void CloseRequest(L4Protocol*);      // Close request from peer
  virtual void Closed(L4Protocol*);            // Connection has closed
  virtual void ConnectionComplete(L4Protocol*);// Connection request succeeded
  virtual void ConnectionFailed(L4Protocol*);  // Connection request failed
  // Called from base application class
  virtual void StartApp();            // Called at time specified by Start
  virtual void StopApp();             // Called at time specified by Stop
  virtual void AttachNode(Node*);     // Note which node attached to
  virtual Application* Copy() const;  // Make a copy of the application
  // Browser application methods
  
  //Doc:Method
  TCP* GetTCP() { return tcp; }
    //Doc:Desc   Get the TCP prototype used for this browser.
    //Doc:Return Pointer to TCP prototype

  //Doc:Method
  void ConcurrentConnections(Count_t);// Specify number concurrent conn's
    //Doc:Desc Specifies the number of concurrent connections for this browser.
    //Doc:Arg1 Count of concurrent connections.

  //Doc:Method
  void SetStatistics(Statistics* s) { stats = s;}
    //Doc:Desc Set a statistics object used for logging web response time.
    //Doc:Arg1 A pointer to any statistics object (such as a histogram).

  void SetDebugStats(Statistics* s) { dbstats = s;}
  //Doc:Method
  void IdleTime(Time_t i) { idleTime = i;}
    //Doc:Desc Specify the idle time.  The browser will stop initiating requests
    //Doc:Desc at this time, allowing the simulation to (eventually) finish.
    //Doc:Arg1 Idle time in simulation seconds.

  void DBDump();                      // Debug data dump
  void Verbose(bool v = true) { verbose = v;}

public:
  //Doc:Method
  void        ThinkTimeBound(Time_t b) { thinkTimeBound = b;}
    //Doc:Desc Specify a bound on the think time.  The empirical distributions
    //Doc:Desc for think time allow a fairly large think time, which
    //Doc:Desc can result in a simulation where most of the browsers
    //Doc:Desc are thinking.  This allows an arbitrary upper bound
    //Doc:Desc on the think time.
    //Doc:Arg1 Think time bound in simulation seconds.

  //Doc:Method
  static void InitializeRandom(Seed_t s = Random::RANDOM_SEED);
    //Doc:Desc Specify a seed for the random number generators.
    //Doc:Arg1 Random seed.

  static void DefaultConcurrentConnections(Count_t);
private:
  // Find the connection info given the protocol pointer
  BrowserConnection* FindConnection(L4Protocol*);
  void               RequestWebPage();// Request the next web page
  void               RequestOneFile(BrowserConnection*); // Request object
  void               ObjectComplete(BrowserConnection*); // Object completed
public:
  AppWebBrowserEvent* pendingEvent;   // Pending SendData event
  Node*              node;            // Attached node
  TCP*               tcp;             // TCP to use for connections
  Statistics*        stats;           // Stats object for logging response time
  Statistics*        dbstats;         // Stats for debugging
  const IPAddrVec_t  servers;         // List of servers
  Random*            selectServer;    // Server selector RNG
  Count_t            concurrentConn;  // Number of concurrent connections
  ConnectionVec_t    connections;     // Existing connections
  IPAddr_t           serverIP;        // IP Address of last server used
  Count_t            consecutive;     // Remaining number of consecutive req
  Count_t            nObjects;        // Number of web objects this page
  Count_t            nObjectsRq;      // Number of web objects requested
  Count_t            nObjectsRx;      // Number of web objects received
  Time_t             thinkTimeBound;  // Bound on think time
  Time_t             idleTime;        // Time to stop requesting
  bool               verbose;         // True for verbose debugging
public:
  // To save memory, we use shared RNG's for the Mah RV's
  static HttpPrimaryRequest*   primaryRequest;
  static HttpSecondaryRequest* secondaryRequest;
  static HttpPrimaryReply*     primaryReply;
  static HttpSecondaryReply*   secondaryReply;
  static HttpFilesPerPage*     filesPerPage;
  static HttpConsecutivePages* consecutivePages;
  static HttpThinkTime*        thinkTime;

  // Default concurrent connection count
  static Count_t               defaultConcurrentConn;

  // Statistics for debugging
  static Count_t               activeBrowsers;
  static Count_t               activeConnections;
  static Count_t               completedConnections;
  static Count_t               failedConnections;
  static Count_t               startedThisPeriod;
  static Count_t               completedThisPeriod;
  static Time_t                totRespTime;
  static DCount_t              totRespSize;
  static Count_t               totObjectsPerPage;
  static Count_t               totBrowserSessions;
  static Time_t                largestRespTime;
  static Count_t               largestRespSize;
  static Time_t                totWaitTime;
  static Time_t                largestWaitTime;
  static Count_t               waitCount;
  static DCount_t              totReqBytes;   // Total size of requests
  static DCount_t              totRespBytes;  // Total size of responses
  static Count_t               cleanConnections; // Number completed cn's with no retx or timeout
  static Count_t               dirtyConnections; // Number completed cn's not clean
  static Count_t               clientTimeouts;
  static Count_t               serverTimeouts;
};

#endif

