/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Helper classes for command line argument processing
// George F. Riley, Georgia Tech, Winter 2002

#include <iostream>

#include "args.h"

using namespace std;

// Arg static member variables
ArgVec_t    Arg::args;      // The expected arguments list
StringVec_t Arg::nonEquiv;  // The non-equivalenced argument list

Arg::Arg(const string& n, long& v, long d)
  : argName(n), argType(Int)
{
  v = d; // Set default value
  pArg = (void*)&v;
  args.push_back(*this);
}

Arg::Arg(const string& n, unsigned long& v, unsigned long d)
  : argName(n), argType(Long)
{
  v = d; // Set default value
  pArg = (void*)&v;
  args.push_back(*this);
}

Arg::Arg(const string& n, double& v, double d)
  : argName(n), argType(Double)
{
  v = d; // Set default value
  pArg = (void*)&v;
  args.push_back(*this);
}

Arg::Arg(const string& n, string& v, const string& d)
  : argName(n), argType(String)
{
  v = d; // Set default value
  pArg = (void*)&v;
  args.push_back(*this);
}

// Arg static methods
bool Arg::ProcessArgs(int argc, char** argv)
{
  for (int i = 1; i < argc; ++i)
    { // Process each argument
      string s(argv[i]);
      string::size_type e = s.find_first_of('=');
      if (e == string::npos)
        { // No equal sign, add to non-equiv list
          nonEquiv.push_back(s);
        }
      else
        {
          if ((e+1) == s.length())
            { // Nothing after the = sign
              cout << "Missing equivalence for parameter " << s << endl;
              exit(1);
            }
          string f = s.substr(0, e);
          string v = s.substr(e+1, string::npos);
          ArgVec_t::iterator j = args.begin();
          bool found = false;
          for(; j != args.end() && !found; ++j)
            {
              //cout << "checking " << j->argName << " vs " << f << endl;
              if (f == j->argName)
                { // Found it
                  switch (j->argType) {
                    case Int :
                      {
                        long* pI = (long*)j->pArg;
                        *pI = atol(v.c_str());
                        found = true;
                      }
                      break;
                    case Long :
                      {
                        unsigned long* pU = (unsigned long*)j->pArg;
                        *pU = atol(v.c_str());
                        found = true;
                      }
                      break;
                    case Double :
                      {
                        double* pD = (double*)j->pArg;
                        *pD = atof(v.c_str());
                        found = true;
                      }
                      break;
                    case String :
                      {
                        string* pS = (string*)j->pArg;
                        *pS = v;
                        found = true;
                      }
                      break;
                    }
                }
            }
          if (!found)
            {
              cout << "Unknown equivalence parameter " << s << endl;
              exit(1);
            }
        }
    }
  return true;
}

StringVec_t::size_type Arg::NumberNonEquiv()
{
  return nonEquiv.size();
}

string Arg::NonEquiv(StringVec_t::size_type i)
{
  if (NumberNonEquiv() < i)
    {
      cout << "No non-equivalent arg index " << i << endl;
      exit(1);
    }
  return nonEquiv[i];
}

bool Arg::Specified(const string& v)
{ // True if specified string v was a non-equivalenced argument
  for (StringVec_t::size_type i = 0; i < NumberNonEquiv(); ++i)
    {
      if (NonEquiv(i) == v) return true;
    }
  return false;
}

#ifdef MAIN
int main(int argc, char** argv)
{
  unsigned long q1Limit;
  signed   long testI;
  double        testD;
  string        testS;

  Arg("q1", q1Limit, 1);
  Arg("qi", testI, 2);
  Arg("qd", testD, 5.0);
  Arg("qs", testS, "This is a test");

  Arg::ProcessArgs(argc, argv);
  cout << "q1L " << q1Limit
       << " testI " << testI
       << " testD " << testD
       << " testS " << testS << endl;
  StringVec_t::size_type nneq = Arg::NumberNonEquiv();
  cout << "Number non-equiv " << nneq << endl;
  for (StringVec_t::size_type i = 0; i < nneq; ++i)
    {
      cout << "nneq " << i << " is " << Arg::NonEquiv(i) << endl;
    }
}
#endif

#ifdef BOEING_ARGS
void usage_ospf6_config_arguments(char** argv, struct ospf6_config_args *ca) {
  cout << "Usage:  " << argv[0] <<endl;
  cout << "Basic Config:" <<endl;
  cout << "       [seed=value] [tag=string]"<< endl;
  cout << "       [start_time=value] [stop_time=value]"<< endl;
  cout << "       [trace] [etrace] [ospfv3log]"<< endl;
  cout << "       [tracetopo=value] [tracetopoonly=value]"<< endl;
  cout << "Data Traffic Config:" << endl;
  cout << "       [pktsize=value] [pktrate=value]"<< endl;
  cout << "Scenario Config: "<< endl;
  cout << "       [num_nodes=value] [radio_range=value]" << endl;
  cout << "       [alpha=value] [grid_length=value]" << endl;
  cout << "       [pause_time=value] [velocity=value]"<< endl;
  cout << "Mobility Groups (MG)s:" << endl;
  cout << "       [mobility_groups=value] [rr_[N]=value]" << endl;
  cout << "       [mg_nNodes_[N]=value]" << endl;
  cout << "       [v_max[N]=value] [v_min_[N]=value]" << endl;
  cout << "       [p_max[N]=value] [p_min_[N]=value]" << endl;
  cout << "MAC Config:" <<endl;
  cout << "       [mac_protocol=value] [alpha=value]"<< endl;
  cout << "OSPFv3 Config:" << endl;
  cout << "       [wireless_interface=value]" << endl;
  cout << "       [wireless_flooding=value]"<<endl;
  cout << "       [HelloInterval=value] "<< endl;
  cout << "       [DeadInterval=value] "<< endl;
  cout << "       [RxmtInterval=value] "<< endl;
  cout << "       [MinLSInterval=value] "<< endl;
  cout << "       [MinLSArrival=value] "<< endl;
  cout << "OR MANET Interface Config:" << endl;
  cout << "       [PushbackInterval=value]"<< endl;
  cout << "       [AckInterval=value] "<< endl;
  cout << "       [SmartPeering]" << endl; 
  cout << "       [UnsynchAdj]" << endl; 
  cout << "       [diff_hellos]" << endl; 
  cout << "OSPF-MPR MANET Interface Config:" << endl;
  cout << "       [TopoReduc]"<< endl;
  cout << "MDR MANET Interface Config:" << endl;
  cout << "       [BackupWaitInterval=value] "<< endl;
  cout << "       [AckInterval=value] "<< endl;
  cout << "       [TwoHopRefresh=value] "<< endl;
  cout << "       [HelloRepeatCount=value] "<< endl;
  cout << "       [AdjConnectivity=value] "<< endl;
  cout << "       [LSAFullness=value] "<< endl;
  cout << "       [diff_hellos]" << endl; 
  cout << "       [NonPersistentMDR] "<<endl;
  cout << "Cross Layer Interface:" << endl;
  cout << "       [cli]" << endl;
  cout << endl;
  cout << "############# Options Description ################"<<endl<<endl;
  cout << "Basic Config:" <<endl;
  cout << "     seed             : random seed, default: " << ca->seed << endl;
  cout << "     tag              : unique string to append to trace/statfile "
                << endl;
  cout << "     start_time       : statistics discarded before this time, "
                << "default: " << ca->start_time << " sec" <<  endl;
  cout << "     stop_time        : simulation stop time, default: "
                << ca->stop_time << " sec" << endl;
  cout << "     trace            : enable per-packet tracing, default: off"
                << endl;
  cout << "     etrace           : enable extended tracing, default: off"
                << endl;
  cout << "     tracetopo        : topology tracing, default:  off"
                << endl;
  cout << "     tracetopoonly    : topology tracing only, default:  off"
                << endl;
  cout << "     ospfv3log        : enable ospfv3 logging, default: off"
                << endl;
  cout << endl;
  cout << "Data Traffic Config:" << endl;
  cout << "     pktsize          : udp packet size default: "
                << ca->pktsize << " bytes" << endl;
  cout << "     pktrate          : udp pkts across whole network, "
                << "default: " << ca->pktrate << " pkts/sec" << endl;
  cout << endl;
  cout << "Scenario Config:" << endl;
  cout << "     num_nodes          : number of nodes, default: "
                << ca->nNodes << endl;
  cout << "     radio_range        : max range of radio, default: "
                << ca->radio_range << " m" << endl;
  cout << "     wireless link alpha: (0,1), default: " << ca->alpha <<endl;
  cout << "     grid_length        : length of square grid, default: "
                << ca->grid_length << " m" << endl;
  cout << "     velocity           : speed of movement, default: "
                << ca->velocity << " m/s" << endl;
  cout << "     pause_time         : duration of pause after movement, "
                << "default: " << ca->pause_time << " sec" << endl;
  cout << "Mobility Groups (MG)s:" << endl;
  cout << "     mobility_groups   : number of MGs, default: 0"
       << endl;
  cout << "     mg_nNodes_[N]       : number of nodes in MG N, default: 0."
       << endl;
  cout << "                           The number of nodes in all mobility" 
       << endl;
  cout << "                           groups must be less than or equal to"
       << endl;
  cout << "                           num_nodes." << endl;
  cout << "     rr_[N]            : maximum radio range of MG N, default: "
                << ca->radio_range << " m" << endl;
  cout << "     v_max_[N]          : maximum velocity of MG N, default: " 
                << ca->velocity << " m/s" << endl;
  cout << "     v_min_[N]          : minimum velocity of MG N, default: 0" 
                << " m/s" << endl;
  cout << "     p_max_[N]          : maximum pause time of MG N, default: " 
                << ca->pause_time << " sec" << endl;
  cout << "     p_min_[N]          : minimum pause time of MG N, default: 0" 
                << " sec" << endl;
  cout << endl;
  cout << "MAC Config:" <<endl;
  cout << "     mac_protocol : type of mac wireless protocol "<< endl;
  cout << "                    IEEE802.11=0, USAP=1, Simple_802.11=2, default: "
                << ca->mac_protocol <<endl;
  cout << endl;
  cout << "OSPFv3 Config:" << endl;
  cout << "     wireless_interface : type of interface pnum[j]=1, manet=2, "
                << "default: " << ca->w_int_type <<endl;
  cout << "     wireless_flooding  : flooding broadast=0, mpr=1, mdr=2, "
                << "default: " << ca->w_flooding <<endl;
  cout << endl;
  cout << "     HelloInterval      : ospf6 HelloInterval, default: "
                << ca->HelloInterval << " sec" << endl;
  cout << "     DeadInterval       : ospf6 DeadInterval, default: "
                << ca->DeadInterval << " sec" << endl;
  cout << "     RxmtInterval       : ospf6 RxmtInterval, default: "
                << ca->RxmtInterval << " sec" << endl;
  cout << "     MinLSInterval      : ospf6 MinLSInterval, default: "
                << ca->MinLSInterval << " sec" << endl;
  cout << "     MinLSArrival       : ospf6 MinLSArrival, default: "
                << ca->MinLSArrival << " sec" << endl;
  cout << endl;
  cout << "OR MANET Interface Config:" << endl;
  cout << "     PushbackInterval   : Overlapping Relays timer, default: "
                << ca->PushbackInterval << " msec" << endl; 
  cout << "     AckInterval        : Overlapping Relays timer, default: "
                << ca->AckInterval << " msec" << endl;
  cout << "     SmartPeering        : enable Smart Peering, "
                "default: off" <<  endl;
  cout << "     UnsynchAdj        : enable Unsynchronized Adjacencies, "
                "default: off" <<  endl;
  cout << "     diff_hellos        : enable Incremental Hellos, "
                "default: off" <<  endl;
  cout << endl;
  cout << "OSPF-MPR MANET Interface Config:" << endl;
  cout << "     TopoReduc        : enable Topology Reduction, "
                "default: off" <<  endl;
  cout << endl;
  cout << "MDR MANET Interface Config:" << endl;
  cout << "     BackupWaitInterval : MDR BackupWaitInterval, default: "
                << ca->BackupWaitInterval << " msec" << endl;
  cout << "     AckInterval        : MDR AckInterval, default: "
                << ca->AckInterval << " msec" << endl;
  cout << "     TwoHopRefresh      : MDR full Hello every TwoHopRefresh "
                "Hellos,  default: " << ca->TwoHopRefresh << endl;
  cout << "     HelloRepeatCount    : MDR HelloRepeatCount, default: "
                << ca->HelloRepeatCount << endl;
  cout << "     AdjConnectivity    : 0=Fully, 1=Uniconnected, 2=Biconnected,"
                " default: " <<  ca->AdjConnectivity << endl;
  cout << "     LSAFullness        : 0=only adjacent neighbors," << endl;
  cout << "                          1=min-hop routing, "
       <<                           "2=min-hop2 routing," << endl;
  cout << "                          2=full LSAs from MDR/MBDRs, " 
       <<                           "3=full LSAs, "
       <<                           "default: " << ca->LSAFullness << endl;
  cout << "     diff_hellos        : enable MDR Differential Hellos, "
                "default: off" <<  endl;
  cout << "     NonPersistentMDR   : disable MDR persistent mdr/mbdrs, "
                " default: off" <<  endl;
  cout << endl;
}

bool set_ospf6_config_arguments(int argc, char** argv, 
                             struct ospf6_config_args *ca)
{
 
  struct ospf6_config_args default_ca;
  int j, nNodes_total = 0;
  string mg_nNodes = "mg_nNodes_" , mg_nNodes_n;
  string rr = "rr_" , rr_n;
  string v_max = "v_max_", v_max_n;
  string v_min = "v_min_", v_min_n;
  string p_max = "p_max_", p_max_n;
  string p_min = "p_min_", p_min_n;
  string num[mobility_groups_max]={"1","2","3","4","5","6","7","8","9","10"};

  // Basic configuration
  default_ca.tag = "";    // string tag for appending output files
  default_ca.seed = 1;               // Random seed
  default_ca.start_time = 1800;   // statistics discarded for start_time sec
  default_ca.stop_time = 3600;    // seconds
  default_ca.tracetopo = 0;       // seconds
  default_ca.tracetopoonly = 0;   // seconds
  Arg("tag", ca->tag, default_ca.tag);
  Arg("seed", ca->seed, default_ca.seed);
  Arg("start_time", ca->start_time, default_ca.start_time);
  Arg("stop_time", ca->stop_time, default_ca.stop_time);
  Arg("tracetopo", ca->tracetopo, default_ca.tracetopo);
  Arg("tracetopoonly", ca->tracetopoonly, default_ca.tracetopoonly);

  default_ca.pktsize = 40;           // CBR packet size (bytes)
  default_ca.pktrate = 10;           // CBR packets per second overall
                                      //   in the simulator
  Arg("pktrate", ca->pktrate, default_ca.pktrate);
  Arg("pktsize", ca->pktsize, default_ca.pktsize);

  // Scenario configuration
  default_ca.nNodes = 20;           // Number of nodes in the topology
  default_ca.grid_length = 500.0; // meters (square grid)
  default_ca.pause_time = 40;     // random waypoint pause time (seconds)
  default_ca.velocity = 10;       // random waypoint velocity (meters/sec)
  default_ca.radio_range = 250;        // meters (usually not changed)
  default_ca.alpha = 0.5;           // wireless link grey area control
  default_ca.mobility_groups = 0;   // number of mobility groups
  Arg("num_nodes", ca->nNodes, default_ca.nNodes);
  Arg("grid_length", ca->grid_length, default_ca.grid_length);
  Arg("pause_time", ca->pause_time, default_ca.pause_time);
  Arg("velocity", ca->velocity, default_ca.velocity);
  Arg("radio_range", ca->radio_range, default_ca.radio_range);
  Arg("alpha", ca->alpha, default_ca.alpha);

  
  Arg("mobility_groups", ca->mobility_groups, default_ca.mobility_groups);
  
  for (j = 1; j <= mobility_groups_max; j++)
  {
    mg_nNodes_n = mg_nNodes + num[j-1];
    rr_n = rr + num[j-1]; 
    v_max_n = v_max + num[j-1]; 
    v_min_n = v_min + num[j-1]; 
    p_max_n = p_max + num[j-1]; 
    p_min_n = p_min + num[j-1]; 

    Arg(mg_nNodes_n.c_str(), ca->mg_nNodes[j-1], 0);
    Arg(rr_n.c_str(), ca->rr[j-1], default_ca.radio_range);
    Arg(v_max_n.c_str(), ca->v_max[j-1], default_ca.velocity);
    Arg(v_min_n.c_str(), ca->v_min[j-1], 0);
    Arg(p_max_n.c_str(), ca->p_max[j-1], default_ca.pause_time);
    Arg(p_min_n.c_str(), ca->p_min[j-1], 0);
  } 

  // MAC layer configuration
  default_ca.mac_protocol = ieee802_11; // MAC protocol type
  Arg("mac_protocol", (long int &)ca->mac_protocol, 
                      (long int)default_ca.mac_protocol);

  // OSPF configuration
  default_ca.w_int_type = 1;    // interface type:  1=pnum[j], 2=manet
  default_ca.w_flooding = 1;     //flooding 0=broadcast, 1=mpr, 2=mdr
  default_ca.HelloInterval = 10;            // hello interval (seconds)
  default_ca.DeadInterval = 4 * default_ca.HelloInterval; //dead interval (secs)
  default_ca.RxmtInterval = 5;             // retransmit interval (seconds)
  default_ca.MinLSInterval = 5;             // min LSA Interval (seconds)
  default_ca.MinLSArrival = 1;         // min LSA Arrival (seconds)
  default_ca.cost = 1;                          // ospfv3 interface cost
  default_ca.interface_jitter = 100;            // jitter of pkt sends (msec)
  default_ca.interface_flood_delay = 100;       // LSA coalescing timer (msec)
  Arg("wireless_interface", ca->w_int_type, default_ca.w_int_type);
  Arg("wireless_flooding", ca->w_flooding, default_ca.w_flooding);
  Arg("HelloInterval", ca->HelloInterval, default_ca.HelloInterval);
  Arg("DeadInterval", ca->DeadInterval, default_ca.DeadInterval);
  Arg("RxmtInterval", ca->RxmtInterval, default_ca.RxmtInterval);
  Arg("MinLSInterval", ca->MinLSInterval, default_ca.MinLSInterval);
  Arg("MinLSArrival", ca->MinLSArrival, default_ca.MinLSArrival);
  ca->DeadInterval = default_ca.DeadInterval;
  ca->cost = default_ca.cost;
  ca->interface_jitter = default_ca.interface_jitter;
  ca->interface_flood_delay = default_ca.interface_flood_delay;

  //Common to both OR and MDR
  default_ca.AckInterval = 1800;            // ack interval (msec)
  Arg("AckInterval", ca->AckInterval, default_ca.AckInterval);

  //OR specific
  default_ca.PushbackInterval = 2000;             // pushback interval (msec)
  Arg("PushbackInterval", ca->PushbackInterval, default_ca.PushbackInterval);

  //MDR specific
  default_ca.BackupWaitInterval = 2000;           // backupwait interval (msec)
  default_ca.TwoHopRefresh = 3; //TwoHopRefresh (hellos)
  default_ca.HelloRepeatCount = 3;       //HelloRepeatCount
  default_ca.AdjConnectivity = 2;     // Level of adjacencies between neighbors
  default_ca.LSAFullness = 2;         //Level of LSA Fullness
  Arg("BackupWaitInterval",ca->BackupWaitInterval,default_ca.BackupWaitInterval);
  Arg("TwoHopRefresh", ca->TwoHopRefresh, default_ca.TwoHopRefresh);
  Arg("HelloRepeatCount", ca->HelloRepeatCount, default_ca.HelloRepeatCount);
  Arg("AdjConnectivity", ca->AdjConnectivity, default_ca.AdjConnectivity);
  Arg("LSAFullness", ca->LSAFullness, default_ca.LSAFullness);

  Arg::ProcessArgs(argc, argv);
  //Specified must be done after ProcessArgs()
  ca->trace = Arg::Specified("trace") ? true : false;
  ca->etrace = Arg::Specified("etrace") ? true : false;
  ca->ospfv3log = Arg::Specified("ospfv3log") ? true : false;
  ca->diff_hellos = Arg::Specified("diff_hellos") ? true : false;
  ca->smart_peering = Arg::Specified("SmartPeering") ? true : false;
  ca->unsynch_adj = Arg::Specified("UnsynchAdj") ? true : false;
  ca->NonPersistentMDR = Arg::Specified("NonPersistentMDR") ? true : false;

  //OSPF-MPR specific
  ca->topo_reduc = Arg::Specified("TopoReduc") ? true : false;

  //Force USAGE Print out
  if (Arg::Specified("help") ||
      Arg::Specified("-help") ||
      Arg::Specified("-h") ||
      Arg::Specified("h") ||
      Arg::Specified("?"))
  {
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }

  if (ca->mobility_groups > mobility_groups_max)
  {
    cout << "Must be less than 10 mobility groups.  You entered: "
         << ca->mobility_groups << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }

  for (j = 0; j < ca->mobility_groups; j++)
  {
    nNodes_total+=ca->mg_nNodes[j];
    if (ca->v_max[j] < ca->v_min[j])
    {
      cout << endl << "Mobility group " << j+1 << " v_max " << ca->v_max[j] <<
                      "must be less than v_min" << ca->v_min;
      usage_ospf6_config_arguments(argv, &default_ca);
      return false;
    }
    else if (ca->p_max[j] < ca->p_min[j])
    {
      cout << endl << "Mobility group " << j+1 << " p_max " << ca->p_max[j] <<
                      "must be less than p_min" << ca->p_min[j];
      usage_ospf6_config_arguments(argv, &default_ca);
      return false;
    }
  }
  if (nNodes_total > ca->nNodes)
  {
    cout << endl <<"The total number of nodes defined in the mobility groups, " 
         << nNodes_total << ", must be less or equal to num_nodes: "
         << ca->nNodes << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }

  //Check a couple of input values
  std::string w_int_type[3] = {"broadcast", "point-to-multipoint", "manet"};
  std::string w_flooding[3] = {"broadcast", "mpr", "mdr"};
  std::string AdjConnectivity[3] = {"fully", "uniconnected", "biconnected"};
  std::string LSAFullness[5] = {"minimal LSAs (only adjacent neighbors)", 
                                "partial LSAs for min-hop routing", 
                                "partial LSAs for min-hop2 routing", 
                                "full LSAs from MDR/MBDRs",
                                "full LSAs (all routable neighbors)"};
  if (ca->w_int_type < 0 || ca->w_int_type > 2)
  {
    cout << "bad interface type: " << ca->w_int_type << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }
  if (ca->w_flooding < 0 || ca->w_flooding > 2)
  {
    cout << "bad flooding type: " << ca->w_flooding << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }

  if (ca->AdjConnectivity < 0 || ca->AdjConnectivity > 2)
  {
    cout << "bad value for AdjConnectivity: " << ca->AdjConnectivity << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }

  if (ca->LSAFullness < 0 || ca->LSAFullness > 4)
  {
    cout << "bad value for LSAFullness: " << ca->LSAFullness << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }

  std::string mac_protocol;
  if (ca->mac_protocol  == ieee802_11)
    mac_protocol = "IEEE802.11";
  else if (ca->mac_protocol == usap)
    if (usap_is_enabled)
      mac_protocol = "USAP";
    else
    {
      cout << "USAP???" <<endl
           << "Please enable USAP by adding USAP_ADDON flag in Makefile."<< endl
           << "Then make clean; make depend; make valid-dbg" <<endl;
      exit(0);
    }
  else if (ca->mac_protocol == simple)
    mac_protocol = "Simple 802.11";
  else
  {
    cout << "bad mac protocol type: " << ca->mac_protocol << endl;
    usage_ospf6_config_arguments(argv, &default_ca);
    return false;
  }
  cout << endl;

  //Print out the values being used
  cout << "################# Arg Values #################" << endl;
  cout << "Basic Config:" <<endl;
  cout << "     seed             : " << ca->seed << endl;
  cout << "     tag              : \"" << ca->tag <<"\"" << endl;
  cout << "     start_time       : " << ca->start_time << " sec" <<  endl;
  cout << "     stop_time        : " << ca->stop_time << " sec" << endl;
  cout << "     trace            : " << (ca->trace?"T":"F") << endl;
  cout << "     etrace           : " << (ca->etrace?"T":"F") << endl;
  cout << "     tracetopo        : " << ca->tracetopo << " sec " << endl;
  cout << "     tracetopoonly    : " << ca->tracetopoonly << " sec " << endl;
  cout << "     ospfv3log        : " << (ca->ospfv3log?"T":"F")<< endl;
  cout << "Data Traffic Config:" << endl;
  cout << "     pktsize          : " << ca->pktsize << " bytes" << endl;
  cout << "     pktrate          : " << ca->pktrate << " pkts/sec" << endl;
  cout << "Scenario Config:" << endl;
  cout << "     num_nodes          : " << ca->nNodes << endl;
  cout << "     radio_range:       : " << ca->radio_range << " m" << endl;
  cout << "     wireless link alpha: " << ca->alpha <<endl;
  cout << "     grid_length        : " << ca->grid_length << " m" << endl;
  cout << "     velocity           : " << ca->velocity << " m/s" << endl;
  cout << "     pause_time         : " << ca->pause_time << " sec" << endl;
  cout << "Mobility Groups (MG)s:" << endl;
  cout << "     mobility_groups    : " << ca->mobility_groups << endl;
  for (j = 0; j < ca->mobility_groups; j++)
  {
  cout << "      mg_nNodes_" << j+1 << "       : " << ca->mg_nNodes[j] << endl;
  cout << "       rr_" << j+1 << "             : " << ca->rr[j] << endl;
  cout << "       v_max_" << j+1 << "          : " << ca->v_max[j] << endl;
  cout << "       v_min_" << j+1 << "          : " << ca->v_min[j] << endl;
  cout << "       p_max_" << j+1 << "          : " << ca->p_max[j] << endl;
  cout << "       p_min_" << j+1 << "          : " << ca->p_min[j] << endl;
  }
  cout << "MAC Config:" <<endl;
  cout << "     mac_protocol       : " << mac_protocol <<endl;
  cout << "OSPFv3 Config:" << endl;
  cout << "     wireless_interface : " << w_int_type[ca->w_int_type] <<endl;
  cout << "     wireless_flooding  : " << w_flooding[ca->w_flooding] <<endl;
  cout << "     HelloInterval      : " << ca->HelloInterval << " sec" << endl;
  cout << "     DeadInterval       : " << ca->DeadInterval << " sec" << endl;
  cout << "     RxmtInterval       : " << ca->RxmtInterval << " sec" << endl;
  cout << "     MinLSInterval      : " << ca->MinLSInterval << " sec" << endl;
  cout << "     MinLSArrival       : " << ca->MinLSArrival << " sec" << endl;
  if (ca->w_flooding == 1) {
  cout << "OR MANET Interface Config:" << endl;
  cout << "     PushbackInterval   : " <<ca->PushbackInterval <<" msec" <<endl;
  cout << "     AckInterval        : " << ca->AckInterval << " msec" << endl;
  cout << "     SmartPeering       : " << (ca->smart_peering?"T":"F")<<  endl;
  cout << "     UnsynchAdj         : " << (ca->unsynch_adj?"T":"F")<<  endl;
  cout << "     diff_hellos        : " << (ca->diff_hellos?"T":"F")<<  endl;
  }
  if (ca->w_flooding == 1) { //For now still 1(OR) but soon will be 3(OSPF-MPR)
  cout << "OSPF-MPR MANET Interface Config:" << endl;
  cout << "     TopoReduc          : " << (ca->topo_reduc?"T":"F") <<endl;
  }
  if (ca->w_flooding == 2) {
  cout << "MDR MANET Interface Config:" << endl;
  cout << "     BackupWaitInterval : "<<ca->BackupWaitInterval<<" msec" << endl;
  cout << "     AckInterval        : " << ca->AckInterval << " msec" << endl;
  cout << "     TwoHopRefresh      : " << ca->TwoHopRefresh << " Hellos" <<endl;
  cout << "     HelloRepeatCount   : " << ca->HelloRepeatCount << endl;
  cout << "     AdjConnectivity    : " << AdjConnectivity[ca->AdjConnectivity] << endl;
  cout << "     LSAFullness        : " << LSAFullness[ca->LSAFullness] << endl;
  cout << "     diff_hellos        : " << (ca->diff_hellos?"T":"F")<<  endl;
  cout << "     NonPersistentMDR   : " << (ca->NonPersistentMDR?"T":"F")<< endl;
  }
  cout << endl;
  return true;
}
#endif //BOEING_ARGS


