/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Address Resolution Protocol Class
// Mohamed Abd Elhafez.  Georgia Tech, Spring 2004

// Implements the Address Resolution Protocol, Layer 3


#include <iostream>

#include "debug.h"
#include "arp.h"
#include "l2proto802.3.h"
#include "node.h"
#include "droppdu.h"
#include "hex.h"

using namespace std;


ARPHeader::ARPHeader() : 
  hwmacaddrtype(1),
  protoaddrtype(0x800),
  hwmacaddrsize(6),
  protoaddrsize(4),
  op(0),
  srcmacaddr(MACAddr::NONE),
  srcipaddr(),
  dstmacaddr(MACAddr::NONE),
  dstipaddr(),
  uid(0)
{
}

ARPHeader::ARPHeader(const ARPHeader& r) : 
  hwmacaddrtype(r.hwmacaddrtype),
  protoaddrtype(r.protoaddrtype),
  hwmacaddrsize(r.hwmacaddrsize),
  protoaddrsize(r.protoaddrsize),
  op(r.op),
  srcmacaddr(r.srcmacaddr),
  srcipaddr(r.srcipaddr),
  dstmacaddr(r.dstmacaddr),
  dstipaddr(r.dstipaddr),
  uid(r.uid)
{
}

ARPHeader::~ARPHeader()
{
}

Size_t ARPHeader::Size() const
{
  Count_t sz = 28;
  return sz;
}

void ARPHeader::Trace(Tfstream& tos, Bitmap_t b)
 {
 
  tos << " L3-ARP-";
  if(op == ARP::REQUEST) tos << "REQ" ;
  else if(op == ARP::REPLY) tos << "RPL";
  tos << " " << srcipaddr;
  tos << " " << dstipaddr;
  tos << " " << srcmacaddr;
  tos << " " << dstmacaddr;
  tos << " " << uid;
 
}



// IP Protocol Class Methods

ARP::ARP() 
{
  rxtimeout = DEFAULT_RX_TIMEOUT;
}


void ARP::ARPRequest(Interface* i, Packet* p, int type)
{
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
  cout << "Sending arp req for IP " << IPAddr::ToDotted(iphdr->dst) << endl;
  
  // check the arp cache for the destination IP
  ARPCache_t::iterator j = arpcache.find(iphdr->dst);
  if (j != arpcache.end()) // found then just send the packet
    {
      cout << "Entry " << IPAddr::ToDotted((*j).first)  << " expires at " <<(*j).second.lifetime << endl;

      if((*j).second.lifetime > Simulator::Now() ) // not Expired
	{
	  cout << "Entry found mac is ";
	  if ((*j).second.state == ARPEntry::RESOLVED)  // We already know the MAC address
	    {
	      //	      cout << (*j).second.mac << " forwarding pkt to iface" << endl;
	      // Send the packet to the destination mac
	      i->Send(p, (*j).second.mac, type);
	    }
	  else if  ((*j).second.state == ARPEntry::PENDING) // We havent resolved the IP address yet
	    {
	      cout << "Unresolved yet queuing packet" << endl;

	      // Queue the packet for later
	      (*j).second.pQueue.push(PacketPair_t( p, type));
	    }
	  return;
	}
     //  else  // entry is expired
// 	{
// 	  cout << "Entry " << IPAddr::ToDotted((*j).first)  << " Is expired" << endl;
// 	  // send the packet to the old address
// 	  //i->Send(p, (*j).second.mac, type);
// 	}
    }
 
  cout << "Entry not found or expired, creating new" << endl;

  // ARP entry doesnt exist or has expired
  // create cache entry
  ARPEntry e = ARPEntry(ARPEntry::PENDING, MACAddr::NONE, Simulator::Now() + timeout, 1);
  e.pQueue.push(PacketPair_t(p, type));  // Queue the packet for later
  arpcache[iphdr->dst] = e;              // store the new entry
  
  // prepare arp request packet
  Packet* pkt;
#ifdef HAVE_QT
  pkt = new ColoredPacket(Qt::green);  // green packets
#else
  pkt = new Packet();  
#endif
  
  ARPHeader* arphdr = new ARPHeader();
  arphdr->op = ARP::REQUEST;
  arphdr->dstmacaddr = MACAddr::NONE;  // The mac address is gonna be filled in the reply
  arphdr->dstipaddr = iphdr->dst;
  arphdr->srcmacaddr = i->GetMACAddr();
  arphdr->srcipaddr = i->GetIPAddr();
  arphdr->uid = pkt->uid;
  
  pkt->PushPDU(arphdr);
  if (Trace::Enabled()) i->GetNode()->TracePDU(this, arphdr);        // Add to trace file
  
  // schedule retransmition
  Simulator::instance->Schedule(new ARPEvent(ARPEvent::RETRANSMIT, pkt->Copy(), iphdr->dst, i), rxtimeout, this);

  // give packet to iface to send it
  i->Send(pkt,  MACAddr(0xffff, 0xffffffff), ARP_PACKET);  // ARP request packets are broadcast
}


void ARP::DataIndication(Interface* i,Packet* p)
{
  ARPHeader* hdr = (ARPHeader*)p->PopPDU();

  cout << "Got ARP Packet " << p->uid << " For IP " << IPAddr::ToDotted(hdr->dstipaddr) << " my IP is " <<  i->GetIPAddr() << endl;

  if (Trace::Enabled()) i->GetNode()->TracePDU(this, hdr);        // Add to trace file
  
  // is this ARP packet for my IP Address ?
  if(hdr->dstipaddr == i->GetIPAddr())
    {
      ARPCache_t::iterator j = arpcache.find(hdr->srcipaddr); // Do we already have the source entry in our cache ?
      if (j != arpcache.end()) // found
	{
	  //cout << "Entry found updating it " << endl;

	  // Uodating the ARP entry
	  (*j).second.lifetime = Simulator::Now() + timeout;
	  (*j).second.mac = hdr->srcmacaddr;  // Updating the mac address in case the IP Addresses are not static

	  if((*j).second.state == ARPEntry::PENDING)
	    {
	      (*j).second.state = ARPEntry::RESOLVED;

	      // Now that we have resolved the MAC address we send all queued packets
	      while ( !(*j).second.pQueue.empty() ) // send the queued pkts
		{
		  PacketPair_t pp = (*j).second.pQueue.front();
		  i->Send(pp.first, hdr->srcmacaddr, pp.second);
		  (*j).second.pQueue.pop(); // Remove from the queue
		}
	    }
	  (*j).second.retries = 0;
	}
      else
	{
	  // create cache entry
	  ARPEntry e = ARPEntry(ARPEntry::RESOLVED, hdr->srcmacaddr, Simulator::Now() + timeout, 0);
	  arpcache[hdr->srcipaddr] = e;
	}
      
      if(hdr->op == ARP::REQUEST)
	{
	  // prepare arp reply packet
	  Packet* pkt;
#ifdef HAVE_QT
	  
	  pkt = new ColoredPacket(Qt::blue);  // blue packets
#else
	  pkt = new Packet(); 
#endif
	  
	  ARPHeader* arphdr = new ARPHeader();
	  arphdr->op = ARP::REPLY;
	  arphdr->dstmacaddr = hdr->srcmacaddr;
	  arphdr->dstipaddr = hdr->srcipaddr;
	  arphdr->srcmacaddr = i->GetMACAddr();
	  arphdr->srcipaddr = i->GetIPAddr();
	  arphdr->uid = pkt->uid;

	  pkt->PushPDU(arphdr);
	  if (Trace::Enabled()) i->GetNode()->TracePDU(this, arphdr);        // Add to trace file
	  // give packet to iface to send it
	  i->Send(pkt, arphdr->dstmacaddr, ARP_PACKET);  // ARP reply packets are unicast
	}
    }
  else
    {
      // just drop the packet as it is not addressed to this host
      cout << "dropping packet " << p->uid << endl;
      DropPDU d("L3-NA", p);
      i->GetNode()->TracePDU(nil, &d);
      DEBUG(0,(cout << "Packet Dropped " << endl));
      Stats::pktsDropped++;
      delete p;
    }
}
  
void ARP::Handle(Event* e, Time_t t)
{
  ARPEvent* ae = (ARPEvent*)e;
  ARPCache_t::iterator j;
  switch(ae->event) {
  case ARPEvent::RETRANSMIT :
    j = arpcache.find(ae->dst);
    if (j != arpcache.end()) // found
      {
	if((*j).second.state == ARPEntry::PENDING)
	  {
	    if( ((*j).second.retries++ > MAX_RETRIES) || ((*j).second.lifetime < Simulator::Now()) )
	      {
		while ( !(*j).second.pQueue.empty() ) // delete the queued pkts
		  {
		    PacketPair_t pp = (*j).second.pQueue.front();
		    //cout << "Dropping packet " << pp.first->uid << endl;
		    DropPDU d("L3-NE", pp.first);
		    ae->iface->GetNode()->TracePDU(nil, &d);
		    DEBUG(0,(cout << "Packet Dropped " << endl));
		    Stats::pktsDropped++;
		    delete pp.first;
		    (*j).second.pQueue.pop();
		}
		// delete entry
		arpcache.erase(j);
	      }
	    else
	      {
		//cout << "Resending the arp request packet " << ae->p->uid << " at time " << t << endl;

		// Send the arp request one more time
		Simulator::instance->Schedule(new ARPEvent(ARPEvent::RETRANSMIT ,ae->p->Copy(), ae->dst, ae->iface), rxtimeout, this);
		ae->iface->Send(ae->p,  MACAddr(0xffff, 0xffffffff), ARP_PACKET);
	      }
	  }
      }
    break;
  default:
    break;
  }
  delete ae;
}



