// Circular message passing queue, pointer references
// George F. Riley, Georgia Tech, Winter 2001

// Define pointer classes, for putting data in/out of CQueues
// Behave as pointers, with ++ and * operators
// Need a separate one for input pointers and output pointers

#ifndef __CQPTR_H__
#define __CQPTR_H__
#include "cq.h"

using namespace std;

#ifdef OLD_JUST_TESTING
class CQBase {
public:
  CQBase(CQueue& cq, char* pBase)
    : m_cq(cq), m_base(pBase), m_current(cq.pFirst) { }
  CQBase(const CQBase& c)
    : m_cq(c.CQ()), m_base(c.Base()), m_current(c.Current()) { }
  CQueue& CQ() const { return m_cq;}
  size_t Current() const { return m_current;}
  void   Current(size_t c) { m_current = c;}
  char*  Base() const { return m_base;}
protected:
  CQueue& m_cq;
  char*   m_base;
  size_t  m_current;
};
#endif

//template <class T>class CQPtr : public CQBase {
template <class T>class CQPtr {
public :
  //  CQPtr(CQueue& cq, char* pBase) : CQBase(cq, pBase) { }
  CQPtr(CQueue& cq, char* pBase)
    : m_cq(cq), m_base(pBase) { }
  CQPtr(const CQPtr& c)
    : m_cq(c.CQ()), m_base(c.Base()), m_current(c.Current()) { }
  int Full(int k, int verbose)
    { // Will buffer hold k additional items?
      int r;
      size_t pOut = m_cq.pOut;
      if (m_current < pOut)
        {
          r = !((m_current+sizeof(T)*k) < pOut);
          if (r)
            { // Debug..log returning full and why
              return(r);
            }
        }
      if ((m_current + sizeof(T)*k) < m_cq.pLimit)
        return 0;//No wrap, not full
      r =  !(((m_current + sizeof(T)*k) - (m_cq.pLimit - m_cq.pFirst)) <
             pOut);
      if (r)
        {
        }
      return(r);
    }
  int Full(int k)
    { // Will buffer hold k additional items?
      return Full(k, 0);
    }
  int Full() { return Full(1);}
  int Empty() { Empty(m_cq.pIn);}
  int Empty(size_t pIn) { return m_current == pIn;}
  void Round(int m)
    {
      if (!m) return; // No alignment needed
      while((m_current - m_cq.pFirst) & m)
        {
          if(Full())
            {
              int x = Full(1,1); // Call again for verbose
              cout << "Buffer overflow on alignment x " << x << endl;
              if (x)
                exit(1);
            }
          m_base[m_current++] = 0;
          if (m_current >= m_cq.pLimit) m_current = m_cq.pFirst;
        }
    }
  CQueue& CQ() const { return m_cq;}
  size_t  Current() const { return m_current;}
  void    Current(size_t c) { m_current = c;}
  char*   Base() const { return m_base;}
  void Adjust()
    {
      if (m_current >= m_cq.pLimit)
        m_current -= (m_cq.pLimit - m_cq.pFirst);
    }
  CQPtr& operator++()
    { m_current += sizeof(T); Adjust(); return *this;}
  CQPtr& operator++(int)
    { CQPtr& tmp = *this; m_current+=sizeof(T); Adjust(); return tmp;}
  CQPtr& operator+=(int k)
    { m_current += sizeof(T)*k; Adjust(); return *this;}
  T& operator*() {
    return *((T*)(&m_base[m_current]));
  }
  operator char*() { return &m_base[m_current];} // char* typecast
  virtual void Update() = 0; // Update pIn or pOut
  virtual void Memcpy(T* s, size_t lth) = 0; // Bulk copy to/from CQueue
  void DBDump()
    {
      cout << "pFirst " << m_cq.pFirst
           << " pIn " << m_cq.pIn << " pOut " << m_cq.pOut
           << " pLimit " << m_cq.pLimit << endl;
      cout << "m_base " << m_base << "  m_current " << m_current << endl;
    }
protected:
  CQueue& m_cq;
  char*   m_base;
  size_t  m_current;
};

template <class T> class CQOutPtr : public CQPtr<T> {
public:
  CQOutPtr(CQueue& cq, char* pBase)
    : CQPtr<T>(cq, pBase)
  { CQPtr<T>::m_current = cq.pOut;}
  CQOutPtr(const CQOutPtr& c) : CQPtr<T>(c) { } // Copy constructor
public:
  void Update() { (this->CQPtr<T>::m_cq).pOut = CQPtr<T>::m_current; }
  void Memcpy(T* s, size_t lth)
    { // Copy data out of a CQueue to specified target
      if ((CQPtr<T>::m_current + lth*sizeof(T)) >= (this->CQPtr<T>::m_cq).pLimit)
        { // Copy first part
          memcpy((char*)s, (char*)&CQPtr<T>::m_base[CQPtr<T>::m_current], ((this->CQPtr<T>::m_cq).pLimit-CQPtr<T>::m_current));
          lth -= ((this->CQPtr<T>::m_cq).pLimit-CQPtr<T>::m_current)/sizeof(T);
          s += ((this->CQPtr<T>::m_cq).pLimit-CQPtr<T>::m_current)/sizeof(T);
          CQPtr<T>::m_current = (this->CQPtr<T>::m_cq).pFirst;
        }
      if (lth > 0)
        { // Copy last (only) part
          memcpy((char*)s, (char*)&CQPtr<T>::m_base[CQPtr<T>::m_current], lth*sizeof(T));
        }
      CQPtr<T>::m_current += lth*sizeof(T);
      CQPtr<T>::Adjust();
    }
};

template <class T> class CQInPtr : public CQPtr<T> {
public:
  CQInPtr(CQueue& cq, char* pBase)
    : CQPtr<T>(cq, pBase)
  { CQPtr<T>::m_current = cq.pIn;}
  CQInPtr(const CQInPtr& c) : CQPtr<T>(c) { } // Copy constructor
public:
  void Update() { (this->CQPtr<T>::m_cq.pIn) = CQPtr<T>::m_current; }
  void Memcpy(T* s, size_t lth)
    {
      if ((CQPtr<T>::m_current + lth*sizeof(T)) >= (this->CQPtr<T>::m_cq).pLimit)
        { // Copy first part
          memcpy((char*)&CQPtr<T>::m_base[CQPtr<T>::m_current], (char*)s, ((this->CQPtr<T>::m_cq).pLimit-CQPtr<T>::m_current));
          lth -= ((this->CQPtr<T>::m_cq).pLimit-CQPtr<T>::m_current)/sizeof(T);
          s += ((this->CQPtr<T>::m_cq).pLimit-CQPtr<T>::m_current)/sizeof(T);
          CQPtr<T>::m_current = (this->CQPtr<T>::m_cq.pFirst);
        }
      if (lth > 0)
        { // Copy last (only) part
          memcpy((char*)&CQPtr<T>::m_base[CQPtr<T>::m_current], (char*)s, lth*sizeof(T));
        }
      CQPtr<T>::m_current += lth*sizeof(T);
      CQPtr<T>::Adjust();
    }
};

#endif
