/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// DropTail Queue class with IPA derivatives calculation
// George F. Riley, Georgia Tech, Fall 2002

#ifndef __droptail_ipa_h__
#define __droptail_ipa_h__

#include "droptail.h"
#include "timer.h"

class IPAEvent;
class IPATimer;

//Doc:ClassXRef
class DropTailIPA : public DropTail, public Timer {
public :
  DropTailIPA();
  DropTailIPA(Count_t l);            // Specify limit
  DropTailIPA(const DropTailIPA&);   // Copy constructor
  ~DropTailIPA();                    // Destructor
  // Time functions
  virtual void Timeout(TimerEvent*); // Called when timer expires
  // DropTail functions
  bool Enque(Packet*);               // Return true if enqued, false if dropped
  Packet*  Deque();                  // Return next element (NULL if none)
  void     SetLimit(Count_t l);      // Set queue limit
  Queue*   Copy() const;             // Make a copy
  virtual void ResetStats();         // Reset the statistics to zero
  // New functions for IPA queues
  virtual  NCount_t DL1dB1();        // Get dl1db1 derivative
  virtual  Time_t   DQ1dB1();        // Get dq1db1 derivative
  virtual  Time_t   PartialDQ1();    // Get partial (from start of lbp) dQ1
  virtual  NCount_t DL2dB1();        // Get dl2db1 derivative
  virtual  Time_t   DQ2dB1();        // Get dq2db1 derivative
  virtual  void     Stage1(DropTailIPA*); // Set stage 1 peer
  virtual  void     Stage2(DropTailIPA*); // Set stage 2 peer
  virtual  Count_t  Losses();             // Return number of losses
  virtual  DCount_t Workload();           // Total buffer workload
  virtual  bool     LBP();                // True if currently in a LBPeriod
  virtual  void     LWeight(Mult_t w);    // Set loss weight
  virtual  void     QWeight(Mult_t w);    // Set queuing weight
  virtual  Mult_t   LWeight();            // Get the loss weight
  virtual  Mult_t   QWeight();            // Get the queue weight
  virtual  void     AdjustPeriod(Time_t); // Set the adjustment period
  virtual  void     AdjustStart(Time_t);  // Set the starting time for adjusts
  virtual  void     Gain(Mult_t g);       // Set the gain factor for adjusts
private:
  bool    busyPeriod; 
  bool    lossyBusyPeriod;
  Time_t  bpStart;           // Time the busy period started
  Time_t  lbpStart;          // Time the lpb started
  Count_t lossyBusyPeriods;  // Count of lossy busy periods
  Time_t  sumLossy;          // Sum of time from start of LBP to end
  bool    full;              // True if queue is full
  // The next two are used for two stage models
  DropTailIPA* stage1;       // Points to stage 1 queue from stage 2
  DropTailIPA* stage2;       // Points to stage 2 queue from stage 1
  // Statisics
  Count_t  totalLosses;      // Total packet losses
  // Information for two stage derivatives
  Count_t  tmp;              // Temporary to store for dL2/dB1
  Count_t  accumulator;      // Accumulator for dL2/dB1
  Count_t  aspCount;         // Count of active switchover points
  DCount_t dQ2;              // Secondary Q workload derivative dQ2/dB1
  Time_t   ppStart;          // Partial period start
  Count_t  ppMult;           // Partial period multiplier
  // Information used for periodic adjustments
  Mult_t   qLim;             // Floating point queue limit
  Mult_t   wL;               // Weight for loss
  Mult_t   wQ;               // Weight for queuing delay
  Time_t   lastAdjust;       // Time of last adjustment
  Time_t   adjustPeriod;     // Length of adjustment period
  Time_t   adjustStart;      // Time to start adjustments
  Mult_t   gain;             // Gain factor for adjustments
  NCount_t cumdL1;           // Cumulative dL1/dB1 at last adjustment
  Time_t   cumdQ1;           // Cumulative dQ1/dB1 at last adjustment
  NCount_t cumdL2;           // Cumulative dL2/dB1 at last adjustment
  Time_t   cumdQ2;           // Cumulative dQ2/dB1 at last adjustment
  IPATimer* adjustTimer;     // Points to currently active IPA timer
  IPAEvent* adjustEvent;     // Points to currently active IPA event
};

#endif

