/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - DropTail Queue class
// George F. Riley.  Georgia Tech, Spring 2002

#define DEBUG_MASK 0x08
#include "debug.h"
#include "droptail.h"
#include "interface.h"
#include "link.h"

using namespace std;

bool DropTail::Enque(Packet* p)
{
  DBG((Stats::pktsEnqued++));
  if (!Detailed()) UpdateSize(); // Maintain virtual size
  UpdateAverage();
  totEnq++;
  if (verbose && 0)
    {
      cout << "Enquing pkt time " << Simulator::Now()
           << " size " << size
           << " pktsize " << p->Size()
           << " pktCount " << pktCount 
           << " pktLimit " << pktLimit << endl;
    }
  DEBUG(0,(cout << "Enquing pkt time " << Simulator::Now()
           << " size " << size
           << " pktsize " << p->Size()
           << " pktCount " << pktCount 
           << " pktLimit " << pktLimit << endl));
  if (pktLimit && (pktCount == pktLimit))
    { // pktLimit non-zero limits total number of pkts allowed
      DEBUG(3,(cout << "Dropping pkt, count " << pktCount
               << " pktlimit " << pktLimit
               << " time " << Simulator::Now() 
               << endl));
      dropCount++;
      Stats::pktsDropped++;
      if (verbose)
        {
          cout << "Dropping pkt, count " << pktCount
               << " pktlimit " << pktLimit
               << " time " << Simulator::Now() 
               << endl;
        }
      return false; // Full
    }
  pktCount++; // Count number of packets in queue
  if (CheckForcedLoss() || (size + p->Size()) > Limit())
    {
      DEBUG(2,(cout << "Dropping pkt, size " << size
               << " pSize " << p->Size()
               << " limit " << Limit()
               << " time " << Simulator::Now() 
               << endl));
      dropCount++;
      pktCount--;
      Stats::pktsDropped++;
      if (verbose)
        {
          cout << "Dropping pkt, size " << size
               << " pSize " << p->Size()
               << " limit " << Limit()
               << " time " << Simulator::Now() 
               << endl;
        }
      return false; // Full
    }
  size += p->Size();                     // Note byte count added
  if (Detailed())
    {
      pkts.push_back(p);      // Save the packet if detailed
    }
  else
    { // Non-Detailed
      if (!interface) return false;      // Can't handle w/o interface
      if (!ndDeq) ndDeq = new NDDeq_t(); // Allocate the Deque
      Time_t startTx = Simulator::Now(); // Time that this pkt starts transmit
      if (!ndDeq->empty())
        { // Not empty, starts after last one finishes
          NDInfo& nd = ndDeq->back();    // Last entry in queue
          startTx = nd.stopTx;           // Time that last one finishes
        }
      Link* link = interface->GetLink();
      if (!link) return false;           // Link must exist
      Time_t txTime =  ((Rate_t)p->Size() * 8) / link->Bandwidth();
      NDInfo  newND(startTx, startTx + txTime, p->Size());
      ndDeq->push_back(newND);           // Add to the deque
    }
  DEBUG(1,(cout << " newsize " << size << endl));
  return true;                           // Not full, enqued
}

Packet* DropTail::Deque()
{ // Deque is never called for non-detailed queues
  if (!Detailed()) return nil;            // Always nil if not detailed
  DBG((Stats::pktsDequed++));
  UpdateAverage();
  if (Length() == 0) return nil;         // Nothing enqued
  Packet* p = pkts.front();              // Get head of queue
  pkts.pop_front();                      // Remove it
  size -= p->Size();                     // Note byte count removed
  pktCount--;                            // Note pkt removed
  return p;                              // Return the packet
}

Count_t DropTail::Length()
{
  if (!Detailed())UpdateSize();
  return (Count_t)size;
}

Count_t DropTail::LengthPkts()
{
  if (!Detailed())UpdateSize();
  return pktCount;
}

void   DropTail::SetInterface(Interface* i)
{
  interface = i;
}

Queue* DropTail::Copy() const
{ // Make a copy of this queue
  return new DropTail(*this);
}

void DropTail::SetLimit(Count_t l)
{
  limit = l;
}

Time_t   DropTail::QueuingDelay()
{
  if (Detailed()) return Queue::QueuingDelay();
  UpdateSize();
  // Not detailed, use ndDeq list
  Time_t now = Simulator::Now();
  // Find when last packet will finish
  if (!ndDeq) return 0;  // No pending pkts
  if (ndDeq->empty()) return 0; // No pending pkts
  NDInfo& nd = ndDeq->back();   // Newest entry
  return (nd.stopTx - now);     // Time till this one finishes
}


bool DropTail::Check(Size_t s, Packet*)
{ // Check if sufficient buffer space available
  if(!Detailed())UpdateSize();
  if (Length() == 0) return true; // Always true if empty
  if (pktLimit)
    { // Using packet limit, not bytes
      return pktCount < pktLimit;
    }
  // Queue limit in bytes
  return ((Length() + s) < Limit());
}

// Private methods
void DropTail::UpdateSize()
{ // Adjust size to account for packets transmitted since last update
  if (!ndDeq) return; // Nothing to do if no non-detailed deque
  Time_t now = Simulator::Now();
  while(!ndDeq->empty())
    {
      NDInfo& nd = ndDeq->front(); // Oldest entry
      if (nd.startTx > now) return;      // Starts in future
      if (nd.stopTx  > now)
        { // Stops in future.  This means the packet would have
          // been dequeued at this point, but the link is still
          // busy transmitting it.  We should reduce the size
          // of the queue by the size of the packet, but leave
          // this entry in the ndDeq so we know when the link is
          // free.
          if (!nd.updated)
            {
              size -= nd.size;
              pktCount--;
              nd.updated = true;
            }
          return; // done for now
        }
      // This packet has finished, remove it
      if (!nd.updated)
        {
          size -= nd.size;
          pktCount--;
        }
      ndDeq->pop_front(); // Remove from queue and continue searching
    }
}
  


