/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Dual Queue Class Definition
// George F. Riley.  Georgia Tech, Spring 2003

// The "dual" queue implements a private queue for TCP traffic
// and another queue for all others.  The two queues are scheduled
// in a work-conserving round-robin method.

//#define DEBUG_MASK 0x02
#include "debug.h"
#include "dualqueue.h"
#include "ipv4.h"
#include "tcp.h"
#include "simulator.h"

using namespace std;

// DualQueue constructors

DualQueue::DualQueue() 
  : Queue(true), limit(Queue::DefaultLength()), pktLimit(0),
    lastTCP(true), tcpEmpty(0), udpEmpty(0)
{
  pTCP = Queue::Default()->Copy();
  pNonTCP = Queue::Default()->Copy();
}

DualQueue::DualQueue(Count_t l) 
  : Queue(true), limit(l), pktLimit(0),
    lastTCP(true), tcpEmpty(0), udpEmpty(0)
{
  pTCP = Queue::Default()->Copy();
  pTCP->SetLimit(l);
  pNonTCP = Queue::Default()->Copy();
  pNonTCP->SetLimit(l);
}

// Specify the queues for tcp and non-tcp
DualQueue::DualQueue(const Queue& t, const Queue& n) : Queue(true)
{
  pTCP = t.Copy();
  pNonTCP = n.Copy();
  limit = t.GetLimit();
  pktLimit = t.GetLimitPkts();
  lastTCP = true;
}

// Copy constructor
DualQueue::DualQueue(const DualQueue& c)
    : Queue(c), limit(c.limit), pktLimit(c.pktLimit)
{
  pTCP = c.TCPQueue()->Copy();
  pNonTCP = c.NonTCPQueue()->Copy();
  lastTCP = c.LastTCP();
}

// Destructor
DualQueue::~DualQueue()
{
  if (pTCP) delete pTCP;
  if (pNonTCP) delete pNonTCP;
}

  // Inherited methods from Queue
bool DualQueue::Enque(Packet* p)
{ // Return true if enqued, false if not
  DEBUG0((cout << "DQ " << this << " enquing size " << p->Size() << endl));
  bool b;
  if (IsTCP(p))
    {
      b = pTCP->Enque(p);
      DEBUG(1,(cout << "DQ " << this 
               << " enqTCP, size " << pTCP->LengthPkts() 
               << " time " << Simulator::Now() << endl));
      if (!b) DEBUG0((cout << "DQ " << this << " dropping TCP pkt" << endl));
      return b;
    }
  // Not TCP
  b = pNonTCP->Enque(p);
  DEBUG(1,(cout << "DQ " << this 
           << " enqNonTCP, size " << pNonTCP->LengthPkts() 
           << " time " << Simulator::Now() << endl));
  if (!b) DEBUG(1,(cout << "DQ " << this << " dropping NonTCP pkt" << endl));
  return b;
}

Packet* DualQueue::Deque()
{ // Return next element (NULL if none)
  bool useTCP;
  if (lastTCP)
    { // Try to use Non-TCP next, if non-empty
      useTCP = (pNonTCP->LengthPkts() == 0); // UDP is empty
      if (useTCP) udpEmpty++;                // Count times UDP gave up slot
    }
  else
    {
      useTCP = (pTCP->LengthPkts());         // TCP is not empty
      if (!useTCP) tcpEmpty++;               // Count times TCP gave up slot
    }
  if (useTCP)
    {
      DEBUG(1,(cout << "DQ " << this << " deque TCP" 
              << " size " << pTCP->LengthPkts() 
               << " time " << Simulator::Now() << endl));
      lastTCP = true;
      return pTCP->Deque();
    }
  else
    {
      DEBUG(1,(cout << "DQ " << this << " deque NonTCP" 
              << " size " << pNonTCP->LengthPkts() 
               << " time " << Simulator::Now() << endl));
      lastTCP = false;
      return pNonTCP->Deque();
    }
}

// Most of the following should not be used for DualQueues.
// Rather, the values should be retrived for the TCP/NonTCP queues directly
Count_t DualQueue::Length()
{ // How many bytes are enqued?
  return pTCP->Length() + pNonTCP->Length();
}

Count_t DualQueue::LengthPkts()
{ // How many packets are enqued?
  return pTCP->LengthPkts() + pNonTCP->LengthPkts();
}

Queue* DualQueue::Copy() const
{ // Make a copy of this queue
  return new DualQueue(*this);
}

void DualQueue::SetLimit(Count_t l)
{ // Set new limit
  pTCP->SetLimit(l);
  pNonTCP->SetLimit(l);
}

Count_t DualQueue::GetLimit() const
{  // Get limit (bytes)
  return pTCP->GetLimit() + pNonTCP->GetLimit(); // ?? Is this right?
}

void DualQueue::SetLimitPkts(Count_t l)
{// Set packets limit
  pTCP->SetLimitPkts(l);
  pNonTCP->SetLimitPkts(l);
  return;
}

Count_t DualQueue::GetLimitPkts() const
{ // Set packets limit
  return pTCP->GetLimitPkts() + pNonTCP->GetLimitPkts(); // ?? Is this right?
}

bool DualQueue::Check(Size_t s, Packet* p)
{ // Test if buffer space available
  if (IsTCP(p)) return pTCP->Check(s, p);
  return pNonTCP->Check(s, p);
}

bool DualQueue::Detailed()
{ // True if detailed model w/ local queing
  return true;
}

void DualQueue::Detailed(bool d)
{ // Set detailed on/off
  pTCP->Detailed(d);
  pNonTCP->Detailed(d);
}

Time_t DualQueue::QueuingDelay()
{ // Calculate queuing delay at current lth
  return 0.0; // Not sure here
}

DCount_t DualQueue::TotalWorkload()
{ // Return the byte-seconds (workload)
  return pTCP->TotalWorkload() + pNonTCP->TotalWorkload();
}

void DualQueue::SetInterface(Interface* i)
{ // Let the q know which interface
  pTCP->SetInterface(i);
  pNonTCP->SetInterface(i);
}

Count_t DualQueue::DropCount()
{ // Number of dropped packets
  return pTCP->DropCount() + pNonTCP->DropCount();
}

Count_t DualQueue:: EnqueueCount()
{ // Number of enqueued packets
  return pTCP->EnqueueCount() + pNonTCP->EnqueueCount();
}

void DualQueue::CountEnq(Packet* p)
{ // Count an enque
  if (IsTCP(p)) return pTCP->CountEnq(p);
  return pNonTCP->CountEnq(p);
}

void DualQueue::ResetStats()
{ // Reset the statistics to zero
  pTCP->ResetStats();
  pNonTCP->ResetStats();
  tcpEmpty = 0;
  udpEmpty = 0;
}

// Methods specific to DualQueue
void DualQueue::TCPQueue(const Queue& q)
{
  if (pTCP) delete pTCP;
  pTCP = q.Copy();
}

void DualQueue::NonTCPQueue(const Queue& q)
{
  if (pNonTCP) delete pNonTCP;
  pNonTCP = q.Copy();
}

// Private Methods
bool   DualQueue::IsTCP(Packet* p)
{
  if (!p) return false; // Can't tell if no packet
  // We need to inspect the PDU's to see this is TCP packet or not
  Count_t offset = 0;
  PDU* pdu;
  while((pdu = p->PeekPDU(offset++)))
    {
      Layer_t l = pdu->Layer();
      switch(l) {
        case 0 : // Unknown
          break;
        case 3 : // IP
          {
            Count_t v = pdu->Version(); // Get version
            switch(v) {
              case 4 :
                {
                  // Process IPV4
                  IPV4Header* h = (IPV4Header*)pdu;
                  return h->protocol == 6;
                }
                break;
              case 6 :
                // Process IPV6, Not implemented yet
                return false; // Need to code this
                break;
            }
          }
          break;
        case 4 : // TCP or UDP
          {
            L4PDU* l4 = (L4PDU*)pdu;
            return l4->Proto() == 6;
          }
          break;
        }
    }
  // Can't find, assume not TCP
  return false;
}
