/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Dumbbell topology
// George F. Riley.  Georgia Tech, Fall 2002

// Define the dumbbell topology using point-to-point links

#include <math.h>

#include "dumbbell.h"
#include "node.h"
#include "linkp2p.h"
#include "mask.h"

using namespace std;

// Constructors

Dumbbell::Dumbbell(Count_t l, Count_t r, Mult_t m, SystemId_t id)
{
  ConstructorHelper(l, r, m, IPADDR_NONE, IPADDR_NONE, Linkp2p::Default(), id);
}

Dumbbell::Dumbbell(Count_t l, Count_t r, Mult_t m,
                   const Linkp2p& link, SystemId_t id)
{
  ConstructorHelper(l, r, m, IPADDR_NONE, IPADDR_NONE, link, id);
}

Dumbbell::Dumbbell(Count_t l, Count_t r, Mult_t m,
                   IPAddr_t lip, IPAddr_t rip, SystemId_t id)
{
  ConstructorHelper(l, r, m, lip, rip, Linkp2p::Default(), id);
}

Dumbbell::Dumbbell(Count_t l, Count_t r, Mult_t m,
                   IPAddr_t lip, IPAddr_t rip,
                   const Linkp2p& link, SystemId_t id)
{
  ConstructorHelper(l, r, m, lip, rip, link, id);
}

// Access functions
Node* Dumbbell::Left(Count_t c)
{ // Get specified left side leaf node
  if (c >= leftCount) return nil;
  return Node::GetNode(first + c + 2);
}

Node* Dumbbell::Right(Count_t c)
{ // Get Specified right side leaf node
  if (c >= rightCount) return nil;
  return Node::GetNode(first + leftCount + c + 2); // Allow for the two routers
}

Node* Dumbbell::Left()
{ // Get left side router
  return Node::GetNode(first);    // Left router is first
}

Node* Dumbbell::Right()
{ // Get right side router
  return Node::GetNode(first + 1); // Right router is second
}

Linkp2p* Dumbbell::LeftLink(Count_t c)
{ // Get link from left leaf to left rtr
  Node* n = Left(c);
  if (!n) return nil;
  return (Linkp2p*)n->GetLink(Left());
}

Linkp2p* Dumbbell::RightLink(Count_t c)
{ // Get link from right leaf to right rtr
  Node* n = Right(c);
  if (!n) return nil;
  return (Linkp2p*)n->GetLink(Right());
}

Linkp2p* Dumbbell::LeftLink()
{ // Get link from left rtr to right rtr
  return (Linkp2p*)Left()->GetLink(Right());
}

Linkp2p* Dumbbell::RightLink()
{// Get link from right rtr to left rtr
  return (Linkp2p*)Right()->GetLink(Left());

}

Queue*   Dumbbell::LeftQueue(Count_t c)
{ // Get queue from left leaf to left rtr
  Node* n = Left(c);
  if (!n) return nil;
  return n->GetQueue(Left());
}

Queue*   Dumbbell::RightQueue(Count_t c)
{ // Get queue from right leaf to right rtr
  Node* n = Right(c);
  if (!n) return nil;
  return n->GetQueue(Right());
}

Queue*   Dumbbell::LeftQueue()
{ // Get queue from left rtr to right rtr
  return Left()->GetQueue(Right());
}

Queue*   Dumbbell::RightQueue()
{ // Get queue from left rtr to right rtr
  return Right()->GetQueue(Left());
}

void     Dumbbell::BoundingBox(const Location& ll, const Location& ur)
{
  Meters_t xDist = ur.X() - ll.X();
  Meters_t yDist = ur.Y() - ll.Y();
  Meters_t xAdder = xDist / 3.0;
  //Meters_t yAdderL = yDist / (LeftCount() + 1.0);
  //Meters_t yAdderR = yDist / (RightCount() + 1.0);
  Angle_t  thetaL = M_PI / (LeftCount() + 1.0);
  Angle_t  thetaR = M_PI / (RightCount() + 1.0);

  // Notify simulator of bounds, for max x/y computations
  Simulator::instance->NewLocation(ll.X(), ll.Y());
  Simulator::instance->NewLocation(ur.X(), ur.Y());

  // Place the left router
  Location lrl(ll.X() + xAdder, ll.Y() + yDist/2.0);
  Node* ln = Left();
  ln->SetLocation(lrl);
  
  // Place the right router
  Location rrl(ll.X() + xAdder * 2.0, ll.Y() + yDist/2.0);
  Node* rn = Right();
  rn->SetLocation(rrl);

  // Place the left nodes
  Angle_t theta = -M_PI_2 + thetaL;
  for (Count_t l = 0; l < LeftCount(); ++l)
    {
      // Make them in a circular pattern to make all line lengths the same
      // Special case when theta = 0, to be sure we get a straight line
      if ((LeftCount() % 2) == 1)
        { // Count is odd, see if we are in middle
          if (l == (LeftCount() / 2))
            {
              theta = 0.0;
            }
        }
      Node* n = Left(l);
      Location lnl(lrl.X() - cos(theta) * xAdder,
                   lrl.Y() + sin(theta) * xAdder);
      // Insure did not exceed bounding box
      if (lnl.Y() < ll.Y()) lnl.Y(ll.Y());
      if (lnl.Y() > ur.Y()) lnl.Y(ur.Y());
      n->SetLocation(lnl);
      theta += thetaL;
    }
  // Place the right nodes
  theta = -M_PI_2 + thetaR;
  for (Count_t r = 0; r < RightCount(); ++r)
    {
      // Special case when theta = 0, to be sure we get a straight line
      if ((RightCount() % 2) == 1)
        { // Count is odd, see if we are in middle
          if (r == (RightCount() / 2))
            {
              theta = 0.0;
            }
        }
      Node* n = Right(r);
      Location rnl(rrl.X() + cos(theta) * xAdder,
                   rrl.Y() + sin(theta) * xAdder);
      // Insure did not exceed bounding box
      if (rnl.Y() < ll.Y()) rnl.Y(ll.Y());
      if (rnl.Y() > ur.Y()) rnl.Y(ur.Y());
      n->SetLocation(rnl);
      theta += thetaR;
    }
}

// Private methods
void Dumbbell::ConstructorHelper(Count_t left, Count_t right,
                                 Mult_t m, 
                                 IPAddr_t leftIP, IPAddr_t rightIP,
                                 const Linkp2p& link, SystemId_t id)
{
  leftCount = left;
  rightCount = right;
  first = Node::nextId;
  // First left and right routers, then left side, then right side
  Node* lr = new Node(id);
  Node* rr = new Node(id);
  DEBUG0((cout << "DB LR id " << lr->Id() << endl));
  DEBUG0((cout << "DB RR id " << rr->Id() << endl));

  // Connect the routers, with reduced bandwidth
  DEBUG0((cout << "Constructing Dumbbell bottleneck, link bw "
          << link.Bandwidth() << endl));
  Linkp2p routerLink(link);
  routerLink.Bandwidth(link.Bandwidth() * m);
  DEBUG0((cout << "Router Link bw " << routerLink.Bandwidth() << endl));
  lr->AddDuplexLink(rr, routerLink);
  
  // Set default routes
  lr->DefaultRoute(rr);
  rr->DefaultRoute(lr);
  
  for (Count_t i = 0; i < left; ++i)
    {
      Node* ln = new Node(id);
      DEBUG0((cout << "DB LN id " << ln->Id() << endl));
      IPAddr_t ip = IPADDR_NONE;
      if  (leftIP != IPADDR_NONE)
        { // Assign specified ip addresses
          ip = leftIP;
        }
      ln->AddDuplexLink(lr, link, ip);
      // Add a default route
      ln->DefaultRoute(lr);
      if (leftIP != IPADDR_NONE) leftIP ++; // Advance to next ip
    }
  for (Count_t i = 0; i < right; ++i)
    {
      Node* rn = new Node(id);
      DEBUG0((cout << "DB RN id " << rn->Id() << endl));
      IPAddr_t ip = IPADDR_NONE;
      if  (rightIP != IPADDR_NONE)
        { // Assign specified ip addresses
          ip = rightIP;
        }
      rn->AddDuplexLink(rr, link, ip);
      // Add a default route
      rn->DefaultRoute(rr);
      if (rightIP != IPADDR_NONE) rightIP ++; // Advance to next ip
    }
  last = Node::nextId;
}

