/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Dumbbell topology
// George F. Riley.  Georgia Tech, Fall 2002

// Define the dumbbell topology using point-to-point links

#ifndef __dumbbell_h__
#define __dumbbell_h__

#include "common-defs.h"
#include "node.h"

class Node;
class Linkp2p;
class Queue;

//Doc:ClassXRef
class Dumbbell {
public:
  //Doc:Class Class {\tt Dumbbell} creates a topology model with the familiar
  //Doc:Class {\tt dumbbell} configuration.  Some number of nodes on the
  //Doc:Class left and right side of the topology communicate  through
  //Doc:Class a single bottleneck link.

  //Doc:Method
  Dumbbell(Count_t, Count_t, Mult_t, SystemId_t = 0);
    //Doc:Desc Constructor for dumbbell.
    //Doc:Desc  The default point--to--point link
    //Doc:Desc (with the default bandwidth and delay) is used for the leaf node
    //Doc:Desc links.
    //Doc:Desc No \IPA{s} are assigned to any links.
    //Doc:Arg1 Count of nodes on the left side of the topology.
    //Doc:Arg2 Count of nodes on the right side of the topology.
    //Doc:Arg3 Multiplier for the link bandwidth of the bottleneck link.
    //Doc:Arg3 For example, a multiplier  of 0.1 indicates the bandwidth
    //Doc:Arg3 of the bottleneck is one tenth of the bandwidth of the
    //Doc:Arg3 leaf node links.
    //Doc:Arg4 Responsible system ID for distributed simulation

  //Doc:Method
  Dumbbell(Count_t, Count_t, Mult_t, const Linkp2p&, SystemId_t = 0);
    //Doc:Desc Constructor for dumbbell, specifying a point--to--point
    //Doc:Desc link to use as a template for leaf node links.
    //Doc:Desc No \IPA{s} are assigned to any links.
    //Doc:Arg1 Count of nodes on the left side of the topology.
    //Doc:Arg2 Count of nodes on the right side of the topology.
    //Doc:Arg3 Multiplier for the link bandwidth of the bottleneck link.
    //Doc:Arg4 Link object to copy for leaf node lnks.
    //Doc:Arg5 Responsible system ID for distributed simulation

  //Doc:Method
  Dumbbell(Count_t, Count_t, Mult_t, 
           IPAddr_t leftIP, IPAddr_t rightIP, SystemId_t = 0);
    //Doc:Desc Constructor for dumbbell.
    //Doc:Desc  The default point--to--point link
    //Doc:Desc (with the default bandwidth and delay) is used for the leaf node
    //Doc:Desc links.
    //Doc:Desc \IPA{s} are assigned to leaf nodes based
    //Doc:Desc on the arguments specified.
    //Doc:Arg1 Count of nodes on the left side of the topology.
    //Doc:Arg2 Count of nodes on the right side of the topology.
    //Doc:Arg3 Multiplier for the link bandwidth of the bottleneck link.
    //Doc:Arg4 Starting \IPA\ for left side leaf nodes.
    //Doc:Arg4 The \IPA\ is incremented
    //Doc:Arg4 by one  for each created left side leaf.
    //Doc:Arg5 Starting \IPA\ for right side leaf nodes.
    //Doc:Arg5 The \IPA\ is incremented
    //Doc:Arg5 by one  for each created right side leaf.
    //Doc:Arg6 Responsible system ID for distributed simulation

  //Doc:Method
  Dumbbell(Count_t, Count_t, Mult_t, 
           IPAddr_t leftIP, IPAddr_t rightIP,         // Specify ip addresses
           const Linkp2p&, SystemId_t = 0); // Specify link for leafs
    //Doc:Desc Constructor for dumbbell, specifying a point--to--point
    //Doc:Desc link to use as a template for leaf node links, and
    //Doc:Desc \IPA{s} for  leaf nodes.
    //Doc:Arg1 Count of nodes on the left side of the topology.
    //Doc:Arg2 Count of nodes on the right side of the topology.
    //Doc:Arg3 Multiplier for the link bandwidth of the bottleneck link.
    //Doc:Arg4 Starting \IPA\ for left side leaf nodes.
    //Doc:Arg4 The \IPA\ is incremented
    //Doc:Arg4 by one  for each created left side leaf.
    //Doc:Arg5 Starting \IPA\ for right side leaf nodes.
    //Doc:Arg5 The \IPA\ is incremented
    //Doc:Arg5 by one  for each created right side leaf.
    //Doc:Arg6 Link object to copy for leaf node lnks.
    //Doc:Arg7 Responsible system ID for distributed simulation

  //Doc:Method
  Node* Left(Count_t);                // Get specified left side leaf node
    //Doc:Desc Get a pointer to a left side leaf node.
    //Doc:Arg1 Index for left side node to return.
    //Doc:Return Pointer to the specified left side leaf node.

  //Doc:Method
  Node* Right(Count_t);               // Get Specified right side leaf node
    //Doc:Desc Get a pointer to a right side leaf node.
    //Doc:Arg1 Index for right side node to return.
    //Doc:Return Pointer to the specified right side leaf node.

  //Doc:Method
  Node* Left();                       // Get left side router
    //Doc:Desc Get a pointer to the left side bottleneck router.
    //Doc:Return Pointer to left side bottleneck router.

  //Doc:Method
  Node* Right();                      // Get right side router
    //Doc:Desc Get a pointer to the right side bottleneck router.
    //Doc:Return Pointer to right side bottleneck router.

  //Doc:Method
  Linkp2p* LeftLink(Count_t);         // Get link from left leaf to left rtr
    //Doc:Desc Get a pointer to the link from the specified leaf node to 
    //Doc:Desc bottleneck router.
    //Doc:Arg1 Index for left side node.
    //Doc:Return Pointer to specified left side leaf to router link.

  //Doc:Method
  Linkp2p* RightLink(Count_t);        // Get link from right leaf to right rtr
    //Doc:Desc Get a pointer to the link from the specified left side leaf node
    //Doc:Desc to bottleneck router.
    //Doc:Arg1 Index for right side node.
    //Doc:Return Pointer to specified right side leaf to router link.

  //Doc:Method
  Linkp2p* LeftLink();                // Get link from left rtr to right rtr
    //Doc:Desc Get a pointer to the link from the left side router
    //Doc:Desc to right side router.
    //Doc:Return Pointer to specified router to router link.

  //Doc:Method
  Linkp2p* RightLink();               // Get link from right rtr to left rtr
    //Doc:Desc Get a pointer to the link from the right side router
    //Doc:Desc to left side router.
    //Doc:Return Pointer to specified router to router link.

  //Doc:Method
  Queue*   LeftQueue(Count_t);        // Get queue from left leaf to left rtr
    //Doc:Desc Get a pointer to the queue from the specified left leaf node to 
    //Doc:Desc bottleneck router.
    //Doc:Arg1 Index for left side node.
    //Doc:Return Pointer to specified left side leaf to router queue.

  //Doc:Method
  Queue*   RightQueue(Count_t);       // Get queue from right leaf to right rtr
    //Doc:Desc Get a pointer to the queue from the specified right leaf node 
    //Doc:Desc to bottleneck router.
    //Doc:Arg1 Index for right side node.
    //Doc:Return Pointer to specified left side leaf to router queue.

  //Doc:Method
  Queue*   LeftQueue();               // Get queue from left rtr to right rtr
    //Doc:Desc Get a pointer to the queue from the left side router
    //Doc:Desc to right side router.
    //Doc:Return Pointer to specified router to router queue.

  //Doc:Method
  Queue*   RightQueue();              // Get queue from right rtr to left rtr
    //Doc:Desc Get a pointer to the queue from the right side router
    //Doc:Desc to left side router.
    //Doc:Return Pointer to specified router to router queue.

  //Doc:Method
  Count_t  LeftCount() { return leftCount;}
    //Doc:Desc Return the count of the  number left side leaf nodes.
    //Doc:Desc     //Doc:Arg1 Count of left side leaf nodes.

  //Doc:Method
  Count_t  RightCount(){ return rightCount;}
    //Doc:Desc Return the count of the  number right side leaf nodes.
    //Doc:Desc     //Doc:Arg1 Count of right side leaf nodes.

  //Doc:Method
  void BoundingBox(const Location&, const Location&);
    //Doc:Desc Define a bounding box for this dumbbell object. 
    //Doc:Desc The bounding box
    //Doc:Desc is used to assign node locations, which are used during
    //Doc:Desc the simulation animation.
    //Doc:Arg1 Lower left corner of bounding box.
    //Doc:Arg2 Upper right corner of bounding box.

public:
  NodeId_t   first;      // There is a global node list, we don't need one 
  NodeId_t   last;       // Just log first/last+1 entries
  Count_t    leftCount;  // Number of left side nodes
  Count_t    rightCount; // Number of right side nodes
private:
  void ConstructorHelper(Count_t, Count_t, Mult_t, 
                         IPAddr_t, IPAddr_t,
                         const Linkp2p&, SystemId_t);
};

#endif


