/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Duplex Link class
// George F. Riley.  Georgia Tech, Spring 2002

#include <iostream>

#include "debug.h"
#include "duplexlink.h"
#include "l2proto802.3.h"
#include "node.h"
#include "interface.h"
#include "link-ghost.h"
#include "link-rti.h"

using namespace std;

// Constructors
DuplexLink::DuplexLink(Node* ln, Node* rn)
  : local (nil), remote(nil)
{
  AddSimplexLinks(ln, IPADDR_NONE, MASK_ALL, Link::Default(),
                  rn, IPADDR_NONE, MASK_ALL, Link::Default());
}

DuplexLink::DuplexLink(Node* ln, const Linkp2p& ll,
                       Node* rn, const Linkp2p& rl)
  : local (nil), remote(nil)
{ // Local/Remote nodes/links
  AddSimplexLinks(ln, IPADDR_NONE, MASK_ALL, ll,
                  rn, IPADDR_NONE, MASK_ALL, rl);
}

DuplexLink::DuplexLink(Interface* li,
                       Interface* ri)
  : local ((Linkp2p*)Link::Default().Copy()),
    remote((Linkp2p*)Link::Default().Copy())
{ // Note..this does not work for ghosts..need to fix
  li->SetLink(local);
  ri->SetLink(remote);
  local->SetPeer(ri);
  remote->SetPeer(li);
}

DuplexLink::DuplexLink(Interface* li, const Linkp2p& ll,
                       Interface* ri, const Linkp2p& rl)
  : local ((Linkp2p*)ll.Copy()),
    remote((Linkp2p*)rl.Copy())
{ // Note..this does not work for ghosts..need to fix
  li->SetLink(local);
  ri->SetLink(remote);
  local->SetPeer(ri);
  remote->SetPeer(li);
}

DuplexLink::DuplexLink(Node* ln, IPAddr_t la, Mask_t lm,
                       Node* rn, IPAddr_t ra, Mask_t rm)
  : local(nil), remote(nil)
{
  AddSimplexLinks(ln, la, lm, Link::Default(),
                  rn, ra, rm, Link::Default());
}

DuplexLink::DuplexLink(Node* ln, IPAddr_t la, Mask_t lm,  const Linkp2p& ll,
                       Node* rn, IPAddr_t ra, Mask_t rm,  const Linkp2p& rl)
  : local(nil), remote(nil)
{
  AddSimplexLinks(ln, la, lm, ll, rn, ra, rm, rl);
}

DuplexLink::~DuplexLink()
{ // Does NOT delete the member pointers. The link and interface
  // objects are owned by the associated nodes and deleted when
  // they are deleted.
}

// Member functions
void DuplexLink::AddSimplexLinks(
     Node* ln, IPAddr_t la, Mask_t lm, const Link& ll,
     Node* rn, IPAddr_t ra, Mask_t rm, const Link& rl)
{
  localif  = ln->AddInterface(L2Proto802_3(), la, lm);
  remoteif = rn->AddInterface(L2Proto802_3(), ra, rm);
  bool localReal = ln->IsReal();
  bool remoteReal = rn->IsReal();
  if (localReal && remoteReal)
    { // Normal case
      local = ll.Copy();
      remote = rl.Copy();
    }
  if (!localReal)
    { // Local node is ghost, use ghost link
      local = new LinkGhost();
      DEBUG0((cout << "Using Ghost from localNode " << ln->Id()
              << " to node " << rn->Id() << endl));
    }
  
  if (!remoteReal)
    { // Remote node is ghost, use ghost link
      remote = new LinkGhost();
      DEBUG0((cout << "Using Ghost from remoteNode " << rn->Id()
              << " to node " << ln->Id() << endl));
    }
  
  if (localReal && !remoteReal)
    { // Need RTI Link
      Rate_t bw = ll.Bandwidth();
      Time_t delay = ll.Delay();
      local = new LinkRTI(la, lm, localif, bw, delay );
      DEBUG0((cout << "Using RTI from localNode " << ln->Id()
              << " to node " << rn->Id() << endl));
    }
      
  if (remoteReal && !localReal)
    { // Need RTI Link
      Rate_t bw = rl.Bandwidth();
      Time_t delay = rl.Delay();
      remote = new LinkRTI(ra, rm, remoteif, bw, delay);
      DEBUG0((cout << "Using RTI from remoteNode " << rn->Id()
              << " to node " << ln->Id() << endl));
    }
      
  localif->SetLink(local);
  remoteif->SetLink(remote);
  local->SetPeer(remoteif);
  remote->SetPeer(localif);
  DEBUG0((cout << "DuplexLink::AddSimplexLinks, node " << ln->Id()
          << " if " << localif << endl));
  DEBUG0((cout << "DuplexLink::AddSimplexLinks, node " << rn->Id()
          << " if " << remoteif << endl));
}


