/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Class for EIGRP routing application
// George F. Riley, Georgia Tech, Spring 2004 (and Kunal Punjabi)

#include <iostream>

#define  DEBUG_MASK 0x01
#include "debug.h"
#include "eigrp.h"

using namespace std;

// EIGRP Header Methods
EIGRPHeader::EIGRPHeader()
    : msgSeq(0), mTxt(EMsgUnknown), mSrc(IPADDR_NONE), mDst(IPADDR_NONE),
        mType(MsgUnknown)
{
}

EIGRPHeader::EIGRPHeader(EMsg_t t)
    : msgSeq(0), mTxt(EMsgUnknown), mSrc(IPADDR_NONE), mDst(IPADDR_NONE),
        mType(t)
{
}

#ifdef DONT_NEED
EIGRPHeader::EIGRPHeader(const EIGRPHeader& c)
    : msgSeq(c.msgSeq), mTxt(c.mTxt), mSrc(c.mSrc), mDst(c.mDst),
        mType(c.mType)
{
}
#endif

EIGRPHeader::~EIGRPHeader()
{
}

PDU* EIGRPHeader::Copy()
{
  return new EIGRPHeader(*this);
}

Size_t  EIGRPHeader::SSize()
{ // Size needed for serialization
  return 16; // 4 items, 4 bytes each
}

char*   EIGRPHeader::Serialize(char*, Size_t&)    // Serialize to a buffer
{
  return nil; // code later
}

char*   EIGRPHeader::Construct(char*, Size_t&)    // Construct from buffer
{
  return nil; // code later
}


void    EIGRPHeader::Trace(Tfstream&, Bitmap_t)   // Trace contents of this pdu
{
  return; // code later
}

// Retx info methods

ReTxInfo::ReTxInfo()
{
}

ReTxInfo::~ReTxInfo()
{
  delete e; // Don't need the copy of the header anymore
  if (retxEvent)
    { // Need to cancel the timer
      Simulator::instance->Cancel(retxEvent);
      delete retxEvent;
      retxEvent = nil;
    }
}

// Neighbor Methods
void Neighbor::Dump()
{
}

void Neighbor::AddReTx(EIGRPHeader* e)
{
  DEBUG0((cout << "EIGRP " << myEIGRP->myNode->Id()
          << " neighbor " << (string)IPAddr(nbIP)
          << " AddReTx, seq " << e->msgSeq << endl));
  
  ReTxMap_t::iterator i = reTx.find(e->msgSeq);
  if (i != reTx.end())
    {
      cout << "Huh?  Removing duplicate retx msg seq in EIGRP" << endl;
      delete i->second;
      reTx.erase(i);
    }
  EIGRPEvent* ev = new EIGRPEvent(EIGRPEvent::RETX);
  ev->n = this;        // Neighbor
  ev->s = e->msgSeq;   // Sequence
  
  ReTxInfo*   retx = new ReTxInfo();
  retx->e = (EIGRPHeader*)e->Copy(); // Save a copy for later re-use
  retx->retxCount = 0; // Count number of retries
  retx->retxEvent = ev;// Save the event pointer for deletion
  // Insert the retx info  inthe queue
  DEBUG0((cout << "EIGRP " << myEIGRP->myNode->Id()
          << " neighbor " << (string)IPAddr(nbIP)
          << " adding retx seq " << e->msgSeq << endl));
  reTx.insert(ReTxMap_t::value_type(e->msgSeq, retx));
  // Schedule the retx event
  Simulator::instance->Schedule(ev, myEIGRP->GetReTxTimeout(), myEIGRP);
}

void Neighbor::RemoveReTx(MsgSeq_t s)
{ // An ack has been received, remove the retx entry for this
  DEBUG0((cout << "EIGRP " << myEIGRP->myNode->Id() 
          << " removing retx seq " << s << endl));
  ReTxMap_t::iterator i = reTx.find(s);
  if (i == reTx.end())
    { // Not found, just ignore (this in fact can happen during startup)
      return;
    }
  delete i->second; // This deletes the l4 header, and cancels the retx event
  reTx.erase(i);
}

ReTxInfo* Neighbor::GetReTx(MsgSeq_t s)
{ // Retrieve the retx info for the specified seq number
  ReTxMap_t::iterator i = reTx.find(s);
  if (i == reTx.end())
    {
      cout << "Huh?  Can't find retx msg seq in EIGRP::retx" << endl;
      return nil;
    }
  return i->second;
#ifdef MOVE_THIS
  // At this point, the retx timer has been deleted, so we need a new one
  EIGRPEvent* ev = new EIGRPEvent(EIGRPEvent::RETX);
  ev->n = this;        // Neighbor
  ev->s = s;           // Sequence
  
  ReTxInfo* retx = i->second;
  retx->retxEvent = ev;
#endif  
}

// EIGRP Methods  
EIGRP::EIGRP()
    : nextif(nil), reTxTimeout(1.0), helloTimeout(5.0), 
      status(DEAD), pHelloEvent(nil), nextSeq(0)
{
}

EIGRP::~EIGRP()
{
}

void EIGRP::AttachNode(Node* n)
{
  if (!n->IsReal()) return; // Ghost node in distributed simulation. do nothing
  myNode = n;
  n->InsertProto(4, EIGRP_PROTO, this); // Insert into protocol graph
}

// Handler Methods
void EIGRP::Handle(Event* ev, Time_t)
{
  EIGRPEvent* e = (EIGRPEvent*)ev;
  ReTxInfo* retx;
  
  switch(e->event) {
    case EIGRPEvent::START:
      cout << "EIGRP " << myNode->Id() << " handling START event " << endl;
      Alive();    // Set to alive
      SendHello();// Send the hellos
      break;

    case EIGRPEvent::STOP:
      cout << "EIGRP " << myNode->Id() << " handling STOP event " << endl;
      Dead();     // Set to dead, which stops responding to messages
      break;
      
    case EIGRPEvent::HELLO:
      cout << "EIGRP " << myNode->Id() << " handling HELLO event " << endl;
      pHelloEvent = nil;   // We are going to delete this one
      SendHello();         // Send the hello message
      break;

    case EIGRPEvent::RETX:
      cout << "EIGRP " << myNode->Id() << " handling RETX event " << endl;
      retx = e->n->GetReTx(e->s); // Get the re-tx info
      if (!retx)
        {
          cout << "HuH?  Can't find retx info on RETX event" << endl;
        }
      else
        { 
          if (++retx->retxCount < 3)
            {
              // Schedule the retx event
              Simulator::instance->Schedule(ev, reTxTimeout, this);
              ev = nil; // Since we re-used the event, we don't delete below
              // And send the new packet
              Packet* p = new Packet();
              p->PushPDU(retx->e->Copy());
              IPV4ReqInfo r(retx->e->mDst, IPADDR_NONE, 1, 88);
              IPV4::Instance()->DataRequest(myNode, p, &r);
              DEBUG0((cout << "EIGRP " << myNode->Id()
                      << " retransmitting seq "
                      << e->s << endl));
            }
          else
            { // retransmitted too many times, just give up
              retx->retxEvent = nil; // Deleting this event below, not here
              e->n->RemoveReTx(e->s);
              e->n->Down();  // And set the neighbor down
            }
        }
      break;
  }
  delete ev;
}

  
  
#ifdef SIM
void EIGRP::DataIndication(Node* n, Packet* p, IPAddr_t ip,
                           L3Protocol& l3, Interface* i, IPAddr_t dst)
#else
void EIGRP::DataIndication(Node* n, Packet* p, IPAddr_t ip,
                           L3Protocol& l3, Interface* i)
#endif //SIM
{ // Got a packet from L3
  DEBUG0((cout << "EIGRP " << myNode->Id() << " DataInd " << endl));
  if (lossRNG)
    {
      Random_t l = lossRNG->Value();
      if (l < lossProb)
        { // Randomly loose packet for debugging reliable Unicast
          cout << "EIGRP " << myNode->Id() 
               << " randomly loosing packet" << endl;
          delete p;
          return;
        }
    }
  
          
  EIGRPHeader* e = (EIGRPHeader*)p->PeekPDU();
  if (e->Layer() != 4 || e->Proto() != EIGRP_PROTO)
    {
      cout << "HuH?  EIGRP receive, not EIGRP header" << endl;
      return;
    }
   
  if (Status() == ALIVE)
    { // Ignore all incoming msgs if not alive
      Neighbor*  nb = GetNeighbor(ip, i);// Find or create neighbor
      nb->status = Neighbor::UP;   // Set status up if we get a message
      EIGRPHeader eigrp;
      switch(e->mTxt) {
        case  EMsgReliable :
          {
            // Need to reply
            EIGRPHeader* e1 = new EIGRPHeader();
            Packet* p1 = new Packet();
            e1->msgSeq = e->msgSeq;
            e1->mTxt = EMsgAck;
            e1->mSrc = nb->myIP;
            e1->mDst = ip;
            p1->PushPDU(e1);
            // Send the ack
            IPV4ReqInfo r(ip, IPADDR_NONE, 1, 88); // Info for l3
            IPV4::Instance()->DataRequest(myNode, p1, &r);
          }
          break;
        case EMsgAck :
          // Got an ack, remove the retx info
          nb->RemoveReTx(e->msgSeq);
          break;
        default:
          break;
      }
    }
  delete p; // Delete the packet
}

#ifdef CODE_THIS_LATER
void EIGRP::retransmit(ReTxEvent* e)
{
  DEBUG0((cout << "Got retx timeout time " << Simulator::Now()
          << " for seq " << e->m_seq
          << " to " << (string)IPAddr(e->m_ip) << endl;
  Neighbor& n = getNeighbor(e->m_ip);
  if (e->m_retx == RETX_LIMIT)
    { // Too many retries, declare this neighbor down and exit
      n.Down();
      if(1)printf("Too Many Retries, declaring neighbor %d/%d down\n",
                  e->m_np.first, e->m_np.second);
      return;
    }
  MsgEventMap_t::iterator i = n.m_sent.find(e->m_seq);
  if (i == n.m_sent.end())
    { // ?
      printf("HuH? EIGRP::retransmit can't find msg %ld\n",
             e->m_seq);
      exit(1);
    }
  if(0)printf("ReTxTimeout, copying msg %f\n", Now());
  RouteMsg msg = i->second.first;
  if(0)printf("ReTxTimeout erasing %f\n", Now());
  n.m_sent.erase(i);
  uCast(msg, EMsgReliable, e->m_ip, n, e); // Resend
  if(0)printf("Exiting ReTxTimeout %f\n", Now());
}
#endif

#ifdef DONT_NEED
void EIGRP::Timeout(TimerEvent*)
{
  DEBUG0((cout << "EIGRP timeout" << endl));
}
#endif

#ifdef CODE_THIS_LATER
int EIGRP::command(int argc, const char*const* argv)
{
	// printf("BASE COMMAND\n");
  if (argc == 2)
    {
      if (strcmp(argv[1], "alive") == 0)
        { // Set status to alive
					Alive();
					return TCL_OK;
				}
      if (strcmp(argv[1], "dead") == 0)
        { // Set status to dead
					Dead();
					return TCL_OK;
				}
		}
	return Agent::command(argc, argv);
}
#endif

void EIGRP::McastUnreliable(EIGRPHeader* p)
{
  mCast(p, EMsgUnreliable);
}

void EIGRP::McastReliable(EIGRPHeader* p)
{
  mCast(p, EMsgReliable);
}

void EIGRP::UcastUnreliable(EIGRPHeader* p, IPAddr_t ip)
{
  //uCast(p, EMsgUnreliable);
}

void EIGRP::UcastReliable(EIGRPHeader* p, IPAddr_t)
{
  //uCast(p, EMsgReliable);
}

void EIGRP::StartAt(Time_t t)
{
  if (!myNode)
    {
      cout << "Attempt to start EIGRP with no local node" << endl;
      return;
    }
  
  if (!myNode->IsReal()) return; // No action for ghost nodes
  // Schedule the start event
  EIGRPEvent* e = new EIGRPEvent(EIGRPEvent::START);
  cout << "Starting eigrp " << myNode->Id() << " at time " << t << endl;
  Simulator::instance->Schedule(e, t - Simulator::Now(), this);
}

void EIGRP::StopAt(Time_t t)
{
  if (!myNode)
    {
      cout << "Attempt to stop EIGRP with no local node" << endl;
      return;
    }
  
  if (!myNode->IsReal()) return; // No action for ghost nodes
  // Schedule the stop event
  EIGRPEvent* e = new EIGRPEvent(EIGRPEvent::STOP);
  DEBUG0((cout << "Stopping eigrp " << myNode->Id()
          << " at time " << t << endl));
  Simulator::instance->Schedule(e, t - Simulator::Now(), this);
}

void EIGRP::SetLossProbability(Random_t lp)
{
  if (lossRNG) delete lossRNG;
  lossProb = lp;
  lossRNG = new Uniform(); // [0..1)
}


#ifdef CODE_THIS_LATER
void  EIGRP::recvMessage(RouteMsg& m, IPAddr_t ip, 
                         NPPair_t f, IF_t i)  // msg recv
{ // Should be overridden in subclass
	printf("EIGRP %s got msg on if %d from %d/%d\n",
				 name(), i, f.first, f.second);
}
#endif

Neighbor* EIGRP::GetNeighbor(IPAddr_t ip, Interface* i)
{
  NeighborMap_t::iterator n = neighbors.find(ip);
  if (n != neighbors.end())
    { // Found it.  Make it UP (in case it's presently down)
      n->second->Up();
      return n->second; // Return it
    }
  
  // Does not exist, create one
  Neighbor* nb = new Neighbor();
  nb->myIP = i->GetIPAddr(); // IPAddress from received interface
  nb->nbIP = ip;             // Neighbor's IP
  nb->iface = i;             // Interface to use to talk to this neighbor
  nb->myEIGRP = this;        // Owning EIGRP
  neighbors.insert(NeighborMap_t::value_type(ip, nb));
  DEBUG0((cout << "EIGRP " << myNode->Id()
          << " added nb " << (string)IPAddr(ip) << endl));
  return nb;
}


void EIGRP::mCast(EIGRPHeader* e, EMCast_t t)
{ // Common for rel/unreliable
  e->mTxt = t; // Set message transmission type in EIGRP Header
  e->msgSeq = nextSeq++; // Set sequence number
  // Need to save the packet for re-tx if reliable desired
  if (t == EMsgReliable)
    { // Save a copy on every neighbor
      for (NeighborMap_t::iterator i = neighbors.begin();
           i != neighbors.end(); ++i)
        { // Save this info for later retransmission
          Neighbor* n = i->second;
          if (n->IsDown()) continue; // Not if neighbor is down
          n->AddReTx(e);
        }
    }
  // We use broadcast with ttl of 1, since we don't have
  // multicast at the moment.
  Packet* p = new Packet();
  p->PushPDU(e);
  IPV4ReqInfo r(IPAddrBroadcast, IPADDR_NONE, 1, 88); // Info for l3
  IPV4::Instance()->DataRequest(myNode, p, &r);
}

   
void EIGRP::uCast(EIGRPHeader* e, EMCast_t t, IPAddr_t i, Neighbor* n)
{
#ifdef CODE_THIS_LATER
  if (Status() == ALIVE)
    { // Send only if alive
      if(0)printf("uCast time %f\n", Now());
      Packet* p    = allocpkt();
      hdr_ip*  ip  = hdr_ip::access(p);
      hdr_eigrp* e = hdr_eigrp::access(p);
      hdr_cmn* c   = hdr_cmn::access(p);

      // kunal
      /* for colors*/
      ip->flowid()=fid;
      /* for colors*/
      // kunal

      /* Talal, comment it out */
      //ip->daddr()  = n.m_nodeport.first;
			
      /* Talal */
      // use the Hier IP address instead of the IP address	
      NeighborHierIPMap_t::iterator u = m_neighborsHierIP.find(to);
      ip->daddr() = u->second;
      /* Talal */

      ip->dport()  = n.m_nodeport.second;
      c->size()    = m.Size();
      if (pEv)
        { // Use seq from re-tx event
          e->seq() = pEv->m_seq;
        }
      else
        {
          e->seq() = n.m_seq++;
        }
      e->type()    = t;
      e->src()     = n.m_myip;

      /* Talal, comment it out */
      //e->dst()     = to;
      /* Talal */

      /* Talal */
      // use the Hier IP address instead of the IP address	
      NeighborHierIPMap_t::iterator v = m_neighborsHierIP.find(to);
      e->dst() = v->second;
      /* Talal */

      if (t == EMsgReliable)
        { // Save a copy for possible re-tx
          // First see if already in
          MsgEventMap_t::iterator i = n.m_sent.find(e->seq());
          if (i != n.m_sent.end())
            {
              if(0)printf("Erasing duplicate from m_sent %f\n", Now());
              n.m_sent.erase(i); 
            }
          if (pEv)
            {
              pEv->m_retx++; // Increment re-tx counter
            }
          else
            {
              pEv = new ReTxEvent(to, n.m_nodeport, e->seq());
            }
          if(0)printf("Inserting msg in sent queue %f\n", Now());
          n.m_sent.insert(MsgSeqEventPair_t(e->seq(),
                                            MsgEventPair_t(m, pEv)));
          // Schedule a re-tx event.
          if(0)printf("Scheduling retx event\n");
          m_pReTxTimer->Schedule(m_timeout, pEv);
        }
      if(0)printf("EIGRP Sending, target is %p daddr %d\n",
                  target_, ip->daddr());
      // Pass the message to the peer to be used when the datagram arrives
      n.m_peer->acceptMessage(m, n.m_myip, m_myNP, e->seq());
#ifdef DEBUG_DROP
      // Enable this to test behavior when packets are lost
      double pr = drand48(); // this is just debug test test re-tx
      if (pr < 0.2)
        {
          Packet::free(p);   // debug
        }
      else
        {
          send(p, NULL);
        }
#else
      send(p, NULL);
#endif
    }
#endif
}

void EIGRP::uCast(EIGRPHeader* e, EMCast_t t, IPAddr_t to)
{ // Called when neighbor not known
  Neighbor* n = GetNeighbor(to, nil);
  uCast(e, t, to, n);
}

void EIGRP::SendHello()
{ // Send the hello messages using unreliable multicast
  HelloHeader* h = new HelloHeader();
  //McastUnreliable(h);
  McastReliable(h); // For testing, use reliable hellos
  if (pHelloEvent)
    { // Cancel any pending
      Simulator::instance->Cancel(pHelloEvent);
      delete pHelloEvent;
      pHelloEvent = nil;
    }
  // Schedule another hello at 'timeout' seconds in future
  pHelloEvent = new EIGRPEvent(EIGRPEvent::HELLO);
  Simulator::instance->Schedule(pHelloEvent, helloTimeout, this);
}

  
#ifdef CODE_LATER
// ReTxTimer methods
void ReTxTimer::Schedule(double delay, ReTxEvent*)
{
  DEBUG0((cout << "ReTxTimer::Schedule" << endl));
#ifdef CODE_LATER
  Scheduler::instance().schedule(this, e, delay);
#endif
}

void ReTxTimer::Cancel(ReTxEvent*)
{
#ifdef CODE_lATER
  Scheduler::instance().cancel(e);
#endif
}
#endif

IPAddr_t EIGRP::GetIPAddress(Interface* interface)
{ 
#ifdef DO_WE_NEED_THIS
  for (NeighborMap_t::iterator i = m_neighbors.begin();
       i != m_neighbors.end();  ++i)
    {
      if(i->second.m_if==interface) 
        {
          //	printf("found the interface- IPAddress Pair\n");
          return(i->second.m_myip);
        }
    }
#endif
  return 0;
}



void EIGRP::DisplayNeighborTable()
{
  for (NeighborMap_t::iterator i = neighbors.begin(); i != neighbors.end();  ++i)
    {
      cout << "Neighbors, if " << i->second->iface
           << " IP " << (string)IPAddr(i->second->myIP) << endl;
    }
}

