/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Class for EIGRP routing protocol
// George F. Riley, Georgia Tech, Spring 2004 (and Kunal Punjabi)


#ifndef __EIGRP_H__
#define __EIGRP_H__


#include <vector>
#include <map>
#include <string>

#include "pdu.h"
#include "l4protocol.h"
#include "protocol.h"
#include "l4demux.h"
#include "handler.h"
#include "rng.h"

#define EIGRP_PROTO 88

class Interface;
typedef unsigned long MsgSeq_t; // Message sequence number
typedef double        Metric_t; // For route metrics
typedef enum { // Enumerat the multicast transmission types
    EMsgUnknown,
    EMsgReliable,
    EMsgUnreliable,
    EMsgAck
} EMCast_t;
typedef enum { // Enumerate the EIGRP Message Types
    MsgUnknown,
    MsgHello 
} EMsg_t;
    

// Define the packet header info for eigrp.
// This is the basic information common to all EIGRP messages.
// This is sub-classed by the more specific EIGRP messages (eg. Hello, etc).
// Since EIGRP is conceptually a Layer 4 protocol, we subclass L4PDU.
class EIGRPHeader : public L4PDU {
public:
  EIGRPHeader();
  EIGRPHeader(EMsg_t);
  //EIGRPHeader(const EIGRPHeader&);
  virtual ~EIGRPHeader();
  Size_t  Size() const { return 16;}
  Layer_t Layer() { return 4;}
  Proto_t Proto() { return EIGRP_PROTO;}
  PDU*    Copy();  
  // Serialization
  Size_t  SSize();                     // Size needed for serialization
  char*   Serialize(char*, Size_t&);   // Serialize to a buffer
  char*   Construct(char*, Size_t&);   // Construct from buffer
  // Packet tracing
  void    Trace(Tfstream&, Bitmap_t);  // Trace the contents of this pdu
public:
  MsgSeq_t        msgSeq;// Sequence number
  EMCast_t        mTxt;  // Message transmission type
  IPAddr_t        mSrc;  // Source IP address
  IPAddr_t        mDst;  // Dest IP address
  EMsg_t          mType; // Type of message  
};

// Subclass EIGRPHeader for each of the message types
class HelloHeader : public EIGRPHeader 
{
public:
  HelloHeader() : EIGRPHeader(MsgHello) {}
};

//typedef std::map<MsgSeq_t, EIGRPHeader*>   MsgMap_t; // Used for message re-tx
#ifdef ADD_BACK_IN_AS_NEEDED
// Define a map by msg-sequence to message and event
class ReTxEvent;
typedef std::pair<RouteMsg, ReTxEvent*>    MsgEventPair_t;
typedef std::map<MsgSeq_t, RouteMsg>       MsgMap_t;
typedef MsgMap_t::value_type               MsgSeqPair_t;
typedef std::map<MsgSeq_t, MsgEventPair_t> MsgEventMap_t;
typedef MsgEventMap_t::value_type          MsgSeqEventPair_t;
class EIGRPBaseAgent;
#endif

class EIGRPEvent;
class ReTxInfo {
  // Information stored in each neighbor for re-transmission
public:
  ReTxInfo();
  ~ReTxInfo();
public:
  EIGRPHeader* e;
  Count_t      retxCount; // Number of re-tx's
  EIGRPEvent*  retxEvent; // Pending re-tx event
};

typedef std::map<MsgSeq_t, ReTxInfo*> ReTxMap_t;

class EIGRP;
class Neighbor {
public:
  typedef enum { UP, DOWN } Status_t;
  Neighbor() : myIP(IPADDR_NONE), nbIP(IPADDR_NONE), iface(nil),
      seq(0), status(UP), metric(100000), myEIGRP(nil) {}
  
  Status_t Status() { return status;}
  void Up()    { status = UP;}
  bool IsUp()  { return status == UP;}
  void Down()  { status = DOWN;}
  bool IsDown(){ return status == DOWN;}
  void Dump();        // Debug
  void AddReTx(EIGRPHeader*);    // Add an entry in the re-tx queue
  void RemoveReTx(MsgSeq_t);     // Remove an entry in re-tx queue (ack rx)
  ReTxInfo* GetReTx(MsgSeq_t);   // Get ReTx info for the specified sequence
public:
  IPAddr_t      myIP;    // Src IP Address used when sending to this neighbor
  IPAddr_t      nbIP;    // Dst IP of this neighbor
  Interface*    iface;   // Interface used to communicate with this neighbor
  MsgSeq_t      seq;     // Next seq. to send to this peer
  Status_t      status;  // Status of this neighbor
  Metric_t      metric;  // Metric for routing to this neighbor
  ReTxMap_t     reTx;    // Retransmit queue for reliable transfers
  EIGRP*        myEIGRP; // Points to owning EIGRP
};

// Map of neighbors, indexed by IP Address
typedef std::map<IPAddr_t, Neighbor*> NeighborMap_t;
typedef NeighborMap_t::value_type     NeighborPair_t;

class EIGRPEvent : public Event {
public:
  typedef enum { START, STOP, HELLO, RETX } EIGRPEvent_t;
  EIGRPEvent() { }
  EIGRPEvent(Event_t ev):Event(ev){}
  virtual ~EIGRPEvent() { }
public:
  Neighbor* n;  // Which neighbor (if retx)
  MsgSeq_t  s;  // Which sequence (if retx)
};


// Definition of EIGRP protocol class.
// EIGRP is conceptually a layer 4 protocol, but since its not
// designed to support layer 5 applications, we don't subclass L4Protocol
// here, since most of the L4Protocol api is designed to support
// applications.
class EIGRP : public Handler, public L4Demux {
public:
  typedef enum { ALIVE, DEAD } Status_t;
  typedef enum { RETX_LIMIT = 16 };
  EIGRP();
  virtual ~EIGRP();
  void  AttachNode(Node*);  // Attach this protocol to a node
  // Handler methods
  void Handle(Event*, Time_t);
  
  // Upcalls from L3 protocol, via L4Demux
#ifdef SIM
  void DataIndication(Node*, Packet*, IPAddr_t, L3Protocol&,Interface*, IPAddr_t = 0);
#else
  void DataIndication(Node*, Packet*, IPAddr_t, L3Protocol&,Interface*);
#endif //SIM
private:
  void McastUnreliable(EIGRPHeader*);
  void McastReliable(EIGRPHeader*);
  void UcastUnreliable(EIGRPHeader* , IPAddr_t);
  void UcastReliable(EIGRPHeader*, IPAddr_t);
public:
  Status_t  Status() { return status;}
  void      Alive()  { status = ALIVE;}
  bool      IsAlive(){ return status == ALIVE;}
  void      Dead()   { status = DEAD;}
  bool      IsDead() { return status == DEAD;}
  void      StartAt(Time_t);
  void      StopAt(Time_t);
  void      SetHelloInterval(Time_t t) { helloTimeout = t;}
  void      SetReTxInterval(Time_t t) { reTxTimeout = t;}
  Time_t    GetReTxTimeout() { return reTxTimeout; }
  void      SetLossProbability(Random_t);
private:
  Neighbor* GetNeighbor(IPAddr_t ip, Interface*);
  IPAddr_t  GetIPAddress(Interface*);
  void      DisplayNeighborTable(); // debug

private:
  void      mCast(EIGRPHeader*, EMCast_t); // Common for rel/unrel
  // Two versions of uCast, one if neighbor known (avoids lookup)
  void      uCast(EIGRPHeader*, EMCast_t, IPAddr_t, Neighbor*);
  void      uCast(EIGRPHeader*, EMCast_t, IPAddr_t);
  void      SendHello();
private:
  NeighborMap_t neighbors;
  Interface*    nextif;      // What is this?
  Time_t        reTxTimeout; // Retransmit time period (seconds)
  Time_t        helloTimeout;// Hello timeout period
  Status_t      status;
  EIGRPEvent*   pHelloEvent; // Non-nil if a pending Hello event
  MsgSeq_t      nextSeq;     // Next sequence to use for messages
public:
  Node*         myNode;      // Attached node
  Random*       lossRNG;     // For debugging
  Random_t      lossProb;    // For debugging
 private:
  static IPAddr_t m_nextip;  // Used to make up IP addresses if not known
};

#endif

