/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Global stat control
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __estat_h__
#define __estat_h__

#include <fstream>
#include <vector>
#include <string>

#include "common-defs.h"

typedef std::vector<int>  vNodeIds;

class Node;
class Protocol;
class PDU;
class Packet;
class Interface;

typedef enum {
  UDP_SENT = 0,
  UDP_BYTE_SENT,
  UDP_FWD,
  UDP_BYTE_FWD,
  UDP_RECV,
  UDP_BYTE_RECV,
  POWER,
  POWER_COUNT,
  CONTENTION_RATIO,
  CONTENTION_RATIO_COUNT,

  STAT_LENGTH
}stats;

class EStat {
public:
  typedef enum { ENABLED, DISABLED, DEFAULT } EStatStatus;
  typedef std::vector<EStatStatus> ETSVec_t;
public:
  //Doc:Method
  EStat();
    //Doc:Desc Default constructor, no arguments.

  ~EStat();

  //Doc:Method
  void LayerOn();               // Enable all stats
    //Doc:Desc Enable tracing for all protocol layers.

  //Doc:Method
  void LayerOn(Layer_t);        // Enable for specific layer
    //Doc:Desc Enable tracing for a specified protocol layer.
    //Doc:Arg1 Layer to enable (2, 3, or 4).

  //Doc:Method
  void LayerOff();              // Disable all stats
    //Doc:Desc Enable tracing for all protocol layers.

  //Doc:Method
  void LayerOff(Layer_t);       // Disable for specific layer
    //Doc:Desc Disable tracing for a specified protocol layer.
    //Doc:Arg1 Layer to disable (2, 3, or 4).

  //Doc:Method
  bool IsLayerOn(Layer_t);      // Enquire stat status (on/off) by layer
    //Doc:Desc Test whether tracing is on or off by layer.
    //Doc:Arg1 Layer to test.
    //Doc:Return True if tracing enabled for this layer.
 
  //Doc:Method
  bool IsEnabled();        // Enquire stat file opened
    //Doc:Desc Check if stat file globally enabled.
    //Doc:Return True if enabled.

	//Doc:Method
void NodeOn();            //Enables tracing on all nodes
	//Doc:Desc Enable tracing for all nodes.

	//Doc:Method
void NodeOn(int id); // Enable for specific node
	//Doc:Desc Enable tracing for a specified node.
	//Doc:Arg1 Node to enable (1, 2, 3, 4,...).

	//Doc:Method
void NodeOff(int id);// Disable for specific node
  //Doc:Desc Disable tracing for a specified node.
  //Doc:Arg1 Node to Disable (1, 2, 3, 4,...).
		
 	//Doc:Method
  bool IsNodeOn(int id);      // Enquire stat status (on/off) by layer
    //Doc:Desc Test whether tracing is on or off by layer.
    //Doc:Arg1 Layer to test.
    //Doc:Return True if tracing enabled for this layer.
	
	std::ofstream& GetOS() { return os;} // The stat file object
	void EStatPacket(Packet*, Node*, Layer_t , char *str = "?");     
	void EStatIPV4(PDU *pdup, std::string str);

	void SetExtended(bool e) {extended = e;}
	bool IsExtended() {return extended;}
	
  //Doc:Method
  static EStat* Instance();// Get the global stat object
    //Doc:Desc Get a pointer to the single, global stat object.
    //Doc:Return Pointer to stat object.

  //Doc:Method
  static bool   Enabled() { return enabled;} // True if tracing enabled
    //Doc:Desc Test if tracing is enabled.
    //Doc:Return True if tracing enabled.
	
	void EStatPrint();
	void CollectEStats(const char *name, Time_t time);

 void UpdateAveragePower(double power);
 void UpdateContentionRatio(double ratio);

private:
  //Doc:Method
  bool Open(const char*);  // Open the stat file
    //Doc:Desc Create a stat file.
    //Doc:Return     //Doc:Arg1 File name.
    //Doc:Return True if successfully opened.

  //Doc:Method
  void Close();            // Close the stat file
    //Doc:Desc Close the stat file.  Should be called when the simulation completes.
	//
  static EStat* instance;  // The single global stat object
  static bool   enabled;   // Global enabled flag
  static bool   extended;  // extended extended tracing flag
  static int stat_event_level;
  ETSVec_t  stat;          // On/Off status by layer
  std::ofstream os;             // File for logging
  Node*    node;           // Current node
  vNodeIds vnodeids;
  std::string statfile;
  double statistics[STAT_LENGTH];
  Time_t start_stat_time;
};

#endif

