/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ethernet LAN class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the Ethernet LAN Class

#ifndef __ethernet_h__
#define __ethernet_h__

#include "common-defs.h"
#include "link-real.h"
#include "ratetimeparse.h"
#include "l4protocol.h"
#include "interface-ethernet.h"

// The Ethernet model has two detail levels:
// PARTIAL does model collisions and insures only one node is transmitting
//      at once.  However, collisions are detected instantaneously,
//      (all nodes can hear the carrier with zero delay).  This is
//      more realistic than the NONE detail level, at the expense of
//      slight computational overhead.
// FULL models collisions and time delay between nodes on the LAN.
//      When using this model, the nodes MUST specify an X/Y position
//      (meters) to allow for propogation delay calculations.

typedef std::vector<Node*>  NodeVec_t; 

//Doc:ClassXRef
class Ethernet : public LinkReal {
  //Doc:Class Class {\tt Ethernet} defines the Ethernet Link object. It is
  //Doc:Class derived from the {\tt Link} object. The ethernet object
  //Doc:Class creates an ethernet with a user-defined number of nodes.
  
public:
  typedef enum { NONE, PARTIAL, FULL } Detail_t;
  // Constructors
  // This one specifies how many nodes
  //Doc:Method
  Ethernet(Count_t,               // Node count
           IPAddr_t, Mask_t,      // IP addr start, mask
           SystemId_t = 0,
           Rate_t r = Rate("10Mb"), 
           Time_t dly = Time("1us"),
           Detail_t d = PARTIAL);
  //Doc:Desc The constructor for the {\tt Ethernet} object that creates an
  //Doc:Desc an ethernet, with a given number of nodes and the associated
  //Doc:Desc interfaces. This constructor assumes that we create a subnet
  //Doc:Desc without any default gateway.
  //Doc:Arg1 The number of nodes in this ethernet
  //Doc:Arg2 The beginning IP address of this ethernet
  //Doc:Arg3 The subnet mask to be used for this ethernet.
  //Doc:Arg4 Responsible system ID for distributed simulation
  //Doc:Arg5 The Transmission Rate of the link in bps. (may use
  //Doc:Arg5 ratetimeparse for simplicity)
  //Doc:Arg6 The link delay in secs. (may use ratetimeparse for simplicity)
  //Doc:Arg7 A detail type whether PARTIAL, FULL or NONE

  //Doc:Method
  Ethernet(Count_t,               // Node count
           IPAddr_t, Mask_t,      // IP addr start, mask
           const L4Protocol& l4,  // Layer 4 protocol to bind to each node 
           SystemId_t = 0,
           Rate_t r = Rate("10Mb"), 
           Time_t dly = Time("1us"),
           Detail_t d = PARTIAL);
  //Doc:Desc The constructor for the {\tt Ethernet} object that creates an
  //Doc:Desc an ethernet, with a given number of nodes and the associated
  //Doc:Desc interfaces. This constructor assumes that we create a subnet
  //Doc:Desc without any default gateway.  This one also binds
  //Doc:Desc   the specified l4 protocol to each node (presumably a TCP Server)
  //Doc:Arg1 The number of nodes in this ethernet
  //Doc:Arg2 The beginning IP address of this ethernet
  //Doc:Arg3 The subnet mask to be used for this ethernet.
  //Doc:Arg4 The layer 4 protocol to bind to each node
  //Doc:Arg5 Responsible system ID for distributed simulation
  //Doc:Arg6 The Transmission Rate of the link in bps. (may use
  //Doc:Arg6 ratetimeparse for simplicity)
  //Doc:Arg7 The link delay in secs. (may use ratetimeparse for simplicity)
  //Doc:Arg8 A detail type whether PARTIAL, FULL or NONE

  // This one specifies a single "gateway" node for packet forwarding
  // This is quite memory efficient, as all nodes other than the
  // gateway just use a single default route to the gateway
  // The gateway node is assumed to already exist

  //Doc:Method
  Ethernet(Count_t,  Node*,  // Node count, gateway node
           IPAddr_t, Mask_t, // IP addr start, mask
           SystemId_t = 0,
           Rate_t r = Rate("10Mb"),
           Time_t dly = Time("1us"),
           Detail_t d = PARTIAL);
  //Doc:Desc This constructor creates an {\tt Ethernet} object that has a
  //Doc:Desc gateway node and all of the rest of the nodes use this as a
  //Doc:Desc default route. The gateway node is assumed to already exist.
  //Doc:Arg1 The number of nodes in this ethernet
  //Doc:Arg2 Pointer to the gateway node object
  //Doc:Arg3 The beginning IP address of this ethernet
  //Doc:Arg4 The subnet mask to be used for this ethernet.
  //Doc:Arg5 Responsible system ID for distributed simulation
  //Doc:Arg6 The Transmission Rate of the link in bps. (may use
  //Doc:Arg6 ratetimeparse for simplicity)
  //Doc:Arg7 The link delay in secs. (may use ratetimeparse for simplicity)
  //Doc:Arg8 A detail type whether PARTIAL, FULL or NONE
  
  //Doc:Method
  Ethernet(Count_t,  Node*,        // Node count, gateway node
           IPAddr_t, Mask_t,       // IP addr start, mask
           const L4Protocol& l4,   // Layer 4 protocol to bind to each node 
           SystemId_t = 0,
           Rate_t r = Rate("10Mb"),
           Time_t dly = Time("1us"),
           Detail_t d = PARTIAL);
  //Doc:Desc This constructor creates an {\tt Ethernet} object that has a
  //Doc:Desc gateway node and all of the rest of the nodes use this as a
  //Doc:Desc default route. The gateway node is assumed to already exist.
  //Doc:Desc Also specifies a layer 4 protocol object to bind at each node.
  //Doc:Arg1 The number of nodes in this ethernet
  //Doc:Arg2 Pointer to the gateway node object
  //Doc:Arg3 The beginning IP address of this ethernet
  //Doc:Arg4 The subnet mask to be used for this ethernet.
  //Doc:Arg5 The layer 4 protocol to bind to each node
  //Doc:Arg6 Responsible system ID for distributed simulation
  //Doc:Arg7 The Transmission Rate of the link in bps. (may use
  //Doc:Arg7 ratetimeparse for simplicity)
  //Doc:Arg8 The link delay in secs. (may use ratetimeparse for simplicity)
  //Doc:Arg9 A detail type whether PARTIAL, FULL or NONE
  
  ~Ethernet();      // Destructor
  // Link methods
  //Doc:Method
  virtual Link*  Copy() const;               // All subclasses must have Copy
  //Doc:Desc This is a copy constructor

  //Doc:Method
  virtual Interface* GetPeer(Packet*);       // Get peer interface
  //Doc:Desc This method returns the destination interface by looking
  //Doc:Desc at the destination MAC address of the packet.
  //Doc:Arg1 The packet to be transmitted.
  //Doc:Return The destination interface.

  //Doc:Method
  virtual Interface* GetPeer(Count_t);   // Get peer interface for peer index
  //Doc:Desc Get the interface pointer for the remote endpoint of
  //Doc:Desc this link for the specified peer index
  //Doc:Arg1 Which peer desired (0 <- i < NeighborCount())
  //Doc:Return A pointer to the receiving interface for the specified peer

  virtual void  SetPeer(Interface*) ;

  //Doc:Method
  virtual bool   Transmit(Packet*, Interface*, Node*); // Transmit a packet
  //Doc:Desc Transmit a packet.
  //Doc:Arg1 The packet to transmit.
  //Doc:Arg2 A pointer to the local (transmitting) interface
  //Doc:Arg3 A pointer to the local (transmitting) node.
  //Doc:Return True if the transmit succeeded, false if packet was dropped.
  virtual  bool Transmit(Packet*, Interface*, Node*, Rate_t);

  //Doc:Method
  Count_t  Size();                      // Return number of nodes
  //Doc:Desc Returns the number of nodes on this link
  //Doc:Return number of nodes
  
  //Doc:Method
  Count_t  PeerCount();                      // Return number of peers
  //Doc:Desc Returns the number of peers on this link
  //Doc:Return number of peers

  //Doc:Method
  IPAddr_t PeerIP(Count_t);                  // Get peer IP Address (if known)
  //Doc:Desc Since each node on this link is also associated with a count
  //Doc:Desc this returns the IP address of the n'th  peer
  //Doc:Arg1 the count index n
  //Doc:Return the corresopnding IPAddress

  //Doc:Method
  IPAddr_t NodePeerIP(Node*);                // Get peer IP Address (if known)
  //Doc:Desc Return the IP address of the node if it is on this link and
  //Doc:Desc is found
  //Doc:Arg1 The pointer to the node object
  //Doc:Return The IP Address

  //Doc:Method
  bool     NodeIsPeer(Node*);                // Find out if node is a peer
  //Doc:Desc Find out if node is a peer
  //Doc:Arg1 The Node
  //Doc:Return boolean result of the test

  //Doc:Method
  Count_t  NodePeerIndex(Node*);             // Get peer index for node
  //Doc:Desc Returns the index of the node in this link
  //Doc:Arg1 The pointer to the node
  //Doc:Return The index of the peer

  //Doc:Method
  IPAddr_t DefaultPeer(Interface*);          // Get IP Addr of default peer
  //Doc:Desc Find the default peer's IP Address, this will find the default
  //Doc:Desc gateway's IP Address in the same subnet
  //Doc:Arg1 The interface whose default peerIP is to be known
  //Doc:Return The default peer's IP
  
  //Doc:Method 
  Count_t  NeighborCount(Node*);             // Number of routing neighbors
  //Doc:Desc Returns the number of neighbors on this link which are routers
  //Doc:Desc technically, count of routing peers
  //Doc:Arg1 Node querying neighbors (significant for Ethernet links)
  //Doc:Return The number of routing peers.

  // Build neighbor list

  //Doc:Method
#ifdef SIM
  virtual void     Neighbors(NodeWeightVec_t&, Interface*, bool);
#else
  void     Neighbors(NodeWeightVec_t&, Interface*, bool);
#endif //SIM
  //Doc:Desc Build the vector of neighbor,interface list
  //Doc:Arg1 The NodeWightVec element where we will fill in the neighbors
  //Doc:Arg2 The interface, with respect to which we need to find these
  //Doc:Arg2 neighbors
  //Doc:Arg3 boolean to specify if we need all the neighbors OR if we need
  //Doc:Arg3 only routing peers
  
  //Doc:Method
  void     AllNeighbors(NodeWeightVec_t&);   // Build neighbor list
  //Doc:Desc AllNeighbors is used by broadcasts even if they are
  //Doc:Desc not routing peers. Use this to find all the peers
  //Doc:Arg1 The NodeWeightVec element, with respect to which,
  //Doc:Arg1 we need all the peers in this link
  

  //Doc:Method
  MACAddr  IPToMac(IPAddr_t);                // Convert peer IP to peer MAC
  //Doc:Desc Convert peer IP to peer MAC
  //Doc:Arg1 The IP Address whose MAC is to be determined
  //Doc:Return The corresponding MAC address


  //Doc:Method
  bool     RxBroadcast() { return receiveOwnBroadcast;}
  // Ethernet Methods
  //Doc:Desc   Checks whether nodes receive their own broadcasts on this link
  //Doc:Return Return true if node receive their own broadcasts;
  //Doc:Return false otherwise

  //Doc:Method
  void       AddNode(Node*);                 // Add existing node to LAN
  //Doc:Desc Add a new node to the ethernet
  //Doc:Arg1 The pointer to the node to be added

  
  //Doc:Method
  Node*      GetNode(Count_t);               // Get the i'th node in network
  //Doc:Desc  Get the node by its index
  //Doc:Arg1  The index to the list of nodes
  //Doc:Return Pointer to the node
  
  //Doc:Method
  Interface* GetIf(Count_t);                 // Get the i'th interface
  //Doc:Desc Get the interface of the node connected to this link by its index
  //Doc:Arg1 the index of the node
  //Doc:Return the interface of the node that attaches to this link

  
  //Doc:Method 
  IPAddr_t   GetIP(Count_t);                 // Get the i'th ip address
  //Doc:Desc Get IP of the node by the index of the node
  //Doc:Arg1 int index of the node
  //Doc:Return IP address if the index is within range, else IPADDR_NONE
  
  //Doc:Method
  Interface* GetIfByMac(MACAddr);            // Get interface by MAC addr
  //Doc:Desc Get a pointer to the interface by its MAC address
  //Doc:Arg1 the MAC address
  //Doc:Return pointer to the interface
  
  //Doc:Method
  Node*      Gateway();                      // Get the gateway node
  //Doc:Desc Returns the gateway node of this LAN
  //Doc:Return Pointer to the gateway node
  
  //Doc:Method
  void       SetLeafQLimit(Count_t);         // Set q limit for all leafs
  //Doc:Desc Method to set the queue limits of all the nodes in this link
  //Doc:Desc to a specified value
  //Doc:Arg1 the size of each queue
  
  //Doc:Method
  void       RxOwnBroadcast(bool rob = true) { receiveOwnBroadcast = rob;}
  //Doc:Desc   Checks whether nodes receive their own broadcasts on this link
  //Doc:Return Return true if node receive their own broadcasts;
  //Doc:Return false otherwise
    
public:
  NodeId_t   first;    // Since there is a global node list, we don't need one 
  NodeId_t   last;     // Just log first/last+1 entries
  IPAddr_t   firstIP;  // First IP Address
  Mask_t     mask;
  IPAddr_t   nextIP;   // Next available IP Address
  MACAddr    firstMac; // First MAC Address
  Detail_t   detail;   // Level of detail
  Node*      gw;       // Points to gateway node if specified
  Interface* gwIf;     // Gateway interface
  bool       receiveOwnBroadcast; // True if nodes receive their own broadcasts
  NodeVec_t nodes;     // Vector of new nodes
  Count_t   size;      // Number of connected nodes
  L4Protocol* l4proto;  // L4 protocol to bind to each node
private:
  void ConstructorHelper(Count_t,          // Node count
                         Node*,            // Gateway node
                         IPAddr_t, Mask_t, // IP addr start, mask
                         SystemId_t,       // Responsible System for dist sim
                         L4Protocol* l4,   // Layer 4 protocol to bind to each 
                         Rate_t r, Time_t dly, Detail_t d);
};


#endif


