/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Global trace control
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __etrace_h__
#define __etrace_h__

#include <fstream>
#include <vector>

#include "common-defs.h"
#include "tfstream.h"
#include "timer.h" // for position tracing

typedef std::vector<int>  vNodeIds;

class Node;
class Protocol;
class PDU;
class Packet;
class Interface;

//Doc:ClassXRef
//class ETrace : public TimerHandler {
class ETrace : public TimerHandler {
  //Doc:Class Class {\tt ETrace} defines the behavior of the packet tracing
  //Doc:Class feature of \GTNS.
public:
  typedef enum { ENABLED, DISABLED, DEFAULT } ETraceStatus;
  typedef std::vector<ETraceStatus> ETSVec_t;
  typedef bool (*Callback_t)(Node*,Protocol*,PDU*,bool);
public:
  //Doc:Method
  ETrace();
    //Doc:Desc Default constructor, no arguments.

  ~ETrace();
  //Doc:Method
  bool Open(const char*);  // Open the trace file
    //Doc:Desc Create a trace file.
    //Doc:Return     //Doc:Arg1 File name.
    //Doc:Return True if successfully opened.

  //Doc:Method
  void Close();            // Close the trace file
    //Doc:Desc Close the trace file.  Should be called when the simulation completes.

  //Doc:Method
  void LayerOn();               // Enable all traces
    //Doc:Desc Enable tracing for all protocol layers.

  //Doc:Method
  void LayerOn(Layer_t);        // Enable for specific layer
    //Doc:Desc Enable tracing for a specified protocol layer.
    //Doc:Arg1 Layer to enable (2, 3, or 4).

  //Doc:Method
  void LayerOff();              // Disable all traces
    //Doc:Desc Enable tracing for all protocol layers.

  //Doc:Method
  void LayerOff(Layer_t);       // Disable for specific layer
    //Doc:Desc Disable tracing for a specified protocol layer.
    //Doc:Arg1 Layer to disable (2, 3, or 4).

  //Doc:Method
  bool IsLayerOn(Layer_t);      // Enquire trace status (on/off) by layer
    //Doc:Desc Test whether tracing is on or off by layer.
    //Doc:Arg1 Layer to test.
    //Doc:Return True if tracing enabled for this layer.
 
  //Doc:Method
  bool IsEnabled();        // Enquire trace file opened
    //Doc:Desc Check if trace file globally enabled.
    //Doc:Return True if enabled.

	//Doc:Method
void NodeOn();            //Enables tracing on all nodes
	//Doc:Desc Enable tracing for all nodes.

	//Doc:Method
void NodeOn(int id); // Enable for specific node
	//Doc:Desc Enable tracing for a specified node.
	//Doc:Arg1 Node to enable (1, 2, 3, 4,...).

	//Doc:Method
void NodeOff(int id);// Disable for specific node
  //Doc:Desc Disable tracing for a specified node.
  //Doc:Arg1 Node to Disable (1, 2, 3, 4,...).
		
 	//Doc:Method
  bool IsNodeOn(int id);      // Enquire trace status (on/off) by layer
    //Doc:Desc Test whether tracing is on or off by layer.
    //Doc:Arg1 Layer to test.
    //Doc:Return True if tracing enabled for this layer.
	
void TraceTopo(Time_t interval);
Time_t& TraceTopo() {return tracetopo;}
void StartTraceTopo();

  Tfstream& GetTS() { return os;} // The trace file object
  void AppendEOL();        // Add the end of line if needed
  void TraceEvent(Node*, char *str, int level);
	void SetTraceEventLevel(int i);
	void TracePacket(Packet*, Node*, Interface*, Layer_t , char *str = "?");     

	void SetExtended(bool e) {extended = e;}
	bool IsExtended() {return extended;}
	
  void Trace8023(Tfstream&, PDU*);
  void Trace80211(Tfstream&, PDU*);
#ifdef USAP_ADDON
  void TraceUsap(Tfstream&, PDU*);
#endif //USAP_ADDON
  void TraceIPV4(Tfstream&, PDU*);
  void TraceOspf6(Tfstream&, PDU*);
	void TraceOspf6Hello(Tfstream& os, struct ospf6_header *oh);
	void TraceOspf6Dbdesc(Tfstream& os, struct ospf6_header *oh);
	void TraceOspf6Lsr(Tfstream& os, struct ospf6_header *oh);
	void TraceOspf6Lsu(Tfstream& os, struct ospf6_header *oh);
	void TraceOspf6Lsa(Tfstream& os, struct ospf6_header *oh);
	void TraceOspf6LsaHeader(Tfstream& os, struct ospf6_lsa_header *header);
 void TraceOspf6LLS(Tfstream& os, struct ospf6_LLS_header * lls_header);
	void TraceOspf6Tlv(Tfstream& os, struct ospf6_TLV_header * tlv_header);

  void NewNode(Node*);     // Log a new node for trace line
  void SetNode(Node*);     // Set the node for subsequent traces

  //Doc:Method
  void IPDotted(bool d = true) { os.IPDotted(d);}
    //Doc:Desc Specify that all \IPA{s} are to be logged in dotted notation.
    //Doc:Desc If not, trace in 32 bit hex.
    //Doc:Arg1 True if dotted desired.

  //Doc:Method
  void TimePrecision(Count_t d) { os.TimePrecision(d);}
    //Doc:Desc Specify the digits of precision in the time field for
    //Doc:Desc the trace file.
    //Doc:Arg1 Number of digits of precision.

  //Doc:Method
  static ETrace* Instance();// Get the global trace object
    //Doc:Desc Get a pointer to the single, global trace object.
    //Doc:Return Pointer to trace object.

  //Doc:Method
  static bool   Enabled() { return enabled;} // True if tracing enabled
    //Doc:Desc Test if tracing is enabled.
    //Doc:Return True if tracing enabled.

  //Doc:Method
  static void AddCallback(Callback_t cb) { callback = cb;}
    //Doc:Desc Add a callback to be called on every trace decision.
    //Doc:Desc The callback is called AFTER the normal decisioning process,
    //Doc:Desc and the result of that is passsed to the callback.  Also
    //Doc:Desc passed are the node pointer, protocol pointer, and pdu pointer.
    //Doc:Desc The callback must be a funtion returning a bool (which is the
    //Doc:Desc final decision on whether the PDU is traced or not) and
    //Doc:Desc with four arguments, Node*, Protocol*, PDU*, bool.
    //Doc:Desc See definition of Callback_t in trace.h.
    //Doc:Arg1 Callback function to add.

  void SetTraceOnTime(Time_t a) { trace_on_time = a; }
  void SetTraceOffTime(Time_t a) { trace_off_time = a; }

public:
  static Callback_t callback;

private:
  static ETrace* instance;  // The single global trace object
  static bool   enabled;   // Global enabled flag
	static bool   extended;  // extended extended tracing flag
	static int trace_event_level;
  ETSVec_t  trace;          // On/Off status by layer
  Tfstream os;             // File for logging
  Node*    node;           // Current node
	vNodeIds vnodeids;
  Time_t trace_on_time;
  Time_t trace_off_time;

  // Below is for Position tracing
  Time_t tracetopo;            // for position tracing
  Timer timer;
  TimerEvent* timeoutevent;
  void Timeout(TimerEvent*);
  void TraceTopology();
};

#endif

