/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Circular Queue for Sequential Events
// George F. Riley.  Georgia Tech, Fall 2002

// Defines an efficient circular queue for events that are inserted
// and removed in timestamp order.

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "eventcq.h"

using namespace std;

// Constructor
EventCQ::EventCQ() 
  : events(DefaultInitialSize, nil),
    limit(DefaultInitialSize), in(0), out(0),
    timeLastIn(0)
{
}

#ifdef  MOVED_TO_H
// Data access functions
void EventCQ::push_back(EventPair* e)
{
#ifdef EXTRA_CHECK
  if (e->time < timeLastIn)
    {
      cout << "HuH?  EventCQ::push_back out of order" << endl;
      cout << "Time " << e->time << " lastIn " << timeLastIn << endl;
      exit(1);
    }
#endif
  timeLastIn = e->time;
  EventCQVec_t::size_type newin = NextIn();
  if (newin == out)
    {
      Grow();               // Need more room
      newin = NextIn();
    }
  events[in] = e;           // Store at the in location
  in = newin;               // Advance in
}

void EventCQ::pop_front()
{ // Remove front element
  out = NextOut(); 
}

EventPair* EventCQ::front()
{ // Get a reference to the front object
  return events[out];
}

EventPair* EventCQ::next()
{
  return events[in];
}

Count_t EventCQ::size()
{
  if (in >= out) return in - out;
  return limit - (out - in);
}

EventPair* EventCQ::operator[](Count_t i)
{
  EventCQVec_t::size_type k = out + i;
  while(k >= limit) k-=limit;
  return events[k];
}

// Private Methods
EventCQVec_t::size_type EventCQ::NextIn()
{ // Get the advanced in pointer, modulo limit
  EventCQVec_t::size_type r = in + 1;
  if (r != limit) return r;
  return 0;
}

EventCQVec_t::size_type EventCQ::NextOut()
{ // Get the advanced out pointer, modulo limit
  EventCQVec_t::size_type r = out + 1;
  if (r != limit) return r;
  return 0;
}
#endif
  
void EventCQ::Grow()
{ // Increase size to next power of 2
  EventCQVec_t::size_type oldLim = limit;
#undef  VERBOSE
#ifdef  VERBOSE
  cout << "eventcq " << this << " dumping orig before grow" << endl;
  cout << "In " << in << " out " << out << endl;
  for(EventCQVec_t::size_type i = 0; i < limit; ++i)
    {
      if (events[i])
        {
          cout << "time " << events[i]->time 
               << " event " << events[i]->event << endl;
        }
      else
        {
          cout << "nil";
          
        }
    }
#endif
  limit *= 2;
  DEBUG0((cout << "Growing EvenvCQ to size " << limit << endl));
  events.reserve(limit); // Reserve new space
  //EventPair evp(0.0, nil);
  events.insert(events.end(),  oldLim, nil);
  if (out != 0)
    { // out == 0 is simplest case, no moving around needed
      for (EventCQVec_t::size_type i = 0; i != in; ++i)
        { // Move data to newly reserved locations
          events[oldLim + i] = events[i];
          events[i] = nil;
        }
      in = oldLim + in;
    }
#ifdef VERBOSE
  cout << "Dumping orig after grow" << endl;
  for(EventCQVec_t::size_type i = 0; i < limit; ++i)
    {
      if (events[i])
        {
          cout << "time " << events[i]->time
               << " event " << events[i]->event << endl;
        }
      else
        {
          cout << "nil" << endl;
        }
    }
  cout << "New In " << in << " out " << out << endl;
#endif
}
