/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - FTP Client Class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the FTP Client Class

#define DEBUG_MASK 0x01
#include "debug.h"
#include "ftp-client.h"
#include "simulator.h"

using namespace std;

// FTPAction methods
// Constructor
FTPAction::FTPAction()
  : action(NONE), peerIP(IPADDR_NONE),
    reqSz(0), fileSz(0),
    reqRv(0), fileRv(0),
    sleepTime(0), sleepRv(0),
    nRepeat(0), cRepeat(0), stepRepeat(0)
{
}

FTPAction::FTPAction(Actions_t a)
  : action(a), peerIP(IPADDR_NONE),
    reqSz(0), fileSz(0),
    reqRv(0), fileRv(0),
    sleepTime(0), sleepRv(0),
    nRepeat(0), cRepeat(0), stepRepeat(0)
{
}

FTPAction::FTPAction(const FTPAction& c)
  : action(c.action), peerIP(c.peerIP),
    reqSz(c.reqSz), fileSz(c.fileSz),
    reqRv(0), fileRv(0),
    sleepTime(c.sleepTime), sleepRv(0),
    nRepeat(c.nRepeat), cRepeat(c.cRepeat), stepRepeat(c.stepRepeat)
{
  // Copy each random variable if it exists
  if (c.reqRv)   reqRv   = c.reqRv->Copy();
  if (c.fileRv)  fileRv  = c.fileRv->Copy();
  if (c.sleepRv) sleepRv = c.sleepRv->Copy();
}

// Assignment Operator
FTPAction& FTPAction::operator=(const FTPAction& r)
{
  if (&r != this)
    { // ignore self assignment
      action  = r.action;
      peerIP  = r.peerIP;
      reqSz   = r.reqSz;
      fileSz  = r.fileSz;
      delete reqRv;
      if (r.reqRv)  reqRv  = r.reqRv->Copy();
      else          reqRv  = NULL;
      delete fileRv;
      if (r.fileRv) fileRv = r.fileRv->Copy();
      else          fileRv = NULL;
      sleepTime = r.sleepTime;
      delete sleepRv;
      if (r.sleepRv) sleepRv = r.sleepRv->Copy();
      else           sleepRv = NULL;
      nRepeat    = r.nRepeat;
      cRepeat    = r.cRepeat;
      stepRepeat = r.stepRepeat;
    }
  return *this;
}
FTPAction::~FTPAction()
{
  if (reqRv)   delete reqRv;
  if (fileRv)  delete fileRv;
  if (sleepRv) delete sleepRv;
}

void FTPAction::Reset()
{
  cRepeat = 0;
}

// FTPClient methods
FTPClient::FTPClient(const TCP& t)
  : state(CLOSED), currentStep(0), started(false),
    l4Proto((TCP*)t.Copy()), currentAction(0), timeoutEvent(0)
{
  l4Proto->AttachApplication(this);
}

FTPClient::~FTPClient()
{
  delete l4Proto;
}

// TimerHandler 
void FTPClient::Timeout(TimerEvent* ev)
{
  DEBUG(0,(cout << "FTPClient::Timeout time " 
        << Simulator::Now() << endl));
  timeoutEvent = NULL;
  NextAction();
}


// Upcalls from L4 protocol
void FTPClient::Receive(Packet*,L4Protocol*)
{ // Data received
}

void FTPClient::Sent(Count_t s, L4Protocol*)
{ // Data has been sent
  DEBUG(0,(cout << "FTPClient::Sent " << s << " bytes" << endl));
  getputCurrent += s;
  if (getputCurrent >= getputSize)
    { // Time for next
      switch (state) {
        case SENDING_GET :
          DEBUG(1,(cout << "Time to get file" << endl));
          state = GETTING;
          l4Proto->Send(currentAction->fileSz);
          break;
        case GETTING :
          DEBUG(1,(cout << "Time to advance to next action" << endl));
          state = OPEN;
          NextAction();
          break;
        case SENDING_PUT :
          DEBUG(1,(cout << "Time to put file" << endl));
          state = PUTTING;
          l4Proto->Send(currentAction->fileSz);
          break;
        case PUTTING :
          DEBUG(1,(cout << "Time to advance to next action" << endl));
          state = OPEN;
          NextAction();
          break;
        default:
          cout << "HuH?  FTPClient got sent in bad state " << state << endl;
          return;
        }
    }
}

void FTPClient::Closed(L4Protocol*)
{ // Connection has closed
  DEBUG(0,(cout << "FTPClient::Closed" << endl));
  if (state != CLOSING)
    {
      cout << "HuH? FTPClient::Closed in bad state " << state << endl;
    }
  state = CLOSED;
}

void FTPClient::ConnectionComplete(L4Protocol*)
{ // Connection request succeeded
  if (state != OPENING)
    {
      cout << "HuH?  FTPClient::ConnectionComplete when not opening" << endl;
      return;
    }
  state = OPEN;
  DEBUG(0,(cout << "FTPClient::ConnectionComplete at time " 
        << Simulator::Now() << endl));
  NextAction();
}

void FTPClient::ConnectionFailed(L4Protocol*)
{  // Connection request failed
  if (state != OPENING)
    {
      cout << "HuH?  FTPClient::ConnectionFailed when not opening" << endl;
      return;
    }
  DEBUG(0,(cout << "FTPClient::ConnectionFailed at time " 
        << Simulator::Now() << endl));
  state = CLOSED;
  // If the connection fails, we just do nothing (no more actions)
}


// FTP Action processors
// All actions return step number, that can be used as a repeat target
Count_t FTPClient::Open(IPAddr_t ip)
{ // Open session with peer
  FTPAction a(FTPAction::OPEN);
  a.peerIP = ip;
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Get(Count_t req, Count_t file)
{ // Get, size of req, size of file
  FTPAction a(FTPAction::GET);
  a.reqSz = req;
  a.fileSz = file;
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Get(const Random& reqRv, const Random& fileRv)
{ // Get, size of req/file (random)
  FTPAction a(FTPAction::GET);
  a.reqRv = reqRv.Copy();
  a.fileRv = fileRv.Copy();
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Put(Count_t req, Count_t file)
{ // Get, size of req, size of file
  FTPAction a(FTPAction::PUT);
  a.reqSz = req;
  a.fileSz = file;
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Put(const Random& reqRv, const Random& fileRv)
{ // Get, size of req/rile (random)
  FTPAction a(FTPAction::PUT);
  a.reqRv = reqRv.Copy();
  a.fileRv = fileRv.Copy();
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Close()
{ // Close connection
  FTPAction a(FTPAction::CLOSE);
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Sleep(Time_t sleepTime)
{ // Sleep for specified time
  FTPAction a(FTPAction::SLEEP);
  a.sleepTime = sleepTime;
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Sleep(Random* sleepRv)
{ // Sleep for random time
  FTPAction a(FTPAction::SLEEP);
  a.sleepRv = sleepRv;
  actions.push_back(a);
  return actions.size() - 1;
}

Count_t FTPClient::Repeat(Count_t step, Count_t count)
{ // Repeat steps, starting at specified
  FTPAction a(FTPAction::REPEAT);
  a.stepRepeat = step;
  a.nRepeat = count;
  actions.push_back(a);
  return actions.size() - 1;
}


// Application methods
void FTPClient::StartApp()
{ // Called at time specified by Start
  for (ActionVec_t::size_type i = 0; i < actions.size(); ++i)
    {
      actions[i].Reset();
    }
  currentStep = 0;
  started = true;
  cout << "FTPClient started" << endl;
  NextAction();
}

void FTPClient::StopApp()
{ // Called at time specified by Stop
  started = false;
  if (timeoutEvent) timer.Cancel(timeoutEvent);
  timeoutEvent = NULL;
}

// FTPClient, private methods
FTPClient::ActionStatus_t FTPClient::NextAction()
{ // Returns true if successful, false if failed
  DEBUG(1,(cout << "FTPClient::NextAction, started " 
       << started 
       << " currentStep " << currentStep
           << " size " << actions.size() << endl));
  if (!started) return FINISHED;
  if (currentStep >= actions.size()) return FINISHED; // No more action
  currentAction = &actions[currentStep++];
  switch (currentAction->action) {
    case FTPAction::OPEN :
      if (state != CLOSED) return FAILED;
      l4Proto->Connect(currentAction->peerIP, FTP_PORT);
      state = OPENING;
      DEBUG(1,(cout << "FTPClient::NextAction, Opening" << endl));
      return SUCCESS;
      break;
    case FTPAction::GET  :
      if (state != OPEN) return FAILED;
      state = SENDING_GET;
      if (currentAction->reqRv)
        { // Use the random variable
          currentAction->reqSz = (Count_t)currentAction->reqRv->Value();
        }
      if (currentAction->fileRv)
        { // Use the random variable
          currentAction->fileSz = (Count_t)currentAction->fileRv->Value();
        }
      getputSize = currentAction->reqSz;
      getputCurrent = 0;
      l4Proto->Send(currentAction->reqSz);
      break;
    case FTPAction::PUT  :
      if (state != OPEN) return FAILED;
      state = SENDING_PUT;
      if (currentAction->reqRv)
        { // Use the random variable
          currentAction->reqSz = (Count_t)currentAction->reqRv->Value();
        }
      if (currentAction->fileRv)
        { // Use the random variable
          currentAction->fileSz = (Count_t)currentAction->fileRv->Value();
        }
      getputSize = currentAction->reqSz;
      getputCurrent = 0;
      l4Proto->Send(currentAction->reqSz);
      break;
    case FTPAction::CLOSE :
      if (state != OPEN) return FAILED;
      state = CLOSING;
      l4Proto->Close();
      break;
    case FTPAction::SLEEP : 
      // Schedule the sleep timer
      if (currentAction->sleepRv)
        { // Use the random variable
          currentAction->sleepTime = currentAction->sleepRv->Value();
        }
      timeoutEvent = new TimerEvent();
      timer.Schedule(timeoutEvent, currentAction->sleepTime, this);
      break;
    case FTPAction::REPEAT :
      currentAction->cRepeat++;
      DEBUG(0,(cout << "FTPAction::REPEAT, cRepeat "
               << currentAction->cRepeat << ", nRepeat "
               << currentAction->nRepeat << endl));
      if ((!currentAction->nRepeat) ||
          currentAction->cRepeat < currentAction->nRepeat)
        {
          currentStep = currentAction->stepRepeat;
        }
      return NextAction();
      break;
    default:    // Some compilers complain about missing enum values
      break;
  }
  return FAILED;
}


