/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Grid Network Object
// George F. Riley.  Georgia Tech, Spring 2002

// Define an object for a m by n grid of nodes

#include <iostream>

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "common-defs.h"
#include "grid.h"
#include "link.h"
#include "node.h"
#include "duplexlink.h"

using namespace std;

// Constructors

Grid::Grid(Count_t mx, Count_t my,  // X and Y grid size
           IPAddr_t firstIP,        // First IP Address (may be NOADDR)
           const Linkp2p& lk,       // Link object to use
           SystemId_t id)           // Responsible system for distributed sim
  : xSize(mx), ySize(my)
{
  IPAddr_t nextIP = firstIP;        // Starting IP Address
  first = Node::nextId;             // Note first node id
  for (Count_t y = 0; y < my; ++y)
    {
      for (Count_t x = 0; x < mx; ++x)
        {
          DEBUG0((cout << "Nextip is " << (string)IPAddr(nextIP) << endl));
          Node* n = new Node(id);
          Node* ln = LeftNeighbor(x, y);
          if (ln) 
            { // left node exists, create duplex link
              if (nextIP != IPADDR_NONE)
                { // IP Address used, create the links with ip addresses
                  DuplexLink(ln, nextIP,   MASK_ALL, lk,
                             n,  nextIP+1, MASK_ALL, lk);
                  nextIP += 2; // Advance for next link
                }
              else
                { // No ip addresses
                  DuplexLink(ln, lk, n, lk);
                }
            }
          Node* bn = BottomNeighbor(x, y);
          if (bn) 
            { // Bottom node exists, create duplex link
              if (nextIP != IPADDR_NONE)
                { // IP Address used, create the links with ip addresses
                  DuplexLink(bn, nextIP,   MASK_ALL, lk,
                             n,  nextIP+1, MASK_ALL, lk);
                  nextIP += 2; // Advance for next link
                }
              else
                { // No ip addresses
                  DuplexLink(bn, lk, n, lk);
                }
            }
        }
    }
}

Node* Grid::GetNode(Count_t x, Count_t y)
{
  if (x >= xSize || y >= ySize) return nil;
  Count_t i = y * xSize + x;
  return Node::GetNode(first + i);
}

void Grid::BoundingBox(const Location& ll, const Location& ur)
{
  Meters_t xDist = ur.X() - ll.X();
  Meters_t yDist = ur.Y() - ll.Y();
  Meters_t xAdder = xDist / xSize;
  Meters_t yAdder = yDist / ySize;
  Meters_t yLoc = yDist / 2;
  for (Count_t y = 0; y < ySize; ++y)
    {
      Meters_t xLoc = xDist / 2;
      for (Count_t x = 0; x < xSize; ++x)
        {
          GetNode(x, y)->SetLocation(Location(xLoc, yLoc));
          xLoc += xAdder;
        }
      yLoc += yAdder;
    }
}

// Private Methods
Node*    Grid::LeftNeighbor(Count_t x, Count_t y)
{ // Get pointer to left neighbor
  if (x == 0) return nil;   // No left neighbor
  return GetNode(x - 1, y); // Left neighbor
}

Node*    Grid::RightNeighbor(Count_t x, Count_t y)
{ // Get pointer to right neighbor
  return GetNode(x + 1, y); // Right neighbor (GetNode verifies range)
}

Node*    Grid::BottomNeighbor(Count_t x, Count_t y)
{ // Get pointer to bottom neighbor
  if (y == 0) return nil;   // No bottom neighbor
  return GetNode(x, y - 1); // Bottom neighbor
}

Node*    Grid::TopNeighbor(Count_t x, Count_t y)
{ // Get pointer to top neighbor
  return GetNode(x, y + 1); // Top neighbor (GetNode verifies range) 
}

