/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Grid Network Object
// George F. Riley.  Georgia Tech, Spring 2002

// Define an object for a m by n grid of nodes
// Grids are 2-dimensional array of nodes, with point-to-point
// links connecting nodes.  The default p2p link is used.
// A starting IP Address may be specified, which is automatically
// incremented as new links are created.  If IPADDR_NONE is specified,
// the ip addresses are not used.

#include "common-defs.h"
#include "linkp2p.h"

class Link;
class Node;

//Doc:ClassXRef
class Grid {
  //Doc:Class Defines an object for a m by n grid of nodes
  //Doc:Class Grids are 2-dimensional array of nodes, with point-to-point
  //Doc:Class links connecting nodes.  The default p2p link is used.
  //Doc:Class A starting IP Address may be specified, which is automatically
  //Doc:Class incremented as new links are created.
  //Doc:Class If \IPAN\ is specified,
  //Doc:Class the ip addresses are not used.
public:
  // Constructors
  //Doc:Method
  Grid(Count_t, Count_t,               // X and Y grid size
       IPAddr_t firstIP = IPADDR_NONE, // First IP Address
       const Linkp2p& = Linkp2p::Default(),
       SystemId_t = 0);
    //Doc:Desc Constructor for Grid object.
    //Doc:Arg1 Count of nodes on the grid x--axis.
    //Doc:Arg2 Count of nodes on the grid y--axis.
    //Doc:Arg3 \IPA\ for node at x = 0, y = 0.
    //Doc:Arg4 Link object to use for links.
    //Doc:Arg5 Responsible system ID for distributed simulation

  // Node access
  //Doc:Method
  Node*    GetNode(Count_t, Count_t);  // Get specified node
    //Doc:Desc Get a pointer to a specified node in the grid.
    //Doc:Arg1 Node location X--coordinate.
    //Doc:Arg1 Node location Y--coordinate.

  //Doc:Method
  void BoundingBox(const Location&, const Location&);
    //Doc:Desc Define a bounding box for this grid object. 
    //Doc:Desc The bounding box
    //Doc:Desc is used to assign node locations, which are used during
    //Doc:Desc the simulation animation.
    //Doc:Arg1 Lower left corner of bounding box.
    //Doc:Arg2 Upper right corner of bounding box.

private:
  Count_t  xSize;                // Size of grid, x direction
  Count_t  ySize;                // Size of grid, y direction
  NodeId_t first;                // Id of first node in grid
private:
  Node*    LeftNeighbor(Count_t, Count_t);   // Get pointer to left neighbor
  Node*    RightNeighbor(Count_t, Count_t);  // Get pointer to right neighbor
  Node*    BottomNeighbor(Count_t, Count_t); // Get pointer to bottom neighbor
  Node*    TopNeighbor(Count_t, Count_t);    // Get pointer to top neighbor
};


