/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Histogram Class
// George F. Riley.  Georgia Tech, Spring 2002

// Defines a Histogram and CDF statistics collection

#ifndef __histogram_h__
#define __histogram_h__

#include <iostream>
#include <vector>

#include "common-defs.h"
#include "statistics.h"

//Doc:ClassXRef
class Histogram : public Statistics {
  //Doc:Class Class {\tt Histogram} defines a statistics collection
  //Doc:Class object that tracks samplees in fixed size buckets
  //Doc:Class and counts the number of samples falling in each
  //Doc:Class bucket interval.
public:
  //Doc:Method
  Histogram();
    //Doc:Desc Default constructor, no arguments.  Not particularly
    //Doc:Desc useful, since the default constructor sets the maximum
    //Doc:Desc measurement to zero, resulting in all measurements
    //Doc:Desc being out of range.

  //Doc:Method
  Histogram(double);
    //Doc:Desc Construct with specified maximum measurement.
    //Doc:Arg1 Maximum measurement.

  //Doc:Method
  Histogram(double, Count_t, double = 0);
    //Doc:Desc Construct with specified maximum value and bin count.
    //Doc:Desc Optional third parameter with minimum value
    //Doc:Desc This is the preferred constuctor for Histograms.
    //Doc:Arg1 Maximum measurement.
    //Doc:Arg2 Desired bin count.
    //Doc:Arg2 Minimum measurement

  virtual ~Histogram();
  //Doc:Method
  void Bins(double, Count_t, double = 0); // Max, number of bins, min
    //Doc:Desc Specify the maximum value and bin count.
    //Doc:Arg1 Maximum measurement.
    //Doc:Arg2 Desired bin count.
    //Doc:Arg3 Minimum measurement.

  bool Record(double);
  void Reset();
  Statistics* Copy() const;   // Make a copy of the statistics

  //Doc:Method
  void Log(std::ostream&, const char* hdr = NULL, const char sep = ',');
    //Doc:Desc Record the statistics on an output stream.  The
    //Doc:Desc histogram is recorded with two values per line, the minimum 
    //Doc:Desc value for a bin, and the count  of measurements in that bin.
    //Doc:Arg1 Output stream to record  the  statistics on.
    //Doc:Arg2 Header line to add to beginning of the statistics log
    //Doc:Arg2 (if not {\tt nil}).
    //Doc:Arg3 Separator  character between statistics elements.

  // Log in CDF format   
  //Doc:Method
  void CDF(std::ostream&, const char* hdr = NULL, const char sep = ',');
    //Doc:Desc Record the statistics on an output stream, in Cumulative 
    //Doc:Desc Distribution Function (CDF) form.  The CDF is recorded with 
    //Doc:Desc two values per line, the maximum value for a bin, and the
    //Doc:Desc cumulative  probability that  the measurement is less than
    //Doc:Desc that value.
    //Doc:Arg1 Output stream to record  the  statistics on.
    //Doc:Arg2 Header line to add to beginning of the statistics log
    //Doc:Arg2 (if not {\tt nil}).
    //Doc:Arg3 Separator  character between statistics elements.

public:
  std::vector<Count_t> bins;
  double               min;
  double               max;
  Count_t              outOfRange;
};
#endif



