/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Histograph Class
// George F. Riley.  Georgia Tech, Fall 2003

// Defines a Histogram and CDF statistics collection
// and displays graphically using Qt.

#include <iostream>

#include "debug.h"
#include "histograph.h"

using namespace std;

Histograph::Histograph()
    : Histogram(), initialized(false), showCDF(false),
      updateRate(1), updateCountdown(1)
#ifdef HAVE_QT
      , app(nil), canvas(nil), view(nil)
#endif
{
}

Histograph::Histograph(double m)
    : Histogram(m), initialized(false), showCDF(false),
      updateRate(1), updateCountdown(1)
#ifdef HAVE_QT
      , app(nil), canvas(nil), view(nil)
#endif
{
}

Histograph::Histograph(double m, Count_t b, double mn)
    : Histogram(m, b, mn), initialized(false), showCDF(false),
      updateRate(1), updateCountdown(1)
#ifdef HAVE_QT
      , app(nil), canvas(nil), view(nil)
#endif
{
}

Histograph::Histograph(const Histograph& c) // Copy constructor
    : Histogram(c), initialized(false), showCDF(c.showCDF),
      updateRate(c.updateRate), updateCountdown(c.updateRate)
#ifdef HAVE_QT
      , app(nil), canvas(nil), view(nil)
#endif
{
}


Histograph::~Histograph()
{
#ifdef HAVE_QT
  if (app) delete app;
#endif
}

Statistics* Histograph::Copy() const   // Make a copy of the histograph
{
  return new Histograph(*this);
}

void Histograph::UpdateRate(Count_t c )
 {
   updateRate = c;
 }

// Overridden methods from Histogram
bool Histograph::Record(double v)
{
  bool r = Histogram::Record(v);
  if (!updateCountdown || --updateCountdown == 0)
    { // Time to update
      if (showCDF)
        {
          UpdateCDF();
        }
      else
        {
          UpdateHisto();
        }
      updateCountdown = updateRate;
    }
  return r;
}

void Histograph::Reset()
{
  Histogram::Reset();
  if (showCDF)
    {
      UpdateCDF();
    }
  else
    {
      UpdateHisto();
    }
}

// Private methods
void Histograph::UpdateCDF()
{
#ifdef HAVE_QT
  if (!initialized) Initialize();
  Count_t total = 0;
  Count_t nBins = bins.size();
  for (Count_t i = 0; i < nBins; ++i)
    {
      total += bins[i];
    }
  total += outOfRange;
  double  x = textRect.width(); // Initial x coordinate
  double  xAdder = (double)(canvasXY.x() - textRect.width()) / bins.size();
  Count_t cumulative = 0;
  Count_t usableY = canvasXY.y() - (int)(textRect.height() * 1.5);
  QPoint prior((int)x, canvasXY.y() - (int)(textRect.height()));

  for (Count_t i = 0; i < nBins; ++i)
    { //  Draw the cdf
      if (items.size() <= i)
        { // too small, add empty
          items.push_back(nil);
        }
      cumulative += bins[i];
      x += xAdder;
      double height = (double)cumulative / total * usableY;
      QPoint p1((int)x, (int)(canvasXY.y()-int(textRect.height())-height));
      // check if same, and if so don't change (code later)
      if (items[i]) delete items[i];
      QCanvasLine* l = new QCanvasLine(canvas);
      items[i] = l;
      l->setPoints(prior.x(), prior.y(), p1.x(), p1.y());
      l->setPen(Qt::black);
      l->show();
      prior = p1;
    }
  
  // Draw the registration lines
  const Count_t n = 10;
  for (Count_t i = 0; i <= n; ++i)
    {
      double v  = (double)i / n;
      Count_t h = (Count_t)(usableY * v);
      
      // labels and lines always same size, so we take a shortcut here
      if (i < labels.size())
        {
          delete labels[i];
          delete lines[i];
        }
      else
        {
          labels.push_back(nil);
          lines.push_back(nil);
        }
      QString qs = QString("%1").arg(v);
      labels[i] = new QCanvasText(qs, canvas);
      labels[i]->moveBy(0,
                        canvasXY.y() - (int)(textRect.height()*1.5) - h);
      labels[i]->show();
      lines[i] = new QCanvasLine(canvas);
      lines[i]->setPen(Qt::lightGray);
      lines[i]->setPoints(textRect.width(), 
                          canvasXY.y() - (int)(textRect.height()) - h,
                          canvasXY.x(), 
                          canvasXY.y() - (int)(textRect.height()) - h);
      lines[i]->show();
    }
  canvas->update(); // Show the updates
  
  while(app->hasPendingEvents())
    {
      app->processEvents(1);
    }
#endif
}

void Histograph::UpdateHisto()
{
#ifdef HAVE_QT
  if (!initialized) Initialize();
  if (bins.size())
    { // If bins not empty
      Count_t largest = 0;
      for (Count_t i = 0; i < bins.size(); ++i)
        {
          if (bins[i] > largest) largest = bins[i];
        }
      
      // Compute height of each bar
      Count_t nBins = bins.size();
      Count_t usableY = canvasXY.y() - (int)(textRect.height() * 1.5);
      for (Count_t b = 0; b < nBins; ++b)
        {
          Count_t height = (Count_t)(usableY * (bins[b]/(double)largest));
          DEBUG0((cout << "Height of bin " << b << " is " << height << endl));
          QCanvasRectangle* rect = (QCanvasRectangle*)items[b];
          rect->setSize(rect->width(), -height);
        }
      // Draw the registration lines and labels
      Count_t n = 10;
      if (largest < n) n = largest;
      double each = (double)largest / n;
      
      for (Count_t i = 0; i <= n; ++i)
        {
          double v  = each * i;
          Count_t h = (Count_t)(usableY * v / largest);
          Count_t intV = (Count_t)v;

          // labels and lines always same size, so we take a shortcut here
          if (i < labels.size())
            {
              delete labels[i];
              delete lines[i];
            }
          else
            {
              labels.push_back(nil);
              lines.push_back(nil);
            }
          QString qs = QString("%1").arg(intV);
          labels[i] = new QCanvasText(qs, canvas);
          labels[i]->moveBy(textRect.width() -
                            labels[i]->boundingRect().width(),
                            canvasXY.y() - (int)(textRect.height()*1.5) - h);
          labels[i]->show();
          lines[i] = new QCanvasLine(canvas);
          lines[i]->setPen(Qt::lightGray);
          lines[i]->setPoints(textRect.width(), 
                              canvasXY.y() - (int)(textRect.height()) - h,
                              canvasXY.x(), 
                              canvasXY.y() - (int)(textRect.height()) - h);
          lines[i]->show();
        }
      canvas->update(); // Show the updates
    }
  
  while(app->hasPendingEvents())
    {
      app->processEvents(1);
    }
#endif
}

void Histograph::Initialize()
{
#ifdef HAVE_QT
  int argc = 1;
  char* argv[] = { "GTNetS", nil};
  app = new QApplication(argc, argv);

  canvasXY = QPoint(640, 640);
  QString qs;
  if (showCDF)
    {
      qs = QString("%1").arg(0.9);
    }
  else
    {
      qs = QString("%1").arg(12345);
    }
  // Fudge the Y size by one and a half times
  // text height.  This allows for a bottom row
  // of box labels and half of the x-axis label
  // on top
  QCanvasText txt(qs, canvas);
  textRect = txt.boundingRect();
  canvasXY.setY(canvasXY.y() + (int)(textRect.height() * 1.5));
  // Adjust x by size of left side labels
  canvasXY.setX(canvasXY.x() + textRect.width());

  Count_t xAdder = 0;
  if (showCDF)
    { // Need to finish coding this...
    }
  else
    {
      // xAdder is width of each bar (integer pixels)
      xAdder = (canvasXY.x() - textRect.width()) / bins.size();
      // Since possibly not evenly divisible, resize canvas
      // to be correct size. Fudge two extra pixels
      xAdder+=2; // two extra pixels between bars
      //canvasXY.setX(textRect.width() + xAdder * bins.size() + xAdder);
      // And fudge the X size by the width of a 5 digit label
      canvasXY.setX(textRect.width() + xAdder * bins.size() + 4); // + 4???
    }
  
  // Create the canvas
  canvas = new QCanvas(canvasXY.x()-4, canvasXY.y());
  canvas->setDoubleBuffering(true);

  // Create the bars for the histogram or lines for CDF
  for (Count_t i = 0; i < bins.size(); ++i)
    {
      if (showCDF)
        {
          QCanvasLine* cp = new QCanvasLine(canvas);
          items.push_back(cp);
          // Will set endpoints later
        }
      else
        {
          QCanvasRectangle* cp = new QCanvasRectangle(canvas);
          // Set  zero height for now, will adjust later
          cp->setSize(xAdder, 0);
          // And move to correct position
          cp->moveBy(textRect.width() + xAdder * i,
                     canvasXY.y() - textRect.height());
          cp->setPen(Qt::black);
          cp->setBrush(Qt::red);
          cp->show();
          items.push_back(cp);
        }
    }
  
  view   = new QCanvasView(canvas);
  //view->setHScrollBarMode(QScrollView::AlwaysOff);
  //view->setVScrollBarMode(QScrollView::AlwaysOff);
  view->resize(canvasXY.x(), canvasXY.y());
  view->show();
  app->setMainWidget(view);
  while(app->hasPendingEvents())
    {
      app->processEvents(1);
    }
#endif
  initialized = true;
}
