/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Histogram Class
// George F. Riley.  Georgia Tech, Spring 2002

// Defines a Histogram and CDF statistics collection
// that dynamically draws he histogram using Qt

#ifndef __histograph_h__
#define __histograph_h__

#include <vector>

#include "common-defs.h"
#include "histogram.h"

#ifdef HAVE_QT
#include <qobject.h>
#include <qapplication.h>
#include <qcanvas.h>
typedef  std::vector<QCanvasPolygonalItem*> ItemVec_t;
typedef  std::vector<QCanvasText*>          TextVec_t;
typedef  std::vector<QCanvasLine*>          LineVec_t;
#endif

//Doc:ClassXRef
class Histograph : public Histogram {
  //Doc:Class Class {\tt Histograph} defines a statistics collection
  //Doc:Class object that tracks samples in fixed size buckets
  //Doc:Class and counts the number of samples falling in each
  //Doc:Class bucket interval.  A graphical display is produced
  //Doc:Class and updated on each sample.
public:
  //Doc:Method
  Histograph();
    //Doc:Desc Default constructor, no arguments.  Not particularly
    //Doc:Desc useful, since the default constructor sets the maximum
    //Doc:Desc measurement to zero, resulting in all measurements
    //Doc:Desc being out of range.

  //Doc:Method
  Histograph(double);
    //Doc:Desc Construct with specified maximum measurement.
    //Doc:Arg1 Maximum measurement.

  //Doc:Method
  Histograph(double, Count_t, double = 0);
    //Doc:Desc Construct with specified maximum value and bin count.
    //Doc:Desc Optional third parameter with minimum value
    //Doc:Desc This is the preferred constuctor for Histographs.
    //Doc:Arg1 Maximum measurement.
    //Doc:Arg2 Desired bin count.
    //Doc:Arg2 Minimum measurement

  Histograph(const Histograph&); // Copy constructor
  virtual ~Histograph();

  //Doc:Method
  void UpdateRate(Count_t c);
    //Doc:Desc By default, the histograph is re-drawn on every new
    //Doc:Desc "Record" call.
    //Doc:Desc Specifying an UpdateRate will re-draw less often
    //Doc:Desc (every "c" Record calls), improving performance
    //Doc:Arg1 Desired update rate.

  //Doc:Method
  void ShowCDF(bool s) { showCDF = s;}
    //Doc:Desc Specify that the graph should be displayed in CDF format.
    //Doc:Arg1 True if CDF format desired.
  
  // Over-ridden methods from histogram
  bool Record(double);
  void Reset();
  Statistics* Copy() const;   // Make a copy of the statistics
private:
  void UpdateCDF();
  void UpdateHisto();
  void Initialize();
private:
  bool    initialized;
  bool    showCDF;
  Count_t updateRate;
  Count_t updateCountdown;
#ifdef HAVE_QT
  QPoint        canvasXY;
  QApplication* app;
  QCanvas*      canvas;
  QCanvasView*  view;
  QRect         textRect;
  ItemVec_t     items;
  TextVec_t     labels;
  LineVec_t     lines;
#endif
};
#endif



