/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - ICMP Base Class
// George F. Riley.  Georgia Tech, Summer 2004

#include <iostream>

#include "debug.h"
#include "icmp.h"
#include "protograph.h"
#include "application.h"

#ifdef HAVE_QT
#include <qcolor.h>
#endif

using namespace std;

// Static variables
bool           ICMP::enabled = false;
bool           ICMP::quench = false;
ICMP*          ICMP::instance = nil;
ICMPCallback_t ICMP::callbacks;

// ICMP Methods
ICMP::ICMP()
{ // ICMP has only static variables, so no initialization needed
}

ICMP::~ICMP()
{ // Destructor, nothing needed
}

#ifdef SIM
void ICMP::DataIndication(Node* n, Packet* p, IPAddr_t ip,
                          L3Protocol& l3, Interface* i, IPAddr_t dst)
#else
void ICMP::DataIndication(Node* n, Packet* p, IPAddr_t ip,
                          L3Protocol& l3, Interface* i)
#endif //SIM
{ // Called when an ICMP packet is received at an addressed host.
  // ICMP finds the callback(s) for the node specified and calls it.
  // The IPV4 packet has been popped, the ICMP header is still on the stack
  //cout << "Received ICMP packet" << endl;
  IPV4Header* ipv4 = (IPV4Header*)p->PeekPDU(-1);
  ICMPHeader* icmp = (ICMPHeader*)p->PeekPDU();
  // Trace it
  //Trace::Instance()->NewNode(n);
  n->TracePDU(this, icmp, p, "+");
  // First see if any automatic replies are needed
  DEBUG0((cout << "ICMP::DataIndication, packet " << p << endl));
  switch(icmp->type) {
    case ICMP::ECHO :
      DEBUG0((cout << "ICMP::DataInd::ECHO from " << (string)IPAddr(ip)
              << " to " << (string)IPAddr(ipv4->dst)
              << " p " << p << endl));
      EchoReply(n, p, ipv4, icmp);
      break;
    case ICMP::ECHO_REPLY :
    case ICMP::DESTINATION_UNREACHABLE :
    case ICMP::SOURCE_QUENCH :
    case ICMP::TIME_EXCEEDED :
      break; // No action needed for these
    case ICMP::TIMESTAMP :
      TimestampReply(n, p, ipv4, icmp);
      break;
      
    case ICMP::TIMESTAMP_REPLY :
    case ICMP::REDIRECT :
    case ICMP::INFORMATION_REQUEST :
    case ICMP::INFORMATION_REPLY :
    case ICMP::PARAMETER_PROBLEM :
      // None of these are implemented
      break;
  }
  
  // Find the callback (if any)
  ICMPCallback_t::iterator it = callbacks.find(n);
  if (it != callbacks.end())
    { // Found it
      ApplVec_t& apps = it->second;
      for (ApplVec_t::size_type k = 0; k < apps.size(); ++k)
        { // Notify each application in the callback list
          // NOTE!  The applications ARE responsible for the
          // packet, as a copy of the packet is forwarded.
          // This subroutine deletes the original packet.
          Application* app = apps[k];
          app->Receive(p->Copy(), nil, 0);
        }
    }
  delete p;
}
#ifdef NOT_USED
// All of the following unused.
NCount_t ICMP::Send(Size_t){return 0;}
NCount_t ICMP::Send(char*, Size_t){return 0;}
NCount_t ICMP::Send(Data&){return 0;}
NCount_t ICMP::SendTo(Size_t, IPAddr_t, PortId_t){return 0;}
NCount_t ICMP::SendTo(char*, Size_t, IPAddr_t, PortId_t){return 0;}
NCount_t ICMP::SendTo(Data&, IPAddr_t, PortId_t){return 0;}
bool     ICMP::Connect(IPAddr_t, PortId_t){return false;}
bool     ICMP::Close(){return false;}

L4Protocol* ICMP::Copy() const
{ // Make a copy of this protocol
  return new ICMP(*this);
}
#endif

// Static methods
void ICMP::Enable(bool e)
{
  if (e)
    { // enabling
      if (enabled) return;
      if (instance) return;
      instance = new ICMP();
      enabled = true;
      ProtocolGraph::CreateCommon();
      // Register ICMP in the protocol graph
      ProtocolGraph::common->Insert(4, (Proto_t)ProtoNumber, instance);
    }
  else
    { // disabling
      if (!enabled) return;
      if (!instance) return;
      ProtocolGraph::common->Remove(4, (Proto_t)ProtoNumber);
      delete instance;
      instance = nil;
      enabled = false;
    }
}

void ICMP::EnableQuench(bool e)
{
  if (!enabled) return;
  quench = e;
}


// Methods for each ICMP message generation, used mostly by IPV4
void ICMP::DestinationUnreachable(Node* n, Packet* p, IPV4Header* ip, DU_t c)
{ // Prepare a ICMP DestinationUnreachable packet and forward to IPV4
  if (!n->ICMPEnabled()) return; // This node does not support ICMP
  if (!enabled) return;
  //  cout << "sending du at time " << Simulator::Now() << endl;
#ifdef HAVE_QT
  ColoredPacket* p1 = new ColoredPacket(Qt::red);
#else
  Packet* p1 = new Packet();
#endif
  ICMPHeader* icmp = new ICMPHeader(DESTINATION_UNREACHABLE);
  icmp->code = c;
  p1->PushPDU(icmp); // Add the ICMP pdu
  // Trace it
  //Trace::Instance()->NewNode(n);
  n->TracePDU(ICMP::Instance(), icmp, p1, "-");
  //cout << "ICMP pckt frm " << IPAddr::ToDotted(ip->src) << " 2 " << IPAddr::ToDotted(ip->dst) << endl;
  IPV4ReqInfo info(ip->src, ip->dst, 64, ProtoNumber);
  IPV4::Instance()->DataRequest(n, p1, &info);
}

void ICMP::TimeExceeded(Node* n, Packet* p, IPV4Header* ip, TE_t c)
{ // Prepare a ICMP TimeExceeded packet and forward to IPV4
  if (!n->ICMPEnabled()) return; // This node does not support ICMP
  if (!enabled) return;
#ifdef HAVE_QT
  ColoredPacket* p1 = new ColoredPacket(Qt::red);
#else
  Packet* p1 = new Packet();
#endif
  ICMPHeader* icmp = new ICMPHeader(TIME_EXCEEDED);
  icmp->code = c;
  p1->PushPDU(icmp); // Add the ICMP pdu
  // Trace it
  //Trace::Instance()->NewNode(n);
  n->TracePDU(ICMP::Instance(), icmp, p1, "-");
  IPV4ReqInfo info(ip->src, IPADDR_NONE, 64, ProtoNumber);
  IPV4::Instance()->DataRequest(n, p1, &info);
}

void ICMP::SourceQuench(Node* n, Packet*, IPV4Header*)
{
  if (!n->ICMPEnabled()) return; // This node does not support ICMP
  if (!enabled) return;
}

void ICMP::EchoReply(Node* n, Packet* p, IPV4Header* ip, ICMPHeader* icmp)
{ // Form an echo reply (PING response)
  if (!n->ICMPEnabled()) return; // This node does not support ICMP
  if (!enabled) return;
#ifdef HAVE_QT
  ColoredPacket* p1 = new ColoredPacket(Qt::red);
#else
  Packet* p1 = new Packet();
#endif
  ICMPHeader* icmp1 = (ICMPHeader*)icmp->Copy();
  icmp1->type = ECHO_REPLY;
  p1->PushPDU(icmp1); // Add the ICMP pdu
  // Trace it
  //Trace::Instance()->NewNode(n);
  n->TracePDU(ICMP::Instance(), icmp1, p1, "-");
  IPV4ReqInfo info(ip->src, IPADDR_NONE, 64, ProtoNumber);
  DEBUG0((cout << "ICMP::EchoReply sending pkt " << p1
          << " to ip " << (string)IPAddr(ip->src) << endl));
  IPV4::Instance()->DataRequest(n, p1, &info);
}

void ICMP::TimestampReply(Node* n, Packet*, IPV4Header*, ICMPHeader*)
{
  if (!n->ICMPEnabled()) return; // This node does not support ICMP
  if (!enabled) return;
}

  
// Static methods used by applications to generate ICMP requests
// Generate an echo message
void ICMP::Echo(Node* n, ICMPHeader* icmp, IPAddr_t dst)
{
  if (!enabled) return;
#ifdef HAVE_QT
  ColoredPacket* p1 = new ColoredPacket(Qt::red);
#else
  Packet* p1 = new Packet();
#endif
  icmp->type = ECHO;
  p1->PushPDU(icmp);
  // Trace it
  //Trace::Instance()->NewNode(n);
  n->TracePDU(ICMP::Instance(), icmp, p1, "-");
  IPV4ReqInfo info(dst, IPADDR_NONE, 64, ProtoNumber);
  DEBUG0((cout << "Creating PING pkt " << p1 
          << " to node " << (string)IPAddr(dst) << endl));
  IPV4::Instance()->DataRequest(n, p1, &info);
}

//Generate timestamp req
void ICMP::Timestamp(Node* n, ICMPHeader* icmp, IPAddr_t dst)
{
  if (!enabled) return;
#ifdef HAVE_QT
  ColoredPacket* p1 = new ColoredPacket(Qt::red);
#else
  Packet* p1 = new Packet();
#endif
  icmp->type = TIMESTAMP;
  p1->PushPDU(icmp);
  // Trace it
  //Trace::Instance()->NewNode(n);
  n->TracePDU(ICMP::Instance(), icmp, p1, "-");
  IPV4ReqInfo info(dst, IPADDR_NONE, 64, ProtoNumber);
  IPV4::Instance()->DataRequest(n, p1, &info);
}


// Request a callback on ICMP message receipt
void ICMP::AddCallback(Node* n, Application* a)
{
  ApplVec_t& apps = callbacks[n];
  // Make sure the app is not already listed
  for (ApplVec_t::size_type i = 0; i < apps.size(); ++i)
    {
      if (apps[i] == a) return;
    }
  // Add it
  apps.push_back(a);
}

// ICMPHeader methods
ICMPHeader::ICMPHeader()
    : type(ICMP::ECHO_REPLY), code(0), identifier(0), sequence(0),
      originateTimestamp(0), receiveTimestamp(0), transmitTimestamp(0),
      ipv4Hdr(nil)
{ // Nothing else needed
}

ICMPHeader::ICMPHeader(ICMP::ICMPMsg_t t)
    : type(t), code(0), identifier(0), sequence(0),
      originateTimestamp(0), receiveTimestamp(0), transmitTimestamp(0),
      ipv4Hdr(nil)
{ // Nothing else needed
}

  
ICMPHeader::ICMPHeader(const ICMPHeader& h)
    : type(h.type), code(h.code),
      identifier(h.identifier), sequence(h.sequence),
      originateTimestamp(h.originateTimestamp),
      receiveTimestamp(h.receiveTimestamp),
      transmitTimestamp(h.transmitTimestamp),
      ipv4Hdr(nil)
{
  if (h.ipv4Hdr) ipv4Hdr = (IPV4Header*)h.ipv4Hdr->Copy();
}

Size_t  ICMPHeader::Size() const
{
  Size_t ipSize = 0;
  Size_t s = 0;
  if (ipv4Hdr) ipSize = ipv4Hdr->Size();
  switch(type) {
    case ICMP::ECHO :
    case ICMP::ECHO_REPLY :
      s = 8;
      break;
    case ICMP::DESTINATION_UNREACHABLE :
    case ICMP::SOURCE_QUENCH :
    case ICMP::TIME_EXCEEDED :
      s = 8 + 8;
      break;
    case ICMP::TIMESTAMP :
    case ICMP::TIMESTAMP_REPLY :
      s = 8 + 12;
      break;
    case ICMP::REDIRECT :
    case ICMP::INFORMATION_REQUEST :
    case ICMP::INFORMATION_REPLY :
    case ICMP::PARAMETER_PROBLEM :
      s = 0; // Not implemented
  }
  return ipSize + s;
}

       
PDU*    ICMPHeader::Copy()
{
  return new ICMPHeader(*this);
}


// Serialization
Size_t  ICMPHeader::SSize()
{ 
  return
      sizeof(unsigned long) +  // type
      sizeof(code) +
      sizeof(identifier) + 
      sizeof(sequence) +
      sizeof(originateTimestamp) +
      sizeof(receiveTimestamp) + 
      sizeof(transmitTimestamp);
}

char*   ICMPHeader::Serialize(char* b, Size_t& sz)
{
  b = SerializeToBuffer(b, sz, (unsigned long&)type);
  b = SerializeToBuffer(b, sz, code);
  b = SerializeToBuffer(b, sz, identifier);
  b = SerializeToBuffer(b, sz, sequence);
  b = SerializeToBuffer(b, sz, originateTimestamp);
  b = SerializeToBuffer(b, sz, receiveTimestamp);
  b = SerializeToBuffer(b, sz, transmitTimestamp);
  return b;
}

char*   ICMPHeader::Construct(char* b , Size_t& sz)
{
  b = ConstructFromBuffer(b, sz, (unsigned long&)type);
  b = ConstructFromBuffer(b, sz, code);
  b = ConstructFromBuffer(b, sz, identifier);
  b = ConstructFromBuffer(b, sz, sequence);
  b = ConstructFromBuffer(b, sz, originateTimestamp);
  b = ConstructFromBuffer(b, sz, receiveTimestamp);
  b = ConstructFromBuffer(b, sz, transmitTimestamp);
  return b;
}

void    ICMPHeader::Trace(Tfstream& tos, Bitmap_t b, Packet* p, const char* s)
{
  tos << " ";
  if (s)
    tos << s;
  tos << "L4-ICMP";
  if (Detail(ICMP::TYPE, b))         tos << " " << type;
  if (Detail(ICMP::CODE, b))         tos << " " << code;
  switch(type){
    case ICMP::ECHO:
    case ICMP::ECHO_REPLY:
      if (Detail(ICMP::IDENTIFIER, b)) tos << " " << identifier;
      if (Detail(ICMP::SEQUENCE,   b)) tos << " " << sequence;
      break;
    case ICMP::TIMESTAMP:
    case ICMP::TIMESTAMP_REPLY:
      if (Detail(ICMP::ORIGIN_TIME,   b))   tos << " " << originateTimestamp;
      if (Detail(ICMP::RECEIVE_TIME,  b))   tos << " " << receiveTimestamp;
      if (Detail(ICMP::TRANSMIT_TIME, b))   tos << " " << transmitTimestamp;
      break;
  }
  return;
}


