/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - InterfaceEthernet class
// Mohamed Abd Elhafez.  Georgia Tech, Spring 2004

// Define class to model network link interfaces
// Implements the Layer 2 protocols

#include <iostream>

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "hex.h"
#include "interface-ethernet.h"
#include "link.h"
#include "queue.h"
#include "droppdu.h"
#include "node.h"
#include "trace.h"
#include "globalstats.h"
#include "ipv4.h"
#include "llcsnap.h"
#include "ratetimeparse.h"

using namespace std;

InterfaceEthernet::InterfaceEthernet(const L2Proto& l2, IPAddr_t i,
				     Mask_t m, MACAddr mac,
				     bool bootstrap)  
  : InterfaceReal(l2, i, m, mac), busyendtime(0), holdtime(0),
    maxbackoff(INITIAL_BACKOFF), backofftimer(0), maxwaittime(0),
    busycount(1), lastpacketsent(0), lastdest(MACAddr::NONE),
    lastrxevent(0), lastchanacq(0), tx_finishtime(0), bCast(false)
{ 
  generator =  new  Uniform(0, INITIAL_BACKOFF);
}

InterfaceEthernet::~InterfaceEthernet()
{
}

void InterfaceEthernet::RegisterRxEvent(LinkEvent* le)
{

  lastrxevent = le;
}

void InterfaceEthernet::AddRxEvent(LinkEvent* le)   // add to the list of rx events , used in case of broadcast
{
  rxEvents.push_back(le);
}

Time_t   InterfaceEthernet::SlotTime()
{
    slottime = SLOT_TIME / GetLink()->Bandwidth();
    return slottime;
}


bool InterfaceEthernet::Send(Packet* p, const MACAddr& dst, int type)
{
  if(!p) return false;
 
  DEBUG0((cout << " calling send from interface ethernet of node " << pNode->Id() << " for packet " << p->uid << endl));
  if (Simulator::Now() < tx_finishtime || Simulator::Now() < holdtime )  // we are already transmitting
      {
	// Resend when we finish the current transmission
	EthernetEvent* laststevent = new EthernetEvent(EthernetEvent::START_TX, p, dst, type);
	Time_t temp = tx_finishtime > holdtime ? tx_finishtime : holdtime; 
	Simulator::instance->Schedule(laststevent, temp - Simulator::Now(), this);
	DEBUG0((cout << "Rescheduling sending packet " << p->uid << " at time " << temp << endl));
	DEBUG0((cout << "txfinish time " << tx_finishtime << " hold time " << holdtime << endl));

	// Tracing
	if(GetL2Proto()->GetTrace() == Trace::ENABLED)
	  {
	    Trace::Instance()->NewNode(pNode);
	    Trace::Instance()->GetTS() << " L2-RAB " << p->retx << " " << p->uid;
	  }
      }
  else if  (Simulator::Now() < busyendtime)  // The link is busy
    {
      // Resense the channel after freetime + random backoff time
      EthernetEvent* laststevent = new EthernetEvent(EthernetEvent::START_TX, p, dst, type);      
      backofftimer =  ceil(generator->Value() * maxbackoff) * slottime;
      Simulator::instance->Schedule(laststevent, backofftimer + busyendtime - Simulator::Now(), this);
      DEBUG0((cout << "Rescheduling sending at time " << busyendtime + backofftimer<< endl));      

      // Tracing
      if(GetL2Proto()->GetTrace() == Trace::ENABLED)
	{
	  Trace::Instance()->NewNode(pNode);
	  Trace::Instance()->GetTS() << " L2-RAB " << p->retx << " " << p->uid;
	}     
    }
  else
    {
      // Store the sent packet in case we need to resend it      
      lastpacketsent = p->Copy();
      lastdest = dst;
      lasttype = type;     
      
      Time_t maxwaittime = 0;      
      
      LLCSNAPHeader* llc = new LLCSNAPHeader(type);
      p->PushPDU(llc);
      
      GetL2Proto()->BuildDataPDU(macAddr, dst, p);
      
      
      if ( !GetQueue()->CheckForcedLoss(false))
	{ // Link is free, and no forced losses presently for the queue
	  //Ethernet* eLink = (Ethernet*)GetLink();
	  // Time_t delay = 0;
	  // schedule FBR to all neighbours	
	  for(NodeTimeVec_t::size_type  i = 0; i < nodeTimes.size(); ++i)
	    {
	      EthernetEvent* evFBR = new EthernetEvent(EthernetEvent::FBR, dst, p->Size());
	      DEBUG0((cout << "Scheduling FBR to node " <<  nodeTimes[i].iface->GetNode()->Id()
		      << " after " << nodeTimes[i].time 
		      << " current time " <<  Simulator::Now() << endl)); 
	      Simulator::instance->Schedule(evFBR, nodeTimes[i].time , nodeTimes[i].iface);
	    }
	  bCast = lastdest.IsBroadcast();

	  
	  // Count as enque/deque to get proper queue statistics
	  Stats::pktsEnqued++;
	  Stats::pktsDequed++;     
	  Time_t txTime = ((Rate_t)p->size * 8) / GetLink()->Bandwidth(); // Time to transmit
	  tx_finishtime = Simulator::Now() + maxwaittime + txTime ;       // Free time for next packet
	  
	  // Tracing
	  if(GetL2Proto()->GetTrace() == Trace::ENABLED)
	    {
	      Trace::Instance()->NewNode(pNode);
	      Trace::Instance()->GetTS() << " L2-RA " << p->retx  << " " << p->uid;
	    }     
	  
	  /* the queue can check if it is full and drop the packet */
	  GetL2Proto()->DataRequest(p);

	  lastchanacq = new EthernetEvent(EthernetEvent::CHAN_ACQ);

	  // Schedule the channel acquired event
	  Simulator::instance->Schedule(lastchanacq, 2 * maxwaittime, this);

	  return true;
	}      
      
      if (!GetQueue()->Enque(p))
	{
	  DropPDU d("L2-QD", p);
	  pNode->TracePDU(nil, &d);
	  Stats::pktsDropped++; // Count dropped packets
	  DEBUG0((cout << "Can't enque, Interface " << this << " dropping pkt" << endl));
	  delete p;     // Delete the packet
	  return false; // Dropped
	}
    }
  return true;
}

  
// Handler methods
void InterfaceEthernet::Handle(Event* e, Time_t t)
{ 
  EthernetEvent* ee = (EthernetEvent*)e; // Convert to right event type
//   L2802_3Header* l2pdu = nil;
//   Interface* src_if =  nil;
  switch (ee->event) {
  
  case EthernetEvent::PACKET_RX :
    DEBUG0((cout << "Packet rx Event at Node " << pNode->Id() << ", time " << t << " packet " <<  ee->p->uid << " Euid " << ee->uid << endl));
    //  // notify sender that it has acquired the chanel
    //     l2pdu = (L2802_3Header*)ee->p->PeekPDU(); // Get a ptr to l2 hdr
    //     src_if = ((Ethernet*)GetLink())->GetIfByMac(l2pdu->src);
    //     Simulator::instance->Schedule(new EthernetEvent(EthernetEvent::PACKET_TX), 0.0, src_if);
    InterfaceReal::Handle(e, t);
    ee = nil;    // do not delete the event here
    break;
  case EthernetEvent::NOTIFY :
    InterfaceReal::Handle(e, t);
    ee = nil;   // do not delete the event here
    break;
  case  EthernetEvent::FBR :
    DEBUG0((cout << "FBR Event at Node " << pNode->Id() << " for mac " << ee->mac << ", time " << t << " Euid " << e->uid << endl));
    HandleFBR(ee, t);
    break;
    
  case EthernetEvent::CLR :  // wait for all sending stations to send the clear signal then set the link status to not busy
    DEBUG0((cout << "CLR Event at Node " << pNode->Id() << ", time " << t << endl));
    if (--busycount == 0)
      {
	busycount = 1;
	busyendtime = t;
      }
    break;
  case EthernetEvent::START_TX :  // Retransmit the packet  
    DEBUG0((cout << "START_TX Event at Node " << pNode->Id() << ", time " << t << " destination " << ee->mac << endl));   
    //cout << "Retransmission " << ee->p->retx <<  " for packet " << ee->p->uid << endl;
    if( ++ee->p->retx > ATTEMPT_LIMIT)  // if we have reached the retransmission limit then drop the packet
      {
	DropPDU d("L2-QD", lastpacketsent);
	pNode->TracePDU(nil, &d);
	Stats::pktsDropped++; // Count dropped packets
	DEBUG0((cout << "Rx Limit reached, Interface " << this << " dropping pkt" << endl));	  
	return; // Dropped
      }
    Send(ee->p, ee->mac, ee->type);    
    break;
 
  case EthernetEvent::UPDATE_NODES : // update the list of nodes and their delays
    DEBUG0((cout << "UPDATE_NODES Event at Node " << pNode->Id() << ", time " << t << endl));
    NeighbourList(nodeTimes);
    break;
  case EthernetEvent::CHAN_ACQ :  // We have succesfully sent the packet so reset the contention window 
    DEBUG0((cout << "Channel acquired Event at Node " << pNode->Id() << ", time " << t << endl));
    maxbackoff = INITIAL_BACKOFF;
    break;
  }
  if (ee) delete ee;
}

  
// Private  methods
void  InterfaceEthernet::NeighbourList(NodeTimeVec_t& ntv)
{ // Make a list of all neighbors. 
  Count_t esize  =  ((Ethernet*)GetLink())->Size();

  maxwaittime = 0;
  Time_t temp = 0;

  for (NodeId_t ni = 0; ni < esize; ++ni)
    {
      Interface* iface =  ((Ethernet*)GetLink())->GetIf(ni);
      if (!iface) cout << "HuH?  InterfaceEthernet::NeighborList nill iface for " 
                       << ni << endl;
      else
	{
	  Node* n = iface->GetNode();
	  if(n != pNode) // interface exists and it's node is not this one
	    {
	      temp = TimeDelay(n);
	      ntv.push_back(NodeIfTime(n, iface, temp ) );
	      maxwaittime = maxwaittime > temp ? maxwaittime : temp;
	    }
	}
    }
 
  if(!slottime) slottime = SLOT_TIME / GetLink()->Bandwidth();  // calculate the slot time in seconds
}



Time_t InterfaceEthernet::TimeDelay(Node* n)
{
  return pNode->Distance(n) / SPEED_LIGHT;
}



void InterfaceEthernet::HandleFBR(EthernetEvent* e, Time_t t)
{
  
  if (t < tx_finishtime) // collision occured
    {
      if( t < holdtime ) return; // if we have already experienced collision then do nothing
      
      // cancel the rx event(s)
      if (bCast)
	{
	  for (LinkEventVec_t::size_type i = 0; i < rxEvents.size(); ++i)
	    {
	      Simulator::instance->Cancel(rxEvents.front());
	      rxEvents.erase(rxEvents.begin());
	    }
	  //bCast = false;
	}
      else
	{
	  //cout << "Cancelling the rx event for packet " << lastrxevent->p->uid << endl;
	  //cout << "Cancelling the rx event  " << lastrxevent->uid << " for time " << lastrxevent->Time()<< endl;
	  
	  Simulator::instance->Cancel(lastrxevent);
	}

      DEBUG0((cout << "Collison occured sending clear event to all nodes " << pNode->Id() << ", time " << t << endl));      
      
      // Tracing
      if(GetL2Proto()->GetTrace() == Trace::ENABLED)
	{
	  Trace::Instance()->NewNode(pNode);
	  Trace::Instance()->GetTS() << " L2-C " << lastpacketsent->uid;
	}
      // Canceling the acquire channel event
      Simulator::instance->Cancel(lastchanacq);

      // schedule CLR to all neighbours	
      for(NodeTimeVec_t::size_type  i = 0; i < nodeTimes.size(); ++i)
	{
	  EthernetEvent* evCLR = new EthernetEvent(EthernetEvent::CLR);
	  Simulator::instance->Schedule(evCLR, nodeTimes[i].time , nodeTimes[i].iface);
	}
      
      // set the new value for the backoff timer and schedule sensing the chanell after the backoff
      tx_finishtime -= (lastpacketsent->Size() * 8)  / GetLink()->Bandwidth();
      maxbackoff *= 2;
      
      if(maxbackoff > BACKOFF_LIMIT)   maxbackoff = BACKOFF_LIMIT; 
      
      backofftimer =  ceil(generator->Value() * maxbackoff) * slottime;
       // hold all packets back
      holdtime = tx_finishtime + backofftimer;

      DEBUG0((cout << "Rescheduling packet for transmit at " << holdtime << " for "<< lastdest << endl));
      EthernetEvent* laststevent = new EthernetEvent(EthernetEvent::START_TX, lastpacketsent, lastdest, lasttype);
      Simulator::instance->Schedule(laststevent, holdtime - Simulator::Now(), this);

      // hold all packets back
      holdtime = tx_finishtime + backofftimer;
      return;
    }
                  
  if (t < busyendtime) // external collision
    {	  
      busycount++;	      
      busyendtime = INFINITE_TIME;   // stay busy until you get the Clear event
      return;
    }      
  // Set the interface mode to busy until the other node finishes its transmission
  Time_t txTime = ((Rate_t)e->size * 8) / GetLink()->Bandwidth(); // Time to transmit
  busyendtime = t + txTime; // Note time link will be free
}

   
   
   
