/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ghost Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define class to model "Ghost" network link interfaces
// Ghosts are used in distributed simulatioonsn to maintain
// connectivity information for nodes assigned to other
// processes.

#include <iostream>

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "interface-ghost.h"
#include "link.h"
#include "hex.h"

using namespace std;

Count_t InterfaceGhost::count = 0;

InterfaceGhost::InterfaceGhost(
    const L2Proto& l2, IPAddr_t i,
    Mask_t m, MACAddr mac, bool bootstrap)  
    : Interface(l2, i, m, mac), pLink(0)
{
  count++;
}

InterfaceGhost::~InterfaceGhost()
{
}

bool InterfaceGhost::Send(Packet* p, IPAddr_t i, int type)
{ // Never called for Ghosts
  return false;
}

// Send without ARP
bool InterfaceGhost::Send(Packet* p, const MACAddr& dst, int type)
{ // Never called for Ghosts
  return false;
}

bool InterfaceGhost::Send(Packet* p, Word_t type)
{ // Never called for Ghosts
  return false;
}

// Handler methods
void InterfaceGhost::Handle(Event* e, Time_t t)
{ // Never called for Ghosts
}

void InterfaceGhost::HandleLLCSNAP(Packet* p)
{
}

// Notifier
void InterfaceGhost::Notify(void* v)
{ // Never called for  ghosts
}

Count_t  InterfaceGhost::PeerCount() const
{ // Number of peers
  return pLink->PeerCount();
}

IPAddr   InterfaceGhost::PeerIP(int npeer) const
{ // Get peer IP addr (if known)
  return pLink->PeerIP(npeer);
}

IPAddr   InterfaceGhost::NodePeerIP(Node* n) const
{ // Get peer IP addr (if known)
  return pLink->NodePeerIP(n);
}

bool     InterfaceGhost::NodeIsPeer(Node* n) const
{ // Find out if specified node is a peer
  if (!pLink) return false;
  return pLink->NodeIsPeer(n);
}

Count_t  InterfaceGhost::NodePeerIndex(Node* n) const
{ // Find out peer index (for NixVector Routing)
  return pLink->NodePeerIndex(n);
}

Link* InterfaceGhost::GetLink() const
{
  return pLink;
}

void InterfaceGhost::SetLink(Link* pl)
{
  pLink = pl;
}

Queue* InterfaceGhost::GetQueue() const
{
  return nil;
}
  
void InterfaceGhost::SetQueue(const Queue&)
{
}

bool InterfaceGhost::QueueDetailed() const
{
  return false;
}

L2Proto* InterfaceGhost::GetL2Proto() const
{
  return nil;
}
  
void InterfaceGhost::SetL2Proto(L2Proto*)
{
}

void InterfaceGhost::AddNotify(NotifyHandler*, void*)
{ // Add a notify client
}

void InterfaceGhost::CancelNotify(NotifyHandler*)
{ // Cancel a notification
}

MACAddr InterfaceGhost::IPToMac(IPAddr_t i) const
{ // Never called for Ghosts
  return MACAddr::NONE;
}

Count_t InterfaceGhost::NeighborCount(Node* n) const
{ // Get number of routing neighbor by asking the link
  // The links do this in different ways, so we ask
  if(pLink) return pLink->NeighborCount(n);
  return 0;
}

void  InterfaceGhost::Neighbors(NodeWeightVec_t& nwv, bool forceAll)
{ // Add to neighbor list by asking the link
  if (pLink) pLink->Neighbors(nwv, this, forceAll);
}

IPAddr_t InterfaceGhost::DefaultPeer()
{
  if (pLink) return pLink->DefaultPeer(this);
  return IPADDR_NONE;
}

bool InterfaceGhost::IsLocalIP(IPAddr_t i) const
{ // Determine if specified is on this subnetwork
  DEBUG0((cout << "Checking local ip " << Hex8(i)
          << " ip " << Hex8(ip)
          << " mask " << Hex8(mask) << endl));
  if ((ip & mask) == (i & mask)) return true;
  return false;
}

Mult_t InterfaceGhost::TrafficIntensity() const
{ // Never called for Ghosts
  return 0.0;
}

// Distributed Simulation Routines
void InterfaceGhost::AddRemoteIP(IPAddr_t, Mask_t, Count_t)
{ // Never called for Ghosts
}



