/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define class to model network link interfaces
// Implements the Layer 2 protocols

#include <iostream>

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "hex.h"
#include "interface-real.h"
#include "link.h"
#include "queue.h"
#include "droppdu.h"
#include "node.h"
#include "trace.h"
#include "globalstats.h"
#include "ipv4.h"
#include "llcsnap.h"
#include "ratetimeparse.h"
#ifdef ETRACE_DEBUG
#include "etrace.h"
#endif

#define DEFAULT_ARP_TIMEOUT 1200  // 20 min

using namespace std;

Count_t InterfaceReal::count = 0;

InterfaceReal::InterfaceReal(const L2Proto& l2, IPAddr_t i,
                             Mask_t m, MACAddr mac, bool bootstrap)  
    : Interface(l2, i, m, mac),
      pLink(0), pQueue(Queue::Default()->Copy()),
      notifEvent(nil),
      totBytesSent(0), pL2Proto(l2.Copy()), defaultPeer(IPADDR_NONE),
      arptimeout(DEFAULT_ARP_TIMEOUT)
{
  pQueue->SetInterface(this); // Some queues need to know attached i/f
  pL2Proto->SetInterface(this); // this l2proto obj is attached to an i/f
  count++;
 if (usearp)
    {
      arpmodule = new ARP();
      arpmodule->SetARPTimeout(arptimeout);
      arpmodule->SetTrace(Trace::ENABLED);
    }
}

InterfaceReal::~InterfaceReal()
{
  delete arpmodule;
}

void InterfaceReal::SetWormContainment(WormContainment* cont)
{
  usewormcontainment = true;
  wormcontainment = cont;
}


bool InterfaceReal::Send(Packet* p, IPAddr_t i, int type)
{
   if(usewormcontainment)
    {
      wormcontainment->ProcessOutPacket(p, i, type, this);
      return true;
    }
   return Xmit(p, i, type);
}

bool InterfaceReal::Xmit(Packet* p, IPAddr_t i, int type)
{
  // Make arp requests for the respective MAC addresses
  if(Interface::usearp)
    {
      arpmodule->ARPRequest(this, p, type);
      return true;
    }
  MACAddr dst = IPToMac(i);
  return Send(p, dst, type);
}


// Send without ARP
bool InterfaceReal::Send(Packet* p, const MACAddr& dst, int type)
{
  if (IsDown())
    { // Interface is down, don't send.  Just drop packet
      DropPDU d("L2-ID", p);
      pNode->TracePDU(nil, &d);
      Stats::pktsDropped++; // Count dropped packets
      delete p;
      return false;
    }
      
  LLCSNAPHeader* llc = new LLCSNAPHeader(type);
  p->PushPDU(llc);

  pL2Proto->BuildDataPDU(macAddr, dst, p);
  PDU* l2pdu = p->PeekPDU();
  if (Trace::Enabled()) GetNode()->TracePDU(pL2Proto, l2pdu, p, "-");

  /* If link is idle (i.e, CSMA) else queue */
  if (!pL2Proto->Busy() && !pQueue->CheckForcedLoss(false))
    { // Link is free, and no forced losses presently for the queue
      // However, we need to check for spoofed loss
      bool spoofLoss = pQueue->CheckSpoofedSource(p);
      if (spoofLoss)
        {
          DEBUG0((cout << "Interface dropping spoofed packet" << endl));
          DropPDU d("L2-FD", p); // Filter drop
          pNode->TracePDU(nil, &d);
          Stats::pktsDropped++; // Count dropped packets
          DEBUG0((cout << "Interface " << this
                  << " dropping pkt due to spoofing detected" << endl));
          if (p->notification)
            { // Sender has requested notification, so we need to
              // schedule one.
              // hack for now just to 1ms...need to calculate queue drain time
              Simulator::instance->AddNotify(
                  p->notification, Time("1ms"), nil);
            }
          delete p;     // Delete the packet
          return false; // Dropped
        }
#ifdef OLD_DONT_NEED
      if (!pQueue->Enque(p)) return false;
      Packet* p1 = pQueue->Deque();
      if (p1 != p)
        {
          cout << "HuH?  Misbehaving queue on non-busy link" << endl;
        }
#endif
      // Count as enque/deque to get proper queue statistics
      Stats::pktsEnqued++;
      Stats::pktsDequed++;
      // IPV4Header*      iphdr = (IPV4Header*)p->PeekPDU(2);
//       cout << "Send:: Packet from " << IPAddr::ToDotted(iphdr->src);
//       cout << " To " <<  IPAddr::ToDotted(iphdr->dst) ;
//       cout << " my IP is " <<IPAddr::ToDotted( GetIPAddr()) << endl;

      //if (notify) pLink->AddNotify(this, NULL); // ?? Is this right??
      /* the queue can check if it is full and drop the packet */
      pL2Proto->DataRequest(p);


      return true;
    }
  
  // Add a notification if none pending
  //if(pQueue->LengthPkts() == 0)	pLink->AddNotify(this, NULL);
  //pLink->AddNotify(this, NULL);
  // Can't transmit now, enque for later
  if (!pQueue->Enque(p))
    {
      DropPDU d("L2-QD", p);
      pNode->TracePDU(nil, &d);
      Stats::pktsDropped++; // Count dropped packets
      DEBUG0((cout << "Interface " << this << " dropping pkt" << endl));
      if (p->notification)
        { // Sender has requested notification, so we need to
          // schedule one.
          // hack for now just to 1ms...need to calculate queue drain time
          DEBUG(1,(cout << "Adding notif for " << p->notification << endl));
          Simulator::instance->AddNotify(p->notification, Time("1ms"), nil);
        }
      delete p;     // Delete the packet
      return false; // Dropped
    }
  return true;
}

// this is a broadcast packet just send it to all peers
bool InterfaceReal::Send(Packet* p, Word_t type)
{
  return Send(p, MACAddr(0xffff, 0xffffffff), type);
}

// Handler methods
void InterfaceReal::Handle(Event* e, Time_t t)
{
  DEBUG0((cout << "InterfaceReal::Handle, time " << t << endl));
  LinkEvent* le = (LinkEvent*)e; // Convert to right event type
  IPV4Header* iphdr;

  Size_t sz = 0;
  switch (le->event) {
  case  LinkEvent::PACKET_RX:
    if (le->fifoEvent)
    { // Should be head of fifo event queue
#undef  EXTRA_CHECKS
#ifdef  EXTRA_CHECKS
      if (evList->empty())
      {
        cout << "HuH?  Interface got fifo event on empty evList" << endl;
        exit(1);
      }
      EventPair* ep = evList->front();
      if (ep->event != le)
      { 
        cout << "HuH?  Interface got fifo event mismatch" << endl;
        cout << "eventcq " << evList << endl;
        cout << "le " << le << " ep.second " << ep->event << endl;
        exit(1);
      }
#endif
      evList->pop_front(); // Remove from fifo event list
      if (!evList->empty())
      { // Need to schedule the next simulator event
        EventPair* ep1 = evList->front();
        Time_t when = ep1->time - Simulator::Now();
        DEBUG0((cout << "Scheduling evlist front " << ep1->event
                     << " ev " << ep1->event->event << endl));
        Simulator::instance->ScheduleEarly(ep1->event, when, this);
      }
    }
    Trace::Instance()->AppendEOL(); // End of log line
    DEBUG0((cout << "Interface " << this 
                 << " Handle node " << pNode->Id() << endl));

    if (IsDown())
    { // Link is down, just drop packet and return
      DropPDU d("L2-ID", le->p);
      pNode->TracePDU(nil, &d);
      Stats::pktsDropped++;
      delete le->p;
      le->p = nil;
      break;  //------->>>>>>>>>
    }

    if(pNode->IsSwitchNode())  // if node is switch
    {
      // Pass the packet to the switch
      pNode->PacketRX(le->p, this);
      break;
    }
    // testing
    iphdr = (IPV4Header*)le->p->PeekPDU(2);
  //   cout << "HANDLE:: Packet from " << IPAddr::ToDotted(iphdr->src);
//     cout << " To " <<  IPAddr::ToDotted(iphdr->dst) ;
//     cout << " my IP is " <<IPAddr::ToDotted( ip) << endl;
//      if(iphdr->protocol == 1)
//       {
// 	cout <<"ICMP packet from " << IPAddr::ToDotted(iphdr->src);
// 	cout << " To " <<  IPAddr::ToDotted(iphdr->dst) << endl;
//       }
    
    PacketRxStart(le);
    sz = le->p->Size();
    if (le->hasBitError)
    { // just drop the packet due to bit error
      DropPDU d("L2-BER", le->p);
      pNode->TracePDU(nil, &d);
      DEBUG(0,(cout << "Packet Dropped due to BER" << endl));
      Stats::pktsDropped++;
      delete le->p;
    }
    else
    {
      //	cout << "port matching = " << useportmatching << endl;
      if(usewormcontainment)
      {
        wormcontainment->ProcessInPacket(le->p, this);
        break;
      }
#ifdef ETRACE_DEBUG
      // Trace "R" event (successful reception)
      PDU *pdu_ = le->p->FindPDU(3,0x800); // IPv4 PDU (if it exists)
      if (pdu_) {
        if ( (((IPV4Header *)pdu_)->protocol == 89) || 
             (((IPV4Header *)pdu_)->protocol == 17) ) {
          ETrace::Instance()->TracePacket(le->p, this->GetNode(), this, 1, "R");
        }
      }
#endif
      GetL2Proto()->DataIndication(le->p);
    }
    PacketRxEnd(le, sz);
    break;
  case LinkEvent::NOTIFY :
    DEBUG(0,(cout << "Got notify event" << endl));
    // Inform the link object that  link is free
    CallNotification(true);
    le = nil; // Not deleted here, since likely reused
    break;
  }
  if (le) delete le;
}

void InterfaceReal::HandleLLCSNAP(Packet* p)
{
  /* Time for the LLC/SNAP*/
  LLCSNAPHeader* llcsnap = (LLCSNAPHeader*)p->PopPDU(); 
  DEBUG(2,(cout << "LLCSnap ind Layer " << llcsnap->Layer()
           << " version " << llcsnap->Version()
           << endl));
  
  switch(llcsnap->snap_ethtype) 
    {
    case 0x0800:
      IPV4::Instance()->DataIndication(this, p);
      break;
    case 0x806:
      arpmodule->DataIndication(this,p);
      break;
    default:
      // Need to add trace file entry for packet drop
      cout << "iface::HandleLLC unknown ethtype "
           <<  llcsnap->snap_ethtype << endl;
      break;
    }
}

// Notifier
void InterfaceReal::Notify(void* v)
{ // Link has finished previous transmit
  if (v) delete (Time_t*)v;
  // if the layer 2 mac trasmits multiple packets for a single transmission ??
  if (pL2Proto->Busy()) return;
  Packet* p = pQueue->Deque();
  if (p)
    { // Packet exists
      pL2Proto->DataRequest(p);
    }
  CallNotification(false); // Send possible queue space available notification
}

Count_t  InterfaceReal::PeerCount() const
{ // Number of peers
  return pLink->PeerCount();
}

IPAddr   InterfaceReal::PeerIP(int npeer) const
{ // Get peer IP addr (if known)
  return pLink->PeerIP(npeer);
}

IPAddr   InterfaceReal::NodePeerIP(Node* n) const
{ // Get peer IP addr (if known)
  return pLink->NodePeerIP(n);
}

bool     InterfaceReal::NodeIsPeer(Node* n) const
{ // Find out if specified node is a peer
  if (!pLink) return false;
  return pLink->NodeIsPeer(n);
}

Count_t  InterfaceReal::NodePeerIndex(Node* n) const
{ // Find out peer index (for NixVector Routing)
  return pLink->NodePeerIndex(n);
}

Link* InterfaceReal::GetLink() const
{
  return pLink;
}

void InterfaceReal::SetLink(Link* pl)
{
  pLink = pl;
}

Queue* InterfaceReal::GetQueue() const
{
  return pQueue;
}
  
void InterfaceReal::SetQueue(const Queue& q)
{
  if (pQueue) delete pQueue; // Delete existing
  pQueue = q.Copy();         // Make a copy to use
  pQueue->SetInterface(this);
}

bool InterfaceReal::QueueDetailed() const
{
  if (pQueue) return pQueue->Detailed();
  return false;
}

void InterfaceReal::AddNotify(NotifyHandler* n,void* v)
{ // Add a notify client
  DEBUG(1,(cout << "IFReal " << this << " AddNotify handler " << n << endl));
  notifications.push_back(Notification(n,v));
  if (!notifEvent && !pQueue->Detailed())
    { // No pending notification event, need to schedule one
      Time_t qd = pQueue->QueuingDelay(); // Time to drain queue
      notifEvent = new LinkEvent(LinkEvent::NOTIFY);
      Simulator::instance->Schedule(notifEvent, qd, this);
    }
}

void InterfaceReal::CancelNotify(NotifyHandler* n)
{ // Cancel a notification
  DEBUG(1,(cout << "iface " << this 
           << " Node " << pNode
           << " CancelNotify handler " << n << endl));
  for (NList_t::iterator i = notifications.begin();
       i != notifications.end(); ++i)
    {
      if (i->handler == n)
        { // Found it
          DEBUG(1,(cout << "iface " << this << " found notif " << n << endl));
          i->handler = nil;
          // return; // Keep looking in case more than one
        }
    }
}

MACAddr InterfaceReal::IPToMac(IPAddr_t i) const
{ // Given a PEER Ip Address, find corresponding MAC Address
  // Indirects to the Link to find this, since different link
  // types do this differently
  if (pLink) return pLink->IPToMac(i);
  return MACAddr::NONE; // If no link
}

Count_t InterfaceReal::NeighborCount(Node* n) const
{ // Get number of routing neighbor by asking the link
  // The links do this in different ways, so we ask
  if(pLink) return pLink->NeighborCount(n);
  return 0;
}

void  InterfaceReal::Neighbors(NodeWeightVec_t& nwv, bool forceAll)
{ // Add to neighbor list by asking the link
  if (pLink) pLink->Neighbors(nwv, this, forceAll);
}

IPAddr_t InterfaceReal::DefaultPeer()
{
  if (defaultPeer != IPADDR_NONE)
    {
      return defaultPeer; // Cached from prior call
    }
  if (pLink) defaultPeer = pLink->DefaultPeer(this);
  return defaultPeer;
}

bool InterfaceReal::IsLocalIP(IPAddr_t i) const
{ // Determine if specified is on this subnetwork
  DEBUG0((cout << "Checking local ip " << Hex8(i)
          << " ip " << Hex8(ip)
          << " mask " << Hex8(mask) << endl));
  if ((ip & mask) == (i & mask)) return true;
  // Check remote IP's (used only for distributed simulations)
  for (IPMaskVec_t::size_type j = 0; j < remoteIPList.size(); ++j)
    {
      const IpMaskCount& ipm = remoteIPList[j];
      if ((ipm.ip & ipm.mask) == (i & ipm.mask)) return true;
    }
  return false;
}

Mult_t InterfaceReal::TrafficIntensity() const
{ // Computed offered load on this link, as fraction of total capacity
  DCount_t maxBytes = Simulator::Now() * pLink->Bandwidth() / 8;
  return totBytesSent / maxBytes;
}

// Distributed Simulation Routines
void InterfaceReal::AddRemoteIP(IPAddr_t i, Mask_t m, Count_t c)
{ // List of remote IP's reached via this interface
  remoteIPList.push_back(IpMaskCount(i, m, c));
}

// Protected methods
void InterfaceReal::PacketRxStart(LinkEvent*)
{
}

void InterfaceReal::PacketRxEnd(LinkEvent*, Size_t)
{
}

// Private  methods
void InterfaceReal::CallNotification(bool resched)
{ // Notify the next non-nil entry in the notification list
  while(1)
    {
      if (notifications.empty())
        {
          if (notifEvent && resched)
            { // Delete pending event
              delete notifEvent;
              notifEvent = nil;
            }
          return;       // No pending
        }
      Notification n = notifications.front();  // Get next notifier
      if (n.handler)
        { // Handler exists
          notifications.pop_front();           // And remove it
          DEBUG(1,(cout << "IFace " << this
                   << " Node " << pNode
                   << " Notifying " << n.handler << endl));
          n.handler->Notify(n.userData);       // Call the notification handler
          break;
        }
      else
        {
          notifications.pop_front();           // Remove null handler, continue
        }
    }
  // If any notifications remaining, need to reschedule notification timer
  if (resched)
    { // Need to schedule another
      Time_t qd = pQueue->QueuingDelay(); // Time to drain queue
      if (!notifEvent)
        notifEvent = new LinkEvent(LinkEvent::NOTIFY);
      Simulator::instance->Schedule(notifEvent, qd, this);
    }
}
