/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define class to model network link interfaces
// Implements the Layer 2 protocols

#ifndef __interface_real_h__
#define __interface_real_h__

#include "common-defs.h"
#include "simulator.h"
#include "protocol.h"
#include "packet.h"
#include "handler.h"
#include "ipaddr.h"
#include "notifier.h"
#include "macaddr.h"
#include "node.h"
#include "link.h"
#include "link-real.h"
#include "eventcq.h"
#include "interface.h"
#include "arp.h"
#include "worm-containment.h"

class L2Proto;
class Node;
class Queue;
class LinkEvent;
class NotifyHandler;

#define CLASS_C_MASK 0xffffff00
#define CLASS_B_MASK 0xffff0000
#define CLASS_A_MASK 0xff000000
#define MASK_NONE    0x00000000

//Doc:ClassXRef
class IpMaskCount { // IP Address, mask, and hop count
public:
  IpMaskCount(IPAddr_t i, Mask_t m, Count_t c = 0) 
      : ip(i), mask(m), hopCount(c) { }
public:
  IPAddr_t ip;
  Mask_t   mask;
  Count_t  hopCount;
};

typedef std::vector<IpMaskCount> IPMaskCountVec_t;

//Doc:ClassXRef
class InterfaceReal : public Interface
{
  //Doc:Class Interface objects are assigned to nodes, and have
  //Doc:Class corresponding link objects to represent the communications
  //Doc:Class links.  Interfaces also have queues  for queuing packets
  //Doc:Class when a transmit is requested when the link is busy,
  //Doc:Class and an associated layer 2 protocol objects
  //Doc:Class Interfaces are typically not allocated directly by
  //Doc:Class a \GTNS\ simulation program, but rather are created
  //Doc:Class indirectly by {\tt AddDuplexLink} and {\tt AddSimplexLink}
  //Doc:Class methods of {\tt Node} objects.

public:
  //Doc:Method
  InterfaceReal(const L2Proto& l2 = L2Proto802_3(),
		IPAddr_t i   = IPADDR_NONE,
		Mask_t   m   = MASK_ALL,
		MACAddr  mac = MACAddr::NONE,
		bool bootstrap = false);
    //Doc:Desc Constructs a new real interface object.
    //Doc:Arg1 The layer 2 protocol type to associate with this interface.
    //Doc:Arg2 The \IPA\ to assign to this interface.
    //Doc:Arg3 The address mask to assign to this interface.
    //Doc:Arg4 The {\tt MAC} address to assign to this interface.

  virtual ~InterfaceReal();

  void SetARPTimeout(Time_t t){ arptimeout = t;}


  //Doc:Method
  bool Send(Packet*, IPAddr_t, int); // Enque and transmit
    //Doc:Desc Send a packet on this interface.
    //Doc:Arg1 Packet to send.
    //Doc:Arg2 \IPA\ of destination
    //Doc:Arg3 LLC/SNAP type
    //Doc:Return True if packet sent or enqued, false if dropped.

 //Doc:Method
  bool Xmit(Packet*, IPAddr_t, int); // Enque and transmit
    //Doc:Desc Send a packet on this interface.
    //Doc:Arg1 Packet to send.
    //Doc:Arg2 \IPA\ of destination
    //Doc:Arg3 LLC/SNAP type
    //Doc:Return True if packet sent or enqued, false if dropped.

  //Doc:Method
  bool Send(Packet*, const MACAddr&, int); // Enque and transmit
    //Doc:Desc Send a packet on this interface.
    //Doc:Arg1 Packet to send.
    //Doc:Arg2 MAC address of destination
    //Doc:Arg3 LLC/SNAP type
    //Doc:Return True if packet sent or enqued, false if dropped.

  //Doc:Method
  bool Send(Packet*, Word_t); // Enque and transmit
    //Doc:Desc Send a broadcast packet on this interface.
    //Doc:Arg1 Packet to send.
    //Doc:Arg2 MAC address of destination
    //Doc:Arg3 LLC/SNAP type
    //Doc:Return True if packet sent or enqued, false if dropped.

  // Handler
  virtual void Handle(Event*, Time_t);
  // Notifier
  virtual void Notify(void*);
  
  // Interface Methods
  //Doc:Method
  Count_t      PeerCount() const;          // Number of peers
    //Doc:Desc Determine how many peer nodes  are associated with this 
    //Doc:Desc interface.
    //Doc:Return Count of peers.

  //Doc:Method
  IPAddr       PeerIP(int npeer) const;    // Get peer IP addr if known
    //Doc:Desc Get the \IPA\ of the specified peer, by peer index.
    //Doc:Arg1 The peer index of the desired peer.
    //Doc:Return The \IPA\ of the specified peer.
    //Doc:Return Returns {\tt IPADDR_NONE} if the peer does not exist.

  //Doc:Method
  IPAddr       NodePeerIP(Node*) const;    // Get peer IP addr if known
    //Doc:Desc Get the \IPA\ of the peer's interface directly connected
    //Doc:Desc to this interface.
    //Doc:Arg1 Node pointer for the desired peer.
    //Doc:Return \IPA\ of the specified peer.
    //Doc:Return Returns {\tt IPADDR_NONE} if the peer does not exist.

  //Doc:Method
  bool         NodeIsPeer(Node*) const;    // Find out if node is a peer
    //Doc:Desc Determine if the specified node is a directly connected
    //Doc:Desc peer on this interface.
    //Doc:Arg1 Node pointer of the desired peer.
    //Doc:Return True if the node is a directly connected peer on this
    //Doc:Return interface.

  //Doc:Method
  Count_t      NodePeerIndex(Node*) const; // Get peer index for node
    //Doc:Desc Get the peer index for the specified node.
    //Doc:Arg1 Node pointer for the desired peer.
    //Doc:Return The peer index of the specified node.

  //Doc:Method
  Link*        GetLink() const;            // Return the link pointer
    //Doc:Desc Get the link object associated with this interface.
    //Doc:Return A pointer to the associated link object.

  //Doc:Method
  void         SetLink(Link*);
    //Doc:Desc Set the associated link to the specified link pointer.
    //Doc:Arg1 Link pointer to set.

  //Doc:Method
  Queue*       GetQueue() const;           // Return the queue pointer
    //Doc:Desc Get the packet queue associated with this interface.
    //Doc:Return A pointer to the associated packet queue.

  //Doc:Method
  void         SetQueue(const Queue& q);          // Set a new queue object
    //Doc:Desc Set a new packet queue object.
    //Doc:Arg1 The type of packet queue  object to use.  This object
    //Doc:Arg1 is copied to this interface, so an anonymous temporary
    //Doc:Arg1 object can be used, or the same object can be passed
    //Doc:Arg1 to multiple interfaces.

  bool         QueueDetailed() const;             // True if queue is detailed

  //Doc:Method
  L2Proto*     GetL2Proto() const { return pL2Proto;} // Layer 2 proto object
    //Doc:Desc Get the layer 2 protocol object associated with this interface.
    //Doc:Return A pointer to the layer 2 protocol object associated with this 
    //Doc:Return interface.

  //Doc:Method
  void         SetL2Proto(L2Proto* l2) { pL2Proto = l2;}
    //Doc:Desc Set a new layer 2 protocol object for this interface.
    //Doc:Return The new layer 2 protocol object to use.

  //Doc:Method
  void         AddNotify(NotifyHandler*,void*);   // Add a notify client
    //Doc:Desc Add a notification client.  The client is notified when
    //Doc:Desc the associated link is free.
    //Doc:Arg1 A pointer to the notify client.
    //Doc:Arg2 A blind pointer to pass back to the notify client.

  //Doc:Method
  void         CancelNotify(NotifyHandler*);      // Cancel pending notif.
    //Doc:Desc Cancel an existing notification request.
    //Doc:Arg1 A pointer to the notifi client to cancel.

  //Doc:Method
  MACAddr      IPToMac(IPAddr_t) const;           // Return mac addr for ip
    //Doc:Desc Determine the {\em MAC} address of the interface
    //Doc:Desc associated with a peer, specified by \IPA.
    //Doc:Arg1 The \IPA\ to search for.
    //Doc:Return The {\em MAC} address of the remote interface associated
    //Doc:Return with the specified \IPA.

  //Doc:Method
  Count_t      NeighborCount(Node*) const;        // Number of rtg neighbors
    //Doc:Desc Get the count of directly connected neighbors on this interface.
    //Doc:Arg1 Node querying neighbors (significant for Ethernet links)
    //Doc:Return Count of directly connect neighbors.

  // Get neighbor list    
  //Doc:Method
  void         Neighbors(NodeWeightVec_t&, bool forceAll = false);
    //Doc:Desc Get a list of directly connected neighbors on this interface.
    //Doc:Arg1 An output parameter, with a vector of nodes and associated
    //Doc:Arg1 link weights directly connected on this link.
    //Doc:Arg2 ForceAll indicates that all neighbors  should be reported.
    //Doc:Arg2 Some link types (ie. {\tt Ethernet} sometimes report
    //Doc:Arg2 zero neighbors, for efficiency in route computations.

  //Doc:Method
  IPAddr_t     DefaultPeer();              // Get default peer from link
    //Doc:Desc Get the \IPA\ of the default peer.
    //Doc:Return The \IPA\ of the default peer.

  //Doc:Method
  bool         IsLocalIP(IPAddr_t) const;  // True if IP is on this i/f
    //Doc:Desc Determine if the specified \IPA\ is a node directly connected
    //Doc:Desc to this interface.
    //Doc:Arg1 \IPA\ to check.
    //Doc:Return True if directly connected \IPA, false if not.

  //Doc:Method
  Mult_t       TrafficIntensity() const;   // TotBytesSent/LineSpeed
    //Doc:Desc Return the traffic intensity on this interface.  Calculated
    //Doc:Desc as the total bytes sent per second divided by the link
    //Doc:Desc bandwidth.
    //Doc:Return The traffic  intensity for this interface.

  // Used by distributed simulation
  //Doc:Method
  void         AddRemoteIP(IPAddr_t, Mask_t, Count_t = 0);// Add to rmt IP list
    //Doc:Desc Add an \IPA\ and mask to the list of the \IPA{s} that
    //Doc:Desc can be reached via the remote link associated with this
    //Doc:Desc interface.
    //Doc:Desc Useful only for distributed simulations.
    //Doc:Arg1 The \IPA\ to add.
    //Doc:Arg2 The address mask to add.
    //Doc:Arg3 Hop count to the destination set

  //Doc:Method
  void         HandleLLCSNAP(Packet *p);
    //Doc:Desc Process the LLCSnap header on a wireless packet.
    //Doc:Arg1 Pointer with packet with LLCSnap header to process.

  //Doc:Method
  virtual bool     IsWireless() const { return false;}// True if wireless interface
  //Doc:Desc Test if this interface is wireless 
  //Doc:Return Always false for InterfaceReal nodes
  
  //Doc:Method
  virtual bool     IsReal() const { return true;}     // True if wireless interface
  //Doc:Desc Test if this interface is real 
  //Doc:Return Always true for InterfaceReal interfaces
  
  virtual bool     IsEthernet() const {return false;}               // True if ethernet interface
  
  void SetWormContainment(WormContainment*);
  WormContainment* GetWormContainment() { return wormcontainment;}

protected:
  // PacketRxStart is called by the event handler for the real interface
  // before calling the l2 data indication.  This
  // which allows the wired and wireless interface implementations
  // to react differently.
  virtual void        PacketRxStart(LinkEvent*);
  // PacketRxDone is called by the event handler for the real interface
  // after calling the l2 data indication.
  virtual void        PacketRxEnd(LinkEvent*, Size_t);

  // Make all data items private, since we need to access via
  // methods, to facilitate real/ghost interface implementation

private:
  Link*            pLink;           // Associated link
  Queue*           pQueue;          // Associated Queue
  NList_t          notifications;   // Deque of pending notifications
  LinkEvent*       notifEvent;      // Pending notification event
  DCount_t         totBytesSent;    // Track total bytes requested to send
  L2Proto*         pL2Proto;        // Associated layer 2 protocol
  IPAddr_t         defaultPeer;     // Used to cache defaultPeer IPAddr
  ARP*             arpmodule;
  Time_t           arptimeout;      // timeout for the arp entries
  WormContainment* wormcontainment; // The worm containment module

 private:
  // The following is for the distibuted version.  List IP addresses 
  // (with masks) that are reached via this interface.  The Ghost Node
  // approach (when implemented) will obviate the need for this
  IPMaskCountVec_t remoteIPList;// List if remote IP/Masks reached via this i/f
private:
  void        CallNotification(bool); // Notify l4 proto of free buffer space
public:
  static Count_t count;    // debug
};

#endif
