/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define class to model wireless network link interfaces

#include <iostream>
//#define DEBUG_MASK 0x01
#include "debug.h"
#include "interface-wireless.h"
#ifdef SIM //ESTAT
#include "estat.h"
#endif //SIM
#include "qtwindow.h"
#include "rng.h"

using namespace std;

// Set the default radio range for all interfaces
Meters_t InterfaceWireless::defaultRadioRange = 100.0;

InterfaceWireless::InterfaceWireless(const L2Proto& l2,
                                     IPAddr_t i,
                                     Mask_t   m,
                                     MACAddr  mac,
				     bool     bootstrap_)
    : InterfaceReal(l2, i, m, mac),
      linkFreeTime(0), evRx(nil), pr(0), linkState(IDLE),
      txPower(0.0), rxPower(0.0), radioRange(defaultRadioRange),
      opmode(ADHOC), rtsTxPow(0), ctsTxPow(0), ackTxPow(0)
{
#ifdef BUGFIX //insure
	last_time = Simulator::Now();
#endif //BUGFIX
  bootstrap = bootstrap_;
  if (bootstrap == true)
    {
      GetL2Proto()->Bootstrap();
    }
  DEBUG0((cout << " Wireless Interface Constructor" << this <<endl));
}

// Notifier
void InterfaceWireless::Notify(void* v)
{ // Link has finished previous transmit
  if (Simulator::Now() >= linkFreeTime)
  {
    linkState = IDLE;
  }
  else
  {
    linkState = CX;
    cout << "HuH? collision while xmitting\n";
  }
  // yj, this is a PHY-TXEND.indication for 802.11
  if (linkState == IDLE)
  {
    PhyInd_t phyInd = PHY_TXEND;
    GetL2Proto()->Notify(&phyInd);
#ifdef HAVE_QT
    // Animate the end of transmission
    if (Simulator::instance->AnimateWirelessTx())
    {
      QTWindow* qtw = Simulator::instance->GetQTWindow();
      qtw->WirelessTxEnd(GetNode(), this, nil, GetNode()->GetRadioRange());
    }
#endif
  }
  // Call parent (wired) notifier to handle the notifications
  InterfaceReal::Notify(v);
}

void InterfaceWireless::PacketRxStart(LinkEvent* le)
{
  if (Simulator::Now() >= linkFreeTime)
    linkState = IDLE;    // reset linkState after rcv is complete

  // Check for the 1st scheduled pkt
  if (evRx && evRx->uid == le->uid)
    evRx = nil;
}

void InterfaceWireless::PacketRxEnd(LinkEvent* le, Size_t sz)
{
  // yj, this is a PHY-RXEND.indication for 802.11
  if (linkState == IDLE)
    {
      PhyInd_t phyInd = le->hasBitError ? PHY_RXEND_ERR : PHY_RXEND;
      GetL2Proto()->Notify(&phyInd);
      // Get size of the packet and update energy levels
      double time = sz/GetLink()->Bandwidth();
      double delta_time = Simulator::Now() - last_time ;
      Node* node = GetNode();
      node->setBattery( node->getBattery() - (time * getRxPower() +
                        (delta_time * node->getComputePower() )) ) ;     
      last_time = Simulator::Now();

//
// if there was a bit error, we don't have a complete packet
// and cannot reliably tell if it was an OSPF packet or not.
// 
      if(le->hasBitError) {
	    Stats::ospfRxErrors++;
        }
     }
}

// Static methods
void InterfaceWireless::SetDefaultRadioRange(Meters_t r)
{
  defaultRadioRange = r;
}


void InterfaceWireless::setOpMode(Opmode_t opmode_)
{
  opmode = opmode_;
  if (bootstrap == true)
    {
      GetL2Proto()->Bootstrap();
    }
}
