/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - IPAddress Class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __ipaddr_h__
#define __ipaddr_h__

#include <string>
#include <algorithm>

#include "common-defs.h"

//Doc:ClassXRef
class IPAddr {
  //Doc:Class Defines an object that specifies an \IPA\ using the familiar
  //Doc:Class "dotted" location.  Objects of class {\tt IPAddr} have
  //Doc:Class an automatic  typecast to type {\tt IPAddr\_t}, so an
  //Doc:Class {\tt IPAddr} object can be used whenever a variable of
  //Doc:Class {\tt IPAddr\_t} is needed.
public:
  //Doc:Method
  IPAddr() : ip(0) { }
    //Doc:Desc Constructor for {\tt IPAddr}.  The \IPA\ defaults to
    //Doc:Desc \IPAN\ (0).

  //Doc:Method
  IPAddr(std::string&);  // Construct from dotted notation
    //Doc:Desc Construct an {\tt IPAddr} object using a string in the
    //Doc:Desc familiar "dotted" notation.
    //Doc:Arg1 String value specifying the desired \IPA\ in dotted notation.

  //Doc:Method
  IPAddr(IPAddr_t i) : ip(i) { }
    //Doc:Desc Construct an {\tt IPAddr} object with the specified
    //Doc:Desc 32 bit address.
    //Doc:Arg1 \IPA\ as a 32 bit unsigned quantity.

  IPAddr(const char*);    // Construct from char to string
         operator IPAddr_t() const { return ip;}  // Typecast to unsigned long
         operator std::string() const;            // Typecast to a string
  bool   operator==(const IPAddr& i) const { return ip == i.ip;}
  bool   operator< (const IPAddr& i) const { return ip < i.ip;}

  //Doc:Method
  static IPAddr_t ToIP(const std::string);
    //Doc:Desc Convert a string (dotted notation) \IPA\ to 32 bit unsigned.
    //Doc:Arg1 String to convert.
    //Doc:Return 32 bit unsigned \IPA.

  //Doc:Method
  static char*    ToDotted(IPAddr_t);
    //Doc:Desc Convert a 32 bit \IPA\ to a string in dotted notation.
    //Doc:Desc \IPA\ to convert.
    //Doc:Arg1 Pointer to ASCIIZ string with the \IPA\ in dotted notation.
    //Doc:Return The buffer pointed to is only valid until another
    //Doc:Return call to {\tt IPAddr} functions.

#ifdef MANET
  static bool IsMulticast(IPAddr_t addr_) { return (addr_ > ToIP("224.0.0.0")); }
#endif

public:
  IPAddr_t ip;
};

// Define a pair for IPAddress/Port
typedef std::pair<IPAddr_t, PortId_t> IPPort_t;

#endif


