/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Internet Protocol Class
// George F. Riley.  Georgia Tech, Spring 2002

// Implements the Internet Protocol, Version 4, Layer 3

// IPHeader Class Methods

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x01
// Uncomment below to enable debug level 1
//#define DEBUG_MASK 0x02
// Uncomment below to enable debug level 2
//#define DEBUG_MASK 0x04
//#define DEBUG_MASK 0x07
//#define DEBUG_MASK 0x08
//#define DEBUG_MASK 0x20
//#define DEBUG_MASK 0x80
#include <iostream>

#include "debug.h"
#include "ipv4.h"
#include "protograph.h"
#include "l2proto802.3.h"
#include "l4demux.h"
#include "routing.h"
#include "routing-nixvector.h"
#include "routing-dsr.h"
#include "routing-nvr.h"
#include "node.h"
#include "droppdu.h"
#include "globalstats.h"
#include "hex.h"
#include "tcp.h"
#include "udp.h"
#include "icmp.h"
#include "routing-aodv.h"

#ifdef SIM
#include "sim.h"
#include "arpa/inet.h"  //for inet_pton
#include "netinet/in.h" //for struct inet6_addr
#include "ospf6_proto.h" //for ALL*ROUTERS
#include "estat.h"
#endif //SIM
#ifdef ETRACE
#include "etrace.h"
#endif //ETRACE

using namespace std;

// Define the pointer for the single ipv4 instance
IPV4*   IPV4::instance = nil;     // Single instance of IPV4 code
IPV4Header* IPV4::bpHeader = nil; // IPV4 header imported from backplane

IPV4Header::IPV4Header() : 
  version(4),
  headerLength(0),
  serviceType(0),
  totalLength(0),
  identification(0),
  flags(0),
  fragmentOffset(0),
  ttl(IPV4::DefaultTTL),
  protocol(0),
  headerChecksum(0),
  src(), dst()
{
}

IPV4Header::IPV4Header(const IPV4Header& r) : 
  version(r.version),
  headerLength(r.headerLength),
  serviceType(r.serviceType),
  totalLength(r.totalLength),
  identification(r.identification),
  flags(r.flags),
  fragmentOffset(r.fragmentOffset),
  ttl(r.ttl),
  protocol(r.protocol),
  headerChecksum(r.headerChecksum),
  src(r.src), dst(r.dst)
{
  // Copy the options
  for (PDUVec_t::size_type i = 0; i < r.options.size(); ++i)
    {
      options.push_back(r.options[i]->Copy());
    }
}


// Construct from a serialized buffer
IPV4Header::IPV4Header(char* b, Size_t& sz, Packet* p)
{
  DEBUG(3, (cout << "Constructing ipv4, initial size " << sz << endl));
  Size_t s = 0;
  b = Serializable::GetSize(b, sz, s);
  DEBUG(3, (cout << "l3pdu size is " << s << " (" << Hex8(s) << ") " << endl));
  s -= sizeof(Size_t);
  sz -= s;
  b = Construct(b, s);
  sz += s; // Add back in residual (subtraced below)
  DEBUG(3,(cout << "Constructing ipv4, final size " << sz << endl));
  p->PushPDUBottom(this);  // Add to packet
  while (s)
    { // Size remaining, must be options
      // Need to peek ahead inthe buffer to find option number
      DEBUG(3,(cout << "IPv4 excess size is " << s << endl));
      char*  sb  = b;
      Size_t thisSize  = 0;
      Size_t ssz = sz;
      b = Serializable::GetSize(b, sz, thisSize);
      DEBUG(3,(cout << "IPv4 nv size is " << thisSize << endl));
      Byte_t optionNumber = 0;
      b = ConstructFromBuffer(b, sz, optionNumber);
      // Put everything back and construct the appropriate options
      b  = sb;
      sz = ssz;
      switch (optionNumber) {
        // Presently only option implemented is NixVector routing
        case 24 :
          //new NixVectorOption(b, sz, p); // Not an option 
          //b += thisSize;
          break;
        }
      s -= thisSize;
    }
  switch (protocol) { // Constuct appropriate l4pdu
    case 6 :  // TCP
      DEBUG(3,(cout << "Constructing TCP header" << endl));
      new TCPHeader(b, sz, p); // Create and push TCP header
      break;
    case 17 : // UDP
      new UDPHeader(b, sz, p); // Create and push UDP header
      DEBUG(3,(cout << "Constructing TCP header" << endl));
      break;
    }
}

IPV4Header::~IPV4Header()
{
  // Delete the options
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    {
      delete options[i];
    }
}

Size_t IPV4Header::Size() const
{
  Count_t sz = 20;
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    {
      sz += options[i]->Size();
    }
  return sz;
}

void IPV4Header::Trace(Tfstream& tos, Bitmap_t b, Packet* p, const char* s)
{
  DEBUG0((cout << "IPV4 tracing src " << (string)src));
  DEBUG0((cout << " dst "             << (string)dst << endl));
  tos << " ";
  if (s)
    tos << s;
  tos << "L3";
  if (Detail(IPV4::VERSION, b))         tos << "-" << version;
  if (Detail(IPV4::HEADERLENGTH, b))    tos << " " << headerLength;
  if (Detail(IPV4::SERVICETYPE, b))     tos << " " << serviceType;
  if (Detail(IPV4::TOTALLENGTH, b))     tos << " " << totalLength;
  if (Detail(IPV4::IDENTIFICATION, b))  tos << " " << identification;
  if (Detail(IPV4::FLAGS, b))           tos << " " << flags;
  if (Detail(IPV4::FRAGMENTOFFSET, b))  tos << " " << fragmentOffset;
  if (Detail(IPV4::TTL, b))             tos << " " << ttl;
  if (Detail(IPV4::PROTOCOL, b))        tos << " " << protocol;
  if (Detail(IPV4::HEADERCHECKSUM, b))  tos << " " << headerChecksum;
  if (Detail(IPV4::SRC, b))             tos << " " << src;
  if (Detail(IPV4::DST, b))             tos << " " << dst;
  if (Detail(IPV4::UID, b))
    { // Get uid from packet
      if (p)
        tos << " " << p->uid;
      else
        tos << " " << 0;
    }
  
  // Need to trace options (code later)
}

Size_t IPV4Header::SSize()
{ // Size needed for serialization
  Size_t r = 
    sizeof(version) + 
    sizeof(headerLength) + 
    sizeof(serviceType) + 
    sizeof(totalLength) + 
    sizeof(identification) + 
    sizeof(flags) + 
    sizeof(fragmentOffset) + 
    sizeof(ttl) + 
    sizeof(protocol) + 
    sizeof(headerChecksum) + 
    sizeof(src) + 
    sizeof(dst);
  DEBUG0((cout << "IPV4Hdr, basic size " << r << endl));
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    { // add the size of the options
      DEBUG0((cout << "ipv4 ssize, opt size " << options[i]->SSize() << endl));
      r += options[i]->SSize() + sizeof(Size_t);
    }
  return r;
}

char*  IPV4Header::Serialize(char* b, Size_t& sz)
{ // Serialize to a buffer
  IPAddr_t srcip = src;
  IPAddr_t dstip = dst;
  b = SerializeToBuffer(b, sz, version);
  b = SerializeToBuffer(b, sz, headerLength);
  b = SerializeToBuffer(b, sz, serviceType);
  b = SerializeToBuffer(b, sz, totalLength);
  b = SerializeToBuffer(b, sz, identification);
  b = SerializeToBuffer(b, sz, flags);
  b = SerializeToBuffer(b, sz, fragmentOffset);
  b = SerializeToBuffer(b, sz, ttl);
  b = SerializeToBuffer(b, sz, protocol);
  b = SerializeToBuffer(b, sz, headerChecksum);
  b = SerializeToBuffer(b, sz, srcip);
  b = SerializeToBuffer(b, sz, dstip);
  // Now serialize the options
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    {
      Size_t s = options[i]->SSize() + sizeof(Size_t);
      b = Serializable::PutSize(b, sz, s);
      b = options[i]->Serialize(b, sz);
    }
  return b;
}

char*  IPV4Header::Construct(char* b, Size_t& sz)
{ // Construct from buffer
  IPAddr_t srcip;
  IPAddr_t dstip;
  b = ConstructFromBuffer(b, sz, version);
  b = ConstructFromBuffer(b, sz, headerLength);
  b = ConstructFromBuffer(b, sz, serviceType);
  b = ConstructFromBuffer(b, sz, totalLength);
  b = ConstructFromBuffer(b, sz, identification);
  b = ConstructFromBuffer(b, sz, flags);
  b = ConstructFromBuffer(b, sz, fragmentOffset);
  b = ConstructFromBuffer(b, sz, ttl);
  b = ConstructFromBuffer(b, sz, protocol);
  b = ConstructFromBuffer(b, sz, headerChecksum);
  b = ConstructFromBuffer(b, sz, srcip);
  b = ConstructFromBuffer(b, sz, dstip);
  src = IPAddr(srcip);
  dst = IPAddr(dstip);
  // Need to construct options...code later
  return b;
}


// IP Protocol Class Methods

IPV4::IPV4()
{
  if (instance == nil)
    { // First instance
      instance = this;
      ProtocolGraph::CreateCommon();
      ProtocolGraph::common->Insert(3, Proto(), this); // Layer 3, proto 0x800
      // Set default tracing options
      DetailOff(HEADERLENGTH);
      DetailOff(SERVICETYPE);
      DetailOff(TOTALLENGTH);
      DetailOff(IDENTIFICATION);
      DetailOff(FLAGS);
      DetailOff(FRAGMENTOFFSET);
      DetailOff(HEADERCHECKSUM);
      DetailOff(OPTIONS);
    }
}

#ifdef SIM
//per interface DataRequest
//Broadcast on only one interface
void IPV4::DataRequest(Node* n, Packet* p, void* v, Interface *i)
{
  IPV4ReqInfo* ipInfo = (IPV4ReqInfo*)v;
	if ((IPAddr_t)ipInfo->dst == OSPF_ALLSPFROUTERS ||
		(IPAddr_t)ipInfo->dst == OSPF_ALLDROUTERS)

  { // broadcast on one interface
    Broadcast(n, p, ipInfo, i);
    return;
  }
	DataRequest(n, p, v);
}

#endif //SIM

void IPV4::DataRequest(Node* n, Packet* p, void* v)
{
  IPV4ReqInfo* ipInfo = (IPV4ReqInfo*)v;
  if ((IPAddr_t)ipInfo->dst == IPAddrBroadcast)

    { // Special case for broadcasts
      Broadcast(n, p, ipInfo);
      return;
    }

  DEBUG0((cout << "Data Request Size " << p->Size()
		  << " time " << Simulator::Now() << endl));
  RoutingEntry re;
  // First see if we only have one interface and one neighbor

#if 0  //SIM  BOEING
  if (n->InterfaceCount() == 1)
    { // See if only one peer on this interface
      const IFVec_t& interfaces = n->Interfaces();
      Interface* i = interfaces[0];
      if (i->NeighborCount(n) == 1)
        { // Ok, just use the single peer
          DEBUG(5,(cout << "V4-DR, single peer, node " << n->Id() << endl));
          re = RoutingEntry(i, i->DefaultPeer());
        }
    }
#endif // Don't know why

  // Next test for NixVector routing, and use if so
  NixVectorOption* nv = nil;
  RoutingNixVector* pnv = n->GetNixRouting();
  if (pnv && !re.interface)
    { // Can't use simple single-neighbor route, 
      // create a nixvector to the destination and use it
      nv = pnv->GetNixVector(n, ipInfo->dst);
      if (nv)
        { // Found or created a valid NixVector
          DEBUG(5,(cout << "DR - New or Cached NV on node " 
                   << (string)IPAddr(n->GetIPAddr()) 
                   << " dst " << (string)IPAddr(ipInfo->dst)
                   << endl));
          DEBUG(5,(nv->Dump()));
          NixBits_t nix = nv->Extract(nv->BitCount(n->NeighborCount()));
          re = n->LookupRouteNix(nix);
        }
    }
  
  if (!re.interface)
    {
      switch (Routing::Default()->Type())
        {
        case Routing::DSR:
        case Routing::NVR:
          n->GetRouting()->DataRequest(n, p, v);
          return;
          break;
#ifdef SIM
        case Routing::OSPF6: 
        {
          n->SetOspf6dGlobals();
          LookupRoute_ospf6sim(ipInfo->src,ipInfo->dst,&re);
          break;
        }
#endif //SIM
        default:
          re = n->LookupRoute(ipInfo->dst);
          break;
        }
    }
  
  //Now try to look up in aodv routing entry if aodv routing
  //exists. if no aodv routing, drop packet.
  //
  if (!re.interface)
    { //try AODV routing.
	  RoutingAODV *pRoutingAODV = n->GetRoutingAODV();
	  if(pRoutingAODV) {
		//do AODV routing lookup
		AODVRoutingEntry *pRt = n->LookupRouteAODV(ipInfo->dst);
		if(pRt && (pRt->Flags==RTF_UP)) {
			cout <<"arrive at aodv routing search, Amasing!\n"<<endl;
			cout <<"Dst Address is "<<(string)IPAddr(ipInfo->dst) <<endl;
			cout <<"pRt->DstIPAddr = " <<(string)IPAddr(pRt->DstIPAddr) <<endl;
			cout <<"pRt->Address = " << (string)IPAddr(pRt->Address) <<endl;
			//assert(pRt->pInterface!=nil); //test purpose 
			re.interface = pRt->pInterface; //to reuse the following header process
			re.nexthop = pRt->Address; //this is the next hop!
		}
		else {//if not found, queue the packet, and do RouteResolve
  			// Prepare the IP header 
			 IPV4Header* iphdr = new IPV4Header();
 			 iphdr->ttl = ipInfo->ttl;
  			 iphdr->protocol = ipInfo->l4proto;
      		 iphdr->src = n->GetIPAddr();	//Fill this first, modify it when rt found
             iphdr->dst = ipInfo->dst;               // Destination address
             //iphdr->uid = p->uid;                    // Set unique id from packet
             p->PushPDU(iphdr);                      // Add header to packet
             iphdr->totalLength = p->Size();         // Size including header
			 // In this callback function, do queue and do RouteResolve.
			 pRoutingAODV->RouteResolve(p, (void *)pRoutingAODV->who);
			 DEBUG0((cout <<"Queue PACKET in AODV and do RouteResolve to Addr"
						  <<(string)IPAddr(ipInfo->dst) << endl));
			 return ;
		}
	  }
	  else { //no aodv routing, still don't know how to route
#ifdef SIM 
		{// Send to etrace, we need layer 3 header for etrace to find packet
  		IPV4Header* iphdr = new IPV4Header();
			iphdr->ttl = ipInfo->ttl;
			iphdr->protocol = ipInfo->l4proto;
			if (ipInfo->src)
      		iphdr->src = ipInfo->src;
  		else
      		iphdr->src = n->GetIPAddr();
  		iphdr->dst = ipInfo->dst;               // Destination address

  		p->PushPDU(iphdr);                      // Add header to packet
  		iphdr->totalLength = p->Size();         // Size including header
			
#ifdef ETRACE
			ETrace::Instance()->TracePacket(p, n, n->GetIfByName("eth1"),3, "d");
#endif //ETRACE
#ifdef SIM //ESTAT 
			EStat::Instance()->EStatPacket(p, n, 3, "d"); 
#endif //SIM
		}
#endif //ETRACE
	    if (Trace::Enabled()) {
	      DropPDU d("L3-SNR", p);
	      n->TracePDU(this, &d);
	    }
	    delete p;
	    Stats::pktsDropped++;
	    return;
	  }
    }//endof trying AODV routing

  // Create and add the IPV4 header
  IPV4Header* iphdr = new IPV4Header();
  iphdr->ttl = ipInfo->ttl;
  iphdr->protocol = ipInfo->l4proto;
  if (ipInfo->src)
    { // Spoofed source requested
      iphdr->src = ipInfo->src;
    }
  else
    {
      iphdr->src = re.interface->GetIPAddr(); // Source address
    }
  if ((IPAddr_t)iphdr->src == IPADDR_NONE)
    { // Use default address from node
      iphdr->src = n->GetIPAddr();
    }
  iphdr->dst = ipInfo->dst;               // Destination address
  //iphdr->uid = p->uid;                    // Set unique id from packet
  if (nv) p->nixVec = nv;                 // Set the nixvector in packet
  p->PushPDU(iphdr);                      // Add header to packet
  iphdr->totalLength = p->Size();         // Size including header
  if (Trace::Enabled()) n->TracePDU(this, iphdr, p, "-");// Add to trace file
#ifdef ETRACE
	ETrace::Instance()->TracePacket(p, n, re.interface,3, "s"); 
#endif //ETRACE
#ifdef SIM //ESTAT 
	EStat::Instance()->EStatPacket(p, n, 3, "s"); 
#endif //SIM
  /* hand it to the interface with the next hop IP */
  re.interface->Send(p, re.nexthop, 0x0800);
}

void IPV4::DataIndication(Interface* i,Packet* p)
{
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU(); // Get ip hdr (don't remove)
  /* Call the call back list  */
  for(CallBackList_t::iterator it = cbList.begin(); it!= cbList.end(); it++)
  {
    (*it)(p);
  }

  // Get the node pointer from interface
  Node* n = i->GetNode();
  if (Trace::Enabled()) n->TracePDU(this, iphdr, p, "+"); // Add to trace file
  // If node specific call backs are enabled.
  if(n->CBEnable())
  {
    const CallBackList_t& cbList = n->GetCBList();
    for(CallBackList_t::const_iterator it = cbList.begin(); 
        it != cbList.end(); it++)
    {
      (*it)(p);
    }
  }
  
  iphdr->ttl--; // Decrement the ttl

  // Process IP options
  if (!iphdr->options.empty())
  {
    for (PDUVec_t::size_type j = 0; j < iphdr->options.size(); ++j)
    {
      IPV4Options* opt = (IPV4Options*)iphdr->options[j];
      switch (opt->Option())
      {
        case DSROptionsHeader::Number:
        case NVROption::Number:
          if (n->GetRouting()->DataIndication(i, p)) return;
          break;
        default:
          cout << "Unknown IP option number " << opt->Option() << endl;
          break;
      }
    } 
  }
  
  // For efficiency, first look for non-empty NixVector.  If exists
  // we can skip a lot of the slower lookups below
  bool useNv = false;
  NixVectorOption* nv = nil;
  RoutingNixVector* pnv = n->GetNixRouting();
  if (p->nixVec)
  {
    useNv = p->nixVec->Remaining();
  }
  
  RoutingEntry re;


  if (useNv)
  {
    nv = p->nixVec;
    NixBits_t nix = nv->Extract(nv->BitCount(n->NeighborCount()));
    DEBUG(5,(cout << "Ind w/ nix at node " << (string)IPAddr(n->GetIPAddr())
                  << " nodeid " << n->Id()
                  << " is " << nix  << endl));
    DEBUG(5,(cout << "After extract at node " 
                  << (string)IPAddr(n->GetIPAddr()) 
                  << " dst " << (string)IPAddr(iphdr->dst)
                  << " nodeid " << n->Id()
                  << " nv size " << nv->size
                  << " nv used " << nv->used
                  << " nv remaining " << nv->Remaining()
                  << endl));
    DEBUG(5,(nv->Dump()));
    re = n->LookupRouteNix(nix);
  }

  if (!re.interface)
  { // For efficiency, see if we have a cached nv, if so use it
    DEBUG(5,(cout << "IP-DI, no If, checking cached nv" << endl));
    if (pnv)
    { // Look for cached nv
      DEBUG(7,(cout << "Looking or cahced nv" << endl));
      nv = pnv->GetCachedNixVector(iphdr->dst);
      if (nv) 
      { // Good, found it.
        DEBUG(5,(cout << "Found cached nv" << endl));
        p->nixVec = nv;
        NixBits_t nix = nv->Extract(nv->BitCount(n->NeighborCount()));
        DEBUG(5,(cout << "Using cached nix at node " << (string)IPAddr(n->GetIPAddr())
                      << " nodeid " << n->Id()
                      << " is " << nix  << endl));
        re = n->LookupRouteNix(nix);  
        DEBUG(5,(cout << "After extract at node " 
                      << (string)IPAddr(n->GetIPAddr()) 
                      << " dst " << (string)IPAddr(iphdr->dst)
                      << " nodeid " << n->Id()
                      << " nv size " << nv->size
                      << " nv used " << nv->used
                      << " nv remaining " << nv->Remaining()
                      << endl));
        DEBUG(5,(nv->Dump()));
        DEBUG(5,(cout << "Used SS cached nv, iface " << re.interface << endl));
      }
    }
  }
  if (!re.interface)
    {
      DEBUG(5,(cout << "IP-DI, no If, checking aodv" << endl));
      //if it is for AODV protocol, it MUST continue to pass up for
      //further processing. 
      // 
      //peek the next protcol data unit
      bool IsAODV = false; 
      if(iphdr->protocol==17)
      {//protocol 17 is UDP
        UDPHeader *pUDPHdr = (UDPHeader *)p->PeekPDU(1);
        if(pUDPHdr->destPort == 564)//UDP port 564 is for AODV
          IsAODV = true; 
      }
      DEBUG(5,(cout << "Node " << n->Id() << " checking localip, dst "
                    << (string)IPAddr(iphdr->dst) << " isLocal "
                    << n->LocalIP(iphdr->dst) << endl));
      // Find out if the packet is addressed to this node or needs forwarding
      if (n->LocalIP(iphdr->dst) || (IPAddr_t)iphdr->dst == IPAddrBroadcast
#ifdef SIM
          || n->LocalIP(iphdr->dst &0x7fffffff) ||  //ipv4 linklocal address
             (IPAddr_t)iphdr->dst == OSPF_ALLSPFROUTERS || 
             (IPAddr_t)iphdr->dst == OSPF_ALLDROUTERS
#endif //SIM
          || IsAODV )
      {

#ifdef SIM
        if (((IPAddr_t)iphdr->dst == OSPF_ALLSPFROUTERS && !i->IsSpfGroup()) ||
 				       ((IPAddr_t)iphdr->dst == OSPF_ALLDROUTERS && !i->IsDrGroup()))
        {
          delete p;
          return;   //XXX HACK that is necessary until multicast is available
        }
#endif //SIM

#ifdef ETRACE
        ETrace::Instance()->TracePacket(p, n, i, 3,"r");
#endif //ETRACE
#ifdef SIM //ESTAT 
        EStat::Instance()->EStatPacket(p, n, 3, "r"); 
#endif //SIM
        // Destined for this node, get the layer 4 port demux  
        p->SkipPDU(); // Remove the header from the pdu list
        L4Demux* l4demux = (L4Demux*)n->LookupProto(4, iphdr->protocol);
          if (!l4demux)
            {
              cout << "HuH?  IPV4 indication with no l4demux object " 
                   << " proto number " << iphdr->protocol
                   << endl;
              delete p;
              return;
            }
#ifdef SIM
					l4demux->DataIndication(n, p, iphdr->src, *this, i, iphdr->dst);
#else
					l4demux->DataIndication(n, p, iphdr->src, *this, i);
#endif //SIM
          return;
        }
    }
// At this point, we need to forward to the next hop
#ifdef SIM
  if (!iphdr->ttl) // move this code upper here to check TTL first
    { // TTL expired, drop and log
      if (Trace::Enabled())
        {
          DropPDU d("L3-TTL", p);
          n->TracePDU(this, &d);
        }
#ifdef ETRACE
      ETrace::Instance()->TracePacket(p, n, i, 3,"df");
#endif //ETRACE
#ifdef SIM //ESTAT 
      EStat::Instance()->EStatPacket(p, n, 3, "df"); 
#endif //SIM
      DEBUG(0,(cout << "TTL Expired" << endl));
      // Notify ICMP to (possibly) send a time exceeded message
      ICMP::TimeExceeded(n, p, iphdr, ICMP::TTL_EXCEEDED);
      // Count stats, delete packet, and return
      Stats::pktsDropped++;
      delete p;
      return;
    }

  if (!re.interface)
    {
      switch (Routing::Default()->Type())
      {
        case Routing::OSPF6: 
				    {
          n->SetOspf6dGlobals();
          LookupRoute_ospf6sim(iphdr->src,iphdr->dst,&re);
          if (!re.interface) 
          { // OSPF can not find nexthop
					       // Count stats, delete packet, and return
#ifdef ETRACE
            ETrace::Instance()->TracePacket(p, n, i, 3, "df");
#endif //ETRACE
#ifdef SIM //ESTAT 
            EStat::Instance()->EStatPacket(p, n, 3, "df"); 
#endif //SIM
            Stats::pktsDropped++;
            delete p;
            return;
          }
					     break;
					   }
        default:
        {
          delete p;
          printf("To use the other routing protocols, please don't use SIM flag\n");
          exit(0);
          return;
        }
      }
    }
#else //SIM old code from GTNets

  if (!re.interface)
    { // No nix vector, try route lookup
      DEBUG(5,(cout << "IP-DI, no If, Trying Route lookup" << endl));
      re = n->LookupRoute(iphdr->dst);
    }
  if (!re.interface)
    { // Still no routing info, test for NixVector routing in use
      DEBUG(5,(cout << "IP-DI, no If, checking nv routing" << endl));
      nv = nil;
      if (pnv)
        { // Create a nix vector
          DEBUG(5,(cout << "IP-DI, no If, getting nixvec" << endl));
          nv = pnv->GetNixVector(n, iphdr->dst);
        }
      DEBUG(7,(cout << "IP::DataInd, pnv " << pnv
               << " nv " << nv << endl));
      if (nv)
        { // Created, use it
   	  DEBUG(5,(cout << "New or Cached NV1 at node " 
                   << (string)IPAddr(n->GetIPAddr()) 
                   << " dst " << (string)IPAddr(iphdr->dst)
                   << " nodeid " << n->Id()
                   << " nv size " << nv->size
                   << " nv used " << nv->used
                   << " nv remaining " << nv->Remaining()
                   << endl));
	  DEBUG(5,(nv->Dump()));
          p->nixVec = nv;
          NixBits_t nix = nv->Extract(nv->BitCount(n->NeighborCount()));
   	  DEBUG(5,(cout << "After Extract, new or Cached NV1 at node " 
                   << (string)IPAddr(n->GetIPAddr()) 
                   << " dst " << (string)IPAddr(iphdr->dst)
                   << " nodeid " << n->Id()
                   << " nv size " << nv->size
                   << " nv used " << nv->used
                   << " nv remaining " << nv->Remaining()
                   << endl));
          re = n->LookupRouteNix(nix);
          DEBUG(5,(cout << "IP::Dataind, nix " << nix
                   << " re.iface " << re.interface << endl));
        }
    }
  if (!re.interface)
    { // Still can't route, use default
      Routing* pRouting = n->GetRouting();
      if (pRouting) re = pRouting->GetDefault();
    }
      
  // Needs forwarding
  if (!iphdr->ttl)
    { // TTL expired, drop and log
      if (Trace::Enabled())
        {
          DropPDU d("L3-TTL", p);
          n->TracePDU(this, &d);
        }
      DEBUG(0,(cout << "TTL Expired" << endl));
      // Notify ICMP to (possibly) send a time exceeded message
      ICMP::TimeExceeded(n, p, iphdr, ICMP::TTL_EXCEEDED);
      // Count stats, delete packet, and return
      Stats::pktsDropped++;
      delete p;
      return;
    }
  //Now try to look up in aodv routing entry if aodv routing exists. 
  //Special NOTE: for wireless, the incoming interface can 
  //              be the outgoing interface.!
  if (!re.interface)
    { //try AODV routing entry lookup 
	  RoutingAODV *pRoutingAODV = n->GetRoutingAODV();
	  if(pRoutingAODV) {
		//do AODV routing lookup
		AODVRoutingEntry *pRt = n->LookupRouteAODV(iphdr->dst);
		if(pRt&&(pRt->Flags==RTF_UP)){ 
			pRt->pInterface->Send(p, pRt->Address, 0x0800);
			return;
		}
		else if(iphdr->dst.ip == IPAddrBroadcast) {
			printf("ospf6 ipv4.cc:  should this ever happen???\n");
			//Not sure if this kind of process is correct or not. 
			//
			IPV4ReqInfo *ipInfo = new IPV4ReqInfo(iphdr->dst,iphdr->src, 
							iphdr->ttl, iphdr->protocol);
			Broadcast(n ,p, ipInfo); //using private method if broadcast.
			delete ipInfo;
			return;
		}
		else {//if not found,drop the packet and log the event. 
			 DEBUG0((cout <<"NO ROUTE, DROP PACKET to Addr"
						  <<(string)IPAddr(iphdr->dst) << endl));
             if (Trace::Enabled()) {
                DropPDU d("L3-NR", p);
                n->TracePDU(this, &d);
             }
             DEBUG(0,(cout << "No routing interface" << endl));
             Stats::pktsDropped++;
             delete p;
			 return ;
		}
	  }
	  else { //still can't route, drop the packet,log it.
	    // Simply host unreachable for now, needs more checking later
	    // Notify ICMP to (possibly) send a time exceeded message
	    ICMP::DestinationUnreachable(n , p, iphdr, ICMP::HOST_UNREACHABLE);

	    //  delete packet, and return
	    if (Trace::Enabled()) {
	      DropPDU d("L3-NR", p);
	      n->TracePDU(this, &d);
	    }
	    DEBUG(0,(cout << "No routing interface" << endl));
	    Stats::pktsDropped++;
	    delete p;
	    return;
	  }
    }//end of trying aodv routing.
  
  if (re.interface == i)
    { // Routing loop (output on same interface as input, log err, drop, return
      if (Trace::Enabled())
        {
          DropPDU d("L3-RL", p);
          n->TracePDU(this, &d);
        }
      DEBUG(0,(cout << "Routing loop" << endl));
      Stats::pktsDropped++;
      delete p;
      return;
    }
#endif // SIM

  // Forward to next hop
  if (Trace::Enabled()) n->TracePDU(this, iphdr, p, "-"); // Add to trace file
#ifdef ETRACE
	ETrace::Instance()->TracePacket(p, n, n->GetIfByName("eth1"),3, "f");
#ifdef SIM //ESTAT 
	EStat::Instance()->EStatPacket(p, n, 3, "f"); 
#endif //SIM
#endif //ETRACE
  re.interface->Send(p, re.nexthop, Proto());
  
#ifdef OLD_SUPERCEEDED
  // Need better..not all l3 interfaces are 802.3
  DEBUG0((cout << "IPV4 Forwarding pkt to if " << re.interface << endl));
  L2Info802_3 l2Info(re.interface->GetMACAddr(),         // Src MAC Addr
                     re.interface->IPToMac(re.nexthop),  // Dst MAC Addr
                     Proto());                           // Protocol
  L2Proto* l2proto = re.interface->GetL2Proto(); // Get l2 protocol obj
  l2proto->DataRequest(re.interface, p, &l2Info); // Pass down to layer2
#endif
}

Interface* IPV4::PeekInterface(Node* n, void* v)
{ // The value passed to ipv4 is an ip address pointer
  IPAddr_t* ip = (IPAddr_t*)v;

  // First see if we only have one interface and one neighbor
  if (n->InterfaceCount() == 1)
    { // See if only one peer on this interface
      const IFVec_t& interfaces = n->Interfaces();
      Interface* i = interfaces[0];
      if (i->NeighborCount(n) == 1)
        { // Ok, just use the single peer
          return i;
        }
    }

  // Find the interface used for this route and pass back to caller
  RoutingEntry re = n->LookupRoute(*ip);
  if (!re.interface)
    { // No known route, test for NixVector routing.
      RoutingNixVector* pnv = n->GetNixRouting();
      if (pnv)
        { // Create a nixvector to the destination and use it
          NixVectorOption* nv = pnv->GetNixVector(n, *ip);
          if (nv)
            { // Found or created a valid NixVector
              NixBits_t nix = nv->Extract(nv->BitCount(n->NeighborCount()));
              re = n->LookupRouteNix(nix);
              delete nv; // Don't need this nixvector at the moment
            }
        }
    }
  return re.interface;
}

#ifdef SIM
// Private Methods
void IPV4::Broadcast(Node* n, Packet* p, IPV4ReqInfo* r, Interface *iface)
{
  // First Common processing (before looping on all interfaces)
  IPV4Header* iphdr = new IPV4Header();
  iphdr->ttl = r->ttl;
  iphdr->protocol = r->l4proto;
  iphdr->dst = r->dst;                    // Destination address
  //iphdr->uid = p->uid;                    // Set unique id from packet
  p->PushPDU(iphdr);                      // Add header to packet
  iphdr->totalLength = p->Size();         // Size including header

  if (iface->GetLink() == nil) return; // Wrong side of simplex link
#ifdef SIM
	if (r->src)
	{ // Spoofed source requested
		iphdr->src = r->src;
	}
	else
	{
   	iphdr->src = iface->GetIPAddr(); // Source address
	}
#else
  iphdr->src = iface->GetIPAddr(); // Source address
#endif //SIM
  if ((IPAddr_t)iphdr->src == IPADDR_NONE)
  { // Use default address from node
    iphdr->src = n->GetIPAddr();
  }
  // Add to trace file
  if (Trace::Enabled()) n->TracePDU(this, iphdr, p, "-");
#ifdef ETRACE
	ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,3, "s");
#endif //ETRACE
#ifdef SIM //ESTAT 
	EStat::Instance()->EStatPacket(p, n, 3, "s"); 
#endif //SIM
  iface->Send(p->Copy(), 0x0800);
#ifdef OLD_SUPERCEEDED
  L2Info802_3 l2Info(iface->GetMACAddr(),         // Src MAC Addr
                         MACAddr(0xffffffff, 0xffffffff), // Broadcast
                         //iface->IPToMac(re.nexthop),  // Dst MAC Addr
                         Proto());                    // Protocol
  L2Proto* l2proto = iface->GetL2Proto();         // Get l2 protocol obj
  l2proto->DataRequest(iface, p->Copy(), &l2Info);// Pass down to layer2
#endif
  delete p; // Since we replicated the packet on each iface, done with this one
}
#endif //SIM

// Private Methods
void IPV4::Broadcast(Node* n, Packet* p, IPV4ReqInfo* r)
{
  // First Common processing (before looping on all interfaces)
  IPV4Header* iphdr = new IPV4Header();
  iphdr->ttl = r->ttl;
  iphdr->protocol = r->l4proto;
  iphdr->dst = r->dst;                    // Destination address
  //iphdr->uid = p->uid;                    // Set unique id from packet
  p->PushPDU(iphdr);                      // Add header to packet
  iphdr->totalLength = p->Size();         // Size including header

  // Now forward the packet to all interfaces
  const IFVec_t ifVec = n->Interfaces();  // Vector of all interface this node
  for (IFVec_t::size_type i = 0; i < ifVec.size(); ++i)
    {
      Interface* iface = ifVec[i];
      if (iface->GetLink() == nil) continue; // Wrong side of simplex link
#ifdef SIM
			if (r->src)
			{ // Spoofed source requested
				iphdr->src = r->src;
			}
			else
			{
      	iphdr->src = iface->GetIPAddr(); // Source address
			}
#else
      iphdr->src = iface->GetIPAddr(); // Source address
#endif //SIM
      if ((IPAddr_t)iphdr->src == IPADDR_NONE)
        { // Use default address from node
          iphdr->src = n->GetIPAddr();
        }
      // Add to trace file
      if (Trace::Enabled()) n->TracePDU(this, iphdr, p, "-");
#ifdef ETRACE
			 ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,3, "s");
#endif //ETRACE
#ifdef SIM //ESTAT 
			 EStat::Instance()->EStatPacket(p, n, 3, "s"); 
#endif //SIM
      iface->Send(p->Copy(), 0x0800);
#ifdef OLD_SUPERCEEDED
      L2Info802_3 l2Info(iface->GetMACAddr(),         // Src MAC Addr
                         MACAddr(0xffffffff, 0xffffffff), // Broadcast
                         //iface->IPToMac(re.nexthop),  // Dst MAC Addr
                         Proto());                    // Protocol
      L2Proto* l2proto = iface->GetL2Proto();         // Get l2 protocol obj
      l2proto->DataRequest(iface, p->Copy(), &l2Info);// Pass down to layer2
#endif
    }
  delete p; // Since we replicated the packet on each iface, done with this one
}

// Static methods
IPV4* IPV4::Instance()
{
  if (!instance) instance = new IPV4();
  return instance;
}

// Backplane related methods
class BPIPV4
{
public:
  typedef enum {
      VERSION,
      HLEN,
      TOS,
      TLEN,
      ID,
      TTL,
      PROTO,
      CKSUM,
      SADDR,
      DADDR,
      OPTIONSSIZE,
      OPTIONSDATA
  } IPV4Items_t;
};

void IPV4::RegisterBackplane(DSHandle_t DST)
{
PRHandle_t prh;
ITHandle_t ith;

  prh = RegisterProtocol("IP", PROTO_REQUIRED, IPV4ExportQuery, nil);
  ith = RegisterItem(prh, "Header Length", ITEM_OPTIONAL, 8,
                     IPV4ExportCallback, (void*)BPIPV4::HLEN,
                     IPV4ImportCallback, (void*)BPIPV4::HLEN,
                     IPV4DefaultCallback, (void*)BPIPV4::HLEN);
  ith = RegisterItem(prh, "Type of Service", ITEM_OPTIONAL, 8,
                     IPV4ExportCallback, (void*)BPIPV4::TOS,
                     IPV4ImportCallback, (void*)BPIPV4::TOS,
                     IPV4DefaultCallback, (void*)BPIPV4::TOS);
  ith = RegisterItem(prh, "Total Length", ITEM_REQUIRED, 32,
                     IPV4ExportCallback, (void*)BPIPV4::TLEN,
                     IPV4ImportCallback, (void*)BPIPV4::TLEN,
                     IPV4DefaultCallback, (void*)BPIPV4::TLEN);
  ith = RegisterItem(prh, "Identification", ITEM_REQUIRED, 16,
                     IPV4ExportCallback, (void*)BPIPV4::ID,
                     IPV4ImportCallback, (void*)BPIPV4::ID,
                     IPV4DefaultCallback, (void*)BPIPV4::ID);
  ith = RegisterItem(prh, "Time to Live", ITEM_OPTIONAL, 8,
                     IPV4ExportCallback, (void*)BPIPV4::TTL,
                     IPV4ImportCallback, (void*)BPIPV4::TTL,
                     IPV4DefaultCallback, (void*)BPIPV4::TTL);
  ith = RegisterItem(prh, "Protocol", ITEM_OPTIONAL, 8,
                     IPV4ExportCallback, (void*)BPIPV4::PROTO,
                     IPV4ImportCallback, (void*)BPIPV4::PROTO,
                     IPV4DefaultCallback, (void*)BPIPV4::PROTO);
  ith = RegisterItem(prh, "Source Address", ITEM_OPTIONAL, 32,
                     IPV4ExportCallback, (void*)BPIPV4::SADDR,
                     IPV4ImportCallback, (void*)BPIPV4::SADDR,
                     IPV4DefaultCallback, (void*)BPIPV4::SADDR);
  ith = RegisterItem(prh, "Destination Address", ITEM_REQUIRED, 32,
                     IPV4ExportCallback, (void*)BPIPV4::DADDR,
                     IPV4ImportCallback, (void*)BPIPV4::DADDR,
                     IPV4DefaultCallback, (void*)BPIPV4::DADDR);
}

int   IPV4::IPV4ExportQuery(char* pSource, int, void*)
{ // Find if an IPV4 header is present in this packet
  Packet*    pPacket = (Packet*)pSource;
  IPV4Header* pHdrIP = nil;
  // Find the IP header
  for (PDUVec_t::size_type i = 0; i < pPacket->top; ++i)
    {
      PDU* pdu = pPacket->PDUs[i];
      if (pdu && (pdu->Layer() == 3) & (pdu->Version() == 4))
        { // Found it
          pHdrIP = (IPV4Header*)pdu;
          break;
        }
    }
  return pHdrIP != nil;
}

int   IPV4::IPV4DefaultCallback(char*, void*)
{
  return 0;
}

int   IPV4::IPV4ExportCallback(char* pSource, char* pRet, int lRet, void* p)
{
Packet*    pPacket = (Packet*)pSource;
BPIPV4::IPV4Items_t  e = (BPIPV4::IPV4Items_t)((int)p);
unsigned long*  pul = (unsigned long*)pRet;
unsigned short* pus = (unsigned short*)pRet;
unsigned char*  puc = (unsigned char*)pRet;

  DEBUG0((cout << "Hello from IPPCallback, item " << e << endl));
  IPV4Header* pHdrIP = nil;
  // Find the IP header
  for (PDUVec_t::size_type i = 0; i < pPacket->top; ++i)
    {
      PDU* pdu = pPacket->PDUs[i];
      if (pdu && (pdu->Layer() == 3) & (pdu->Version() == 4))
        { // Found it
          pHdrIP = (IPV4Header*)pdu;
          break;
        }
    }

  if (!pHdrIP) return 0; // Nope, no IPV4 header here

  switch(e) {
  case BPIPV4::HLEN:
    *puc = pHdrIP->headerLength;
    return sizeof(*puc);

  case BPIPV4::TOS:
    *puc = pHdrIP->serviceType;
    return sizeof(*puc);

  case BPIPV4::TLEN:
    *pus = pHdrIP->totalLength;
    return sizeof(*pus);

  case BPIPV4::ID:
    *pus = pHdrIP->identification;
    return sizeof(*pus);

  case BPIPV4::TTL:
    *puc = pHdrIP->ttl;
    return sizeof(*puc);

  case BPIPV4::PROTO:
    *puc = pHdrIP->protocol;
    return sizeof(*puc);

  case BPIPV4::CKSUM:
    *pus = pHdrIP->headerChecksum;
    return sizeof(*pus);

  case BPIPV4::SADDR:
    *pul = pHdrIP->src;
    return sizeof(*pul);

  case BPIPV4::DADDR:
    *pul = pHdrIP->dst;
    return sizeof(*pul);

  default:
    break;
  }
  return 0;
}

int   IPV4::IPV4ImportCallback(char* pSource, int lSource, char* pDest,void* p)
{
  BPIPV4::IPV4Items_t e = (BPIPV4::IPV4Items_t)((int)p);
  unsigned long*  pul = (unsigned long*)pSource;
  unsigned short* pus = (unsigned short*)pSource;
  unsigned char*  puc = (unsigned char*)pSource;

  if (!bpHeader)
    { // Need a header
      bpHeader = new IPV4Header();
    }

  DEBUG0((cout << "Hello from IPV4ImportCallback, item " << e << endl));
  switch(e) {
  case BPIPV4::HLEN:
    bpHeader->headerLength = *puc;
    break;

  case BPIPV4::TOS:
    bpHeader->serviceType = *puc;
    break;

  case BPIPV4::TLEN:
    bpHeader->totalLength = *pus;
    break;

  case BPIPV4::ID:
    bpHeader->identification = *pus;
    break;

  case BPIPV4::TTL:
    bpHeader->ttl = *puc;
    break;

  case BPIPV4::PROTO:
    bpHeader->protocol = *puc;
    break;

  case BPIPV4::CKSUM:
    bpHeader->headerChecksum = *pus;
    break;

  case BPIPV4::SADDR:
    bpHeader->src = *pul;
    break;

  case BPIPV4::DADDR:
    bpHeader->dst = *pul;
    break;

  default:
    break;
  }
  return 0;
}

