/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Internet Protocol base class
// George F. Riley.  Georgia Tech, Spring 2002

// Implements the Internet Protocol Version 4, Layer 3

#ifndef __ipv4_h__
#define __ipv4_h__

#include <list>

#include "common-defs.h"
#include "l3protocol.h"
#include "pdu.h"
#include "ipaddr.h"
#include "node.h"
#include "backplane.h"


#define IPAddrBroadcast (IPAddr_t)0xffffffff

class IPV4Header;

class IPV4ReqInfo;
//Doc:ClassXRef
class IPV4 : public L3Protocol {
//Doc:Class The class {\tt IPV4} implements the Internet Protocol Version 4
//Doc:Class It derives 	from the class {\tt L3Protocol}. Since IPV4 is a
//Doc:Class a static instance of this is enough for a simulation that uses
//Doc:Class IPv4.
	
public:
  typedef enum { DefaultTTL = 64 } DefTTL_t;
  typedef enum { VERSION, HEADERLENGTH, SERVICETYPE,
                 TOTALLENGTH, IDENTIFICATION,
                 FLAGS, FRAGMENTOFFSET,
                 TTL, PROTOCOL, HEADERCHECKSUM,
                 SRC, DST, UID, OPTIONS } IPV4Trace_t; // Tracing options
  //Doc:Method
  IPV4();
  //Doc:Desc Default constructor

  //Doc:Method
  void DataRequest(Node*, Packet*, void*);     // From upper layer
  //Doc:Desc This method is used by layer 4 protocols like udp/tcp to
  //Doc:Desc hand a packet to IP.
  //Doc:Arg1 Pointer to the node requesting the data transfer
  //Doc:Arg2 Pointer to the packet to be transmitted
  //Doc:Arg3 Any auxiliary information that layer 4 might want to convey
	
#ifdef SIM
  //Doc:Method
  void DataRequest(Node*, Packet*, void*, Interface *i);     // From upper layer
  //Doc:Desc This method is used by layer 4 protocols like udp/tcp to
  //Doc:Desc hand a packet to IP.
  //Doc:Arg1 Pointer to the node requesting the data transfer
  //Doc:Arg2 Pointer to the packet to be transmitted
  //Doc:Arg3 Any auxiliary information that layer 4 might want to convey
#endif //SIM
  
  //Doc:Method
  void DataIndication(Interface*, Packet*);    // From lower layer
  //Doc:Desc This method is used to pass on a packet from any layer 2 protocol
  //Doc:Desc up the stack.
  //Doc:Arg1 A pointer to the interface at which this packet is received
  //Doc:Arg2 A pointer to the packet that was received
  
  //Doc:Method
  Interface* PeekInterface(Node*, void*);      // Find forwarding interface
  //Doc:Desc This method returns the interface pointer. This interface is
  //Doc:Desc chosen based on the node pointer and some arbitrary data
  //Doc:Desc passed to it.
  //Doc:Arg1 the pointer to the assocaited node
  //Doc:Arg2 the data based on which we return the address
  //Doc:Return the pointer to the interface object
  
  //Doc:Method
  virtual Count_t Version() { return 4;}       // Get IP version number
  //Doc:Desc This method returns the version of Internet Protocol
  //DOc:Return returns the version
  
  //Doc:Method
  virtual Proto_t Proto()   { return 0x800;}   // Get the protocol number
  //Doc:Desc This method returns the IANA assigned protocol number
  //Doc:Return the protocol number
  
  //Doc:Method
  void CBListAdd(IPCallBack);
    //Doc:Desc Add a callback to process all  packets.
    //Doc:Arg1 Function pointer to callback to add.

  //Doc:Method
  void CBListDelete(IPCallBack);
    //Doc:Desc Delete an entry from the callback list.
    //Doc:Arg1 Function pointer to callback to delete.

  //Doc:Method
  void PreProcessCB(Packet* p); // loop over the list
    //Doc:Desc   Loops over the callback list, passing the packet to 
    //Doc:Desc each callback function.
    //Doc:Arg1 Packet pointer to pass to each callback.

  CallBackList_t cbList;

private:
  void Broadcast(Node*, Packet*, IPV4ReqInfo*);// Network layer broadcast
#ifdef SIM
  void Broadcast(Node*, Packet*, IPV4ReqInfo*, Interface* i);
#endif //SIM
private:
  static IPV4* instance; // Points to single instance of proto object
public:
  //Doc:Method
  static IPV4* Instance();   // Get a pointer to the single global instance
  //Doc:Desc This method returns the pointer to the global instance
  //Doc:Desc of the IPV4 object
  //Doc:Return the pointer to the global instance

  // Register data items with the dynamic simulation backplane
  static void  RegisterBackplane(DSHandle_t);
  static int   IPV4ExportQuery(char*, int, void*);
  static int   IPV4DefaultCallback(char*, void*);
  static int   IPV4ExportCallback(char*, char*, int, void*);
  static int   IPV4ImportCallback(char*, int, char*, void*);
public:
  // Static members
  static IPV4Header* bpHeader; // IPV4 header imported from backplane
};

//Doc:ClassXRef
class IPV4Header : public PDU {
//Doc:Class	
public :
  //Doc:Method
  IPV4Header();
  //Doc:Desc This is the default constructor
 
  //Doc:Method 
  IPV4Header(const IPV4Header& r);
  //Doc:Desc This constructor takes the reference to an IPV4 header
  
  //Doc:Method
  IPV4Header(char*, Size_t&, Packet*); // Construct from serialized buffer
  //Doc:Desc This method creates a PDU and inserts in the packet from
  //Doc:Desc a serialized buffer
  //Doc:Arg1 pointer to the buffer
  //Doc:Arg2 size of the buffer
  //Doc:Arg3 the pointer to the packet
  
  //Doc:Method
  virtual ~IPV4Header();
  //Doc:Desc Destructor
  
  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc This method returns the size of the PDU
  //Doc:Return the size of the PDU
  
  //Doc:Method
  PDU*    Copy() { return new IPV4Header(*this);}
  //Doc:Desc This method creates a copy of this PDU
  //Doc:Return A pointer to the copy of this PDU
  
  //Doc:Method
  Layer_t Layer() { return 3;}         // Indicate layer 3 pdu
  //Doc:Desc This method returns the layer number it belongs in the ISO model
  //Doc:Return the layer number
  
  //Doc:Method
  Count_t Version() { return 4;}       // Indicate IPV4
  //Doc:Desc Returns the version number of the protocol
  //Doc:Return version 4 of IP
  
  //Doc:Method
  Proto_t Proto() { return 0x800;}        // IPV4 is protocol 0x800
  //Doc:Method Return the IPV4 protocol number
  //Doc:Return 0x800 for IPV4

  void    Trace(Tfstream&, Bitmap_t,  // Trace the contents of this pdu
                     Packet* = nil, const char* = nil);
  //Doc:Desc This method traces the contents of the PDU
  //Doc:Arg1 File output stream descriptor
  //Doc:Arg2 Bitmap that specifies the level of detail of the trace
  //Doc:Arg3 Packet that contains this PDU (optional)
  //Doc:Arg4 Extra text information for trace file (optional)
  
  // Serialization
  //Doc:Method
  Size_t SSize();                   // Size needed for serialization
  //Doc:Desc The actual size of all the PDU contents in terms of bytes for
  //Doc:Desc serialization of the contents.
  //Doc:Return the size
  
  //Doc:Method
  char*  Serialize(char*, Size_t&); // Serialize to a buffer
  //Doc:Desc This method is used to serialize the contents of the PDU
  //Doc:Desc into a serialized character buffer b . the size of this buffer is
  //Doc:Desc in the size arguement
  //Doc:Arg1 This is the pointer to the character buffer
  //Doc:Arg2 a reference to the size variable
  //Doc:Return the pointer to the updated character buffer
  
  //Doc:Method
  char*  Construct(char*, Size_t&); // Construct from buffer
  //Doc:Desc This method is the reverse of the Serialize . It constructs the
  //Doc:Desc PDU object from the character buffer that has been received over
  //Doc:Desc the network
  //Doc:Arg1 This is the pointer to the character buffer
  //Doc:Arg2 a reference to the size variable
  //Doc:Return the pointer to the character buffer of the remaining data
  
public:
  Byte_t   version;
  Byte_t   headerLength;
  Byte_t   serviceType;
  Word_t   totalLength;
  Word_t   identification;
  Byte_t   flags;
  Word_t   fragmentOffset;
  Byte_t   ttl;       // Time to live
  Byte_t   protocol;  // Layer 4 protocols
  Word_t   headerChecksum;
  IPAddr   src;
  IPAddr   dst;
  //Count_t  uid;       // Unique id (not part of IP) (removed by GFR)
  PDUVec_t options;   // IP Options  
};

//Doc:ClassXRef
class IPV4Options : public PDU { // Base class for ipv4 options
//Doc:Class The Class {\tt IPV4Options } defines the IP options part
//Doc:Class of the IP Header. As all other headers this is also derived from
//Doc:Class the class {\tt PDU}
public:
  //Doc:Method
  IPV4Options(Byte_t o) : optionNumber(o) { }
  //Doc:Desc Constructor takes a byte as the arguement
  
public:
  //Doc:Method
  virtual Size_t  Size() const { return 0;}
  //Doc:Desc Returns the size of the PDU
  //Doc:Return size of the PDU
  
  //Doc:Method  
  virtual PDU*    Copy() = 0;
  //Doc:Desc This method creates a copy of this PDU
  //Doc:Return A pointer to the copy of this PDU
  
  //Doc:Method
  virtual void  Trace(Tfstream&, Bitmap_t,     // Trace this pdu 
                     Packet* = nil, const char* = nil) { }
  //Doc:Desc This method traces the contents of the PDU
  //Doc:Arg1 File output stream descriptor
  //Doc:Arg2 Bitmap that specifies the level of detail of the trace
  //Doc:Arg3 Packet that contains this PDU (optional)
  //Doc:Arg4 Extra text information for trace file (optional)
  
  //Doc:Method
  Byte_t  Option() { return optionNumber;}
  //Doc:Desc returns the options of the PDU
  //Doc:Return the options
public:
  Byte_t  optionNumber;
};

//Doc:ClassXRef
class IPV4ReqInfo {  // Information passed from L4 for IP Data Request
//Doc:Class This class defines the data structure that defines the information
//Doc:Class that is passed from layer 4 to layer 3
public:
  IPV4ReqInfo() : dst(IPADDR_NONE), src(IPADDR_NONE) { };
  IPV4ReqInfo(IPAddr_t d, IPAddr_t s, Count_t t, Proto_t p) 
    : dst(d), src(s), ttl(t), l4proto(p) { }
  IPAddr   dst;    // Destination
  IPAddr   src;    // Src IP (if spoofed, 0 if normal lookup)
  Count_t  ttl;    // Time to Live
  Proto_t  l4proto;// Layer 4 protocol
  // Probably need more later
};

#endif

