/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 2 802.3 class
// George F. Riley.  Georgia Tech, Spring 2002

// Implementation for 802.3

#include <iostream>

//#define DEBUG_MASK 0x02
#include "debug.h"
#include "l2proto802.3.h"
#include "interface.h"
#include "ipv4.h"
#include "trace.h"
#include "node.h"
#include "hex.h"
#include "macaddr.h"
#include "llcsnap.h"
#include "link.h"
#ifdef ETRACE
#include "etrace.h"
#endif

using namespace std;

void L2Proto802_3::BuildDataPDU(MACAddr src, MACAddr dst, Packet* p)
{
  L2802_3Header* l2pdu = new L2802_3Header(src, dst, p->Size());
  p->PushPDU(l2pdu); // Add the 802.3 header 
}

void L2Proto802_3::DataRequest(Packet* p)
{
  // Transmit on attached link    
  iface->GetLink()->Transmit(p, iface, iface->GetNode());
#ifdef ETRACE
  ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,2, "s");
#endif
}

void L2Proto802_3::DataIndication(Packet* p)
{
  L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU();
  if (Trace::Enabled()) iface->GetNode()->TracePDU(this, l2pdu, p, "+");
#ifdef ETRACE
  ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,2, "r");
#endif
  p->PopPDU();
  iface->HandleLLCSNAP(p);
}


#ifdef OLD_SUPERCEEDED
void L2Proto802_3::DataIndication(Interface* i, Packet* p)
{
#define SHORT_CIRCUIT
#ifdef  SHORT_CIRCUIT
  p->SkipPDU();
  IPV4::Instance()->DataIndication(i, p);
#ifdef ETRACE
  ETrace::Instance()->TracePacket(p, i->GetNode(), i,2, "r");
#endif
#else
  L2802_3Header* l2pdu = (L2802_3Header*)p->PopPDU();
  DEBUG0((cout << "l2 src "  << (string)l2pdu->src
               << " l2 dst " << (string)l2pdu->dst
               << " l3 proto  " << Hex4(l2pdu->l3proto) << endl));
  if (Trace::Enabled()) i->GetNode()->TracePDU(this, l2pdu);
  // This is not quite right...need to check protocol graph
  if (l2pdu->l3proto == IPV4::Instance()->Proto())
    { // Is IPV4
#ifdef ETRACE
  ETrace::Instance()->TracePacket(p, i->GetNode(), i,2, "r");
#endif
      IPV4::Instance()->DataIndication(i, p);
    }
  else
    {
      cout << "l2Proto802.3, no L3 proto for " << l2pdu->l3proto << endl;
      delete p;
    }
#endif
}
#endif

bool L2Proto802_3::Busy()
{ // For 802.3, the protocol is busy only if link is busy
  Link* l = iface->GetLink();
  return l->Busy();
}

// L2802_3Header methods

// Specifies the size (in bytes) of the 802.3 header when transmitting
// on a link.  Normally, the default 12 is correct.  However, other
// simulators (notably ns2) do not model this, so direct comparisons
// to ns2 are difficult.  Changing this value to 0 (along with the
// corresponding value in the LLCSnap header) will give a more direct
// comparison to ns2.
Size_t L2802_3Header::txSize = 12;


// Construct from serialized buffer
L2802_3Header::L2802_3Header(char* b, Size_t& sz, Packet* p)
{
  DEBUG0((cout << "Constructing l2803-2, initial size " << sz << endl));
  Size_t s = 0;
  b = Serializable::GetSize(b, sz, s);
  DEBUG0((cout << "l2pdu size is " << s << endl));
  s -= sizeof(Size_t);
  sz -= s;
  b = Construct(b, s);
  DEBUG0((cout << "Constructing l2803-2, final  size " << sz << endl));
  p->PushPDUBottom(this);  // Add to packet (bottom of stack)
  new LLCSNAPHeader(b, sz, p);
}

void L2802_3Header::Trace(Tfstream& tos, Bitmap_t b, Packet* p, const char* s)
{  // Trace the contents of this pdu
  tos << " ";
  if (s) 
    tos << s;
  tos << "L2";
  if (Detail(L2Proto802_3::TYPE, b))  tos << "-802.3";
  // Below moved to LLC/SNAP header
  //if (Detail(L2Proto802_3::PROTO, b)) tos << " " << Hex4(l3proto);
  if (Detail(L2Proto802_3::LENGTH, b)) 
    { // Check if packet passed, which we need for length
      if (p)
        tos << " " << p->Size();
    }
  if (Detail(L2Proto802_3::SRC, b))   tos << " " << (string)src;
  if (Detail(L2Proto802_3::DST, b))   tos << " " << (string)dst;
  if (Detail(L2Proto802_3::UID, b))
    {
      if (p) tos << " " << p->uid;
    }
}

Size_t L2802_3Header::SSize()
{
  //return sizeof(src) + sizeof(dst) + sizeof(l3proto);
  return 6 + 6 + 2;
}

char*  L2802_3Header::Serialize(char* b, Size_t& sz)
{ // Serialize to a working buffer
  DEBUG(1,(cout << "Serializing l28023 src " << Hex2(src)
           << " dst " << Hex2(dst)
           << " lth " << length << endl));
  b = SerializeToBuffer(b, sz, src);
  b = SerializeToBuffer(b, sz, dst);
  Word_t p = length;
  b = SerializeToBuffer(b, sz, p);
  return b;
}

char*  L2802_3Header::Construct(char* b, Size_t& sz)
{ // Construct from serialized buffer
  b = ConstructFromBuffer(b, sz, src);
  b = ConstructFromBuffer(b, sz, dst);
  Word_t p = 0;
  b = ConstructFromBuffer(b, sz, p);
  length = p;
  DEBUG0((cout << "L2Hdr::Construct, src " << Hex2(src)
          << " dst " << Hex2(dst)
          << " length " << length << endl));
  return b;
}

