/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 2 802.3 class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __l2proto_802_3_h
#define __l2proto_802_3_h

#include "l2proto.h"
#include "tfstream.h"
#include "macaddr.h"

//Doc:ClassXRef
class L2Proto802_3 : public L2Proto {
//Doc:Class Class {\tt L2Proto802\_3} defines the 802.3 layer 2 protocol. 
//Doc:Class This is derived from class {\tt L2Proto}. It besically implements
//Doc:Class the Medium Access Control portion of layer 2. 
public:
  typedef enum { TYPE, PROTO, LENGTH, SRC, DST, UID } L2Trace_t;  // Tracing options
  //Doc:Method
  L2Proto802_3() { }; // Constructor
  //Doc:Desc The Default Constructor

  //Doc:Method
  virtual   void BuildDataPDU(MACAddr, MACAddr, Packet*);
    //Doc:Desc Builds the layer2 pdu and appends it to the specified packet.
    //Doc:Arg1 Source MAC address
    //Doc:Arg2 Destination MAC address
    //Doc:Arg3 Packet pointer to append this pdu to.

  //Doc:Method
  void DataRequest(Packet*);
  //Doc:Desc This is the real member function is called by the higher layer 
  //Doc:Desc requesting a transfer of  data to a peer for 802.3
  //Doc:Arg1 The packet to be sent
  
  //Doc:Method
  void DataIndication(Packet*);    // from lower layer
  //Doc:Desc This method is called when the direction of protocol stack
  //Doc:Desc traversal is upwards, i.e, a packet has just been received
  //Doc:Desc at the underlying link
  //Doc:Arg1 The packet that has been received

  //Doc:Method
  virtual   bool Busy();                     // True if proto/link is busy
    //Doc:Desc Test if the protocol (or link) is busy processing a packet.
    //Doc:Return True if currently busy.

  //Doc:Method
   L2Proto802_3*  Copy() const { return new L2Proto802_3 (*this); }
  //Doc:Desc This method makes a copy of the protocol object and returns the
  //Doc:Desc pointer to the copy
  //Doc:Return pointer to the copy

  //Doc:Method
   bool IsWireless() const { return false;}
    //Doc:Desc Query if this protocol needs a wireless interface.
    //Doc:Return Always returns false

   void Bootstrap() {}
   
public:
  Word_t       length;

};

//Doc:ClassXRef
class L2802_3Header : public PDU {
//Doc:Class Class {\ttL2802_3Header } defines a PDU (Protocol Data Unit) 
//Doc:Class structure that defines the 802.3 header for the packet
public :
  //Doc:Method
  L2802_3Header() : src(MACAddr::NONE), dst(MACAddr::NONE), length(0) { };
  //Doc:Desc This method is the default constructor for the header. It fills 
  //Doc:Desc all the fields with zeros.
  //Doc:Method
  L2802_3Header(MACAddr s, MACAddr d, Size_t len)
    : src(s), dst(d), length(len) { };
  //Doc:Desc This method takes in three arguements namely, source MAC address
  //Doc:Desc destination MAC address and the layer 3 protocol type and creates
  //Doc:Desc a header PDU object from the same.
  //Doc:Arg1 Source MAC address
  //Doc:Arg2 Destination MAC Address
  //Doc:Arg3 Layer 3 protocol type

  //Doc:Method
  L2802_3Header(char*, Size_t&, Packet*); // Construct from serialized buffer
  //Doc:Desc This method creates an L2 802.3 Header from a serialized character
  //Doc:Desc buffer and pushes it into a packet. This is useful only while
  //Doc:Desc using distributed simulations over a network of workstations.
  //Doc:Arg1 the character buffer pointer
  //Doc:Arg2 the size of the buffer
  //Doc:Arg3 the pointer to the packet 

  //Doc:Method
  Size_t  Size() const { return txSize;}
  //Doc:Desc This method returns the size of the layer 2 header
  //Doc:Return the size of the header

  //Doc:Method
  PDU*    Copy() { return new L2802_3Header(*this);}
  //Doc:Desc This method makes a copy of itself and returns the copy
  //Doc:Return A PDU pointer to a copy of self

  //Doc:Method
  Layer_t Layer() { return 2;}         // Indicate layer 2 pdu
  //Doc:Desc This method returns the layer number it belongs in the ISO model
  //Doc:Return the layer number
#ifdef ETRACE
  virtual Proto_t Proto() { return 3;}   // For 802."3"
  //Doc:Desc Most of the protocols are identified by the numeric values as
  //Doc:Desc listed by IANA. thie method returns that protocol number
  //Doc:Return the protocol number
#endif
 
  //Doc:Method
  void    Trace(Tfstream& tos, Bitmap_t b,   // Trace the contents of this pdu 
                Packet* = nil, const char* = nil); 
  //Doc:Desc This method is used to trace the protocol data to a file (whose
  //Doc:Desc stream descriptor is tos) according to the level of detail 
  //Doc:Desc specified byt the bitmap
  //Doc:Arg1 the file I/O stream descriptor
  //Doc:Arg2 the bitmap
  //Doc:Arg3 Packet the contains this PDU (optional)
  //Doc:Arg4 Extra text information for trace file (optional)

  // Serialization
  //Doc:Method
  Size_t SSize();                   // Size needed for serialization
  //Doc:Method
  char*  Serialize(char*, Size_t&); // Serialize to a buffer
  //Doc:Desc This method serializes the header into a character buffer
  //Doc:Arg1 the pointer to the character buffer
  //Doc:Arg2 the current size of the buffer

  //Doc:Method
  char*  Construct(char*, Size_t&); // Construct from buffer
  //Doc:Desc This method is a helper method for the constructor. It actually 
  //Doc:Desc does the job of constructing a L2 802.3 header from the serialized
  //Doc:Desc character buffer obtained.
  //Doc:Arg1 the pointer to the character buffer
  //Doc:Arg2 the current size of the buffer.
public:
  MACAddr   src;
  MACAddr   dst;
  Word_t    length;
public:
  static Size_t txSize;  // Sie of header when transmitting on link
  static void TxSize(Size_t txs) { txSize = txs; }
  
};

#endif
