/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 4 Protocol base class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the layer 4 protocol interface

#ifdef HAVE_QT
#include <qcolor.h>
#endif

#include "l4protocol.h"
#include "l3protocol.h"
#include "portdemux.h"
#include "queue.h"
#include "application.h"
#include "event.h"

using namespace std;

L4Protocol::L4Protocol()
    : localNode(nil), peerIP(IPADDR_NONE),
      localPort(NO_PORT), peerPort(NO_PORT),
      localApplication(nil), notification(nil),
      l3Proto(IPV4::Instance()), ttl(IPV4::DefaultTTL),
      fid(0), iFace(nil), color(nil),
      extraRxDelay(0), extraTxDelay(0)
{
}

L4Protocol::L4Protocol(Node* n)
    : localNode(n), peerIP(IPADDR_NONE),
      localPort(NO_PORT), peerPort(NO_PORT),
      localApplication(nil), notification(nil),
      l3Proto(IPV4::Instance()), ttl(IPV4::DefaultTTL),
      fid(0), iFace(nil), color(nil),
      extraRxDelay(0), extraTxDelay(0)
{
}

// Copy constructor
L4Protocol::L4Protocol(const L4Protocol& c)
    : Protocol(c), 
      localNode(c.localNode), peerIP(c.peerIP),
      localPort(c.localPort), peerPort(c.peerPort),
      localApplication(nil), notification(c.notification),
      l3Proto(c.l3Proto), ttl(c.ttl),
      fid(c.fid), iFace(c.iFace), color(c.color),
      extraRxDelay(c.extraRxDelay), extraTxDelay(c.extraTxDelay)
{
}

 
L4Protocol::~L4Protocol()
{
#ifdef HAVE_QT
  if (color) delete color;
#endif
}

// Event Handler
void L4Protocol::Handle(Event* e, Time_t)
{
  switch(e->event)
    {
      case DELAYED_RX:
        {
          L4DelayedRxEvent* rx = (L4DelayedRxEvent*)e;
          rx->l4Proto->DataIndication(rx->node, rx->packet, 
                                      rx->ipAddr, rx->fromIF);
          DEBUG0((cout << "Processing delayed tx" << endl));
        }
        break;
      case DELAYED_TX:
        {
          L4DelayedTxEvent* tx = (L4DelayedTxEvent*)e;
          tx->l3Proto->DataRequest(tx->node, tx->packet, &tx->reqInfo);
          if (tx->notifySent && localApplication)
            {
              localApplication->Sent(tx->msgLength, this);
            }
          DEBUG0((cout << "Processing delayed tx" << endl));
        }
        break;
    }
  delete e;  // Done with this event
}

bool L4Protocol::Bind(PortId_t p)
{ // Bind to specific port
  if (!localNode) return false; // Can't bind if no node
  if (p == NO_PORT)
    { // If binding to zero
      Bind();
      return localPort != NO_PORT;
    }
  if (localNode->Bind(ProtocolNumber(), p, this))
    {
      localPort = p;
      return true;
    }
  return false;
}

bool L4Protocol::Bind()
{ // Bind to available port
  if (!localNode) return NO_PORT; // Can't bind if no node
  localPort = localNode->Bind(ProtocolNumber(), this);
  return localPort != NO_PORT;
}

bool L4Protocol::Unbind(Proto_t proto, PortId_t port) // Remove port binding
{
  if (localNode) return localNode->Unbind(proto, port, this);
  return true; 
}

void L4Protocol::SetInterface()
{ // Find the output interface for packets from this flow
  // Query the l4 protocol for the output i/f for the peer ip
  iFace = l3Proto->PeekInterface(localNode, &peerIP);
}

#ifdef HAVE_QT
void L4Protocol::SetColor(const QColor& c)
#else
void L4Protocol::SetColor(const QColor&)
#endif
{
#ifdef HAVE_QT
  if (!color)
    { // No existing color, create it
      color = new QColor(c);
    }
  else
    {
      *color = c;
    }
#endif
}

void    L4Protocol::AddNotification(NotifyHandler* n)
{
  DEBUG0((cout << "L4Proto " << this << " adding notif " << n << endl));
  notification = n;
}

Packet* L4Protocol::NewPacket()
{
  Packet* p = nil;
  
#ifdef HAVE_QT
  if (IsColored()) p = new ColoredPacket(*color);
#endif
  if (!p) p = new Packet();
  p->notification = notification;
  DEBUG0((cout << "l4proto " << this
          << " newPacket, notif " << notification << endl));
  return p;
}

void L4Protocol::AddExtraRxDelay(Time_t d)
{
  extraRxDelay = d;
}

  
void L4Protocol::AddExtraTxDelay(Time_t d)
{
  extraTxDelay = d;
}

  
void L4Protocol::RequestNotification(NotifyHandler* h, void* v)
{
  if (!iFace) SetInterface();  // Get the interface pointer
  if (!iFace) return;          // Packets cannot be routed
  iFace->AddNotify(h, v);
}

void L4Protocol::CancelNotification(NotifyHandler* h)
{
  if (!iFace) return;          // Packets cannot be routed
  DEBUG0((cout << "iFace " << iFace
          << " cancel notif " << h << endl));
  iFace->CancelNotify(h);
}

bool L4Protocol::BufferAvailable(Size_t s, Packet* p)
{ // True if output link buffer avail.
  if (!iFace) SetInterface();
  if (!iFace) return true; // Cannot be found

  Queue* q = iFace->GetQueue();
  if (!q) return true;     // Should not happen
  DEBUG0((cout << "L4proto, BA, limit " << q->GetLimit() 
          << " length " << q->Length() 
          << " avail " << q->GetLimit() - q->Length()
          << endl));
  
  return q->Check(s,p);    // True if sufficient buffer space available
}




