/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ghost Link class
// George F. Riley.  Georgia Tech, Spring 2002

#include "debug.h"
#include "link-ghost.h"
#include "simulator.h"
#include "macaddr.h"

using namespace std;

Count_t LinkGhost::count = 0;

LinkGhost::LinkGhost() 
    : Link(), pPeer(nil)
{
  count++;
}

LinkGhost::~LinkGhost() 
{
}

LinkGhost::LinkGhost(const LinkGhost& c) : Link(c)
{
  pPeer = nil; // Don't copy the peer pointer
}

Link* LinkGhost::Copy() const
{
  return new LinkGhost(*this);
}

MACAddr LinkGhost::IPToMac(IPAddr_t)
{ // Convert peer IP to peer MAC
  // For Ghost, this is just MAC address of peer
  if (pPeer) return pPeer->GetMACAddr();
  return MACAddr::NONE; // Else none
}

Count_t LinkGhost::NeighborCount(Node* n)
{ // Get number of routing neighbors (always 1 for p2p links)
  return 1;
}

void LinkGhost::Neighbors(NodeWeightVec_t& nwv, Interface* i, bool)
{ // Add to neighbor list
  // Ghost has only one peer
  // last param (forceAll) meaningless on Ghost link.
  if (pPeer)
    {
      nwv.push_back(NodeIfWeight(pPeer->GetNode(), i, Weight()));
    }
}

void LinkGhost::AllNeighbors(NodeWeightVec_t& nwv)
{ // Add to neighbor list
  // Ghost has only one peer.  Also note interface is
  // peer, not current as in Neighbors above.
  if (pPeer)
    {
      nwv.push_back(NodeIfWeight(pPeer->GetNode(), pPeer, Weight()));
    }
}

Count_t  LinkGhost::PeerCount()       // Return number of peers
{
  return 1; // Ghost links have exactly 1 peer
}

IPAddr_t LinkGhost::PeerIP(Count_t npeer)  // Get peer IP Address (if known)
{
  if (npeer > 0 || !pPeer)
    return IPADDR_NONE;      // Oops, value out of range
  return pPeer->GetIPAddr(); // Return one and only peer's ip address  
}

IPAddr_t LinkGhost::NodePeerIP(Node* n)  // Get peer IP Address (if known)
{ // Find the peer's ip address for peer n on this link
  // For ghost, there is only one peer, so it must match
  if (!pPeer) return IPADDR_NONE;
  Node* peerNode = pPeer->GetNode();
  if (n != peerNode)
    return IPADDR_NONE;      // Oops, not correct peer
  return pPeer->GetIPAddr(); // Return one and only peer's ip address  
}

bool     LinkGhost::NodeIsPeer(Node* n)
{ // Return true if specified node is a peer
  if (!pPeer) return false;
  Node* peerNode = pPeer->GetNode();
  return (peerNode == n);
}

Count_t  LinkGhost::NodePeerIndex(Node* n)
{
  return 0; // Ghost links always have only one peer
}

IPAddr_t LinkGhost::DefaultPeer(Interface*)
{ // Get IP Addr of default peer
  return PeerIP(0); // For ghostlinks, default is same as only
}

