/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Point to Point Link class
// George F. Riley.  Georgia Tech, Spring 2002

//#define DEBUG_MASK 0x04
//#define DEBUG_MASK 0x08
#include "debug.h"
#include "link-rti.h"
#include "simulator.h"
#include "macaddr.h"
#include "hex.h"
#include "ipv4.h"
#include "routing-nixvector.h"
#include "link.h"
#include "tcp.h"
#include "l2proto802.3.h"
#include "llcsnap.h"

using namespace std;

#ifdef HAVE_RTI
char       LinkRTI::rxbuf[MAX_MSG_LENGTH];
char       LinkRTI::txbuf[MAX_MSG_LENGTH];
Time_t     LinkRTI::minTxTime;
#endif
Count_t    LinkRTI::nUpdates;  // Count number of update messages
Count_t    LinkRTI::nReflects; // Count number of reflect messages
Count_t    LinkRTI::ignoredReflects; // Count number of ignored reflect msgs

// Constructors
LinkRTI::LinkRTI(IPAddr_t i, Mask_t m, Interface* interface) 
    : LinkReal(),
      ip(i), mask(m), peerIP(IPADDR_NONE), iFace(interface), pPeer(nil)
{
  ConstructorHelper(interface);
}

LinkRTI::LinkRTI(IPAddr_t i, Mask_t m, Interface* iface, Rate_t b, Time_t d) 
    : LinkReal( b, d),
      ip(i), mask(m), peerIP(IPADDR_NONE), iFace(iface), pPeer(nil)
{
  ConstructorHelper(iface);
}

LinkRTI::LinkRTI(const LinkRTI& c) 
  : LinkReal(c), ip(c.ip), mask(c.mask), peerIP(c.peerIP),
    peerMAC(c.peerMAC), iFace(c.iFace)
{
  ConstructorHelper(iFace);
}

Link* LinkRTI::Copy() const
{
  return new LinkRTI(*this);
}

bool LinkRTI::Transmit(Packet* p, Interface* i, Node*)
{
#ifdef HAVE_RTI
  Time_t txTime = ((Rate_t)p->Size() * 8) / Bandwidth(); // Time to transmit
  // Compute received time
  Time_t txDelay = Delay();
  Time_t jit = jitter->Value();
  Time_t rxTime = Simulator::Now() + txTime + txDelay + jit;
  if (busy) cout << "HuH?  LinkRTI " << this 
                 << " Transmit on busy link!" << endl;
  busy = true; // Set link busy
  // Create and schedule the packet transmit complete event
  LinkEvent* evTx = new LinkEvent(LinkEvent::PACKET_TX);
  evTx->numberBytes = p->Size();
  Simulator::instance->Schedule(evTx, txTime, this);
  DEBUG0((cout << "LinkRTI " << this << " Transmit, scheduling link free at "
          << txTime + Simulator::Now() << endl));
  struct MsgS* pMyMsg = (struct MsgS*)txbuf;// Use static allocated buffer
  pMyMsg->TimeStamp = rxTime;

  // Find size needed to serialize       
  Size_t sz = p->SSize() + sizeof(Size_t);
  if (sz + sizeof(MsgS) + sizeof(int) > MAX_MSG_LENGTH)
    {
      cout << "HuH?  Packet serialized size exceeds MAX_MSG_LENGTH" << endl;
      delete p;
      return false;
    }


  // Allocate a place for RTIKITnodeid
  unsigned int* pMyNodeId = (unsigned int*)&pMyMsg[1];
  // debug follows
  unsigned int* pMyDebugSize = (unsigned int*)&pMyNodeId[1];
  *pMyDebugSize = p->Size();                // Debug..store size
  // end debug
  char*  b = (char*)&pMyDebugSize[1];       // Data immediately follows nodeid
  Size_t saveSz = sz;                       // Save for RTI Update
  *pMyNodeId = RTIKIT_nodeid;               // Used to ignore own messages
  b = Serializable::PutSize(b, sz, sz);     // Add the total length to buffer
  p->Serialize(b, sz);                      // Serialize the packet
  // Send to peer using RTIKIT
  DEBUG(2,(cout << "Updating" << endl));
  // Debug follows
#undef VERBOSE_DEBUG1
#ifdef VERBOSE_DEBUG1
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU(2); // Get ip hdr (don't remove)
  TCPHeader* tcp = (TCPHeader*)p->PeekPDU(3); // Get the TCP header
  cout << "TX " << this
       << " sip " << (string)IPAddr(iphdr->src)
       << " sp " << tcp->sourcePort
       << " dip " << (string)IPAddr(iphdr->dst)
       << " dp " << tcp->destPort
       << " seq " << tcp->sequenceNumber 
       << " ack "<< tcp->ackNumber
       << " lth " << tcp->dataLength
       << endl;
  // End debug
#endif  
  RTI_UpdateAttributeValues(
      objInstance, pMyMsg,
      saveSz + sizeof(struct MsgS) + sizeof(Size_t) + sizeof(unsigned int),
      0);
  DEBUG(2,(cout << "Done" << endl));
  nUpdates++;
#endif
  DEBUG0((cout << "LinkRTI " << this << " Transmit on link " 
          << (string)IPAddr(ip) 
          << " time " << Simulator::Now() 
          << " pktsize " << p->Size()
          << " busy " << busy
          << endl));
  delete p; // Delete the packet since we forwarded it to peer (or ignored)
  // Add a notification to deque and sent next packet
  AddNotify(iFace, nil);
  return true;
}

bool LinkRTI::Transmit(Packet* p, Interface* i, Node* n, Rate_t)
{
  return Transmit(p, i, n);
}

MACAddr LinkRTI::IPToMac(IPAddr_t)
{ // Convert peer IP to peer MAC
  // For RTI Links we have no idea, so just return None.
  // Really need to implement ARP to do this
  return MACAddr::NONE; // Else none
}

Count_t LinkRTI::NeighborCount(Node*)
{ // Get number of routing neighbors.
  // One if peer known, zero otherwise
  if (pPeer) return 1;
  return 0;
}

void LinkRTI::Neighbors(NodeWeightVec_t& nwv, Interface* i, bool)
{ // Add to neighbor list if peer is known
  if (pPeer)
    {
      nwv.push_back(NodeIfWeight(pPeer->GetNode(), i, Weight()));
    }
}

void LinkRTI::AllNeighbors(NodeWeightVec_t& nwv)
{ // Add to neighbor list
  if (pPeer)
    {
      nwv.push_back(NodeIfWeight(pPeer->GetNode(), pPeer, Weight()));
    }
}

Count_t  LinkRTI::PeerCount()       // Return number of peers
{
  return 1; // Point to point links have exactly 1 peer
}

IPAddr_t LinkRTI::PeerIP(Count_t npeer)  // Get peer IP Address (if known)
{
  if (npeer > 0)
    return IPADDR_NONE;      // Oops, value out of range
  return peerIP;             // Use known value
}

IPAddr_t LinkRTI::NodePeerIP(Node*)    // Get peer IP Address (if known)
{ // Find the peer's ip address for peer n on this link
  // For LinkRTI, we don't know peer nodes, so just return known IP
  return peerIP;             // Return one and only peer's ip address  
}

bool     LinkRTI::NodeIsPeer(Node* n)
{ // Return true if specified node is a peer
  if (!pPeer) return false;
  Node* peerNode = pPeer->GetNode();
  return (peerNode == n);
}

Count_t  LinkRTI::NodePeerIndex(Node* n)
{
  return 0; // RTI Links always have only one peer
}

IPAddr_t LinkRTI::DefaultPeer(Interface*)
{ // Get IP Addr of default peer
  return nil;       // Unknown for RTI Links
}

// debug
bool LinkRTI::IsRTI()
{ // true if rtilink for distributed simulations
  return true;
}

// Private Methods
void LinkRTI::ConstructorHelper(Interface* interface)
{
#ifdef HAVE_RTI
  if (!Simulator::instance) return; // just for debugging, remove later
  
  // Create an object class with the name of the link IP broadcast address
  IPAddr_t bcast = ip & mask | ~mask;
  string   bcastString = (string)IPAddr(bcast);
  cout << "LinkRTI ip is " << (string)IPAddr(ip) 
       << " bcast is " << bcastString << endl;
  RTI_CreateClass((char*)bcastString.c_str());
  // Get the object class handle
  objClass = RTI_GetObjClassHandle((char*)bcastString.c_str());
  if (objClass == NULL)
    {
      printf("Can't get objclasshandle for %s\n", bcastString.c_str());
      exit(0);
    }

  // Publish the class and register our interest
  RTI_PublishObjClass(objClass);
  objInstance = RTI_RegisterObjInstance(objClass);
  // Check if subscription is initialized, and init/subscribe if not
  if (!RTI_IsClassSubscriptionInitialized(objClass))
    { //  Set up a context
      RTI_InitObjClassSubscription (objClass, WhereMessage, iFace);
      // Subscribe to the class
      RTI_SubscribeObjClassAttributes(objClass);
    }
  // Compute the txtime + delay for smallest packet (used for lookahead)
  Time_t delay = Delay();
  Rate_t bw = Bandwidth();
  Time_t totTime = delay + 40 * 8.0 / bw; // Assumes minimum 40 byte pkt
  if (totTime)
    {
      if (totTime < minTxTime || minTxTime == 0) minTxTime = totTime;
    }
#endif
}
  
#ifdef HAVE_RTI
// Static "WhereMessage" procedure
char* LinkRTI::WhereMessage(       // Advise where to store Rx message
      long         MsgSize,        // Size of RX Message
      void*        iface,          // Context info, contains interface pointer
      long         MsgType)        // Type specified by sender
{
  //rxIFace = (Interface*)iface;     // Save for use by ReflectAttributeValues
  DEBUG0((cout << "LinkRTI::WhereMessage, simtime " 
          << Simulator::Now() << endl));
  char* b = (char*)malloc(MsgSize + 32);
  memcpy(b, &iface, sizeof(Interface*));
  return (b+sizeof(Interface*));
}

void ReflectAttributeValues(
                   TM_Time T,
                   struct MsgS* pMsg,
                   long MsgSize,
                   long MsgType)
{
  char* b = (char*)pMsg;
  b -= sizeof(Interface*);
  Interface* iface;
  memcpy(&iface, b, sizeof(Interface*));
  unsigned int* pMyNodeId = (unsigned int*)&pMsg[1];
  if (*pMyNodeId != RTIKIT_nodeid)
    { // Ignore own messages
      LinkRTI::nReflects++;
      unsigned int* pMyDebugSize = (unsigned int*)&pMyNodeId[1];
      char* b = (char*)&pMyDebugSize[1];
      DEBUG0((cout << "LinkRTI::RAV, iface " << iface
              << " size " << MsgSize 
              << " timestamp " << pMsg->TimeStamp
              << " nodeid " << *pMyNodeId
              << " simtime " << Simulator::Now()
              << endl));

      Size_t sz = 0;
      Size_t bsz = MsgSize - sizeof(struct MsgS) - sizeof(unsigned int);
      char* nb = Serializable::GetSize(b, bsz, sz);
      Packet* p = new Packet(nb, sz);
      DEBUG0((cout << "Packet size is " << p->Size() << endl));
      // DEBUG
#undef VERBOSE_DEBUG1
#ifdef  VERBOSE_DEBUG1
      for (unsigned x = 0; x < p->top; ++x)
        {
          cout << "pdu " << x << " layer " << p->PeekPDU(x)->Layer() 
               << " ptr " << p->PeekPDU(x)
               << endl;
        }
      L2802_3Header* l2 = (L2802_3Header*)p->PeekPDU(0);
      LLCSNAPHeader* snap = (LLCSNAPHeader*)p->PeekPDU(1);
      cout << " snap eth " << Hex4(snap->snap_ethtype) << endl;
      // debug follows
      IPV4Header* iphdr = (IPV4Header*)p->PeekPDU(2); // Peek at ip hdr
      TCPHeader* tcp = (TCPHeader*)p->PeekPDU(3); // Get the TCP header
      cout << "RX " << iface
           << " sip " << (string)IPAddr(iphdr->src)
           << " sp " << tcp->sourcePort
           << " dip " << (string)IPAddr(iphdr->dst)
           << " dp " << tcp->destPort
           << " seq " << tcp->sequenceNumber 
           << " ack " << tcp->ackNumber
           << " lth " << tcp->dataLength
           << endl;
      if (p->Size() > 2000)
        {
          cout << "HuH?  Serialized packet bad size" << endl;
          free(pMsg);
          return;
        }
      if (p->Size() != *pMyDebugSize)
        {
          cout << "HuH?  Expected pkt size " << *pMyDebugSize
               << " got " << p->Size() << endl;
        }
      if (p->top > 10)
        {
          cout << "HuH?  Serialized packet, bad top" << endl;
          free(b);
          return;
        }
#endif
      Interface* i = iface;
      Node* n = i->GetNode();                 // Associated node
      IPV4Header* iphdr = (IPV4Header*)p->PeekPDU(2);
      if (!iphdr)
        {
          cout << "HuH?  Serialized packet w/ no IP Header" << endl;
          free(b);
          return;
        }
      
      if (!p->nixVec && !n->LocalRoute(iphdr->dst))
        { // No nix vector and not local route, see if using NV routing
          DEBUG(3,(cout << "LINKRTI::RAV, node " << n
                   << " id " << n->Id()
                   << " ip " << (string)IPAddr(n->GetIPAddr())
                   << endl));
          RoutingNixVector* pnv = n->GetNixRouting();
          if (pnv)
            { // Using nixvector routing, get (or create) an NV to target
              // Get the ipv4 header....we could use a better way to do this
              IPV4Header* iphdr = (IPV4Header*)p->PeekPDU(2);
              // Add the nv to the packet
              p->nixVec = pnv->GetNixVector(n, iphdr->dst);
              DEBUG(3,(cout << "LinkRTI::RAV Adding new Nv to packet, dst "
                       << (string)iphdr->dst 
                       << " nv " << p->nixVec << endl));
              DEBUG(3,(p->nixVec->Dump()));
              DEBUG(3,(cout << "nv sz " << p->nixVec->size
                       << " nv used " << p->nixVec->used << endl));
            }
        }
      // Schedule the packet receive event at the destination interface
      LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p);
      i->ScheduleRx(evRx,pMsg->TimeStamp - Simulator::Now());
    }
  else
    {
      DEBUG(2,(cout << "ignoring self message" << endl));
      LinkRTI::ignoredReflects++;
    }
  free(b);
}


#endif



