/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Point to Point Link class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __linkp2p_h__
#define __linkp2p_h__

#include "link-real.h"
#include "interface.h"
#include "l2proto802.3.h"

//Doc:ClassXRef
class Linkp2p : public LinkReal {
//Doc:Class Class {\tt Linkp2p} defines a serial point to point link object
//Doc:Class derived from the {\tt Link} object 
public:
  
  Linkp2p();
  Linkp2p(Rate_t b, Time_t d);
  Linkp2p(const Linkp2p&);      // Copy constructor
  ~Linkp2p();
  // Methods from Link
  //Doc:Method
  Link*      Copy() const ;     // Make a copy of this link
  //Doc:Desc This method makes a copy of itself and returns a pointer to
  //Doc:Desc the copy
  //Doc:Return Pointer to the new {\tt Link} object

  //Doc:Method
  void       SetPeer(Interface* i) { pPeer = i;}
  //Doc:Desc This method sets a specific interface as the peer
  //Doc:Arg1 The interface to be set as the peer interface
  
  //Doc:Method
  Interface* GetPeer(Packet*) { return pPeer;}
  //Doc:Desc This method looks at the packet's destination MAC address and
  //Doc:Desc returns the interface corresponding to it
  //Doc:Arg1 The packet to be transmitted
  //Doc:Return The destination interface
  
  //Doc:Method
  virtual Interface* GetPeer(Count_t) { return pPeer;}
    //Doc:Desc Get the interface pointer for the remote endpoint of
    //Doc:Desc this link for the specified peer index.
    //Doc:Arg1 Which peer desired (0 <- i < NeighborCount())
    //Doc:Return A pointer to the receiving interface for the specified peer

  //Doc:Method
  bool       Transmit(Packet*, Interface*, Node*);
  //Doc:Desc Transmit a packet.
  //Doc:Arg1 The packet to transmit.
  //Doc:Arg2 A pointer to the local (transmitting) interface
  //Doc:Arg3 A pointer to the local (transmitting) node.
  //Doc:Return True if the transmit succeeded, false if packet was dropped.
  
  //Doc:Method
  bool       Transmit(Packet*, Interface*, Node*, Rate_t);
  //Doc:Desc Transmit a packet.
  //Doc:Arg1 The packet to transmit.
  //Doc:Arg2 A pointer to the local (transmitting) interface
  //Doc:Arg3 A pointer to the local (transmitting) node.
  //Doc:Arg4 Transmitting rate (bits/sec)
  //Doc:Return True if the transmit succeeded, false if packet was dropped.
  
  //Doc:Method
  MACAddr    IPToMac(IPAddr_t);            // Convert peer IP to peer MAC
  //Doc:Desc Convert peer IP to peer MAC
  //Doc:Arg1 The IP Address whose MAC is to be determined
  //Doc:Return The corresponding MAC address
  
  //Doc:Method
  Count_t    NeighborCount(Node*);              // Number of routing neighbors
  //Doc:Desc Returns the number of neighbors on this link which are routers
  //Doc:Desc technically, count of routing peers
  //Doc:Arg1 Node querying neighbors (significant for Ethernet links)
  //Doc:Return The number of routing peers.
  
  // Build neighbor list

  //Doc:Method
  void       Neighbors(NodeWeightVec_t&, Interface*, bool);
  //Doc:Desc Build the vector of neighbor,interface list
  //Doc:Arg1 The NodeWightVec element where we will fill in the neighbors
  //Doc:Arg2 The interface, with respect to which we need to find these
  //Doc:Arg2 neighbors
  //Doc:Arg3 boolean to specify if we need all the neighbors OR if we need
  //Doc:Arg3 only routing peers

  //Doc:Method
  void       AllNeighbors(NodeWeightVec_t&);    // Build neighbor list
  //Doc:Desc AllNeighbors is used by broadcasts even if they are
  //Doc:Desc not routing peers. Use this to find all the peers
  //Doc:Arg1 The NodeWeightVec element, with respect to which,
  //Doc:Arg1 we need all the peers in this link
  
  // Linkp2p only methods
  //Doc:Method
  Count_t    PeerCount();                  // Return number of peers
  //Doc:Desc Returns the number of peers on this link
  //Doc:Return number of peers
  
  //Doc:Method
  IPAddr_t   PeerIP(Count_t);              // Get peer IP Address (if known)
  //Doc:Desc Since each node on this link is also associated with a count
  //Doc:Desc this returns the IP address of the n'th  peer
  //Doc:Arg1 the count index n
  //Doc:Return the corresopnding IPAddress
  
  //Doc:Method
  IPAddr_t   NodePeerIP(Node*);            // Get peer IP Address (if known)
  //Doc:Desc Return the IP address of the node if it is on this link and
  //Doc:Desc is found
  //Doc:Arg1 The pointer to the node object
  //Doc:Return The IP Address
  
  //Doc:Method
  bool       NodeIsPeer(Node*);            // Find out if node is a peer
  //Doc:Desc Find out if node is a peer
  //Doc:Arg1 The Node
  //Doc:Return boolean result of the test
  
  //Doc:Method
  Count_t    NodePeerIndex(Node*);         // Get peer index for node
  //Doc:Desc Returns the index of the node in this link
  //Doc:Arg1 The pointer to the node
  //Doc:Return The index of the peer
  
  //Doc:Method
  IPAddr_t   DefaultPeer(Interface*);      // Get IP Addr of default peer
  //Doc:Desc Find the default peer's IP Address; this will find the default
  //Doc:Desc gateway's IP Address in the same subnet
  //Doc:Arg1 The interface whose default peerIP is to be known
  //Doc:Return The default peer's IP
public:
  Interface* pPeer; // Opposite end of the point to point link
public:
  //Doc:Method
  static Linkp2p& Default();               // Get the default link
  //Doc:Desc Returns the default link
  //Doc:Return a reference to the default point--to--point link
  
  //Doc:Method
  static Linkp2p& Default(const Linkp2p&); // Set new default
  //Doc:Desc Sets a new default linkp2p object, and returns
  //Doc:Arg1 a reference to the new default linkp2p object
  //Doc:Return a reference to the new default point--to--point link

private:
  static Linkp2p* defaultp2p;              // Default p2p link pointer
public:
  static Count_t  count;
};

#endif


