/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Specific waypoint mobility models
// George F. Riley.  Georgia Tech, Spring 2002

// Define the specific waypoint mobility model

#include <math.h>

#include "debug.h"
#include "common-defs.h"
#include "simulator.h"
#include "rng.h"
#include "node.h"
#include "wireless-grid.h"
#include "mobility-specific-waypoint.h"

using namespace std;

// Constructor 
SpecificWaypoint::SpecificWaypoint()
    : lastTime(0), cycleStart(0), loop(false)
{
}

// Copy constructor
SpecificWaypoint::SpecificWaypoint(const SpecificWaypoint& c)
    : Mobility(c), lastTime(c.lastTime), cycleStart(c.cycleStart), loop(c.loop)
{
}

SpecificWaypoint::~SpecificWaypoint()
{
}

void SpecificWaypoint::UpdateLocation(Node* n)
{ // Update node's location to current time
  Time_t now = Simulator::Now();
  if (first)
    { // Set last waypoint as node's initial position
      lastWP = Waypoint(0, n->GetLocation());
      cycleStart = Simulator::Now();
      if (nextIt == waypoints.end())
        { // No next point
          nextWP.time = INFINITE_TIME;
          paused = true;
        }
      else
        {
          nextWP = *nextIt;
          DEBUG0((cout << "Initial nextwp time " << nextWP.time
                  << " x/y " << nextWP.location.X()
                  << " " << nextWP.location.Y() << endl));
          Meters_t d = Distance(lastWP.location, nextWP.location);
          Time_t t = nextWP.time - lastWP.time;
          Meters_t v = d / t;
          // Compute X,Y velocities      
          ComputeVelocities(nextWP.location, lastWP.location, v);
        }
      first = false;
    }
  while(1)
    { // Loop till we find one in the future
      if (paused)
        {
          if (now < nextWP.time)
            {
              lastTime = now;
              return; // Nothing to do
            }
          paused = false;
        }
      // At this point, not paused, move since last update
      DEBUG0((cout << "Not paused, now " << now 
              << " nextTime " << nextWP.time
              << endl));
      if (now < nextWP.time)
        { // Easy case, just move
          Move(n, now - lastTime);
          DEBUG0((cout << "Moved to " << n->GetLocation().X() << ", " 
                  << n->GetLocation().Y() << endl));
          lastTime = now;
          return;
        }
      n->SetLocation(nextWP.location); // Allow for roundoff errors
      lastWP = nextWP;
      nextIt++;  // Next
      if (nextIt == waypoints.end())
        { // No more, just stop here unless looping specified
          if (loop)
            { // First adjust time to arrive by the cycle time
              Time_t cycleTime = Simulator::Now() - cycleStart;
              for (WaypointIt_t it = waypoints.begin();
                   it != waypoints.end(); ++it)
                {
                  it->time += cycleTime;
                }
              nextIt = waypoints.begin();
              cycleStart = Simulator::Now();
            }
          else
            {
              nextWP.time = INFINITE_TIME;
              paused = true;
              lastTime = now;
              return;
            }
        }
      nextWP = *nextIt;
      if (nextWP.time < now)
        {
          n->SetLocation(nextWP.location); // Already there
          lastTime = nextWP.time;
          continue;
        }
      if (nextWP.location == lastWP.location)
        { // Same location, set paused and continue
          paused = true;
          lastTime = now;
          continue;
        }
      Time_t extra = lastWP.time - now; // Time past prior
      // Distance to new waypoint    
      Meters_t d = Distance(lastWP.location, nextWP.location);
      Time_t t = nextWP.time - lastWP.time;
      if (t == 0.0)
        {
          n->SetLocation(nextWP.location); // Zero time, just move
          lastTime = now;
          continue;
        }
      else
        {
          Meters_t v = d / t;
          // Compute X,Y velocities      
          ComputeVelocities(nextWP.location, lastWP.location, v);
          Move(n, extra);
          paused = false;
          lastTime = now;
          return;
        }
    }
}

Mobility* SpecificWaypoint::Copy() const
{ // Make a copy of this object
  return new SpecificWaypoint(*this);
}

void SpecificWaypoint::Initialize()
{ // None necessary
}

void SpecificWaypoint::AddWaypoint(Node* n, const Waypoint& w)
{ // Add a new waypoint
  if (w.time == 0)
    { // Just set current position
      n->SetLocation(w.location);
      return;
    }
  bool firstWP = false;
  Waypoint last;
  if (waypoints.empty())
    { // See if first waypoint is same as node location.  If so, 
      // set paused.
      Location l = n->GetLocation();
      if (l == w.location) paused = true;
      firstWP = true;
    }
  else
    {
      last = waypoints.back();
    }
  if (w.time >= last.time)
    { // Simple case, just append
      waypoints.push_back(w);
    }
  else
    {
      // Need to find correct location
      WaypointIt_t i;
      for (i = waypoints.begin(); i != waypoints.end(); ++i)
        {
          if (w.time < i->time) break; // Found correct location
        }
      waypoints.insert(i, w);
    }
  if (firstWP) nextIt = waypoints.begin();
}

void SpecificWaypoint::MoveToWaypoint(Node* n, const Waypoint& w)
{
}



