/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Base class for mobility models
// George F. Riley.  Georgia Tech, Spring 2002

// Define the base class for all mobility models

#include <math.h>
#include <iostream>

#include "mobility.h"
#include "node.h"

using namespace std;

// Copy constructor
Mobility::Mobility(const Mobility& c)
    : lastTime(c.lastTime), velocityX(c.velocityX), velocityY(c.velocityY),
#ifdef REPEATABLE_MOBILITY
node_(0),
#endif
      velocityZ(c.velocityZ), paused(false), first(true)
{
}

void Mobility::ComputeVelocities(Location& n, Location& o, Meters_t v)
{ // Compute x/y velocities from location o to location n at velocity v
  /* 2-dim
  double theta = atan2(n.Y() - o.Y(), n.X() - o.X()); // Angle
  velocityX = v * cos(theta); // X velocity
  velocityY = v * sin(theta); // Y velocity
  */

  double d_xy = sqrt((n.X() - o.X()) * (n.X() - o.X()) +
                     (n.Y() - o.Y()) * (n.Y() - o.Y()));
  double phi = atan2(n.Z() - o.Z(), d_xy); // Angle
  double theta = atan2(n.Y() - o.Y(), n.X() - o.X()); // Angle
  double v_xy = v * cos(phi);
  velocityX = v_xy * cos(theta); // X velocity
  velocityY = v_xy * sin(theta); // Y velocity
  velocityZ = v * sin(phi);      // Z velocity
}

void Mobility::Move(Node* n, Time_t t)
{ // Move for t seconds at current velocities
  Location l = n->GetLocation();
  n->SetLocation(l.X() + t * velocityX,
                 l.Y() + t * velocityY,
                 l.Z() + t * velocityZ);
}

bool Mobility::IsMoving() const
{
  return !paused;
}

Angle_t Mobility::Direction() const
{
  DEBUG0((cout << "MSWP::Direction, paused " << paused << endl));
  if (paused) return 0; // Nothing if paused
  DEBUG0((cout << "Mobility::Direction, nextWP.y " << nextWP.location.Y()
          << " lastWP.y " << lastWP.location.Y()
          << " nextWP.x " << nextWP.location.X()
          << " lastWP.x " << lastWP.location.X() << endl));
  return atan2(nextWP.location.Y() - lastWP.location.Y(),
               nextWP.location.X() - lastWP.location.X());
}

// Static members
Meters_t Mobility::Distance(Location& n, Location& o)
{
  return sqrt((n.X() - o.X()) * (n.X() - o.X()) +
              (n.Y() - o.Y()) * (n.Y() - o.Y()) +
              (n.Z() - o.Z()) * (n.Z() - o.Z())); // distance
}
