/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Base class for mobility models
// George F. Riley.  Georgia Tech, Spring 2002

// Define the base class for all mobility models

#ifndef __mobililty_h__
#define __mobililty_h__

#include <list>

#include "simulator.h"

#ifdef REPEATABLE_MOBILITY
#define MOBILITY_START_TIME 1 // Seconds
#endif

class Node;
class Location;

//Doc:ClassXRef
class Waypoint {
public:
  Waypoint() : time(0) {}
  Waypoint(Time_t t, Meters_t x, Meters_t y, Meters_t z=0)
    : time(t), location(x,y,z) {}
  Waypoint(Time_t t, const Location& l) : time(t), location(l) {}
public:
  Time_t   time;     // Time to arrive at specified location
  Location location; // Location
};

typedef std::list<Waypoint>      WaypointList_t;
typedef WaypointList_t::iterator WaypointIt_t;


//Doc:ClassXRef
class Mobility {
public:
  Mobility() : lastTime(Simulator::Now()), velocityX(0), velocityY(0),
#ifdef REPEATABLE_MOBILITY
               node_(0), velocityZ(0) {}
#else
               velocityZ(0) {}
#endif
  Mobility(const Mobility&);               // Copy constructor
public:
  virtual void UpdateLocation(Node*) = 0;  // Compute new location
  virtual Mobility* Copy() const = 0;      // Make a copy
  virtual void Initialize() = 0;           // Initialize the mobility
  virtual bool IsMoving() const;           // True if currently moving
  virtual Angle_t Direction() const;       // Angle of motion
#ifdef REPEATABLE_MOBILITY
  Node* node_;
#endif
protected:
  void ComputeVelocities(Location&, Location&, Meters_t); // Compute x/y/z vel.
  void Move(Node*, Time_t);         // Move for t seconds at current velocities
  static Meters_t Distance(Location&, Location&); // Compute distance
  Time_t        lastTime;    // Time last updated
  Meters_t      velocityX;
  Meters_t      velocityY;
  Meters_t      velocityZ;
  // Moved from subclasses, GFR May 11, 2004
  bool            paused;
  bool            first;       // First entry to UpdateLocation
  Waypoint        lastWP;      // Waypoint of previous location
  Waypoint        nextWP;      // Waypoint moving to
};

#endif
