/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ghost Node class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __nodeghost_h__
#define __nodeghost_h__

#include "common-defs.h"
#include "node-impl.h"
#include "node.h"

class Link;
class Node;
class Interface;
class Routing;
class Protocol;
class ProtocolGraph;
class PortDemux;
class Linkp2p;
class Queue;
class Application;

#include "handler.h"
#include "interface.h"

//Doc:ClassXRef
class NodeGhost : public NodeImpl {
  //Doc:Class Ghost nodes are used as placeholders in a distributed
  //Doc:Class simulation, representing those nodes that are mapped
  //Doc:Class to be processed on some other process.  Ghosts allow
  //Doc:Class the calculation of NixVectors, since each simulator
  //Doc:Class has a complete picture of the topology.  Ghost nodes
  //Doc:Class have reduced  state, having only the IP address,
  //Doc:Class list  of interfaces (which are ghost  interfaces),
  //Doc:Class and a neighbor list.

public:
  NodeGhost(Node*);
  ~NodeGhost();
public:
  // IP Address Management is all done by node-impl

  // Real/Ghost test
  bool     IsReal();                     // True if node is real

  // Interface management is all done by node-impl


  // Broadcast a packet
  void       Broadcast(Packet*, Proto_t);// Broadcast packet on all interfaces

  // All link managment done by node-impl.h

  // Queues
  Queue*     GetQueue();                    // Return queue if only one i/f
  Queue*     GetQueue(Node*);               // Get the queue to specified node

  // Applications
  Application* AddApplication(const Application& a);

  // Neighbor management, done by node-impl

  // Routing
  void DefaultRoute(RoutingEntry r);     // Specify default route
  void DefaultRoute(Node*);              // Specify default route
  void AddRoute( IPAddr_t, Count_t, Interface*, IPAddr_t);
  RoutingEntry LookupRoute(IPAddr_t);    // Find a routing entry
  RoutingEntry LookupRouteNix(Count_t);  // Lookup route by Neighbor Index
  Routing::RType_t RoutingType();        // Report type of routing in use
  // LocalRoute used from node-impl
  void       InitializeRoutes();         // Routing initialization (if any)
  Count_t    RoutingFibSize() const;     // Statistics, size of FIB
  // GetNix used from node-impl
  RoutingNixVector* GetNixRouting();     // Returns Nix rtg pointer (if Nix)
  Routing*   GetRouting() { return nil;} // Returns the routing object

  //AODV Routing
  //Specially for AODV routing, it is not compatiable with
  //other route entry.
  //
  AODVRoutingEntry*  LookupRouteAODV(IPAddr_t);
  void               SetRoutingAODV(void *pRouting);
  RoutingAODV*       GetRoutingAODV();  //return the RoutingAODV object pointer.

  // Protocol Graph Interface
  Protocol* LookupProto(Layer_t, Proto_t);// Lookup protocol by layer
  void      InsertProto(Layer_t, Proto_t, Protocol*); // Insert a protocol

  // Port Demux Interface
  bool      Bind(Proto_t, PortId_t, Protocol*);    // Register port usage
  bool      Bind(Proto_t,
                 PortId_t, IPAddr_t,
                 PortId_t, IPAddr_t,
                 Protocol*);                       // Register port usage
  PortId_t  Bind(Proto_t, Protocol*);              // Choose port and register
  bool      Unbind(Proto_t, PortId_t, Protocol*);  // Remove port binding
  bool      Unbind(Proto_t,
                   PortId_t, IPAddr_t,
                   PortId_t, IPAddr_t,
                   Protocol*);                     // Remove port binding
  Protocol* LookupByPort(Proto_t, PortId_t);       // Lookup by port
  Protocol* LookupByPort(Proto_t,
                         PortId_t, IPAddr_t,
                         PortId_t, IPAddr_t);      // Lookup by s/d port/ip

  // Packet tracing interface
  // Trace header if enabled
  bool      TracePDU(Protocol*, PDU*, Packet*, char*);
  void      SetTrace(Trace::TraceStatus);          // Set trace level this node

  // Node location (x, y) interface
  void      SetLocation(Meters_t, Meters_t, Meters_t); // Set x/y/z location
  void      SetLocation(const Location&);          // Set x/y/z location
  void      SetLocationLongLat(const Location&);   // Set x/y/z location
  bool      HasLocation();                         // True if location assigned
  Meters_t  LocationX();
  Meters_t  LocationY();
  Meters_t  LocationZ();
  Location  GetLocation();                         // Return current location
  Location  UpdateLocation();                      // Get updated location
  Mobility* AddMobility(const Mobility&);          // Add a mobility model
  Mobility* GetMobility() const;                   // Get the mobility model
  bool      IsMobile();                            // True if mobility enabled
  bool      IsMoving();                            // True if not pausing
  bool      WirelessTx();                          // True if wireless transmit
  bool      WirelessRx();                          // True if wireless receive
  bool      WirelessCx();                          // True if wireless collid
  bool      WirelessRxMe();                        // True if wireless receive
  bool      WirelessRxZz();                        // True if wireless receive

  // QTWindow information
  void          Show(bool) { }                     // Turn on/off display
  bool          Show() { return false; }           // Always false
  QCanvasItem*  Display(QTWindow*);
  QCanvasItem*  Display(const QPoint&, QTWindow*);
  void          WirelessTxColor(const QColor&);
  const         QColor& WirelessTxColor();
  bool          PushWirelessTx(QCanvasItem*);
  QCanvasItem*  PopWirelessTx();
  void          PixelSize(Count_t);                // Set/query node size
  Count_t       PixelSizeX();
  Count_t       PixelSizeY();
  void          Shape(Node::Shape_t);              // Set/query node shape
  Node::Shape_t Shape();
  void          Color(const QColor&);
  bool          HasColor();
  QColor        Color();
  NodeAnimation* GetNodeAnimation() const;

  Meters_t  Distance(Node*) {return 0;}
  void      BuildRadioInterfaceList(void) { }
  const RadioVec_t& GetRadioInterfaceList() { return emptyIf; }

  void      SetRadioRange(Meters_t) {}
  Meters_t  GetRadioRange() { return 0.0; }

  bool      CBEnable() { return false; }
  void      CBEnable(bool) {}
  void      CBListAdd(IPCallBack) {}
  void      CBListDelete(IPCallBack) {}
  void      PreProcessCB(Packet*) {}
  const CallBackList_t& GetCBList() { return emptyCB; }

  Joules_t  getBattery(void) { return -1.0; }
  void      setBattery(Joules_t) {}
  double    getComputePower(void) { return -1.0; }
  void      setComputePower(double) {}

  // Switching functions
  void IsSwitchNode(bool);
  bool IsSwitchNode() { return false; }

  void PacketRX(Packet*, Interface*);

  // worm containment
  void SetWormContainment(WormContainment*) {};
  WormContainment* GetWormContainment() { return nil; } 

public:
  static Count_t count;
private:
  static RadioVec_t     emptyIf; // EmptyList
  static CallBackList_t emptyCB; // EmptyList
#ifdef DONT_NEED_THIS
  // Data Members
private:
  NodeId_t  id;            // Node id for this node
  IPAddr_t  ipAddr;        // IP Address for this node
  IFVec_t   interfaces;    // All defined network interfaces
protected:
  Node*     pNode;         // Points to corresponding node interface object
#endif
};

#endif

