/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Packet class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __packet_h__
#define __packet_h__

#include <iostream>
#include <vector>

#include "common-defs.h"
#include "pdu.h"
#include "serializable.h"
#include "memreuse.h"

class NotifyHandler;

class NixVectorOption;
//Doc:ClassXRef
class Packet : public Serializable, public ReuseBase {
//Doc:Class Class {\tt Packet} defines a generic packet class which
//Doc:Class is composed of a vector of protocol data units (PDUs). It also
//Doc:Class provides a flexible interface to convert a packet into a
//Doc:Class serialized buffer for transmission over a network. ( to be used
//Doc:Class in distributed simulations )

public :
  //Doc:Method
  Packet();
  //Doc:Desc Default constructor 

  //Doc:Method
  Packet(const Packet& p); 
  //Doc:Desc This constructor takes a reference to a packet and constructs a
  //Doc:Desc new packet with identical contents.
  //Doc:Arg1 reference to a packet p
  
  //Doc:Method
  Packet(char*, Size_t);   // Construct from serialized buffer
  //Doc:Desc This constructor constructs a packet object out of the contents
  //Doc:Desc of a serialized character buffer of a given size.
  //Doc:Arg1 pointer to the character buffer
  //Doc:Arg2 size of the buffer
  
  virtual ~Packet();

  void operator=(const Packet& rhs);

public:
  
  //Doc:Method
  Size_t  Size() { return size;} // Get total size of packet
  //Doc:Desc This method returns the size of the packet contents. It adds the
  //Doc:Desc contents of all its PDUs and returns the sum.
  //Doc:Return the sum

  Size_t  SizeUpdate();
  
  //Doc:Method
  void    PushPDU(PDU* pdu)           // Add a new PDU header
  //Doc:Desc This method is used to inset a protocol data unit (PDU) into
  //Doc:Desc the packet. The PDU is added at the end of the PDU vector.
  //Doc:Arg1 pointer to the pdu object
    {
      if (top == PDUs.size())
        { // Need to add new one
          PDUs.push_back(pdu);
          top++;
        }
      else
        { // Can reuse existing, after deleting
          if (PDUs[top] == pdu)
            { // Replacing with original    
              top++;
            }
          else
            { // New one, delete and re-insert      
              delete PDUs[top];    // Delete previous pdu   
              PDUs[top++] = pdu;   // Replace with this one 
            }
        }
      size += pdu->Size();
    }

  //Doc:Method
  void    PushPDUBottom(PDU*);   // Add a new PDU header at bottom of stack
  //Doc:Desc This method is used to inset a protocol data unit (PDU) into
  //Doc:Desc the packet at the bottom of the stack. The PDU is added at the
  //Doc:Desc beginning of the PDU vector.
  //Doc:Arg1 pointer to the pdu object
  
  //Doc:Method 
  PDU*    PopPDU()               // Get last PDU header
  //Doc:Desc This method is used when we want to pop out a PDU from the
  //Doc:Desc PDU stack and delete it from the packet.
  //Doc:Return A pointer to the PDU object
    {
      if (top == 0) return NULL; // Nope, nothing there
      PDU* p =  PDUs[--top];     // Get the pointer
      size -= p->Size();         // Reduce size
      return p;
    }

  //Doc:Method
  PDU*    PeekPDU()              // Get last PDU header without removal
  //Doc:Desc This method returns the last PDU without removing it from the
  //Doc:Desc packet's vector of PDUs.
  //Doc:Return A pointer to the PDU  object
    {
      if (top == 0) return NULL;  // Nope, nothing there
      return PDUs[top - 1];       // Return without removing
    }
  
  //change Count_t to NCount_t in order to PEEK forward or backward.
  //type conversion for avoiding the warnings.
  //but potentially have problems here!
  //
  //Doc:Method
  PDU*    PeekPDU(NCount_t offset) // Peek at last + offset without removal
  //Doc:Desc This method returns the PDU at 'offset' from the last without
  //Doc:Desc removing it from the packet's vector of PDUs.
  //Doc:Desc NOTE.  A negative offset peeks at PDUS already popped.
  //Doc:Desc NOTE.  A positive offset peeks at PDUS not yet popped.
  //Doc:Return A pointer to the PDU  object	  
    {
      if ((NCount_t)top <= offset) return NULL;  // Nope, nothing there
      return PDUs[top - offset - 1];   // Return without removing
    }

  //Doc:Method
  void    SkipPDU()              // Skip the current pdu
  //Doc:Desc This method skips the current pdu that's at the top of the
  //Doc:Desc PDU stack without deleting the PDU
    {
      if (top == 0) return;      // Nope, nothing there
      PDU* p =  PDUs[--top];     // Get the pointer
      size -= p->Size();         // Reduce size
    }

  //Doc:Method
  PDU*    PushExisting()
  //Doc:Desc This method returns the existing PDUs that have been skipped int
  //Doc:Desc in the {\tt SkipPDU } method
  //Doc:Return pointer to the skipped PDU
    {
      if (top == PDUs.size()) return nil; // None existing
      return PDUs[top++];
    }


  //Doc:Method
  PDU*    FindPDU(Layer_t);
    //Doc:Desc Finds a pdu in the pdu list for the specified layer.
    //Doc:Arg1 Layer to find.
    //Doc:Return PDU for specified layer, nil if none

  //Doc:Method
  PDU*    FindPDU(Layer_t, Proto_t);
    //Doc:Desc Finds a pdu in the pdu list for the specified layer and protocol
    //Doc:Arg1 Layer to find.
    //Doc:Arg1 Protocol to find.
    //Doc:Return PDU for specified layer and protocol, nil if none



  //Doc:Method
  virtual Packet* Copy();                // Make a copy of this packet
  //Doc:Desc This method is a generic copy function that makes a copy of
  //Doc:Desc itself and returns the pointer to the copy.
  //Doc:Return pointer to the copy of the packet
  
  // Serialization interface

  //Doc:Method
  Size_t  SSize();               // Size needed for serialization
  //Doc:Desc This method returns the total size of the payload in this
  //Doc:Desc that needs to be serialized for transmisison over the
  //Doc:Desc network to a remote simulator instance. ( for distributed
  //Doc:Desc simulations )
  //Doc:Return Payload size
  
  //Doc:Method
  char*   Serialize(char*, Size_t&); // Serialize to a working buffer
  //Doc:Desc This method serializes the contents of this packet into
  //Doc:Desc a serial character buffer
  //Doc:Arg1 charater pointer to the buffer
  //Doc:Arg2 reference to the size of the buffer
  //Doc:Return the serialized buffer
  
  //Doc:Method
  char*   Construct(char*, Size_t&); // Construct from serialized buffer
  //Doc:Desc This method does the reverse job of constructing the packet
  //Doc:Desc from a character buffer of a give size
  //Doc:Arg1 the serialized character buffer
  //Doc:Arg2 the size of the serialized buffer
  //Doc:Return the pointer to the buffer

  //Doc:Method
  virtual bool IsColored() { return false; }
    //Doc:Desc Determine if a packet has an associated color (for animation).
    //Doc:Desc We define this and the three color component methods (below) as
    //Doc:Desc virtual, and always return {\tt false} here.
    //Doc:Desc There is a subclass
    //Doc:Desc called {\tt ColoredPacket} that will define these values.
    //Doc:Desc This is
    //Doc:Desc to save memory in packet definitions, as only those packets
    //Doc:Desc assigned will have memory to store it.
    //Doc:Return True if colored.
  
  //Doc:Method
  virtual Color_t R() { return 0; }
    //Doc:Desc Return the Red component of the color.
    //Doc:Return The Red component.

  //Doc:Method
  virtual Color_t G() { return 0; }
    //Doc:Desc Return the Green component of the color.
    //Doc:Return The Green component.
  
  //Doc:Method
  virtual Color_t B() { return 0; }
    //Doc:Desc Return the Blue component of the color.
    //Doc:Return The Blue component.
    
public:
  //Doc:Member
  PDUVec_t PDUs;           // Vector of protocol data unit headers
    //Doc:Desc A vector of protocol data unit (PDU) pointers for this packet.

  //Doc:Member
  PDUVec_t::size_type top; // Current top of PDUs stack
    //Doc:Desc The vector containing the PDU's does not shrink when a PDU is
    //Doc:Desc removed, so we need a value specifying what is the index
    //Doc:Desc of the current {\em top} of the stack.

  //Doc:Member
  Count_t  retx;            // Retransmission counter
  //Doc:Desc Counter specifying how many times this packet was retransmitted
 
  //Doc:Member
  Count_t  uid;            // Each packet has a unique id
    //Doc:Desc Every packet in \GTNS\ has a unique 32 bit identifier,
    //Doc:Desc  allowing the
    //Doc:Desc packet to be easily identified in the trace file.

  //Doc:Member
  Size_t   size;           // Total size in bytes, all pdu's
    //Doc:Desc Total size of this packet.  This size is measured in number 
    //Doc:Desc of bytes
    //Doc:Desc that will be transmitted on a link for this packet.  In \GTNS,
    //Doc:Desc the PDU representing an IPV4 header (for example)
    //Doc:Desc  is substantially larger than
    //Doc:Desc the 20 bytes in a normal IPV4 PDU.  But the size for the IPV4
    //Doc:Desc PDU is claimed to be 20 bytes for purposes of computing
    //Doc:Desc transmission time on a lnk.

  //Doc:Member
  Time_t   time;           // Time this packet was created
    //Doc:Desc Timestamp for when this packet was originally created.

  //Doc:Member
  NixVectorOption* nixVec; // Nix vector (if present)
    //Doc:Desc Pointer to associated \NV\ for this packet (if present).

  //
  //specially for aodv, maybe others also need it
  //
  Time_t   expire;          //the time the packet expires!

  // NotifyHandler object to inform when this packet is transmitted
  NotifyHandler*  notification;  
private:
  //Doc:Member
  static Count_t nextUid;  // Next available uid
    //Doc:Desc Next available unique identifier for a packet.

};

// If we have QT, define a "ColoredPacket" that has an associated color.
#ifdef HAVE_QT
class QColor;
class ColoredPacket : public Packet {
public:
  ColoredPacket();
  ColoredPacket(const QColor&);
  ColoredPacket(const ColoredPacket&);
  Packet* Copy();    // Make a copy of this packet
  bool IsColored();
  Color_t R();
  Color_t G();
  Color_t B();
  void    SetColor(const QColor&);
  const QColor& Color() const;
private:
  QColor* color;
};

#endif

#endif
