/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ring topology

// Monirul I Sharif, Goergia Institute of Technology.

// Define the tree topology using point-to-point links

#ifndef __ring_h__
#define __ring_h__

#include "common-defs.h"
#include "node.h"

class Node;
class Linkp2p;
class Queue;

//Doc:ClassXRef
class Ring {
  //Doc:Class Defines a tree topology using point--to--point links.
public:
  // Size, IP, Mask
  //Doc:Method
  Ring(Count_t, 
       IPAddr_t = IPADDR_NONE);
    //Doc:Desc Construct a {\tt Ring} topology object with the specified
    //Doc:Desc number of nodes,  \IPA\ and address mask.  The
    //Doc:Desc default point--to--point link object is used for links.
    //Doc:Arg1 Number of nodes in the ring.
    //Doc:Arg2 Starting \IPA.  The \IPA\ is incremented by one for each
    //Doc:Arg2 node in the ring.

  // Specify link to use
  //Doc:Method
  Ring(Count_t, const Linkp2p&,
       IPAddr_t = IPADDR_NONE);
    //Doc:Desc Construct a {\tt Ring} topology object with the specified
    //Doc:Desc number of nodes, \IPA\ and address mask, and
    //Doc:Desc the specified point--to--point link.
    //Doc:Arg1 Number of nodes in the ring.
    //Doc:Arg2 A reference to a point--to--point link object to copy for 
    //Doc:Arg2 each link in the ring.
    //Doc:Arg3 Starting \IPA.  The \IPA\ is incremented by one for each
    //Doc:Arg3 node in the ring


  //Doc:Method
  Node* GetNode(Count_t);    // Get node at specified level, index
    //Doc:Desc Return a pointer to the node at the specified level, index.
    //Doc:Arg1 Level number (root is level 0).
    //Doc:Return Pointer to the specfied node.

  void BoundingBox(const Location&, const Location&);
    //Doc:Desc Define a bounding box for this tree object. 
    //Doc:Desc The bounding box
    //Doc:Desc is used to assign node locations, which are used during
    //Doc:Desc the simulation animation.
    //Doc:Arg1 Lower left corner of bounding box.
    //Doc:Arg2 Upper right corner of bounding box.


public:
  NodeId_t   first;      // There is a global node list, we don't need one 
  NodeId_t   last;       // Just log first/last+1 entries
  Count_t    nodeCount;  // Number of nodes
  IPAddr_t   firstNodeIP;// Starting IP Address for nodes
  NodeId_t   firstNodeId;// Starting NodeId for nodes
private:
  void ConstructorHelper(Count_t,
                         const Linkp2p&,
                         IPAddr_t);
};

#endif


