/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Random Number class
// George F. Riley.  Georgia Tech, Spring 2002

#include <math.h>
#include <stdlib.h>
#include <sys/time.h>			// for gettimeofday
#include <unistd.h>
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>       

//#define DEBUG_MASK 0x01
#include "debug.h"
#include "rng.h"

using namespace std;

// Base class methods

bool          Random::useDevRandom = false;  // True if use /dev/random desired
int           Random::devRandom = -1;
unsigned long Random::heuristic_sequence;
Uniform*      Random::globalSeedRNG;
#ifdef BUGFIX
std::list<Random*> Random::_memberList;
#endif
void Random::Seed(Seed_t s)
{ 
  Seed_t seeds[6];
  if (s == RANDOM_SEED)
    { // First check if global seed RNG exists. If so use it.
      if (globalSeedRNG)
        { // This allows a specific sequence of seeds for reproducibility
          seeds[0] = globalSeedRNG->IntValue();
          seeds[1] = globalSeedRNG->IntValue();
          seeds[2] = globalSeedRNG->IntValue();
          seeds[3] = globalSeedRNG->IntValue();
          seeds[4] = globalSeedRNG->IntValue();
          seeds[5] = globalSeedRNG->IntValue();
          Seeds(seeds);
          return;
        }
      // Check if /dev/random exists
      if (useDevRandom && devRandom < 0)
        {
          devRandom = open("/dev/random", O_RDONLY);
        }
      if (devRandom > 0)
        { // Use /dev/random
          DEBUG0((cout << "Seeding with /dev/random" << endl));
          for (int i = 0; i < 6; ++i)
            {
              read(devRandom, &seeds[i], sizeof(seeds[i]));
            }
        }
      else
        { // Seed from time of day (code borrowed from ns2 random seeding)
          // Thanks to John Heidemann for this technique
          DEBUG0((cout << "Seeding with time of day" << endl));
          timeval tv;
          gettimeofday(&tv, 0);
          seeds[0] = (tv.tv_sec^tv.tv_usec^(++heuristic_sequence << 8))
            & 0x7fffffff;
          gettimeofday(&tv, 0);
          seeds[1] = (tv.tv_sec^tv.tv_usec^(++heuristic_sequence << 8))
            & 0x7fffffff;
          gettimeofday(&tv, 0);
          seeds[2] = (tv.tv_sec^tv.tv_usec^(++heuristic_sequence << 8))
            & 0x7fffffff;
          gettimeofday(&tv, 0);
          seeds[3] = (tv.tv_sec^tv.tv_usec^(++heuristic_sequence << 8))
            & 0x7fffffff;
          gettimeofday(&tv, 0);
          seeds[4] = (tv.tv_sec^tv.tv_usec^(++heuristic_sequence << 8))
            & 0x7fffffff;
          gettimeofday(&tv, 0);
          seeds[5] = (tv.tv_sec^tv.tv_usec^(++heuristic_sequence << 8))
            & 0x7fffffff;
        }
      Seeds(seeds);
    }
  else
    {
      DEBUG0((cout << "Seeding with seed " << s << endl));
      seeds[0] = s;
      seeds[1] = s;
      seeds[2] = s;
      seeds[3] = s;
      seeds[4] = s;
      seeds[5] = s;
      Seeds(seeds);
    }
}

void Random::Seeds(Seed_t* seeds)
{
  SetSeed(seeds);
}

Random_t Random::Value()
{
  return RandU01(); // Uniform [0..1)
}

Random* Random::Copy() const
{
  return new Random(*this);
}

// Uniform methods
Random_t Uniform::Value()
{
  return m_min + Random::Value() * (m_max - m_min);
}

Random* Uniform::Copy() const
{
  return new Uniform(*this);
}

// Constant methods
Random_t Constant::Value()
{
  return m_const;
}

IRandom_t Constant::IntValue()
{
  return (IRandom_t)m_const;
}

Random* Constant::Copy() const
{
  return new Constant(*this);
}

// Sequential methods
Sequential::Sequential(Random_t f, Random_t l, Random_t i, Count_t c)
  : m_min(f), m_max(l), m_increment(Constant(i).Copy()), m_consecutive(c),
    m_current(f), m_currentConsecutive(0)
{
}

Sequential::Sequential(Random_t f, Random_t l, const Random& i, Count_t c)
  : m_min(f), m_max(l), m_increment(i.Copy()), m_consecutive(c),
    m_current(f), m_currentConsecutive(0)
{
}

Sequential::Sequential(const Sequential& c)
  : m_min(c.m_min), m_max(c.m_max),
    m_increment(c.m_increment->Copy()), m_consecutive(c.m_consecutive),
    m_current(c.m_current), m_currentConsecutive(c.m_currentConsecutive)
{
}

Random_t Sequential::Value()
{ // Return a sequential series of values
  Random_t r = m_current;
  if (++m_currentConsecutive == m_consecutive)
    { // Time to advance to next
      m_currentConsecutive = 0;
      m_current += m_increment->Value();
      if (m_current >= m_max)
        m_current = m_min + (m_current - m_max);
    }
  return r;
}

Random* Sequential::Copy() const
{
  return new Sequential(*this);
}

// Exponential methods
Random_t Exponential::Value()
{
  Random_t r = -m_mean*log(Random::Value());
  if (m_bound != 0 && r > m_bound) return m_bound;
  return r;
}

Random* Exponential::Copy() const
{
  return new Exponential(*this);
}

// Pareto methods
Random_t Pareto::Value()
{
Random_t scale = m_mean * ( m_shape - 1.0) / m_shape;

  Random_t r = (scale * ( 1.0 / pow(Random::Value(), 1.0 / m_shape)));
  DEBUG0((cout << "Pareto, scale " << scale << " r " << r << endl));
  if (m_bound != 0 && r > m_bound) return m_bound;
  return r;
}

Random* Pareto::Copy() const
{
  return new Pareto(*this);
}

// Normal methods
Random_t Normal::Value()
{
  if (m_nextValid)
    { // use previously generated
      m_nextValid = false;
      return m_next;
    }
  while(1)
    { // See Simulation Modeling and Analysis p. 466 (Averill Law)
      // for algorithm
      Random_t u1 = Random::Value();
      Random_t u2 = Random::Value();
      Random_t v1 = 2 * u1 - 1;
      Random_t v2 = 2 * u2 - 1;
      Random_t w = v1 * v1 + v2 * v2;
      if (w <= 1.0)
        { // Got good pair
          Random_t y = sqrt((-2 * log(w))/w);
          m_next = m_mean + v2 * y * sqrt(m_variance);
          if (fabs(m_next) > m_bound) m_next = m_bound * (m_next)/fabs(m_next);
          m_nextValid = true;
          Random_t x1 = m_mean + v1 * y * sqrt(m_variance);
          if (fabs(x1) > m_bound) x1 = m_bound * (x1)/fabs(x1);
          return x1;
        }
    }
}

Random* Normal::Copy() const
{
  return new Normal(*this);
}

// Empirical methods
Random_t Empirical::Value()
{ // Return a value from the empirical distribution
  // This code based (loosely) on code by Bruce Mah (Thanks Bruce!)
  if (emp.size() == 0) return 0.0; // HuH? No empirical data
  if (!validated) Validate();      // Insure in non-decreasing
  Random_t r = Random::Value();
  if (r <= emp.front().cdf)return emp.front().value; // Less than first
  if (r >= emp.back().cdf) return emp.back().value;  // Greater than last
  // Binary search
  ValueCDFVec_t::size_type bottom = 0;
  ValueCDFVec_t::size_type top = emp.size() - 1;
  while(1)
    {
      ValueCDFVec_t::size_type c = (top + bottom) / 2;
      if (r >= emp[c].cdf && r < emp[c+1].cdf)
        { // Found it
          return Interpolate(emp[c].cdf, emp[c+1].cdf,
                             emp[c].value, emp[c+1].value,
                             r);
        }
      // Not here, adjust bounds
      if (r < emp[c].cdf) top    = c - 1;
      else                bottom = c + 1;
    }
}

Random* Empirical::Copy() const
{
  return new Empirical(*this);
}

void Empirical::CDF(Random_t v, CDF_t c)
{ // Add a new empirical datapoint to the empirical cdf
  // NOTE.   These MUST be inserted in non-decreasing order
  emp.push_back(ValueCDF(v, c));
}

void Empirical::Validate()
{
  ValueCDF prior;
  for (ValueCDFVec_t::size_type i = 0; i < emp.size(); ++i)
    {
      ValueCDF& current = emp[i];
      if (current.value < prior.value || current.cdf < prior.cdf)
        { // Error
          cout << "Empirical Dist error,"
               << " current value " << current.value
               << " prior value "   << prior.value
               << " current cdf "   << current.cdf
               << " prior cdf "     << prior.cdf << endl;
          exit(1);
        }
      prior = current;
    }
  validated = true;
}

Random_t Empirical::Interpolate(CDF_t c1, CDF_t c2,
                                Random_t v1, Random_t v2, CDF_t r)
{ // Interpolate random value in range [v1..v2) based on [c1 .. r .. c2)
  return (v1 + ((v2 - v1) / (c2 - c1)) * (r - c1));
}

// Integer Empirical methods

IRandom_t IntEmpirical::IntValue()
{
  return (IRandom_t)Value();
}

Random* IntEmpirical::Copy() const
{
  return new IntEmpirical(*this);
}


Random_t IntEmpirical::Interpolate(CDF_t c1, CDF_t c2,
                                   Random_t v1, Random_t v2, CDF_t r)
{ // Interpolate random value in range [v1..v2) based on [c1 .. r .. c2)
  return ceil(v1 + ((v2 - v1) / (c2 - c1)) * (r - c1));
}

// Static Methods
void Random::GlobalSeed(Seed_t s0, Seed_t s1, Seed_t s2,
                Seed_t s3, Seed_t s4, Seed_t s5)
{
  Seed_t seeds[6];
  if (globalSeedRNG) delete globalSeedRNG;
  globalSeedRNG = new Uniform(0, 0xffffffff);
  seeds[0] = s0;
  seeds[1] = s1;
  seeds[2] = s2;
  seeds[3] = s3;
  seeds[4] = s4;
  seeds[5] = s5;
  globalSeedRNG->Seeds(seeds);
#ifdef BUGFIX
  Random::updateAllSeeds();
#endif
}

#ifdef BUGFIX
void Random::updateAllSeeds()
{
for (std::list<Random*>::iterator i = _memberList.begin();
	i != _memberList.end(); i++) {
		(*i)->Seed(RANDOM_SEED);
  }
}
#endif
