/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Random Number class
// George F. Riley.  Georgia Tech, Spring 2002

// Random is base class for all random number generators

#ifndef __random_h__
#define __random_h__

#include <vector>
#include <algorithm>

#include "common-defs.h"
#include "RngStream.h"
#ifdef BUGFIX
#include <list>
#endif
typedef double        Random_t;  // Real Random Variable
typedef unsigned long IRandom_t; // Integer Random Variable
typedef double        CDF_t;     // Cumulative distribution Function value
typedef unsigned long Seed_t;

class Uniform;
#ifdef BUGFIX
class Random;
#endif
//Doc:ClassXRef
class Random : public RngStream { 
  //Doc:Class \GTNS\ has a rich set of  random number generators.
  //Doc:Class Class {\tt Random} defines the base class functionalty
  //Doc:Class required for all random number generators.
  //Doc:Class Note: The underlying random number generation method used
  //Doc:Class by \GTNS\ is the {\tt RngStream} code by Pierre L'Ecuyer at
  //Doc:Class the University of Montreal.
public:
  typedef enum { RANDOM_SEED = 0 } RanSeed_t;
public:
  //Doc:Method
  Random() { Seed(RANDOM_SEED);
#ifdef BUGFIX
        _memberList.push_front(this);
#endif
};
    //Doc:Desc Constructor for a random number generator with a random seed.

  //Doc:Method
  explicit Random(Seed_t s) { Seed(s);
#ifdef BUGFIX
        _memberList.push_front(this);
#endif
}
    //Doc:Desc Constructor for a random number generator with a specified seed.
    //Doc:Desc The random numbers used by \GTNS\ actually use a seed value
    //Doc:Desc consisting of 6 32 bit values.  The specified seed is replicated
    //Doc:Desc six times to create the seed.
    //Doc:Arg1 Seed value to use (32 bit unsigned)
  virtual ~Random() { 
#ifdef BUGFIX
        _memberList.remove(this); 
#endif
}
  //Doc:Method
  virtual Random_t  Value();     // Returns a value [0.0 .. 1.0)
    //Doc:Desc Return a floating point random value
    //Doc:Return Floating point random value.

  // Returns integer value
  //Doc:Method
  virtual IRandom_t IntValue(){ return (IRandom_t)Value();}
    //Doc:Desc Return an integer random value.
    //Doc:Return Integer random value.

  //Doc:Method
  void    Seed(Seed_t);
    //Doc:Desc Specify a seed for the random number generators.  If the
    //Doc:Desc specified value is zero, then the RNG is seeded randomly
    //Doc:Desc by one of three methods.  First, if a global seed random 
    //Doc:Desc number generator is specified (See static method
    //Doc:Desc {\tt GlobalSeed})
    //Doc:Desc then the seed RNG is sampled for the seed.
    //Doc:Desc  If {\em UseDevRandom}
    //Doc:Desc is specified, then the linux device {\tt /dev/random} is
    //Doc:Desc sampled for the seed.  If neither of these is specified, then
    //Doc:Desc the time of day is used as a random seed.  
    //Doc:Desc If the specified value is not zero, the six copies of this
    //Doc:Desc are used for the seed.
    //Doc:Arg1 Random seed to use.

  // The Pierre L'Ecuyer RNG's use a sequence of 6 seeds
  //Doc:Method
  void    Seeds(Seed_t*);
    //Doc:Desc Seed the random number generator with six seeds.
    //Doc:Arg1 Pointer to an array of six 32 bit unsigned to use for seeds.

  virtual Random*   Copy() const;
  //Doc:Method
  static  void UseDevRandom(bool udr = true) { useDevRandom = udr;}
    //Doc:Desc Specify whether the linux device {\tt /dev/random} is to
    //Doc:Desc be used for a random seed.
    //Doc:Arg1 True if {\tt /dev/random} desired.

private:
  static  void updateAllSeeds();
  static  bool useDevRandom;  // True if using /dev/random desired
  static  int  devRandom;     // File handle for /dev/random
#ifdef BUGFIX
  static  std::list<Random*> _memberList;
#endif
protected:
  static unsigned long heuristic_sequence;
  static Uniform* globalSeedRNG;
public:
  // Use the global seed to force precisely reproducible results
  // each "RANDOM_SEED" rng will ask the global rng for the
  // next seed sequence, thereby giving the same seeds each time.
  //Doc:Method
  static void GlobalSeed(Seed_t,Seed_t,Seed_t,Seed_t,Seed_t,Seed_t) ;
    //Doc:Desc It is often desirable to create a simulation that uses
    //Doc:Desc random numbers, while at the same time is completely reproducible.
    //Doc:Desc Specifying this set of six random seeds creates a special
    //Doc:Desc random number generator used for all random seeds for
    //Doc:Desc other random number generators.  Thus, specifying
    //Doc:Desc the {tt RANDOM_SEED} value for the seed for each  created
    //Doc:Desc generator will result in a known and reproducible set of 
    //Doc:Desc seeds, and a reproducible set of samples.
    //Doc:Arg1 Seed to use (32 bit unsigned)
    //Doc:Arg2 Seed to use (32 bit unsigned)
    //Doc:Arg3 Seed to use (32 bit unsigned)
    //Doc:Arg4 Seed to use (32 bit unsigned)
    //Doc:Arg5 Seed to use (32 bit unsigned)
    //Doc:Arg6 Seed to use (32 bit unsigned)
};


//Doc:ClassXRef
class Uniform : public Random { // Uniformly distributed random variable
  //Doc:Class Creates a uniformly distributed random number generator.
public:
  //Doc:Method
  Uniform() : m_min(0), m_max(1.0) { }
    //Doc:Desc Creates a uniform random number generator in the
    //Doc:Desc range [0.0 .. 1.0)

  //Doc:Method
  Uniform(Random_t s, Random_t l) : m_min(s), m_max(l) { };
    //Doc:Desc Creates a uniform random number generator with the specified
    //Doc:Desc range.

  Uniform(const Uniform& c) : m_min(c.m_min), m_max(c.m_max) { }
  virtual Random_t Value();
  virtual Random*  Copy() const;
private:
  Random_t m_min;
  Random_t m_max;
};

//Doc:ClassXRef
class Constant : public Random { // A random variable that returns a constant
  //Doc:Class Class {\tt Constant} defines a random number generator that
  //Doc:Class returns the same value every sample.  It's not immediately
  //Doc:Class obvious
  //Doc:Class why such a generator is useful, but it is.  A number of other
  //Doc:Class \GTNS\ objects require a random number generator as parameters
  //Doc:Class to describe the behavior.  For example, application {\tt TCPSend}
  //Doc:Class uses a random number generator to determine how much data
  //Doc:Class to send.  By passing a {\tt Constant} RNG to the {\tt TCPSend}
  //Doc:Class application, it will send a known constant amount.

public:
  //Doc:Method
  Constant() : m_const(0) { }
    //Doc:Desc Construct a {\tt Constant} RNG that return zero every sample.

  //Doc:Method
  Constant(Random_t c) : m_const(c) { };
    //Doc:Desc Construct a {\tt Constant} RNG that returns the specified value
    //Doc:Desc every sample.
    //Doc:Arg1 Constant value for this RNG.

  Constant(const Constant& c) : m_const(c.m_const) { }
  //Doc:Method
  void    NewConstant(Random_t c) { m_const = c;}
    //Doc:Desc Specify a new constant RNG for this generator.
    //Doc:Arg1 Constant value for this RNG.

  virtual Random_t  Value();
  virtual IRandom_t IntValue();
  virtual Random*   Copy() const;
private:
  Random_t m_const;
};

//Doc:ClassXRef
class Sequential : public Random { // Return a sequential list of values
  //Doc:Class Class {\tt Sequential} defines a random number generator that
  //Doc:Class returns a sequential sequence.
public:
  // first, last, increment, consecutive
  //Doc:Method
  Sequential(Random_t f, Random_t l, Random_t i = 1, Count_t c = 1);
    //Doc:Desc Constructor for the {\tt Sequential} RNG.  The four parameters
    //Doc:Desc define the sequence.  For example {\tt Sequential(0, 5, 1, 2)}
    //Doc:Desc returns the sequence 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 0, 0 \ldots
    //Doc:Arg1 First value returned.
    //Doc:Arg2 Last + 1 value returned.
    //Doc:Arg3 Increment.
    //Doc:Arg4 Repeat count for  each  value.

  //Doc:Method
  Sequential(Random_t f, Random_t l, const Random& i, Count_t c = 1);
    //Doc:Desc Constructor for the {\tt Sequential} RNG.
    //Doc:Desc Differs from the first constructor only in the fact that the
    //Doc:Desc increment paramenter is a random variable.
    //Doc:Arg1 First value returned.
    //Doc:Arg2 Last + 1 value returned.
    //Doc:Arg3 Reference to a random variable for the sequence increment.
    //Doc:Arg4 Repeat count for  each  value.

  Sequential(const Sequential& c);
  virtual Random_t Value();
  virtual Random*  Copy() const;
private:
  Random_t m_min;
  Random_t m_max;
  Random*  m_increment;
  Count_t  m_consecutive;
  Random_t m_current;
  Count_t  m_currentConsecutive;
};

//Doc:ClassXRef
class Exponential : public Random { // Exponentially Distributed random var
  //Doc:Class Class {\tt Exponential} defines a random variable with an
  //Doc:Class  exponential distribution.
public:
  //Doc:Method
  Exponential() : m_mean(1.0), m_bound(0) { }
    //Doc:Desc Constructs an exponential random variable  with a mean
    //Doc:Desc value of 1.0.

  //Doc:Method
  explicit Exponential(Random_t m) : m_mean(m), m_bound(0) { };
    //Doc:Desc Constructs an exponential  random variable with the specified
    //Doc:Desc mean value.
    //Doc:Arg1 Mean value for the random variable.

  //Doc:Method
  Exponential(Random_t m, Random_t b) : m_mean(m), m_bound(b) { };
    //Doc:Desc Constructs an exponential randome variable  with the specified
    //Doc:Desc mean value and specified upper limit.  Since exponential
    //Doc:Desc distributions can theoretically return unbounded values,
    //Doc:Desc it is sometimes useful to specify  a fixed upper limit.
    //Doc:Desc Note however that when the upper limit is specified, the
    //Doc:Desc true mean of the distribution is slightly smaller than
    //Doc:Desc the mean value specified.
    //Doc:Arg1 Mean value for the random variable.
    //Doc:Arg2 Upper limit on returned values.

  Exponential(const Exponential& c) : m_mean(c.m_mean), m_bound(c.m_bound) { }
  virtual Random_t Value();
  virtual Random* Copy() const;
private:
  Random_t m_mean;  // Mean value of RV
  Random_t m_bound; // Upper bound on value (if non-zero)
};

//Doc:ClassXRef
class Pareto : public Random { // Pareto distributed random var
  //Doc:Class Class {\tt Pareto} defines a random variable with a
  //Doc:Class pareto distribution.
public:
  //Doc:Method
  Pareto() : m_mean(1.0), m_shape(1.5), m_bound(0) { }
    //Doc:Desc Constructs a pareto random variable  with a mean
    //Doc:Desc value of 1.0 and a shape (alpha) parameter of 1.5.

  //Doc:Method
  explicit Pareto(Random_t m) 
    : m_mean(m), m_shape(1.5), m_bound(0) { };
    //Doc:Desc Constructs a pareto random variable with the specified mean
    //Doc:Desc value and a shape (alpha) parameter of 1.5.
    //Doc:Arg1 Mean value for the distribution.

  //Doc:Method
  Pareto(Random_t m, Random_t s) 
    : m_mean(m), m_shape(s), m_bound(0) { };
    //Doc:Desc Constructs a pareto random variable with the specified mean
    //Doc:Desc value and a shape (alpha) parameter of 1.5.
    //Doc:Arg1 Mean value for the distribution.
    //Doc:Arg2 Shape (alpha) parameter for the distribution.

  //Doc:Method
  Pareto(Random_t m, Random_t s, Random_t b) 
    : m_mean(m), m_shape(s), m_bound(b) { };
    //Doc:Desc Constructs a pareto random variable with the specified mean
    //Doc:Desc value, shape (alpha) parameter of 1.5, and upper bound.
    //Doc:Desc Since pareto 
    //Doc:Desc distributions can theoretically return unbounded values,
    //Doc:Desc it is sometimes useful to specify  a fixed upper limit.
    //Doc:Desc Note however that when the upper limit is specified, the
    //Doc:Desc true mean of the distribution is slightly smaller than
    //Doc:Desc the mean value specified.
    //Doc:Arg1 Mean value for the random variable.
    //Doc:Arg2 Shape (alpha) parameter for the distribution.
    //Doc:Arg3 Upper limit on returned values.

  Pareto(const Pareto& c) 
    : m_mean(c.m_mean), m_shape(c.m_shape), m_bound(c.m_bound) { }
  virtual Random_t Value();
  virtual Random* Copy() const;
private:
  Random_t m_mean;  // Mean value of RV
  Random_t m_shape; // Shape parameter
  Random_t m_bound; // Upper bound on value (if non-zero)
};

//Doc:ClassXRef
class Normal : public Random { // Normally Distributed random var
  //Doc:Class Class {\tt Normal} defines a random variable with an
  //Doc:Class  exponential distribution.
public:
  //Doc:Method
  Normal() : m_mean(0.0), m_variance(1.0), m_bound(INFINITE_VALUE),
                     m_nextValid(false) { }
    //Doc:Desc Constructs an normal random variable  with a mean
    //Doc:Desc value of 0 and variance of 1.

  //Doc:Method
  Normal(Random_t m, Random_t v, Random_t b = INFINITE_VALUE)
      : m_mean(m), m_variance(v), m_bound(b), m_nextValid(false) { };
    //Doc:Desc Constructs an normal random variable with the specified
    //Doc:Desc mean and variance values.
    //Doc:Arg1 Mean value for the random variable.
    //Doc:Arg2 Variance for the random variable.
    //Doc:Arg1 Bound value for the random variable.  The bound is
    //Doc:Arg1 positive.  The Normal is bounded within +- bound.

  Normal(const Normal& c)
      : m_mean(c.m_mean), m_variance(c.m_variance), m_bound(c.m_bound) { }
  virtual Random_t Value();
  virtual Random* Copy() const;
private:
  Random_t m_mean;      // Mean value of RV
  Random_t m_variance;  // Mean value of RV
  Random_t m_bound;     // Bound on value (absolute value)
  bool     m_nextValid; // True if next valid
  Random_t m_next;      // The algorithm produces two values at a time
};

// Value/CDF pair class for Emiprical Distributions
//Doc:ClassXRef
class ValueCDF {
public:
  ValueCDF() : value(0.0), cdf(0.0){ }
  ValueCDF(Random_t v, CDF_t c) : value(v), cdf(c) { }
  ValueCDF(const ValueCDF& c) : value(c.value), cdf(c.cdf) { }
  Random_t value;
  CDF_t    cdf;
};

typedef std::vector<ValueCDF> ValueCDFVec_t;

//Doc:ClassXRef
class Empirical : public Random { // Empirical distribution random var
  //Doc:Class Defines a random variable  that has a specified, empirical 
  //Doc:Class distribution.  The distribution is specified by a
  //Doc:Class series of calls the the {\tt CDF} member function, specifying a
  //Doc:Class value and the probability that the function value is less than
  //Doc:Class the specified value.  When values are requested,
  //Doc:Class a uniform random variable is used to select a probabililty,
  //Doc:Class and the return value is interpreted linerarly between the
  //Doc:Class two appropriate points in the CDF.
public:
  //Doc:Method
  explicit Empirical(Seed_t s = RANDOM_SEED) : Random(s), validated(false) { }
    //Doc:Desc Constructor for the {\tt Empirical} random variables.
    //Doc:Arg1 Optional seed value to use for selecting points.

  virtual ~Empirical() {}
  Empirical(const Empirical& c) : validated(c.validated), emp(c.emp) { }
  virtual Random_t Value();
  virtual Random* Copy() const;
  //Doc:Method
  virtual void     CDF(Random_t, CDF_t);  // Value, prob <= Value
    //Doc:Desc Specifies  a point in the empirical distribution.
    //Doc:Arg1 The function value for this point.
    //Doc:Arg2 Probability that the function is less than or equal
    //Doc:Arg2 to the specified value.
private:
  virtual void     Validate();  // Insure non-decreasing emiprical values
  virtual Random_t Interpolate(CDF_t, CDF_t, Random_t, Random_t, CDF_t);
  bool               validated; // True if non-decreasing validated
  ValueCDFVec_t      emp;       // Empicical CDF
};

//Doc:ClassXRef
class IntEmpirical : public Empirical { // Empirical distribution, integer
  //Doc:Class Defines an empirical distribution where all values are integers.
  //Doc:Class Indentical to {\tt Empirical}, but with slightly different
  //Doc:Class interpolation between points.
  // Identical to Empirical, but slightly different interpolation
public:
  //Doc:Method
  IntEmpirical(Seed_t s = RANDOM_SEED) : Empirical(s) { }
    //Doc:Desc Constructor for {\tt IntEmpirical}
    //Doc:Arg1 Optional seed value to use for selecting points.

  virtual Random* Copy() const;
  virtual IRandom_t IntValue();
  virtual Random_t Interpolate(CDF_t, CDF_t, Random_t, Random_t, CDF_t);
};


#endif
