/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/
// Georgia Tech Network Simulator - AODV Routing class
// George F. Riley.  Georgia Tech, Summer 2003

//
// AODV routing.

#include <string>
#include <iostream>
#include <assert.h>

// Uncomment below to enable debug level 0
// #define DEBUG_MASK 0x01
#include "debug.h"
#include "routing-aodv.h"
#include "mask.h"
#include "ipaddr.h"

using namespace std;

AODVRoutingEntry::AODVRoutingEntry(){
	DBG((Stats::routingEntAllocated++));
	int i;

	ReqTimeout = 0.0;
	ReqCount = 0;
	
	SeqNo = 0;
	Hops = LastHopCount = INFINITY2;
	Expire = 0.0;
	Flags = RTF_DOWN;

	for (i=0; i<MAX_HISTORY; i++)
		DiscLatency[i] = 0.0;
	
	HistoryIndex = 0;
	RequestLastTTL = 0;

	pInterface= 0; 
	DstIPAddr = IPADDR_NONE;
	Address = IPADDR_NONE; //this is actually next hop in ad hoc
}

AODVRoutingEntry::AODVRoutingEntry(Interface *pIf, IPAddr_t a){
	DBG((Stats::routingEntAllocated++));
	int i;

	ReqTimeout = 0.0;
	ReqCount = 0;
	
	SeqNo = 0;
	Hops = LastHopCount = INFINITY2;
	Expire = 0.0;
	Flags = RTF_DOWN;

	for (i=0; i<MAX_HISTORY; i++)
		DiscLatency[i] = 0.0;
	
	HistoryIndex = 0;
	RequestLastTTL = 0;

	pInterface= pIf;
	DstIPAddr = a;
	Address = IPADDR_NONE; //this is the next hop in ad hoc
}
	
AODVRoutingEntry::~AODVRoutingEntry() {
	// Delete each members in NeighborList and PrecursorList;
	 DBG((Stats::routingEntDeleted++));
	 while(!PrecursorList.empty()) {
		AODVPrecursor *pc = PrecursorList.front();
		PrecursorList.pop_front();
		if(pc)
		    delete pc;
	 }
	while(!NeighborList.empty()) {
		AODVNeighbor *nb = NeighborList.front();
		NeighborList.pop_front();
		if(nb)
			delete nb;
	}	
}

void
AODVRoutingEntry::NeighborInsert(IPAddr_t id) {
	//Insert the neighbor id into the list.		
	AODVNeighbor *pNeighbor = new AODVNeighbor(id);
	assert(pNeighbor);
	pNeighbor->Expire = 0;
	NeighborList.push_front(pNeighbor);
}	

AODVNeighbor *
AODVRoutingEntry::NeighborLookup(IPAddr_t id) {
	//look for if the neighbor is in the list
	for(AODVNeighborList_t::iterator i=NeighborList.begin();
					i!=NeighborList.end(); i++ ) {
		if((*i)->Address == id)
			return (*i);
	}
	return nil;
}

void
AODVRoutingEntry::PrecursorInsert(IPAddr_t id) {
	if(PrecursorLookup(id) == NULL) {
		AODVPrecursor *pPrecursor = new AODVPrecursor(id);
		assert(pPrecursor);
		PrecursorList.push_front(pPrecursor);
	}
}

AODVPrecursor *
AODVRoutingEntry::PrecursorLookup(IPAddr_t id) {
	for(AODVPrecursorList_t::iterator i = PrecursorList.begin(); 
					i!=PrecursorList.end(); i++) {
		if((*i)->Address == id)
			return (*i);
	}
	return NULL;
}

void 
AODVRoutingEntry::PrecursorDelete(IPAddr_t id) {
	for(AODVPrecursorList_t::iterator i = PrecursorList.begin();
					i!=PrecursorList.end(); i++) {
		if((*i)->Address == id) {
			AODVPrecursor *pPrecursor = *i;
			PrecursorList.erase(i);
			delete pPrecursor;
			break;
		}
	}
}

// if no parameters, PrecursorDelete is to empty!
void 
AODVRoutingEntry::PrecursorDelete() {
	 while(!PrecursorList.empty()) {
		AODVPrecursor *pPrecursor = PrecursorList.front();
		PrecursorList.pop_front();
		if(pPrecursor)
				delete pPrecursor;
	}
}

bool
AODVRoutingEntry::PrecursorEmpty(void) {
	return PrecursorList.empty();
}

//
// The routing table related functions used by normal applications.
// TODO:
//      Maybe some functions to be erased.
//
void 
RoutingAODV::Default(AODVRoutingEntry r) {
  defaultRoute = new AODVRoutingEntry(r); //TODO: define a copyise contructor.
 // defaultRoute = r; // Set the default route
  DEBUG0((cout << "RoutingAODV::Default if " << r.pInterface
          << " ip " << (string)IPAddr(r.Address) << endl));
}

RoutingAODV::~RoutingAODV() {
	if(defaultRoute)
		delete defaultRoute;
}
//TODO:
//NOTE: Probably network mask will be used later.
// 
AODVRoutingEntry *
RoutingAODV::Add(IPAddr_t a, Mask_t m, Interface* i, IPAddr_t ip){
	assert(i!=NULL);//test purpose.
	AODVRoutingEntry *pRt = new AODVRoutingEntry(i, ip);
  	fib.push_front(pRt); // Insert in FIB
	return pRt;
}

AODVRoutingEntry * 
RoutingAODV::Lookup(Node*, IPAddr_t t) { 
	// Lookup routing entry for specified target address
	for (AODVFIB_t::reverse_iterator ri = fib.rbegin(); 
					ri != fib.rend(); ++ri) {
		 if((*ri)->DstIPAddr == t)
		 	return *ri; // Return the pointer to routing entry
    }
	return NULL; //in aodv no default route. 
}

RoutingAODV* 
RoutingAODV::Clone() {
		
	return new RoutingAODV();
}

RoutingAODV::RType_t RoutingAODV::Type() {
		
  return AODV; //defined in routing.h
   
}

void    RoutingAODV::InitializeRoutes(Node*){ 
	// Initialization (none needed for aodv routing
}

bool    RoutingAODV::NeedInit(){ 
	// Initialization not needed for aodv routing
  return false;
}

Size_t RoutingAODV::Size() const{ 
	// For statistics, return the size (bytes) of the FIB
  return fib.size();
}
//NOTE:
//Normally the following function should be called
//because aodv routing entry maintained by aodv 
//"applications" (application-aodv.{cc,h}
//
void RoutingAODV::Delete(IPAddr_t id) {
	
	AODVRoutingEntry *rt = Lookup(nil, id);
	if(rt->DstIPAddr == id) {
        fib.remove(rt); // delete 
		delete rt; //remove it from memory
	}
	//TODO: add some debug info here.
}
