/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

* $Id: routing-dsr-options.cc,v 1.1.1.1 2006/12/13 13:09:35 dnguyen Exp $

*/

#include "routing-dsr.h"
#include "node.h"

/*
  ------------------------------------------------------
  DSROptionsHeader
  ------------------------------------------------------
*/

DSROptionsHeader::DSROptionsHeader()
    : IPV4Options(Number)
{
}

DSROptionsHeader::DSROptionsHeader(const DSROptionsHeader& r)
    : IPV4Options(Number)
{
  for (PDUVec_t::size_type i = 0; i < r.options.size(); ++i)
    options.push_back(r.options[i]->Copy());
}

DSROptionsHeader::~DSROptionsHeader()
{
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    delete options[i];
}

Size_t DSROptionsHeader::Size() const
{
  Count_t sz = 4;    // fixed portion
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    sz += options[i]->Size();
  return sz;
}

void DSROptionsHeader::Trace(Tfstream& tos, Bitmap_t b)
{
  tos << " DSR";
  for (PDUVec_t::size_type i = 0; i < options.size(); ++i)
    {
      options[i]->Trace(tos, b);
    }
}

/*
  ------------------------------------------------------
  RouteRequestOption
  ------------------------------------------------------
*/

RouteRequestOption::RouteRequestOption(Word_t id, IPAddr_t d, IPAddr_t s)
    : DSROptions(ROUTEREQUEST_ID),
      seqID(id), dst(d), src(s)
{
}

Size_t RouteRequestOption::Size() const
{
  Count_t sz = 8
      /*
	DSROptions::Size()    // 2
	+ sizeof(seqID)       // 2
	+ sizeof(dst)         // 4
      */
      + 4 * routeRecord.size();
  return sz;
}

bool RouteRequestOption::IsIPAdded(IPAddr_t ip)
{
  if (ip == src) return true;
  SourceRoute_t::iterator i;
  for (i = routeRecord.begin(); i != routeRecord.end(); ++i)
    if (ip == *i) return true;
  return false;
}

void RouteRequestOption::Trace(Tfstream& tos, Bitmap_t b)
{
  tos << " REQ";
  tos << " " << seqID;
  tos << " " << src;
  tos << " " << dst;
  for (SourceRoute_t::size_type i = 0; i < routeRecord.size(); ++i)
    {
      if(i)
        tos << "->";
      else
        tos << " ";

      if(ROUTE_DOTTED)
        tos << IPAddr::ToDotted(routeRecord[i]);
      else
        tos << (routeRecord[i] & ROUTE_MASK);

    }
}

/*
  ------------------------------------------------------
  RouteReplyOption
  ------------------------------------------------------
*/

RouteReplyOption::RouteReplyOption(SourceRoute_t sr, Word_t id)
    : DSROptions(ROUTEREPLY_ID),
      recordedRoute(sr), seqID(id)
{
}

Size_t RouteReplyOption::Size() const
{
  Count_t sz = 4
      + 4 * recordedRoute.size();
  return sz;
}

void RouteReplyOption::Trace(Tfstream& tos, Bitmap_t b)
{
  tos << " REP";
  tos << " " << seqID;
  for (SourceRoute_t::size_type i = 0; i < recordedRoute.size(); ++i)
    {
      if(i)
        tos << "->";
      else
        tos << " ";

      if(ROUTE_DOTTED)
        tos << IPAddr::ToDotted(recordedRoute[i]);
      else
        tos << (recordedRoute[i] & ROUTE_MASK);
    }
}

/*
  ------------------------------------------------------
  RouteErrorOption
  ------------------------------------------------------
*/

RouteErrorOption::RouteErrorOption(Byte_t e, IPAddr_t s, IPAddr_t d)
    : DSROptions(ROUTEERROR_ID),
      errorType(e), salvage(0), errorSrc(s), errorDst(d), info(0)
{
}

Size_t RouteErrorOption::Size() const
{
  return 16;
}

void RouteErrorOption::Trace(Tfstream& tos, Bitmap_t b)
{
  tos << " ERR";
  tos << " " << errorType;
  tos << " " << salvage;
  tos << " " << IPAddr::ToDotted(errorSrc);
  tos << " " << IPAddr::ToDotted(errorDst);
  tos << " " << info;
}

/*
  ------------------------------------------------------
  SourceRouteOption
  ------------------------------------------------------
*/

SourceRouteOption::SourceRouteOption(SourceRoute_t sr)
    : DSROptions(SOURCEROUTE_ID),
      salvage(0), segsLeft(0), route(sr), hopIndex(0)
{
}

RoutingEntry SourceRouteOption::LookupNextHop(Node* current)
{
  Interface* out_inf = (Interface*)(current->Interfaces()[0]);
  return RoutingEntry(out_inf, route[hopIndex++]);
}

Size_t SourceRouteOption::Size() const
{
  Count_t sz = 4
      + 4 * route.size();
  return sz;
}

void SourceRouteOption::Trace(Tfstream& tos, Bitmap_t b)
{
  tos << " SR";
  for (SourceRoute_t::size_type i = 0; i < route.size(); ++i)
    {
      if(i)
        tos << "->";
      else
        tos << " ";

      if(ROUTE_DOTTED)
        tos << IPAddr::ToDotted(route[i]);
      else
        tos << (route[i] & ROUTE_MASK);
    }
  tos << " " << salvage;
  tos << " " << segsLeft;
}
