/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Manual Routing class
// George F. Riley.  Georgia Tech, Spring 2002

// Manually computed and specified routes
// Manual routing has no way to delete or aggregate entries

#include <vector>
#include <map>
#include <string>
#include <iostream>

// Uncomment below to enable debug level 0
// #define DEBUG_MASK 0x01
#include "debug.h"
#include "routing.h"
#include "routing-manual.h"
#include "mask.h"
#include "ipaddr.h"

using namespace std;

void RoutingManual::Default(RoutingEntry r)
{
  defaultRoute = r; // Set the default route
  DEBUG0((cout << "RoutingManual::Default if " << r.interface
          << " ip " << (string)IPAddr(r.nexthop) << endl));
}

RoutingEntry RoutingManual::GetDefault()
{
  return defaultRoute;
}

void RoutingManual::Add( IPAddr_t a, Mask_t m, Interface* i, IPAddr_t ip)
{
  fib[m].insert(FIB_t::value_type(a & (Mask_t)Mask(m),
                                  RoutingEntry(i, ip))); // Insert in FIB
}

RoutingEntry RoutingManual::Lookup(Node*, IPAddr_t t)
{ // Lookup routing entry for specified target address
  // Lookup in FIB starting with longest mask working backwards
  for (FIBMap_t::reverse_iterator ri = fib.rbegin(); ri != fib.rend(); ++ri)
    {
      DEBUG0((cout << "RoutingManual::Lookup, checking mask lth " 
              << ri->first << endl));
      FIB_t::iterator i = ri->second.find(t & Mask(ri->first));
      if (i != ri->second.end()) return i->second; // Return the routing entry
    }
  return defaultRoute; // Return the default route
}

RoutingEntry RoutingManual::LookupFromPDU(PDU*)
{// Find from PDU (NixVector)..not used for manual
  return RoutingEntry();
}

Routing* RoutingManual::Clone()
{
  return new RoutingManual();
}

Routing::RType_t RoutingManual::Type()
{
  return MANUAL;
}

void    RoutingManual::InitializeRoutes(Node*)
{ // Initialization (none needed for manual routing
}

bool    RoutingManual::NeedInit()
{ // Initialization not needed for manual routing
  return false;
}

Size_t RoutingManual::Size() const
{ // For statistics, return the size (bytes) of the FIB
  Size_t s = 0;
  for (FIBMap_t::const_iterator i = fib.begin(); i != fib.end(); ++i)
    {
      s += i->second.size() * (sizeof(IPAddr_t) + sizeof(RoutingEntry));
    }
  return s;
}

