/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Event Scheduler
// George F. Riley, Georgia Tech.  Spring 2002

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <math.h>

#ifdef HAVE_PSINFO
// Added for PSC
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#endif

//#define DEBUG_MASK 0x10
#include "debug.h"
#include "simulator.h"
#include "timer.h"
#include "routing.h"
#include "node.h"
#include "globalstats.h"
#include "tcpdemux.h"
#include "udpdemux.h"
#define OSPF6D
#ifdef OSPF6D
#include "ospfdemux.h"
#endif
#include "ipv4.h"
#include "link-rti.h"
#include "application-tcpsend.h"
#include "memreuse.h"
#include "backplane.h"
#include "stringtok.h"
#include "link-real.h"
#include "eventcq.h"
#include "node.h"
#ifdef ETRACE
#include "etrace.h"
#endif
#ifdef SIM //ESTATS
#include "estat.h"
#endif //SIM

#ifdef  HAVE_RTI
extern "C" {
#include "rticore.h"
#include "brti.h"
}
#endif

#ifndef HAVE_QT
#define COMPACT
#endif

#ifndef COMPACT
#include "qtwindow.h"
#endif

using namespace std;

SystemId_t Simulator::systemId = 0;     // System id (f distributed simulation)
bool       Simulator::granted = false;  // True if TimeAdvanceGrant
Time_t     Simulator::grantedTime = 0.0;// Time of last TAG (if using RTI)
Simulator* Simulator::instance;         // Points to single Simulator instance
PHVec_t    Simulator::progressHooks;    // List of callbacks on progress
RectRegion Simulator::boundingBox;      // Min/Max X/Y positions
// Custom background callback
CustomBackground_t Simulator::pCustomBackground = nil;
NodeSelected_t     Simulator::pNodeSelected = nil;
bool       Simulator::recorderEnabled = false;
Mult_t     Simulator::recorderSpeedup = 10.0;

void ProgressTimer::Timeout(TimerEvent* ev)
{ // Process the progress timer expiration
  ProgressEvent* pe = (ProgressEvent*)ev;
  if  (Simulator::progressHooks.empty())
    { // No hooks, just print simple progress message
#ifdef SIM
			struct timeval now;
			struct timeval diff;
			
			gettimeofday(&now, (struct timezone *) NULL);
			timersub(&now, &start_time, &diff);
      cout << "Progress to " << Simulator::Now();
      cout << " sec (sim time), "
        << diff.tv_sec << " sec (real time); "
        << "Events: " << Simulator::TotalEventsScheduled()
				<<endl;
#ifdef DEBUG_MOBILITY
		  const NodeVec_t& nodes = Node::GetNodes();
		  for (NodeVec_t::size_type i = 0; i < nodes.size(); ++i)
		  {
				nodes[i]->UpdateLocation();
				Location L = nodes[i]->GetLocation();
				printf("NODE:%d location (%f,%f)\n", nodes[i]->Id(), L.X(), L.Y());
			}
#endif 			
#else
      cout << "Progress to " << Simulator::Now()  << endl;
#endif //SIM
    }
  else
    { // Call each hook
      for (PHVec_t::iterator i = Simulator::progressHooks.begin();
           i != Simulator::progressHooks.end(); ++i)
        {
          (*i)(Simulator::Now()); // Call the user progress function
        }
    }
  Time_t interval = pe->interval;
  Simulator::instance->progressEvent = nil;
  delete pe;
  // If only remaining event is progress, we are out of events and should exit
  // othersize reschedule the progress event
  if (Simulator::instance->EventListSize() || Simulator::instance->UsingRTI())
    { // Not empty (or distributed sim), schedule new one
      Simulator::instance->ProgressResched(interval);
    }
  else
    {
      Simulator::instance->Halt();
    }
}


Simulator::Simulator() : 
  verbose(false), verbosemod(1000), cleanUp(false), 
  lastTopologyChange(0),
  simtime(0),
  totevs(0), totrm(0), totevp(0), totevc(0),
  evlSize(0),
  halted(false), uid(0),
  progressInterval(0), silent(false), progressEvent(0),
  startRouteTime(0), startRunTime(0),
  qtWin(nil), animate(false), playbackAnimate(false), animateWirelessTx(false),
  smallestX(0), smallestY(0),
  largestX(0), largestY(0),
  firstLocation(true), hasMobility(false),
  usingNER(true)
{
  grantedTime = INFINITE_TIME;
  usingRTI = false;
  instance = this;
  //  hint = eventList.end();
  // Call the initializers for various protocols
  TCPDemux::Initialize();
  UDPDemux::Initialize();
#ifdef OSPF6D
  OSPF6Demux::Initialize();
#endif
  Stats::initialMemory = ReportMemoryUsage();
}

Simulator::Simulator(SystemId_t sysId) : 
  verbose(false), verbosemod(1000), cleanUp(false),
  lastTopologyChange(0),
  simtime(0),
  totevs(0), totrm(0), totevp(0), totevc(0),
  evlSize(0),
  halted(false), uid(0),
  progressInterval(0), silent(false), progressEvent(0),
  startRouteTime(0), startRunTime(0),
  qtWin(nil), animate(false), playbackAnimate(false), animateWirelessTx(false),
  smallestX(0), smallestY(0),
  largestX(0), largestY(0),
  firstLocation(true), hasMobility(false),
  usingNER(true)
{
#ifndef HAVE_RTI
  cout << "Simulator constructor called with sysId param, but no RTI" << endl;
  cout << "This is likely incorrect" << endl;
#endif
  systemId = sysId;
  usingRTI = true;
  instance = this;
  //  hint = eventList.end();
  // Call the initializers for various protocols
  TCPDemux::Initialize();
  UDPDemux::Initialize();
#ifdef HAVE_RTI
  char* argv1[2] = {"gtnets", NULL };
  int   argc1 = 1;
  RTI_Init(argc1, argv1);  // Initilize the RTI kit
#endif
  Stats::initialMemory = ReportMemoryUsage();
  if (!silent)
    {
      if (usingNER) cout << "Using NER Request Method" << endl;
      else cout << "Using TAR Request Method" << endl;
    }
}

Simulator::~Simulator()
{
  DEBUG0((cout << "Hello from Simulator Destructor" << endl));
  instance = nil;
  const NodeVec_t& nodes = Node::GetNodes();
  if (cleanUp)
    {
      for (NodeVec_t::size_type i = 0; i < nodes.size(); ++i)
        {
          DEBUG0((nodes[i]->DumpNV())); // Print debug nixvec info
          delete nodes[i];
        }
      DEBUG0((ReuseBase::DumpQ())); // Print dbg info from memory reuse buffers
    }
  Node::Clear(); // Remove all entries
  Trace::Instance()->Close();
#ifdef ETRACE
  ETrace::Instance()->Close();
#endif
}

void Simulator::Handle(Event* e, Time_t t)
{
  SimulatorEvent* ev = (SimulatorEvent*)e;
  switch(e->event)
    {
      case SimulatorEvent::HALT :
        halted = true;
        if (!silent) cout <<  "Simulator halting at time " << t << endl;
#ifdef SIM
				Finalize();
#endif //SIM
        break;
      case SimulatorEvent::DELETE_OBJECT :
        delete ev->object; // delete the object specified by caller
        break;
      case SimulatorEvent::NOTIFY:
        ((NotifyHandler*)ev->object)->Notify(ev->c); // Call the notification
        break;
      case SimulatorEvent::NODE_DOWN:
        ev->node->Down();
        break;
      case SimulatorEvent::NODE_UP:
        ev->node->Up();
        break;
      case SimulatorEvent::INTERFACE_DOWN:
        ev->iFace->Down();
        break;
      case SimulatorEvent::INTERFACE_UP:
        ev->iFace->Up();
        break;
    }
  delete e;
}

#ifdef SIM
void Simulator::Finalize()
{
  const NodeVec_t& nodes = Node::GetNodes();
  for (NodeVec_t::size_type i = 0; i < nodes.size(); ++i)
  {
		//Finalizing the ospf6 daemon
		OSPF6DInstance * ospf6_inst = nodes[i]->GetOspf6dInst();
		if(ospf6_inst)
			ospf6_inst->Finalize();
	}
}
#endif //SIM

void Simulator::Cancel(Event* p)
{
  EventMap_t::iterator i = eventList.find(KeyPair(p->Time(), p->uid));
  //EventSet_t::iterator i = eventList.find(p);
  if (i != eventList.end())
    {
      eventList.erase(i);
      //      hint = eventList.end(); 
      totrm++;
      totevc++;
      evlSize--;
    }
}

void Simulator::Schedule(Event* p)
{
  Schedule(p, p->Time() - Now());
}

void Simulator::Schedule(Event* p, Time_t t)
{
  Schedule(p, t, p->handler);
}

void Simulator::Schedule(Event* p, Time_t t, Handler* h)
{ // Schedule event for t secs in future
  p->uid = uid++;
  p->Time(t + Now());
  p->handler = h;
  //hint = eventList.insert(
  //           hint,
  //           EventMap_t::value_type(KeyPair(p->time, p->uid), p));
  eventList.insert(
             EventMap_t::value_type(KeyPair(p->Time(), p->uid), p));
  //eventList.insert(p);
  evlSize++;
  totevs++;  
  if (verbose && ((totevs % verbosemod) == 0))
    {
      cout << "STLSched::total " << totevs
           << " current size " << totevs - totrm << endl;
    }
}

void Simulator::Schedule(Event* p, Time_t t, Handler& h)
{
  Schedule(p, t, &h);
}

void Simulator::ScheduleEarly(Event* p, Time_t t, Handler* h)
{ // Schedule event for t secs in future, using STL hint
  p->uid = uid++;
  p->Time(t + Now());
  p->handler = h;
  // For now no difference from normal schedule
  eventList.insert(
             EventMap_t::value_type(KeyPair(p->Time(), p->uid), p));
  //eventList.insert(eventList.begin(), p);
  //eventList.insert(eventList.end(), p); // Just testing!!!!!!!!!!!
  evlSize++;
  totevs++;  
}

void Simulator::Progress(Time_t interval)
{
  if (progressEvent)
    {
      Cancel(progressEvent); // Cancel any pending
      delete progressEvent;  // And remove the event
      progressEvent = nil;
    }
  // Schedule new event for "interval" secs in future
  progressInterval = interval;
  ProgressResched(interval);
}

// Schedule node and interface failures
SimulatorEvent* Simulator::NodeDownAt(Node* n, const Random& r)
{
  Random& r1 = (Random&)r;   // Small trick to get around "const"
  Time_t t = r1.Value();
  if (t < Now()) return nil; // Can't do it in the past
  SimulatorEvent* e = new SimulatorEvent(SimulatorEvent::NODE_DOWN);
  e->node = n;
  Schedule(e, t - Now(), this);
  return e;
}

SimulatorEvent* Simulator::NodeUpAt(Node* n, const Random& r)
{
  Random& r1 = (Random&)r;   // Small trick to get around "const"
  Time_t t = r1.Value();
  if (t < Now()) return nil; // Can't do it in the past
  SimulatorEvent* e = new SimulatorEvent(SimulatorEvent::NODE_UP);
  e->node = n;
  Schedule(e, t - Now(), this);
  return e;
}


SimulatorEvent* Simulator::InterfaceDownAt(Interface* i, const Random& r)
{
  Random& r1 = (Random&)r;   // Small trick to get around "const"
  Time_t t = r1.Value();
  if (t < Now()) return nil; // Can't do it in the past
  SimulatorEvent* e = new SimulatorEvent(SimulatorEvent::INTERFACE_DOWN);
  e->iFace = i;
  Schedule(e, t - Now(), this);
  return e;
}


SimulatorEvent* Simulator::InterfaceUpAt(Interface* i, const Random& r)
{
  Random& r1 = (Random&)r;   // Small trick to get around "const"
  Time_t t = r1.Value();
  if (t < Now()) return nil; // Can't do it in the past
  SimulatorEvent* e = new SimulatorEvent(SimulatorEvent::INTERFACE_UP);
  e->iFace = i;
  Schedule(e, t - Now(), this);
  return e;
}

Event* Simulator::DeQueue()		// next event (removes from q)
{
  EventMap_t::iterator i = eventList.begin();
  //EventSet_t::iterator i = eventList.begin();
  if (i == eventList.end()) return nil; // Empty on deque
  Event* p = i->second;
  //Event* p = *i;
  eventList.erase(i);
  //  hint = eventList.end();
  totrm++;
  evlSize--;
  return p;
}

Event* Simulator::PeekEvent()
{ // Return next event without removal
  EventMap_t::iterator i = eventList.begin();
  //EventSet_t::iterator i = eventList.begin();
  if (i == eventList.end()) return nil; // Empty on deque
  return eventList.begin()->second;
  //return *(eventList.begin());
}

void Simulator::PrintStats()
{
  if(!silent)
    cout << "Simulator::PrintStats, totevs " << totevs
         << " totrm " << totrm 
         << " totevp " << totevp 
         << " totevc " << totevc 
         << " size " << eventList.size() << endl;
}

Count_t extractCount;
extern int PRTIKIT_UsingMCAST;
void Simulator::Run()
{
#ifdef SIM
	gettimeofday(&progressTimer.start_time, (struct timezone *) NULL);
#endif //SIM
#ifdef HAVE_RTI
  // Perform some RTI related initialization
  if (usingRTI)
    {
      if (!silent) cout << "Using lookahead " << LinkRTI::minTxTime << endl;
      if (usingNER)
        RTI_SetLookAhead(LinkRTI::minTxTime);
      else
        RTI_SetLookAhead(0);  // for the TAR version (default)
      RTIKIT_Barrier();  // Wait for others to become ready also
    }
#endif

  halted = false;
  Stats::modelMemory = ReportMemoryUsage();
  startRouteTime = GetSec(); // Log CPU time used before route computations
  DEBUG0((cout << "Simulator::Run, needInit() " 
          << Routing::Default()->NeedInit() << endl));
  if (Routing::Default()->NeedInit())
    { // Need to initialize routes at each node
      const NodeVec_t& n = Node::GetNodes(); // Get the global list of nodes
      for (NodeVec_t::size_type i = 0; i < n.size(); ++i)
        { // Call the initializer for each node
          n[i]->InitializeRoutes();
        }
    }
  DEBUG0((cout << "Done initializing routes..." << endl));
  Stats::routingMemory = ReportMemoryUsage();
  startRunTime = GetSec(); // Log CPU time used before RUN
#ifdef HAVE_RTI
  if (usingRTI)
    {
      while(!halted)
        {
          Event* e;
          e = PeekEvent();
          Time_t     requestTime = INFINITE_TIME;
          if (e) requestTime = e->Time();
          if (requestTime > grantedTime)
            { // Need to ask RTI if ok to advance
              granted = false;
              DEBUG(4,(cout << "Requesting adv to " << requestTime
                       << endl));
              if (usingNER)
                RTI_NextEventRequest(requestTime);
              else
                RTI_TimeAdvanceRequest(grantedTime + LinkRTI::minTxTime);
              DEBUG(4,(cout << "Ticking after TAR" << endl));
              while(!granted) Core_tick(); // Wait for approval
              DEBUG(4,(cout << "Ticked, grantedTime " << grantedTime << endl));
            }
          if (grantedTime == INFINITE_TIME) break;
          DEBUG(4,(cout << "Granted to " << grantedTime << endl));
          e = PeekEvent(); // Get fresh copy in case new message arrived
          if (e && e->Time() < simtime)
            cout << "HuH?  Time backing up in Simulator::run()" << endl;
          if (e && e->Time() <= grantedTime)
            {
              e = DeQueue();          // Ok to process, remove from queue
              if (e->handler)
                {
                  simtime = e->Time();
                  totevp++;
                  DEBUG(4,(cout << "Handling ev at time " << e->Time() << endl));
                  e->handler->Handle(e, e->Time()); // Process the event
                }
              else
                {
                  if (!silent)
                    {
                      cout << "Skipping event " << e->uid
                           << "  at time " << e->Time()
                           << ", no handler " << endl;
                    }
                }
              if (!usingNER) Core_tick(); // Process message deliveries
            }
#ifndef COMPACT
          if (animate) qtWin->ProcessEvents();
#endif
        }
    }
  else
    { // Not using RTI, more efficient version
      while(!halted)
        {
          Event* e = DeQueue();
          if (!e) break; // No more events
          if (e->handler)
            {
              simtime = e->Time();
              totevp++;
              DEBUG(4,(cout << "Handling ev at time " << e->Time() << endl));
              e->handler->Handle(e, e->Time()); // Process the event
            }
          else
            {
                  if (!silent)
                    {
                      cout << "Skipping event " << e->uid
                           << "  at time " << e->Time()
                           << ", no handler " << endl;
                    }
            }
#ifndef COMPACT
          if (animate) qtWin->ProcessEvents();
#endif
        }
    }
#else
  // No RTI version
  while(!halted)
    {
      Event* e = DeQueue();
      if (!e) break; // No more events
      if (e->handler)
        {
          if (e->Time() < simtime) 
            cout << "HuH? Time backing up from " << simtime
                 << " to " << e->Time() << endl;
          simtime = e->Time();
          totevp++;
          DEBUG(4,(cout << "Handling ev at time " << e->Time() << endl));
          e->handler->Handle(e, e->Time()); // Process the event
        }
      else
        {
                  if (!silent)
                    {
                      cout << "Skipping event " << e->uid
                           << "  at time " << e->Time()
                           << ", no handler " << endl;
                    }
        }
#ifndef COMPACT
      //if (animate) qtWin->ProcessEvents();
#endif
    }
#endif
  DEBUG0((cout << "Simulator::Run exit event loop" << endl));
  if (!EventListSize() && !halted)
    if(!silent)cout << "Simulator::Run, event list exhausted" << endl;
  if (usingRTI) if (!silent) cout << " grantedTime " << grantedTime << endl;
#ifdef HAVE_RTI       
  // Now keep trying to advance to INFINITE_TIME, to insure all others have exited
  if (usingRTI)
    {
      while(grantedTime < INFINITE_TIME)
        {
          granted = false;
          RTI_NextEventRequest(INFINITE_TIME);
          while(!granted) Core_tick(); // Wait for approval
        }
      // And enter the final barrier
      RTIKIT_FinalBarrier();
      // Hack per Kalyan
      for (int x = 0; x < 100; ++x)
        {
          struct timeval tv;
          tv.tv_sec = 0;
          tv.tv_usec = 10000;
          FM_extract(~0);
          select(0, NULL, NULL, NULL, &tv); // 10ms delay
        }
      // End hack
      if (!silent)
        { 
          cout << "Total RTI updates "  << LinkRTI::nUpdates << endl;
          cout << "Total RTI reflects " << LinkRTI::nReflects << endl;
          cout << "Total RTI ignored " << LinkRTI::ignoredReflects << endl;
          TM_PrintStats(); // Print debug statistics
        }
    }
#endif
  Stats::finalMemory = ReportMemoryUsage();
}

void Simulator::StopAt(Time_t t)
{
SimulatorEvent* e = new SimulatorEvent(SimulatorEvent::HALT);

 // Schedule the halt event
 Schedule(e, t, this);
}

void Simulator::Halt()
{
  halted = true;
}


#ifdef HAVE_RTI
// The following are defined
extern unsigned long FM_nodeid;
extern unsigned long FM_numnodes;
#endif

void Simulator::UseBackplane()
{
  //#ifdef HAVE_RTI
  // Initialize the backplane interface
  DSHandle_t dsimHandle = InitializeDSim(systemId);
  // Register the TCP and IP data items
  TCP::RegisterBackplane(dsimHandle);
  IPV4::RegisterBackplane(dsimHandle);
  // Notify the backplane that registration is complete
  //RegistrationComplete(systemId, FM_numnodes); remove for debugging
  //#endif
}




void Simulator::DeleteObject(Object* o)
{ // In C++, objects cannot delete themselves (ie. can't say "delete this")
  // Since our design is such that the objects (such as protocol objects)
  // know when they should be deleted, we provide an interface
  // for the simulator to delete them by scheduling an event at
  // current time, which would be the next event scheduled.
  SimulatorEvent* ev = new SimulatorEvent(SimulatorEvent::DELETE_OBJECT);
  ev->object = o;
  Schedule(ev, 0.0, this); // Schedule for 0 secs in future
}

SystemId_t Simulator::SystemId()
{ // Return the system id
  return systemId;
}

void Simulator::ProgressHook(ProgressHook_t ph)
{ // Add a new progress callback
  progressHooks.push_back(ph);
}

void Simulator::AddNotify(NotifyHandler* n, Time_t t, void* c)
{ // Schedule a notification event at t seconds in the future
  SimulatorEvent* ev = new SimulatorEvent(SimulatorEvent::NOTIFY);
  ev->object = (Object*)n;
  ev->c = c;
  Schedule(ev, t, this);
}

void Simulator::TopologyChanged()
{ // Note time of last topology change.  Used by NixVector validation
  lastTopologyChange = Now();
}

// debug
bool Simulator::IsScheduled(Event* p )
{
  EventMap_t::iterator i = eventList.find(KeyPair(p->Time(), p->uid));
  //EventSet_t::iterator i = eventList.find(p);
  return i != eventList.end();
}

void Simulator::DisplayTopology()
{
#ifdef HAVE_QT
  if (!qtWin) qtWin = new QTWindow();
  qtWin->DisplayTopology();
  delete qtWin; // Destroy qt app and windows before continuing
  qtWin = nil;
#endif
}

void Simulator::DisplayTopologyAndReturn()
{
#ifdef HAVE_QT
  if (!qtWin) qtWin = new QTWindow();
  qtWin->DisplayTopologyAndReturn();
  qtWin->ProcessEvents();
#endif
}

void Simulator::UpdateTopology()
{
#ifdef HAVE_QT
  if (qtWin)
    {
      qtWin->UpdateTopology();
    }
#endif
}

void Simulator::StartAnimation(Time_t when, bool initialPause)
{
#ifdef HAVE_QT
  if (!qtWin) qtWin = new QTWindow();
  animate = true;
  qtWin->DisplayTopologyAndReturn();
  // Schedule the start animation event
  QTEvent* ev;
  if (initialPause)
    {
      ev = new QTEvent(QTEvent::PAUSE);
    }
  else
    {
      ev = new QTEvent(QTEvent::START);
    }
  Simulator::instance->Schedule(ev, when - Simulator::Now(), qtWin);
#endif
}

void Simulator::StopAnimation(Time_t when)
{
#ifdef HAVE_QT
  if (!qtWin) qtWin = new QTWindow();
  // Schedule the stop animation event
  QTEvent* ev = new QTEvent(QTEvent::STOP);
  Simulator::instance->Schedule(ev, when - Simulator::Now(), qtWin);
#endif
}

void Simulator::PauseAnimation(Time_t when)
{
#ifdef HAVE_QT
  if (!qtWin) qtWin = new QTWindow();
  // Schedule the pause animation event
  QTEvent* ev = new QTEvent(QTEvent::PAUSE);
  Simulator::instance->Schedule(ev, when - Simulator::Now(), qtWin);
#endif
}

void Simulator::AnimationUpdateInterval(Time_t t)
{ // Set the initial animation update interval
#ifdef HAVE_QT
  if (!qtWin) qtWin = new QTWindow();
  qtWin->AnimationUpdateRate(t);
#endif
}

void Simulator::AnimateWirelessTx(bool a)
{
  animateWirelessTx = a;
}

bool Simulator::AnimateWirelessTx()
{
  if (!animate) return false; // False if no animation selected
  return animateWirelessTx;
}

bool Simulator::AddBackgroundMap(const string& mapName, const RectRegion& r)
{ // Create a one entry vector and use the overloaded AddBackgroundMap function
  StringVec_t sv;
  sv.push_back(mapName);
  return AddBackgroundMap(sv, r);
}

bool Simulator::AddBackgroundMap(const StringVec_t& maps, const RectRegion& r)
{
#ifdef HAVE_QT
  LocationVecVec_t locVecVec;
  for (StringVec_t::size_type i = 0; i < maps.size(); ++i)
    {
      //ProcessOneMap(maps[i], r, locVecVec, minX, maxX, minY, maxY);
      ProcessOneMap(maps[i], r, locVecVec);
    }
  
  boundingBox = FindBoundingBox(locVecVec);
  
  // Now we adjust the longitude value by cos(latitude) for correct
  // projections on the map.  This is done by taking a centerline
  // of the bounding box, and adjusting the offsets from the centerline
  // by cos(lat).
  Meters_t centerLine = boundingBox.LowerLeft().X() + 
      (boundingBox.UpperRight().X() - boundingBox.LowerLeft().X()) / 2;
  
  for (LocationVecVec_t::size_type i = 0; i < locVecVec.size(); ++i)
    {
      LocationVec_t& lv = locVecVec[i];
      for (LocationVec_t::size_type j = 0; j < lv.size(); ++j)
        {
          Meters_t x = lv[j].X();
          Meters_t y = lv[j].Y();
          Meters_t delta = x - centerLine;
          delta *= cos(y * DegreesToRadians);
          DEBUG0((cout << "orig x " << x 
                  << " y " << y
                  << " cl " << centerLine
                  << " adjusted x " << delta + centerLine << endl));
          lv[j].X(centerLine + delta);
        }
    }
  
  // Bounding box has changed, so find it again
  boundingBox = FindBoundingBox(locVecVec);

  // Specify min/max and then display the lines
  NewLocation(boundingBox.LowerLeft());
  NewLocation(boundingBox.UpperRight());
  
  if (!silent)
    {
      cout << "minX "  << boundingBox.LowerLeft().X()
           << " minY " << boundingBox.LowerLeft().Y()
           << " maxX " << boundingBox.UpperRight().X()
           << " maxY " << boundingBox.UpperRight().Y()
           << " locVecVec.size " << locVecVec.size() 
           << endl;
    }
  // Now display each of the polylines
  if (!qtWin) qtWin = new QTWindow();
  for (Count_t i = 0; i < locVecVec.size(); ++i)
    {
      qtWin->AddBackgroundLines(locVecVec[i]);
    }
#endif
  return true;
}

bool Simulator::PlaybackTraceFile(const char* fn)
{  // Specify the name of the playback trace file
  //  This should be called before the topology is
  // created, since it sets the Interface Mac to IF logging
  // which is neede by the trace file playback
#ifdef HAVE_QT
  playbackFile.open(fn, ios::in);
  if (!playbackFile)
    {
      cout << "Can't open playback trace file " << fn << endl;
      return false;
    }
  playbackAnimate = true;
  Interface::recordMacAddresses = true;
  return true;
#else
  cout << "Cannot playback without QTWIN defined.  Need to rebuild" << endl;
  return false;
#endif
}

// Definitions for Playback mode
bool    Simulator::StartPlayback()
{
#ifdef HAVE_QT
  simtime = 0.0;
  Time_t msPerFrame = 1000/24; // 24 fps

  if (!playbackAnimate)
    {
      cout << "No playback file specified, can't start playback" << endl;
      return false;
    }
  DisplayTopologyAndReturn();
  qtWin->PlaybackMode(true); // Notify qtwindow we are in playback mode
  qtWin->PlaybackPause();

  while(!qtWin->HasQuit())
    {
      FillEventsFromTrace(simtime + 0.1);
      qtWin->UpdateTopology(false);
      simtime += qtWin->PlaybackRate() * (double)msPerFrame / 1000.0;
      DEBUG0((cout << "Simtime updated to " << simtime << endl));
    }
  return true;
#else
  cout << "Cannot playback without QTWIN defined.  Need to rebuild" << endl;
  return false;
#endif
}

void    Simulator::CustomBackground(CustomBackground_t cb)
{
  pCustomBackground = cb;
}

void    Simulator::NodeSelectedCallback(NodeSelected_t ns)
{
  pNodeSelected = ns;
}

void    Simulator::EnableAnimationRecorder(bool e)
{
  recorderEnabled = e;
}

void    Simulator::RecorderMPEGSpeedup(Mult_t sp)
{
  recorderSpeedup = sp;
}

void    Simulator::FillEventsFromTrace(Time_t t)
{ // Read the trace file to the first event with time > t
  // and populate the pending events list
  while(playbackFile)
    {
      string s;

      getline(playbackFile, s);
      DEBUG0((cout << "Playback line " << s << endl));
      vector<string> tokens;
      stringtok(tokens, s, " "); // Break into tokens
      if (tokens.size() < 2) continue; // ? empty ?
      Time_t thisTime = atof(tokens[0].c_str());
      string node = tokens[1];
      for (vector<string>::size_type i = 2; i < tokens.size(); ++i)
        {
          if (tokens[i].find("L2") != string::npos)
            {
              DEBUG0((cout << "Found l2, i " << i
                      << " token " << tokens[i] << endl));
              Count_t size = atol(tokens[i+1].c_str());
              MACAddr::MACAddr_t src = atol(tokens[i+2].c_str());
              MACAddr::MACAddr_t dst = atol(tokens[i+3].c_str());
              //Count_t            uid = atol(tokens[i+4].c_str());
              bool isRx;
              Interface* iface;
              if (tokens[i][0] == '+')
                {  // This is a rx
                  iface = Interface::GetByMac(dst);
                  isRx = true;
                  if (!iface)
                    cout << "HuH? Can't find iface for mac " << dst << endl;
                  else if(iface->GetMACAddr() != MACAddr(0, dst))
                    cout << "HuH?  MAC Mismatch, dst " << dst
                         << " ifmac " << iface->GetMACAddr() << endl;
                }
              else
                {
                  iface = Interface::GetByMac(src);
                  isRx = false;
                }
              if (iface)
                { // Interface exists, process event if target node displayed
                  bool nodeDisplayed = false;
                  Node* n = iface->GetNode();
                  if (n) nodeDisplayed = n->Show();
                  if (isRx && nodeDisplayed && thisTime >= Now())
                    { // Put in pending rx queue if node displayed
                      // and not in the past
                      Packet* p = new Packet();
                      p->size = size; // Set size
                      LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p);
                      iface->ScheduleRx(evRx, thisTime);
                      DEBUG0((cout << "Scheduling future rx event time "
                              << thisTime
                              << " iface " << iface << endl));
                    }
                }
              i += 3;
            }

        }
      if (thisTime > t) return;
    }
}




// Statistics
Time_t  Simulator::SetupTime()
{ // Wall clock time for initialization
  return startRouteTime; // This is CPU time used before the "Run" was called
}

Time_t Simulator::RouteTime()
{ // Time to calculate routes
  return startRunTime - startRouteTime;
}

Time_t  Simulator::RunTime()
{// CPU clock time for Run()
  return GetSec() - startRunTime;
}

Time_t  Simulator::TotalTime()
{ // Total time of execution
  return RunTime() + RouteTime() + SetupTime();
}

#ifdef HAVE_PSINFO
static int pagesize = 0;
#endif

Size_t  Simulator::ReportMemoryUsage()
{
  pid_t  pid;
  char   work[4096];
  FILE*  f;
  char*  pCh;

  pid = getpid();
  sprintf(work, "/proc/%d/stat", (int)pid);
  f = fopen(work, "r");
  if (f == NULL)
    {
#ifdef HAVE_PSINFO
      DEBUG(0,(cout << "Can't open " << work 
               <<" trying ioctl method" << endl));
      if (!pagesize)
        {
          pagesize = getpagesize();
        }
      char w[255];
      sprintf(w, "/proc/%d", pid);
      int p = open(w, O_RDONLY, 0);
      if (p < 0)
        {
          DEBUG(0,(cout << "Huh? Can't open "
                   << string(w) << endl));
          return(0);
        }
      struct prpsinfo prps;
      int i = ioctl(p, PIOCPSINFO, &prps);
      if (i)
        {
          DEBUG(0,(cout << "HuH? Can't get ioctl in memusage" << endl));
          return(0);
        }
      return(prps.pr_rssize * pagesize);
#else
      DEBUG(0,(cout << "Can't open " << work << endl));
#endif
      return(0);
    }
  fgets(work, sizeof(work), f);
  fclose(f);
  strtok(work, " ");
  for (int i = 1; i < 23; i++) 
    {
      pCh = strtok(NULL, " ");
    }
  //return(atoll(pCh));
  return(atol(pCh));
}

#ifdef OLD_ONE
Size_t  Simulator::ReportMemoryUsage()
{
  pid_t  pid;
  char   work[4096];
  FILE*  f;
  char*  pCh;

  pid = getpid();
  sprintf(work, "/proc/%d/stat", (int)pid);
  f = fopen(work, "r");
  if (f == NULL)
    {
      cout << "Can't open " << work << endl;
      return(0);    
    }
  fgets(work, sizeof(work), f);
  fclose(f);
  strtok(work, " ");
  for (int i = 1; i < 23; i++) 
    {
      pCh = strtok(NULL, " ");
    }
  return(atol(pCh));
}
#endif

Size_t Simulator::ReportMemoryUsageMB()
{ // Returns usage in MB!
  long long u = ReportMemoryUsage();
  return((u + 500000) / 1000000 );
}

// Grid Location Management
void Simulator::NewLocation(Meters_t x, Meters_t y)
{
  if (firstLocation)
    {
      smallestX = x;
      smallestY = y;
      largestX  = x;
      largestY  = y;
      firstLocation = false;
      return;
    }
  smallestX = min(smallestX, x);
  smallestY = min(smallestY, y);
  largestX  = max(largestX, x);
  largestY  = max(largestY, y);
}

void Simulator::NewLocation(const Location& l)
{
  NewLocation(l.X(), l.Y());
}

Meters_t  Simulator::SmallestX()
{
  return smallestX;
}

Meters_t  Simulator::SmallestY()
{
  return smallestY;
}

Meters_t  Simulator::LargestX()
{
  return largestX;
}

Meters_t  Simulator::LargestY()
{
  return largestY;
}

bool Simulator::UsingRTI()
{
  return usingRTI;
}

// Private Methods
void Simulator::ProgressResched(Time_t interval)
{
  progressEvent = new ProgressEvent(interval);;
  progressEvent->event = 0; // ! HuH? why do I need this..
  progressTimer.Schedule(progressEvent, interval);
}

Time_t Simulator::GetSec()
{ // Get cpu usage time, rounded to seconds
  struct rusage u;
  getrusage (RUSAGE_SELF, &u);
  Time_t us = u.ru_utime.tv_usec + u.ru_stime.tv_usec;
  Time_t s  = u.ru_utime.tv_sec + u.ru_stime.tv_sec;
  return s + us / 1000000.0;
}

bool  Simulator::ProcessOneMap(std::string n, const RectRegion& r,
                               LocationVecVec_t& lv)
{
  ifstream is(n.c_str());
  if (!is)
    {
      cout << "Can't open map file " << n << endl;
      return false;
    }
  LocationVec_t locs;

  // Keep a vector of the vectors.  We need to find then
  // entire set of points to set the min/max points
  // for meters to pixels conversions.  So we first
  // build the entire set, set the min/max points,
  // then display the lines.
  string s;
  getline(is, s);
  vector<string> tokens;

  while(is)
    {
      int lineCount = 0;
      int rank = 1;
      // process the header line
      DEBUG0((cout << "Hdr line is " << s << endl));
      tokens.clear();
      stringtok(tokens, s);
      
      if (s.find("segment")  != string::npos)
        { // CIA World Databank format
          lineCount = atol(tokens[5].c_str());
          rank = atol(tokens[3].c_str());
          DEBUG0((cout << "CIA lineCount " << lineCount
                  << " rank " << rank << endl));
        }
      else
        { // Simple format, just a line count on first line
          if (tokens.size() != 1)
            {
              cout <<  "Unrecognized map file format" << endl;
              return false;
            }
          lineCount = atol(tokens[0].c_str());
        }
      bool skipAll = false;
  
      for (int i = 0; i < lineCount; ++i)
        {
          double lat;
          double lon;
          getline(is, s);
          tokens.clear();
          stringtok(tokens, s);
          if (tokens.size() != 2)
            {
              cout <<  "Unrecognized map file format, token size" 
                   << tokens.size() << endl;
              return false;
            }
          lat = atof(tokens[0].c_str());
          lon = atof(tokens[1].c_str());
          if (!r.Empty() && !r.InRegion(Location(lon, lat)))
            { // Not in bounding rect, terminate this polyline
              skipAll = true;
            }
          if (lat == 0 && lon == 0)
            { // Discontinuous,  save and clear the vector
              lv.push_back(locs);
              DEBUG0((cout << "size " << locs.size() << endl));
              locs.clear();
            }
          else
            {
              // Convert to meters from the 0,0 location
              // (equator, zero meridian)
              //lon *= (DegreesToMeters / cos(lat * DegreesToRadians));
              // Need a better adjustment for longitude..code later
              //lon *= DegreesToMeters;
              //lat *= DegreesToMeters;
              if (rank == 1 && !skipAll)
                { // Only process rank 1 entries from CIA Database
                  //mnx = min(lon, mnx);
                  //mny = min(lat, mny);
                  //mxx = max(lon, mxx);
                  //mxy = max(lat, mxy);
              
                  DEBUG0((cout << "Lat " << lat << " lon  " << lon << endl));
                  locs.push_back(Location(lon, lat));
                }
            }
        }
      
      if (skipAll) locs.clear();
      if (locs.size()) lv.push_back(locs);
      locs.clear();
      if(is)getline(is, s); // Get Next Header Line
      if (s.length() == 0) break; // allow for empty last line
    }
  is.close();
  return true;
}

RectRegion Simulator::FindBoundingBox(LocationVecVec_t& lvv)
{
  Meters_t minX = 1e307;
  Meters_t minY = 1e307;
  Meters_t maxX = -1e307;
  Meters_t maxY = -1e307;
  for (LocationVecVec_t::size_type i = 0; i < lvv.size(); ++i)
    {
      LocationVec_t& lv = lvv[i];
      for (LocationVec_t::size_type j = 0; j < lv.size(); ++j)
        {
          Meters_t x = lv[j].X();
          Meters_t y = lv[j].Y();
          minX = min(x, minX);
          minY = min(y, minY);
          maxX = max(x, maxX);
          maxY = max(y, maxY);
        }
    }
  return RectRegion(Location(minX, minY), Location(maxX, maxY));
}

// Static Methods
Time_t Simulator::Now()
{
  if (!instance) return(0.0);  // Simulation has not yet started
  return instance->simtime;
}

#ifdef HAVE_RTI
// RTIKit requires this exact name
void TimeAdvanceGrant (TM_Time T)
{
 Simulator::grantedTime = T; // More efficient to use below
 //TM_Recent_LBTS(&Simulator::grantedTime); // Use most recent LBTS value
 DEBUG0((cout << "Granted to " << Simulator::grantedTime << endl));
 Simulator::granted = true;
}

// Two dummy functions needed by BRTI but not used here

extern "C" {
void MyHereProc (long MsgSize, struct MsgS *Msg) {}
void RequestRetraction(CoreRetractionHandle) { }
}


#endif


