/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Star topology
// George F. Riley.  Georgia Tech, Fall 2002

// Define a star topology using point-to-point links

#include <iostream>

#include "debug.h"
#include "star.h"
#include "node.h"
#include "linkp2p.h"
#include "mask.h"

using namespace std;

// Constructors

Star::Star(Count_t l, IPAddr_t i, SystemId_t id)
{
  ConstructorHelper(l, nil, i, Linkp2p::Default(), id);
}

Star::Star(Count_t l, const Linkp2p& link, IPAddr_t i, SystemId_t id)
{
  ConstructorHelper(l, nil, i, link, id);
}

Star::Star(Count_t l, Node* n, IPAddr_t i, SystemId_t id)
{
  ConstructorHelper(l, n, i, Linkp2p::Default(), id);
}

Star::Star(Count_t l, Node* n, const Linkp2p& link, IPAddr_t i, SystemId_t id)
{
  ConstructorHelper(l, n, i, link, id);
}

// Access functions
Node* Star::GetHub()
{ // Get left side router
  return hub;
}

Node* Star::GetLeaf(Count_t c)
{ // Get specified leaf node
  if (c >= leafCount) return nil;
  return Node::GetNode(first + c);
}

Linkp2p* Star::HubLink(Count_t c)
{ // Get link from hub to specified leaf
  Node* n = GetLeaf(c);
  if (!n) return nil;
  return (Linkp2p*)hub->GetLink(n);
}

Linkp2p* Star::LeafLink(Count_t c)
{ // Get link from leaf to hub
  Node* n = GetLeaf(c);
  if (!n) return nil;
  return (Linkp2p*)n->GetLink(hub);
}

Queue*   Star::HubQueue(Count_t c)
{ // Get queue from hub to leaf
  Node* n = GetLeaf(c);
  if (!n) return nil;
  return hub->GetQueue(n);
}

Queue*   Star::LeafQueue(Count_t c)
{ // Get queue from hub to leaf
  Node* n = GetLeaf(c);
  if (!n) return nil;
  return n->GetQueue(hub);
}

void Star::BoundingBox(const Location& ll, const Location& ur,
                       Angle_t startAngle, Angle_t arcSize)
{
  // Make copies of the passed in locations, as we may have 
  // to change these.
  Location lll(ll);
  Location lur(ur);
  
  Meters_t xDist = lur.X() - lll.X();
  Meters_t yDist = lur.Y() - lll.Y();
  Meters_t arcDist = xDist / 2;
  Node*    hub = GetHub();

  if (hub->HasLocation())
    { // The hub already has a location.  Use this as the bounding box
      // as a relative bb, with the hub at the center.
      Location hl = hub->GetLocation();
      lll = Location(hl.X() - xDist / 2, hl.Y() - yDist / 2);
      lur = Location(hl.X() + xDist / 2, hl.Y() + yDist / 2);
    }
    
  // Notify simulator of bounds, for max x/y computations
  Simulator::instance->NewLocation(lll.X(), lll.Y());
  Simulator::instance->NewLocation(lur.X(), lur.Y());

  Location hubLoc(lll.X() + xDist / 2, lll.Y() + yDist / 2);
  GetHub()->SetLocation(hubLoc);
  // Now position the leaf nodes
  Angle_t theta = startAngle - arcSize / 2;
  Angle_t adder = arcSize / LeafCount();
  for (Count_t i = 0; i < LeafCount(); ++i)
    {
      Location leafLoc(hubLoc.X() + cos(theta) * arcDist,
                       hubLoc.Y() + sin(theta) * arcDist);
      GetLeaf(i)->SetLocation(leafLoc);
      theta += adder;
    }
}

// Private methods
void Star::ConstructorHelper(Count_t l, Node* h,
                             IPAddr_t ip,
                             const Linkp2p& link,
                             SystemId_t id)
{
  leafCount = l;
  hub = h;
  int adder = 1;
  
  if (!hub)
    { // No hub specified, need to allocate
      hub = new Node(id);
      DEBUG0((cout << "ST HubId " << hub->Id() << endl));
    }
  first = Node::nextId;
  if (ip != IPADDR_NONE)
    {
      if (ip & 0x1)
        { // Ends in 1, use ip and ip - 1 for endpoints
          adder = -1;
        }
    }
  for (Count_t i = 0; i < leafCount; ++i)
    {
      Node* ln = new Node(id);
      DEBUG0((cout << "ST NodeId " << ln->Id() << endl));
      if (ip != IPADDR_NONE)
        {
          ln->AddDuplexLink(hub, link, ip, Mask(31), ip + adder, Mask(31));
        }
      else
        {
          ln->AddDuplexLink(hub, link);
        }
      // Add a default route
      ln->DefaultRoute(hub);
      if (ip != IPADDR_NONE) ip += 2; // Advance to next set of ip's
    }
  last = Node::nextId;
}

