/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - TCP-NewReno Class
// George F. Riley.  Georgia Tech, Spring 2002

#include <iostream>

//#define DEBUG_MASK 0x03
#include "debug.h"
#include "tcp-newreno.h"
#include "simulator.h"

using namespace std;

// Constructors
TCPNewReno::TCPNewReno() : TCP() 
{
fastRecMode = false;
fastRecMark = 0;
}

TCPNewReno::TCPNewReno(Node* n) : TCP(n) 
{
fastRecMode = false;
fastRecMark = 0;
}

TCPNewReno::TCPNewReno(const TCPNewReno& c) : TCP(c)
{
fastRecMode = false;
fastRecMark = 0;
}

// L4Protocol Methods
L4Protocol* TCPNewReno::Copy() const
{
  return new TCPNewReno(*this); // Create a copy of this protocol
}

// TCP Overridden methods
void TCPNewReno::DupAck(const TCPHeader& t, Count_t c)
{ // Dup ack received
  //cout << "Reno DupAck " << t.ackNumber << " at time " << Simulator::Now() << endl;
  DEBUG(1,(cout << "Hello from TCPReno::DupAck " << t.ackNumber
           << ", count " << c
           << ", time " << Simulator::Now()
           << endl));
  if (fastRecMode)
    { cWin += segSize;   // Inflate the cWin for every additional DupAck
      SendPendingData(); // Try to send data if permitted by cWin size
    }
  if (c == 3 && !fastRecMode)
    { // Count of three indicates triple dupack has been received
      fastRecMode = true; // Enter the fast recovery mode
      fastRecMark = nextTxSeq;
      //cout << "Entering fast recovery with cWin = " << cWin << " at " << Simulator::Now() <<endl;
      ssThresh = cWin / 2;      // Reduce slow start threshold to half cwin
      if (ssThresh < 2*segSize) ssThresh = 2*segSize;
      Retransmit();        // Retransmit the packet  
      cWin = ssThresh + DCount_t(3*segSize);  // Collapse congestion window to ssThresh
                // the 3* segSize accounts for the DupAcks that have already exited the network
      //cout << " ssThresh set to " << ssThresh << " cWin is " << cWin << endl;
    }
    
}

void TCPNewReno::ReTxTimeout()
{ // Retransmit timeout
  //cout << "Reno Timeout at time " << Simulator::Now() << endl;
  DEBUG0((cout << "Hello from TCPReno::Timeout" 
           << ", time " << Simulator::Now()
          << endl));
  ssThresh = cWin / 2;      // Reduce slow start threshold to half cwin
  if (ssThresh < 2*segSize) ssThresh = 2*segSize;
  cWin = segSize;           // Collapse congestion window
  nextTxSeq = highestRxAck; // Start from highest Ack
  fastRecMode = false;
  fastRecMark = 0;
  Retransmit();             // Retransmit the packet
}

void TCPNewReno::NewAck(Seq ack)
{ // NewAck is called only for "New" (non-duplicate) acks
  // Always cancel any pending re-tx timer on new acknowledgement
  //cout << "Reno NewAck " << ack << " at time " << Simulator::Now() << endl;

  DEBUG(1,(cout << "TCP " << this << "Cancelling retx timer " << endl));
  Count_t numberAck = ack - highestRxAck; // Number bytes ack'ed
  highestRxAck = ack;         // Note the highest recieved Ack
  if (highestRxAck >= nextTxSeq)  CancelTimer(retxTimeout, true);
  if (fastRecMode) {

	  if (highestRxAck < fastRecMark)
	    { cWin += segSize;   // Inflate the cWin for every additional Ack < fastRecMark
	      Retransmit();
	      return;
	  }else { 
   	    //cout << "Exiting Fast Recovery at " << Simulator::Now() << endl;
	    fastRecMode = false;
	    fastRecMark = 0;
	    cWin = ssThresh;
	  }
  } else {
      NewCWnd(numberAck);         // Adjust congestion window
  }

  DEBUG(5,(cout << "TCP " << this << " NewAck " << ack
          << " numberAck " << numberAck << endl));
  totalAck += numberAck;                  // Count for statistics
  if (ack > nextTxSeq) nextTxSeq = ack; // If advanced
  // See if all pending ack'ed; if so we can delete the data
  if (pendingData)
    { // Data exists, see of can be deleted
      if (pendingData->SizeFromSeq(firstPendingSeq, highestRxAck) == 0)
        { // All pending acked, can be deleted
          pendingData->Clear();
          delete pendingData;
          pendingData = nil;
          // Insure no re-tx timer
          CancelTimer(retxTimeout, true);
          // Check if FIN needed
          if (closeOnEmpty) { } // Need to check/code this
        }
    }
  // Notify application of data sent
  Application* a = GetApplication();
  if (a) a->Sent(numberAck, this);
  // Try to send more data
  SendPendingData();
  // See if we need to post a re-tx timer
  if (!retxTimeout &&
      (UnAckDataCount() || state == FIN_WAIT_1 || state == FIN_WAIT_2))
    {
      ScheduleTimer(TCPEvent::RETRANSMIT_PACKET,
                    retxTimeout,
                    rtt->RetransmitTimeout());
    }
}

