/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Bucketed Timers for GTNetS
// George F. Riley, Georgia Tech, Winter 2002

// Bucketed timers use a fifo queue for all timer events for a given "k"
// milliseconds in the future.  This allows the main event list to remain
// relatively small.

#include "timerbuckets.h"
#include "simulator.h"

using namespace std;

#ifdef OLD_NOT_USED
// TimerBucketEvent methods
TBucketEvent::TBucketEvent() : client(nil), clientEvent(nil),
                               evList(nil), expiration(0),
                               cancelled(false)

{
}

TBucketEvent::TBucketEvent(TimerHandler* h, TimerEvent* e,
                           FifoEventList_t* f, Time_t t)
  : client(h), clientEvent(e), evList(f), expiration(t), cancelled(false)
{
}
#endif

// TimerBucket static variables
TimerBucket* TimerBucket::instance;  // A single instance for each simulation
Count_t      TimerBucket::scheduled;
Count_t      TimerBucket::simScheduled; // Number scheduled in simualtor
Count_t      TimerBucket::handled;
Count_t      TimerBucket::cancelled;
Count_t      TimerBucket::skipped;
Count_t      TimerBucket::processed;

// TimerBucket methods
TimerBucket::TimerBucket() : bucketSize(10) // Default 10ms
{
}

void TimerBucket::Handle(Event* e, Time_t t)
{
  handled++;
  TimerEvent* ev = (TimerEvent*)e;
  FifoEventList_t* evl = ev->evList;
  Time_t now = Simulator::Now();
#define EXTRA_CHECKS
#ifdef EXTRA_CHECKS
  // Extra checks for debugging, remove later
  if (!evl->size())
    {
      cout << "HuH?  TimerBucket::Handle w/ empty evl" << endl;
      exit(1);
    }
#endif
  TimerEvent* tme = evl->front();
#ifdef EXTRA_CHECKS
  // More extra checks
  if (tme != ev)
    {
      cout << "HuH?  TimerBucket::Handle, ev mismatch" << endl;
      exit(1);
    }
#endif
  evl->pop_front(); // Remove current
  // Find next sequential event to schedule
  while (!evl->empty())
    { // Check for any new events to schedule
      tme = evl->front();
      if (!tme->cancelled)
        { // Found one, schedule it and exit
          Simulator::instance->Schedule(tme, tme->Time() - now, this);
          simScheduled++;
          break;
        }
      // Event is cancelled, remove and delete
      evl->pop_front();
      delete tme; // Event is cancelled, just delete it
    }
  // Pass current event to handler if not cancelled
  ev->evList = nil; // Note no longer scheduled (in case cancelled again)
  if (!ev->cancelled)
    { // Call the original event handler if not cancelled
      processed++;
      ev->timerHandler->Timeout(ev);
    }
  else
    { // otherwise delete it
      skipped++;
      delete ev;
    }
}

void TimerBucket::BucketSize(Count_t ms)
{
  if (buckets.size())
    {
      cout << "TimerBucket::BucketSize called on non-empty buckets. Aborting."
           << endl;
      exit(1);
    }
  bucketSize = ms;
}

void TimerBucket::Schedule(TimerEvent* ev, Time_t t, TimerHandler* h)
{
  Time_t  now = Simulator::Now();
  Count_t futureMs = (Count_t)(t * 1000.0); // Convert to milliseconds
  Count_t bucket = 0;
  if (futureMs)
    { // 1..10 go in bucket 0, 11..20 in bucket 1, etc
      // assuming bucketSize is 10
      bucket = (futureMs - 1) / bucketSize;
    }
  while (buckets.size() <= bucket)
    { // Expand size of bucket vector to include this one
      buckets.push_back(nil);
    }
  FifoEventList_t* evList = buckets[bucket];
  if (evList == nil)
    { // Need new one
      evList = new FifoEventList_t;
      buckets[bucket] = evList;
    }
  Time_t futureSec = (Time_t)((bucket + 1) * bucketSize) / 1000.0;
  bool empty = evList->empty();
  ev->evList = evList;
  ev->timerHandler = h;
  ev->Time(futureSec + now);
  evList->push_back(ev); // Add to fifo queue
  if (empty)
    { // Need to schedule an event to the main event queue
      Simulator::instance->Schedule(ev, futureSec, this);
      simScheduled++;
    }
  scheduled++;
}

void TimerBucket::Cancel(TimerEvent* ev)
{
  ev->cancelled = true;
  cancelled++;
}

void TimerBucket::DBDump()
{
  cout << "TimerBucket Stats" << endl;
  cout << "scheduled " << scheduled;
  cout << " simScheduled " << simScheduled;
  cout << " handled " << handled;
  cout << " cancelled " << cancelled;
  cout << " skipped " << skipped;
  cout << " processed " << processed;
  cout << endl;
}

// Static methods

TimerBucket* TimerBucket::Instance()
{
  if (!instance) instance = new TimerBucket();
  return instance;
}

