/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Efficient Bucketed timers
// George F. Riley.  Georgia Tech, Winter 2002

// Timer buckets reduce the size of the simulator's pending event list
// by taking advantage of the fact that any timer that is "k" milliseconds
// in the future will always be sequentially after the previously scheduled
// timer that is "k" milliseconds in the future. Thus only the earliest such
// time need be scheduled in the simulator's main event list.  All others
// can be in a fifo queue.  We also observe that in "real" tcp implementations,
// the timers are typically in 10ms units, since most hardware implementations
// perform process switching on 10ms interrupts.  Thus all TCP retransmit
// timers are in units of 10ms, allowing the bucketing of all TCP timers in
// the same [k - k+10ms] range.

#ifndef __timerbuckets_h__
#define __timerbuckets_h__

#include <vector>

#include "timer.h"

typedef std::vector<FifoEventList_t*> BucketVec_t;       // The buckets

class TimerBucket : public Handler {
public:
  TimerBucket();
  // Handler for events
  void Handle(Event*, Time_t);
  void BucketSize(Count_t); // Size in milliseconds of each bucket
  void Schedule(TimerEvent*, Time_t, TimerHandler*);
  void Cancel(TimerEvent*); // Cancel a pending timer
  void DBDump();            // Debug dump statistics
public:
  static TimerBucket* Instance();
private:
  BucketVec_t         buckets;     // The set of FIFO timer buckets
  Count_t             bucketSize;  // Size of each bucket in milliseconds
private:
  static TimerBucket* instance;
  static Count_t      scheduled;
  static Count_t      simScheduled; // Number actually scheduled in simualtor
  static Count_t      handled;
  static Count_t      cancelled;
  static Count_t      skipped;
  static Count_t      processed;
};

#endif
