/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Tree topology
// George F. Riley.  Georgia Tech, Fall 2002

// Define the tree topology using point-to-point links

#include "tree.h"
#include "node.h"
#include "linkp2p.h"
#include "mask.h"

#include <math.h>
#include <stdio.h>

using namespace std;

// Constructors

Tree::Tree(Count_t l, Count_t f, IPAddr_t i, SystemId_t id)
{
  ConstructorHelper(l, f, Linkp2p::Default(), i, id);
}

Tree::Tree(Count_t l, Count_t f, const Linkp2p& link,
           IPAddr_t i, SystemId_t id)
{
  ConstructorHelper(l, f, link, i, id);
}

// Access functions
Node* Tree::GetRoot()
{ // Get the root node
  return GetNode(0,0);
}

Node* Tree::GetNode(Count_t l, Count_t c) // Level, index
{ // Get Specified right side leaf node
  Count_t  nodesThisLevel = 1; // Starting with root
  NodeId_t firstId = first;    // First ID for this level
  for (Count_t thisLevel = 0; thisLevel < l; ++thisLevel)
    {
      firstId += nodesThisLevel;
      nodesThisLevel *= fanout;
    }
  if (c < nodesThisLevel) return Node::GetNode(firstId + c);
  return nil; // Out of range
}

Node* Tree::GetLeaf(Count_t c)
{ // Get specified leaf
  if (c < leafCount) return Node::GetNode(firstLeafId + c);
  return nil; // Out of range
}

Linkp2p* Tree::GetChildLink(Count_t l,  // Level
                            Count_t i,  // Index on level
                            Count_t c)  // Child number
{ 
  if (c >= fanout) return nil;
  Node* n = GetNode(l, i);
  if (!n) return nil;
  Node* cn = GetNode(l + 1, i * fanout + c);
  if (!cn) return nil;
  return (Linkp2p*)n->GetLink(cn);
}

Linkp2p* Tree::GetParentLink(Count_t l, Count_t i)
{ // Get link to parent from specified node
  Node* n = GetNode(l, i);
  if (!n) return nil;
  Node* pn = GetNode(l - 1, i / fanout);
  if (!pn) return nil;
  return (Linkp2p*)n->GetLink(pn);
}

Queue* Tree::GetChildQueue(Count_t l,  // Level
                           Count_t i,  // Index on level
                           Count_t c)  // Child number
{ 
  if (c >= fanout) return nil;
  Node* n = GetNode(l, i);
  if (!n) return nil;
  Node* cn = GetNode(l + 1, i * fanout + c);
  if (!cn) return nil;
  return n->GetQueue(cn);
}

Queue* Tree::GetParentQueue(Count_t l, Count_t i)
{ // Get link to parent from specified node
  Node* n = GetNode(l, i);
  if (!n) return nil;
  Node* pn = GetNode(l - 1, i / fanout);
  if (!pn) return nil;
  return n->GetQueue(pn);
}

void Tree::BoundingBox(const Location& ll, const Location& ur)
{
  vector<Meters_t> myX;
  LocateRow(0, 1, first, ll, ur, myX);
}

void Tree::BoundingBox(const Location& ll, const Location& rr, double angle)
{
  Meters_t xx,yy,x,y,nx,ny;
  Count_t thisLevelCount = 1;

  angle = M_PI/180.0 * angle;  // convert to radians

  BoundingBox(ll, rr);
  xx = (ll.X()+rr.X())/2;
  yy = rr.Y();

  for (Count_t lev = 0; lev < levels; ++lev) {
    DEBUG0((cout<<"Nodes in this level "<<thisLevelCount<<endl));
    for (Count_t count = 0; count < thisLevelCount; ++count) {
      Node *node = GetNode(lev, count);

      DEBUG0((cout<<"In level "<<lev<<" count "<<count<<endl));
      
      x = node->GetLocation().X()-xx;
      y = node->GetLocation().Y()-yy;

      nx = x* cos(angle)-y*sin(angle);
      ny = x* sin(angle)+y*cos(angle);

      nx = nx + xx;
      ny = ny + yy;

      DEBUG0((printf("Location %6.2lf %6.2lf\n", nx,ny)));
      
      node->SetLocation(Location(nx,ny));
      
    }
    thisLevelCount *=fanout;
    DEBUG0((cout<<"Going to the next level"<<endl));
  }

}

// Private methods
void Tree::ConstructorHelper(Count_t l, Count_t f,
                             const Linkp2p& link,
                             IPAddr_t leafIP,
                             SystemId_t id)
{
  first = Node::nextId;
  levels = l;
  fanout = f;
  Node* n = new Node(id);
  Count_t   lastLevelCount = 1;
  NodeId_t  lastLevelFirstId = n->Id();
  IPAddr_t  nextIP = leafIP;

  for (Count_t lev = 1; lev < levels; ++lev)
    { // Create each subsequent level
      bool     lastLevel = (lev == (levels-1));
      Count_t  thisLevelCount = lastLevelCount * fanout;
      NodeId_t thisLevelFirstId = Node::nextId;
      if (lastLevel)
        {
          firstLeafId = Node::nextId;
          leafCount = thisLevelCount;
        }
      for (Count_t c = 0; c < thisLevelCount; ++c)
        { // Create each child
          Node* parent = Node::GetNode(lastLevelFirstId + c / fanout);
          Node* child = new Node(id);
          if (lastLevel)
            { // Create link with IP Address
              if (nextIP == IPADDR_NONE)
                { // No IP specified
                  child->AddDuplexLink(parent, link);
                }
              else
                {
                  child->AddDuplexLink(parent, link,
                                       nextIP++, Mask(32),
                                       IPADDR_NONE, Mask(32));
                }
            }
          else
            { // Create anonymous (no IPAddr) link
              child->AddDuplexLink(parent, link);
            }
          // Add a default route from child to parent
          child->DefaultRoute(parent);
        }
      lastLevelCount = thisLevelCount;
      lastLevelFirstId = thisLevelFirstId;
    }
  last = Node::nextId;
}

void Tree::LocateRow(Count_t level, Count_t leafThisRow, NodeId_t firstId, 
                      const Location& ll, const Location& ur,
                      std::vector<Meters_t>& parentX)
{ // LocateRow recurses through the rows, since each row needs to know
  // the locations of the children to properly position the parent.
  vector<Meters_t> myX;
  if ((level + 1) < levels)
    {
      LocateRow(level+1, leafThisRow * fanout, firstId + leafThisRow,
                 ll, ur, myX);
    }
  else
    { // This is the bottom row.  Calculate X Locations
      Meters_t xSpacing = (ur.X() - ll.X()) / (leafThisRow + 1);
      Meters_t xLoc = xSpacing;
      for (Count_t i = 0; i < leafThisRow; ++i)
        {
          myX.push_back(xLoc);
          xLoc += xSpacing;
        }
    }
  
  const NodeVec_t& nodes = Node::GetNodes();
  Meters_t minX = 0;
  Meters_t ySpacing = (ur.Y() - ll.Y()) / levels;
  Meters_t yLoc = ySpacing / 2.0 + level * ySpacing;
  
  for (Count_t i = 0; i < leafThisRow; ++i)
    {
      Meters_t xLoc = myX[i];
      Node* n = nodes[firstId + i];
      n->SetLocation(ll.X() + xLoc, ur.Y() - yLoc);
      // Maintain stats for higher level x-location
      if ((i % fanout) == 0)
        {
          minX = xLoc;
        }
      if (((i+1) % fanout) == 0)
        {
          Meters_t upX = minX + (xLoc - minX) / 2;
          parentX.push_back(upX);
        }
    }
}

