/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Tree topology
// George F. Riley.  Georgia Tech, Fall 2002

// Define the tree topology using point-to-point links

#ifndef __tree_h__
#define __tree_h__

#include "common-defs.h"
#include "node.h"
#include "mask.h"

class Node;
class Linkp2p;
class Queue;

//Doc:ClassXRef
class Tree {
  //Doc:Class Defines a tree topology using point--to--point links.
public:
  // Levels, Fanout, IP
  //Doc:Method
  Tree(Count_t, Count_t,
       IPAddr_t = IPADDR_NONE, SystemId_t = 0);
    //Doc:Desc Construct a {\tt Tree} topology object with the specified
    //Doc:Desc number of levels, fanout, \IPA\ and address mask.  The
    //Doc:Desc default point--to--point link object is used for links.
    //Doc:Arg1 Number of levels in the tree.
    //Doc:Arg2 Fanout of each level (number of child  nodes for each parent).
    //Doc:Arg3 Starting \IPA.  The \IPA\ is incremented by one for each
    //Doc:Arg3 node in the tree.
    //Doc:Arg4 Responsible system ID for distributed simulation

  // Specify link to use
  //Doc:Method
  Tree(Count_t, Count_t, const Linkp2p&,
       IPAddr_t = IPADDR_NONE, SystemId_t = 0);
    //Doc:Desc Construct a {\tt Tree} topology object with the specified
    //Doc:Desc number of levels, fanout, \IPA\ and address mask, and
    //Doc:Desc the specified point--to--point link.
    //Doc:Arg1 Number of levels in the tree.
    //Doc:Arg2 Fanout of each level (number of child  nodes for each parent).
    //Doc:Arg3 A reference to a point--to--point link object to copy for 
    //Doc:Arg3 each link in the tree.
    //Doc:Arg4 Starting \IPA.  The \IPA\ is incremented by one for each
    //Doc:Arg5 node in the tree.
    //Doc:Arg6 Responsible system ID for distributed simulation

  //Doc:Method
  //Doc:Method
  Node* GetRoot();                    // Get root node
    //Doc:Desc Return a pointer to the root node of the tree.
    //Doc:Return Pointer to root node.

  //Doc:Method
  Node* GetNode(Count_t, Count_t);    // Get node at specified level, index
    //Doc:Desc Return a pointer to the node at the specified level, index.
    //Doc:Arg1 Level number (root is level 0).
    //Doc:Arg2 Index, range 0 to number of nodes in the level - 1.
    //Doc:Return Pointer to the specfied node.

  //Doc:Method
  Node* GetLeaf(Count_t);             // Get specified leaf
    //Doc:Desc Return a pointer to the specified leaf node.
    //Doc:Arg1 Desired leaf node indes, counting from zero.
    //Doc:Return Pointer to the specified leaf node.

  // Get link from specified node to specified child
  //Doc:Method
  Linkp2p* GetChildLink(Count_t, Count_t, Count_t);
    //Doc:Desc Get a pointer to the link from the specified node to its
    //Doc:Desc specified child.
    //Doc:Arg1 Level number (root is level 0).
    //Doc:Arg2 Index, range 0 to number of nodes in the level - 1.
    //Doc:Arg3 Child index, range 0 to fanout - 1.
    //Doc:Return Pointer to the specfied link.

  // Get link from specified node to parent
  //Doc:Method
  Linkp2p* GetParentLink(Count_t, Count_t);
    //Doc:Desc Get a pointer to the link from the specified node to its parent.
    //Doc:Arg1 Level number (root is level 0).
    //Doc:Arg2 Index, range 0 to number of nodes in the level - 1.
    //Doc:Return Pointer to the specfied link.

  // Get queue from specified node to specified child
  //Doc:Method
  Queue*   GetChildQueue(Count_t, Count_t, Count_t);
    //Doc:Desc Get a pointer to the queue from the specified node to its
    //Doc:Desc specified child.
    //Doc:Arg1 Level number (root is level 0).
    //Doc:Arg2 Index, range 0 to number of nodes in the level - 1.
    //Doc:Arg3 Child index, range 0 to fanout - 1.
    //Doc:Return Pointer to the specfied queue.

  // Get queue from specified node to parent
  //Doc:Method
  Queue*   GetParentQueue(Count_t, Count_t);
    //Doc:Desc Get a pointer to the queue from the specified node
    //Doc:Desc to its parent.
    //Doc:Arg1 Level number (root is level 0).
    //Doc:Arg2 Index, range 0 to number of nodes in the level - 1.
    //Doc:Return Pointer to the specfied queue.

  //Doc:Method
  Count_t  LeafCount() { return leafCount;}
    //Doc:Desc Get a count of the number of leaf nodes in this tree.
    //Doc:Return Count of leaf nodes.

  //Doc:Method
  void BoundingBox(const Location&, const Location&);
    //Doc:Desc Define a bounding box for this tree object. 
    //Doc:Desc The bounding box
    //Doc:Desc is used to assign node locations, which are used during
    //Doc:Desc the simulation animation.
    //Doc:Arg1 Lower left corner of bounding box.
    //Doc:Arg2 Upper right corner of bounding box.

  void BoundingBox(const Location&, const Location&, double angle);


public:
  NodeId_t   first;      // There is a global node list, we don't need one 
  NodeId_t   last;       // Just log first/last+1 entries
  Count_t    levels;     // Number of levels
  Count_t    fanout;     // Children per node
  Count_t    leafCount;  // Number of leaf nodes
  IPAddr_t   firstLeafIP;// Starting IP Address for leaf
  NodeId_t   firstLeafId;// Starting NodeId for leaf
private:
  void ConstructorHelper(Count_t, Count_t,
                         const Linkp2p&,
                         IPAddr_t,
                         SystemId_t);
  void LocateRow(Count_t, Count_t, NodeId_t,
                  const Location&, const Location&, 
                  std::vector<Meters_t>&);
};

#endif


