/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// the wireless lan layer

#include <iostream>

#define DEBUG_MASK 0x01

#include "debug.h"
#include "wlan.h"
#include "node.h"
#include "queue.h"
#include "propagation.h"
#include "qtwindow.h"
//#include <time.h>
//#include "mystat.h"

#ifdef SIM
#include "sim.h"
#include "ipv4.h"
#include "estat.h"
#include "ospf6d/ospf6_message.h"
#include "lib/linklist.h"
#include "ospf6d/ospf6_top.h"
#include "ospf6d/ospf6_area.h"
#include "ospf6d/ospf6_interface.h"
#include "ospf6d/ospf6_neighbor.h"
#include "application-ospf6d.h"
#endif //SIM



#ifdef ETRACE
#include "etrace.h"
#endif

using namespace std;

WirelessLink::WirelessLink(Count_t n,   // Node count
			   IPAddr_t i, Mask_t m,    // IP addr start, mask
#ifdef USAP_ADDON
                           const L2Proto & l2,
#endif // USAP_ADDON
			   L4Protocol* l4,          // L4 protocol to bind to
			   Rate_t r, Time_t dly,
			   Detail_t d, bool bootstrap)
    : LinkReal()
{
#ifdef USAP_ADDON
  ConstructorHelper(n, i, m, l2, l4, r, dly, d, bootstrap);
#else
  ConstructorHelper(n, i, m, l4, r, dly, d, bootstrap);
#endif // USAP_ADDON
  propagation = new PropaTwoRay();
  csThreshold = pow(10, -7.8);    // -78 dBm
  rxThreshold = pow(10, -6.4);    // -64 dBm
  snrThreshold = 10;              // 10 dB
#ifdef BUGFIX
  alpha = 0.5; // ~80% at the specified range
  //alpha = 1; // 100% at the specified range
  rngProp = new Uniform(0, 1); // Keep this on separate RNG stream
#endif // BUGFIX
}

WirelessLink::~WirelessLink()
{ // Destructor, nothing to do
  if (propagation) delete propagation;
#ifdef BUGFIX
  if (rngProp) delete rngProp;
#endif // BUGFIX
}

Interface* WirelessLink::GetPeer(Packet* p)
{
  L2Header802_11* l2pdu = (L2Header802_11*)p->PeekPDU(); // Get a ptr to l2 hdr

  return GetIfByMac(l2pdu->ra);
}

Interface* WirelessLink::GetPeer(Count_t i)
{ // Get the i'th peer's interface
  return GetIf(i);
}

// Link methods
Link* WirelessLink::Copy() const
{
  return new WirelessLink(*this);
}

bool WirelessLink::Transmit(Packet* p, Interface* myIface, Node* node, Rate_t r)
{
  InterfaceWireless* self_if = (InterfaceWireless*)myIface;
  if (self_if->GetLinkState()) {
    // Link is busy
    // This should NOT happen!!!
    cout << "HuH? linkBusy while xmitting linkState " << self_if->GetLinkState()
         << endl;
    return false;
  }

  node->BuildRadioInterfaceList();

  LinkEvent* evTx = new LinkEvent(LinkEvent::PACKET_TX);
  evTx->numberBytes = p->Size();
  Time_t txTime = ((Rate_t)p->Size() * 8) / r;
  Simulator::instance->Schedule(evTx, txTime, this);
  nextFreeTime = Simulator::Now() + txTime;
  Time_t* ptime = new Time_t(nextFreeTime);
  if (self_if->GetLinkFreeTime() < nextFreeTime)
    self_if->SetLinkFreeTime(nextFreeTime);
  else {
    // This should NOT happen!!!
    cout << "HuH? linkFreeTime out of order linkState "
         << self_if->GetLinkState() << endl
         << " self linkFreeTime " << self_if->GetLinkFreeTime()
         << " nextFreeTime " << nextFreeTime
         << " now " << Simulator::Now() << endl;
  }
  self_if->SetLinkState(InterfaceWireless::TX);
  // yj, this is a PHY-TXSTART.indication
  InterfaceWireless::PhyInd_t phyInd = InterfaceWireless::PHY_TXSTART;
  self_if->GetL2Proto()->Notify(&phyInd);
  AddNotify(self_if, ptime);
  // also time to cut the node's energy reserve by appropriate values
  double delta_time = Simulator::Now() - self_if->last_time ;

  L2Header802_11* l2hdr = (L2Header802_11*)p->PeekPDU();
  double ftxPow;
  switch (l2hdr->type) {
  case RTS:
           ftxPow *= self_if->getRTSTxPower();
           break;
  case CTS:
           ftxPow *= self_if->getCTSTxPower();
           break;
  case ACK:
           ftxPow *= self_if->getACKTxPower();
           break;
  default:
           ftxPow = self_if->getTxPower();
           /* Nothing to do here */
  }
  node->setBattery(node->getBattery() - (txTime * ftxPow +
                                  (delta_time * node->getComputePower())));
  self_if->last_time = Simulator::Now();

  if (node->getBattery() < 0) {
    cout << "OUT OF BATTERY POWER SIMULATION FROM HERE ON IS"
         << "TRUELY WHAT IT IS :: A SIMULATION :) "<< endl;
  }
  const RadioVec_t& rIFList = node->GetRadioInterfaceList();
#ifdef ETRACE_DEBUG
  // Trace "S" event (before broadcast replication)
  PDU *pdu_ = p->FindPDU(3,0x800); // IPv4 PDU (if it exists)
  if (pdu_) {
    if ( (((IPV4Header *)pdu_)->protocol == 89) || (((IPV4Header *)pdu_)->protocol == 17) ) {    
      ETrace::Instance()->TracePacket(p, myIface->GetNode(), myIface,1, "S");
    }
	}
#endif
  RadioVec_t::const_iterator i = rIFList.begin();
  for (; i != rIFList.end(); ++i)
    // TBD:Check for BERs because of Radio Propagation Models
    {
      // schedule recv at all i/f except self
      InterfaceWireless* inf = i->inf;
      if (inf == self_if) continue;
 
#ifdef ETRACE_DEBUG
      // Trace "T" event (broadcast replication to all nodes that supposedly
      // are within range or possibly in range)
      PDU *pdu_ = p->FindPDU(3,0x800); // IPv4 PDU (if it exists)
      if (pdu_) {
        if ( (((IPV4Header *)pdu_)->protocol == 89) || 
             (((IPV4Header *)pdu_)->protocol == 17) ) {    
          ETrace::Instance()->TracePacket(p, myIface->GetNode(), myIface,1,"T");
        }
      }
#endif
#ifdef BUGFIX
      // Here is the simplest way to fix the problem
      propagation->SetPower(node->GetRadioRange(), 
                            alpha*rxThreshold + (1.0-alpha)*csThreshold);
#endif //BUGFIX
      double pr = propagation->Pr(i->dist);    // rx power, mW

      if (pr < csThreshold)  // ignore if rx power < CST
        continue;
      bool error = false;
      if (pr < rxThreshold)  // mark as erroneous if rx power < RXT
#ifdef BUGFIX 
      { // power between csThreshold and rxThreshold
        double probability = (pr-csThreshold)/(rxThreshold-csThreshold);
        if (rngProp->Value() >= probability) 
          error = true;
      }
#else 
        error = true;
#endif  //BUGFIX

      Time_t rxTime
        = txTime + SpeedOfLight(node, inf->GetNode()) + jitter->Value();
      Time_t linkFreeTime = Simulator::Now() + rxTime;

      LinkEvent* evRx = nil;
      if (linkFreeTime > inf->GetLinkFreeTime()) 
      {
        // Schedule pkt receive event for this case only
        evRx = new LinkEvent(LinkEvent::PACKET_RX, p->Copy());
        inf->ScheduleRx(evRx, rxTime);
        inf->SetLinkFreeTime(linkFreeTime);
      }
      if (!inf->GetLinkState()) 
      { // link is idle
        // 1st scheduled pkt
        evRx->hasBitError = error;
        inf->SetRxPacketEvent(evRx);
        inf->SetRxPower(pr);
        inf->SetLinkState(InterfaceWireless::RX);
        // yj, this is a PHY-RXSTART.indication
        phyInd = InterfaceWireless::PHY_RXSTART;
        inf->GetL2Proto()->Notify(&phyInd);
#ifdef SIM //ETRACE
        if (error)
        {
          PDU *pdu = p->FindPDU(3,0x800); // IPv4 PDU (if it exists)
#ifdef USAP_ADDON
          if((pdu))
#else
          if((pdu) && (((IPV4Header *)pdu)->protocol == 89))
#endif // USAP_ADDON
          {
            Stats::ospfTxPktsDropped++;
          }
        }
#endif //SIM
#ifdef HAVE_QT
        /* Differentiate RX, RX_ME, and RX_ZZ
           RX_ME means the pkt is for me
           RX_ZZ means the pkt is interference
           These states are mutually exclusive
        */
        if (error)
          inf->SetLinkState(InterfaceWireless::RX_ZZ);
        else 
        {
          L2Header802_11* l2hdr = (L2Header802_11*)p->PeekPDU();
          if (l2hdr->ra == inf->GetMACAddr() || l2hdr->ra.IsBroadcast())
            inf->SetLinkState(InterfaceWireless::RX_ME);
        }
#endif
      }
      else
      { // link is busy

        // compare with rx power of incoming pkt
        if (inf->GetRxPacketEvent() && !inf->GetRxPacketEvent()->hasBitError)
        {
          if (inf->GetRxPower()/pr < snrThreshold)
          { // collision
            inf->GetRxPacketEvent()->hasBitError = true;
            inf->SetLinkState(InterfaceWireless::CX);
#ifdef SIM //ETRACE
            PDU *pdu = p->FindPDU(3,0x800); // IPv4 PDU (if it exists)
            if((pdu) && (((IPV4Header *)pdu)->protocol == 89)) {
              Stats::ospfTxCollisions++;
            }
          }
#endif //SIM
        }
        else if (inf->GetLinkState() != InterfaceWireless::TX) 
        {
          /* Actually, this checking may be unnecessary because
             if a neighbor is xmitting, this node cannot xmit.
             Thus, this node's xmitting assumes no neighbors
             are transmitting.
             Mobility may make this happen?
             interference + interference = collision?
          */
          inf->SetLinkState(InterfaceWireless::CX);
        }
        if (evRx) 
        {
          evRx->hasBitError = true;  // this pkt is an interference
        }
        PDU *pdu = p->FindPDU(3,0x800); // IPv4 PDU (if it exists)
	 
        if((pdu) && (((IPV4Header *)pdu)->protocol == 89)) {
          Stats::ospfTxCollisions++;
        }
      }
    } //for

  //delete p;
  Stats::pktsTransmitted++;
#ifdef HAVE_QT
  // Animate if requested
  if (Simulator::instance->AnimateWirelessTx())
    {
      L2Header802_11* l2hdr = (L2Header802_11*)p->PeekPDU();
      QColor c;
      switch (l2hdr->type)
        {
          case RTS:
            c = Qt::cyan;   break;
          case CTS:
            c = Qt::yellow; break;
          case DATA:
            c = l2hdr->ra.IsBroadcast() ? Qt::black : Qt::blue;  break;
          case ACK:
            c = Qt::green;  break;
        }
      node->WirelessTxColor(c);
      QTWindow* qtw = Simulator::instance->GetQTWindow();
      qtw->WirelessTxStart(node, self_if, nil, node->GetRadioRange());
    }
#endif
  return true;
}

Time_t WirelessLink::SpeedOfLight(Node* n1, Node* n2)
{ // Calculate speed of light delay between nodes
  Meters_t dist = n1->Distance(n2);
  return dist / SPEED_LIGHT;
}

void WirelessLink::ConstructorHelper(
    Count_t n,      // Node count
    IPAddr_t i,
    Mask_t mask,    // IP addr start, mask
#ifdef USAP_ADDON
    const L2Proto &l2,
#endif // USAP_ADDON
    L4Protocol* l4, // L4 protocol to bind to each
    Rate_t r, Time_t dly,
    Detail_t d, bool bootstrap)
{
  
  first = Node::nextId;             // Set the address of first node
  firstMac = MACAddr::next;         // First MAC address
  firstIP = i;                      // First IP Address
  detail = d;                       // Detail level
  bw = r;                           // Bandwidth
  delay = dly;                      // Delay
  for (Count_t j = 0; j < n; ++j)
    { // Create the nodes
      Node* node = new Node();
#ifdef USAP_ADDON
      Interface* iface = node->AddInterface(l2, i,
					    mask, MACAddr::Allocate(),
					    bootstrap);
#else
      Interface* iface = node->AddInterface(L2Proto802_11(), i,
					    mask, MACAddr::Allocate(),
					    bootstrap);
#endif // USAP_ADDON

      iface->SetLink(this);         // Uses this link
      ++i;                          // Advance IP Address
    }
  last = Node::nextId;              // Last id + 1
}

Interface* WirelessLink::GetIfByMac(MACAddr m)
{ // Get the interface corresponding to the specified MAC address
  Node* n = GetNode(m - firstMac);    // First get the node
  if (n) return n->GetIfByLink(this); // Find the interface for this MAC
  return nil;                         // Out of range
}

Count_t WirelessLink::PeerCount()
{ // Return number of peers
  Count_t r = last - first - 1; // Subtract one since not self-peer
  return r;
}

Count_t WirelessLink::NodeCount()
{ // Return number of nodes
  return last - first;
}

IPAddr_t WirelessLink::PeerIP(Count_t npeer)
{ // Get peer IP Address (if known)
  Node* n = GetNode(npeer);  // Get the node
  if (n) return n->GetIPAddr(); // get the node's ip address
  return IPADDR_NONE;        // Problem, return none
}

Count_t WirelessLink::NodePeerIndex(Node* n)
{ // Return the peer index.  Assumed to be in range
  NodeId_t id = n->Id();
  return id - first;
}

IPAddr_t WirelessLink::NodePeerIP(Node* n)
{ // Get peer IP Address for the specified node
  for (Count_t i = 0; i < PeerCount(); ++i)
    {
      Node* peerNode = GetNode(i);
      if (peerNode == n) return GetIP(i); // Found it, return IP Addresss
    }
  return IPADDR_NONE;        // Not found, return none
}

Count_t WirelessLink::NeighborCount(Node*)
{ // Number of routing neighbors
  return PeerCount(); // Otherwise return number of peers
}

void WirelessLink::Neighbors(NodeWeightVec_t& nwv, Interface* i, bool all)
{ // Add to neighbor list
  // For wireless LAN's, return an empty list.
}


void WirelessLink::AllNeighbors(NodeWeightVec_t& nwv)
{ // Make a list of all neighbors.  Note the the "Interface" pointer
  // in the NodeIfWeight vector is the TARGET interface, not the
  // source as in Neighbors above.  This function is used by L2
  // broadcast to determine where to schedule receive events.
  //
  for (NodeId_t ni = first; ni < last; ++ni)
    {
      Interface* iface = GetIf(ni);
      if (!iface) cout << "HuH?  Ethernet::AllNeighbors nill iface for "
                       << ni << endl;
      nwv.push_back(NodeIfWeight(Node::GetNode(ni), GetIf(ni), Weight()));
    }
}

MACAddr WirelessLink::IPToMac(IPAddr_t ip)
{ // Convert peer IP to peer MAC
  if (ip < firstIP)
    {
      return MACAddr::NONE; // Out of range
    }
  if (ip >= (firstIP + (last - first)))
    {
      return MACAddr::NONE;
    }
  Count_t delta = ip - firstIP;
  return firstMac + delta;
}

IPAddr_t WirelessLink::GetIP(Count_t i)
{ // Get the i'th ip address
  if ((first + i) >= (last)) return IPADDR_NONE; // Out of range
  return (IPAddr_t)(firstIP + i);
}

Node* WirelessLink::GetNode(Count_t i)
{ // Get the i'th node
  if ((first + i) >= last) return nil; // Out of range
  return Node::GetNode(first + i);     // Return the specified node
}

Interface* WirelessLink::GetIf(Count_t i)
{ // Get the interface for the i'th node
  Node* n = GetNode(i);
  if (!n) return nil;
  return n->GetIfByLink(this);
}

bool WirelessLink::NodeIsPeer(Node* n)
{ // Return true of node is a peer
  NodeId_t id = n->Id();
  return (id >= first && id < last); // True if in range
}

#ifdef BUGFIX
void WirelessLink::SetAlpha(double a)
{
  alpha = a;
}
#endif // BUGFIX

#ifdef MANET
bool SimpleWirelessLink::Transmit(Packet* p, Interface* myIface, Node* node, Rate_t r) 
{
  InterfaceWireless* self_if = (InterfaceWireless*)myIface;
  Time_t txTime = ((Rate_t)p->Size() * 8) / r;
  // IPv4 PDU (if it exists)
  IPV4Header *ippdu = (IPV4Header *)(p->FindPDU(3,0x800)); 
  if (!ippdu) {
     cout << "Not an IPv4 PDU-- unsupported" << endl;
     exit(1);
  }
#ifdef ETRACE_DEBUG
  // Trace "S" event (before broadcast replication)
  if  ( (ippdu->protocol == 89) || (ippdu->protocol == 17) )     
    ETrace::Instance()->TracePacket(p, myIface->GetNode(), myIface,1, "S");
#endif
   // Here we parallel the handling of packets in the Ethernet model
   // In normal wireless model, we iterate among all nodes
   // Here, if unicast, we send directly, but if multicast, we
   // send only to the neighbors
  if (IPAddr::IsMulticast(ippdu->dst)) {
    // MULTICAST
    //cout << "Multicast packet " << IPAddr::ToDotted(ippdu->dst) << endl;
    const NodeVec_t& nodes = node->GetNodes();
    InterfaceWireless* inf = nil;
    // Iterate over physical neighbors instead of all nodes
#if 0
    // This will not work because it looks at OSPF neighbors
    OSPF6DInstance* ospf6_i =  node->GetOspf6dInst();
    struct ospf6* ospf6_inst = ospf6_i->get_inst_ospf6();
    struct listnode *i, *j, *k;
    for (i = listhead(ospf6_inst->area_list); i; nextnode(i)) {
      struct ospf6_area *oa = (struct ospf6_area *) getdata(i);
      for (j = listhead(oa->if_list); j; nextnode(j)) {
        struct ospf6_interface *oi = (struct ospf6_interface *) getdata(j);
        for (k = listhead(oi->neighbor_list); k; nextnode(k)) {
          struct ospf6_neighbor *on = (struct ospf6_neighbor *) getdata(k);
          Node* n = Node::RouterIdToNode((int)((ntohl(on->router_id)) & 0xff));
        }
      }
    }
#endif
 
    if (node->IsMobile())
      node->UpdateLocation();
    for (NodeVec_t::const_iterator i = nodes.begin(); i != nodes.end(); ++i)
    {
      Node* n = *i;
      if (node->Id() == n->Id()) continue;    // skip self node
      Interface* iface = n->GetIfByNode(n);
      if (iface->IsWireless())
        inf = (InterfaceWireless*)iface;
      else
        continue;
#ifdef ETRACE_DEBUG
  // Trace "T" event (broadcast replication to all nodes that supposedly
  // are within range or possibly in range)
      if ( (ippdu->protocol == 89) || (ippdu->protocol == 17) )     
         ETrace::Instance()->TracePacket(p, myIface->GetNode(), myIface,1, "T");
#endif
      // Here is the simplest way to fix the problem
      propagation->SetPower(node->GetRadioRange(), alpha*rxThreshold + 
        (1.0-alpha)*csThreshold);
      double pr = propagation->Pr(node->Distance(n));    // rx power, mW
      if (pr < csThreshold)  // ignore if rx power < CST
        continue;

      bool error = false;
      if (pr < rxThreshold)  // mark as erroneous if rx power < RXT
			{ // power between csThreshold and rxThreshold
				double probability = (pr-csThreshold)/(rxThreshold-csThreshold);
        if (rngProp->Value() >= probability) error = true;
			}
      Time_t rxTime = txTime + SpeedOfLight(node, inf->GetNode()) + 
        jitter->Value();
      LinkEvent* evRx = nil;
      evRx = new LinkEvent(LinkEvent::PACKET_RX, p->Copy());
      evRx->hasBitError = error;
      inf->SetRxPacketEvent(evRx);
      inf->SetRxPower(pr);
      Simulator::instance->Schedule(evRx, rxTime, inf);
      if (error)
      {
#ifdef USAP_ADDON
        {
#else
	        if(ippdu->protocol == 89) {    
#endif // USAP_ADDON
	          Stats::ospfTxPktsDropped++;
          }
      }
    } //for MULTICAST
  } else {
    // else UNICAST
    L2HeaderSimple* l2pdu = (L2HeaderSimple*)p->PeekPDU();
    MACAddr dst;
    DATA_frame* dataFrame = (DATA_frame*)l2pdu;
    dst = dataFrame->addr1;
    // XXX relies on fact that mac addresses are same as router IDs
    // Normally this would be an iterator here for all nodes and a for loop
    Node* n = Node::RouterIdToNode((int)dst);
    InterfaceWireless* inf = nil;
      Interface* iface = n->GetIfByNode(n);
      if (iface->IsWireless())
        inf = (InterfaceWireless*)iface;
#ifdef ETRACE_DEBUG
  // Trace "T" event (broadcast replication to all nodes that supposedly
  // are within range or possibly in range)
      if ( (ippdu->protocol == 89) || (ippdu->protocol == 17) )     
         ETrace::Instance()->TracePacket(p, myIface->GetNode(), myIface,1, "T");
#endif
      // Here is the simplest way to fix the problem
      propagation->SetPower(node->GetRadioRange(), alpha*rxThreshold + 
        (1.0-alpha)*csThreshold);
      if (n->IsMobile()) 
        n->UpdateLocation();
      double pr = propagation->Pr(node->Distance(n));    // rx power, mW
#ifdef REMOVED_FOR_SCALING
      if (pr < csThreshold)  // ignore if rx power < CST
        continue;
#endif
      bool error = false;
      if (pr < rxThreshold)  // mark as erroneous if rx power < RXT
			{ // power between csThreshold and rxThreshold
				double probability = (pr-csThreshold)/(rxThreshold-csThreshold);
        if (rngProp->Value() >= probability) error = true;
			}

      Time_t rxTime = txTime + SpeedOfLight(node, inf->GetNode()) + 
        jitter->Value();
      LinkEvent* evRx = nil;
      evRx = new LinkEvent(LinkEvent::PACKET_RX, p->Copy());
      evRx->hasBitError = error;
      inf->SetRxPacketEvent(evRx);
      inf->SetRxPower(pr);
      Simulator::instance->Schedule(evRx, rxTime, inf);
      if (error)
      {
#ifdef USAP_ADDON
        {
#else
	        if(ippdu->protocol == 89) {    
#endif // USAP_ADDON
	          Stats::ospfTxPktsDropped++;
          }
      }
  } // else UNICAST

  //delete p;

  Stats::pktsTransmitted++;
#ifdef HAVE_QT
  // Animate if requested
  if (Simulator::instance->AnimateWirelessTx())
    {
      L2Header802_11* l2hdr = (L2Header802_11*)p->PeekPDU();
      QColor c;
      switch (l2hdr->type)
        {
          case RTS:
            c = Qt::cyan;   break;
          case CTS:
            c = Qt::yellow; break;
          case DATA:
            c = l2hdr->ra.IsBroadcast() ? Qt::black : Qt::blue;  break;
          case ACK:
            c = Qt::green;  break;
        }
      node->WirelessTxColor(c);
      QTWindow* qtw = Simulator::instance->GetQTWindow();
      qtw->WirelessTxStart(node, self_if, nil, node->GetRadioRange());
    }
#endif
  return true;
}

// Simple helper function to compare addresses.
// Note, we distinguish between link local IPv4 and regular IPv4 addresses
// by the uppermost v4 nibble; i.e., global scope 10.0.0.1 has link-local
// equivalent of 138.0.0.1
//
// This function sees if the two IP addresses match in their lowest 7 nibbles
bool SimpleWirelessLink::IsAddrMatch(IPAddr_t a, IPAddr_t b) {
  return ( (a & 0x7fffffff) == (b & 0x7fffffff));
} 

#endif
