/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Worm Target Vector Class
// Monirul I Sharif.  Georgia Tech, Fall 2003

// Define a Worm Application class

//#define DEBUG_MASK 0x1

#include "common-defs.h"
#include "rng.h"
#include "debug.h"
#include "wormtargetvector.h"

using namespace std;

WormTargetVector::WormTargetVector(Count_t range) 
  : initialized(false), scanrange(range), localIP(IPADDR_NONE)
{

}

WormTargetVector::WormTargetVector(Mask m)
{
  scanrange = 1<<(32-m.NBits());
}

WormTargetVector* WormTargetVector::Copy() const
{
  return new WormTargetVector(*this);
}

void WormTargetVector::Initialize(IPAddr_t IP)
{
  //cout<<"Initialize called"<<endl;
  localIP = IP;
  initialized = true;
}


IPAddr_t WormTargetVector::Generate()
{
  if (!initialized) 
    cout<<"TargetVector Not Initialized!!"<<endl;
  return IPADDR_NONE;
}

WTVUniform::WTVUniform(Count_t scanrange) : WormTargetVector(scanrange)
{
}


WormTargetVector* WTVUniform::Copy() const
{
  return new WTVUniform(*this);
}

void WTVUniform::Initialize(IPAddr_t IP)
{
  WormTargetVector::Initialize(IP);

  rngT = new Uniform(0, scanrange);
}

IPAddr_t WTVUniform::Generate()
{
  WormTargetVector::Generate();

  return rngT->IntValue();
}


WTVLocalPref::WTVLocalPref(Count_t range, double p,
			   Mask local) : WormTargetVector(range)
{
  localscanrange = 1<<(32-local.NBits());
  prob = p;
}

WTVLocalPref::WTVLocalPref(Mask global, double p,
			   Mask local) : WormTargetVector(global)
{
  localscanrange = 1<<(32-local.NBits());
  prob = p;
}

WormTargetVector* WTVLocalPref::Copy() const
{
  return new WTVLocalPref(*this);
}

void WTVLocalPref::Initialize(IPAddr_t IP)
{
  WormTargetVector::Initialize(IP);

  rngP  = new Uniform(0, 1);
  rngT1 = new Uniform(0, scanrange);
  rngT2 = new Uniform(0, localscanrange);
}


IPAddr_t WTVLocalPref::Generate()
{
  WormTargetVector::Generate();

  IPAddr_t a,b;
  double p = rngP->Value();
  DEBUG0((cout<<"p="<<p<<endl));

  if ( p< prob) {
    a = rngT2->IntValue();
    DEBUG0((cout<<"l="<<(string)IPAddr(a)
	    <<" s="<<scanrange
	    <<" l="<<(string)IPAddr(~(localscanrange-1))<<endl));
    b = localIP&(scanrange-1);
    DEBUG0((cout<<"a="<<(string)IPAddr(b)<<endl));
    b = b&~(localscanrange-1);
    DEBUG0((cout<<"a="<<(string)IPAddr(b)<<endl));
    return a|b;
  } else {
    a = rngT1->IntValue();
    DEBUG0((cout<<"g="<<(string)IPAddr(a)<<endl));
    return a;
  }
}

WTVSequential::WTVSequential(Count_t range) : WormTargetVector(range)
{
}

WTVSequential::WTVSequential(Mask global) : WormTargetVector(global)
{
}

WormTargetVector* WTVSequential::Copy() const
{
  return new WTVSequential(*this);
}

void WTVSequential::Initialize(IPAddr_t IP)
{
  WormTargetVector::Initialize(IP);
  Random *rngT = new Uniform(0, scanrange);

  currentIP = rngT->IntValue();
  delete rngT;
}


IPAddr_t WTVSequential::Generate()
{
  WormTargetVector::Generate();

  IPAddr_t a = currentIP;

  currentIP++;
  if (currentIP>=scanrange)
    currentIP=0;

  return a;
}





WTVSequentialWithLocalPref::WTVSequentialWithLocalPref(Count_t range,
						       double p,
						       Mask local) 
  : WormTargetVector(range)
{
  localscanrange = 1<<(32-local.NBits());
  prob = p;
}

WTVSequentialWithLocalPref::WTVSequentialWithLocalPref(Mask global,
						       double p,
						       Mask local) 
  : WormTargetVector(global)
{
  localscanrange = 1<<(32-local.NBits());
  prob = p;
}


WormTargetVector* WTVSequentialWithLocalPref::Copy() const
{
  return new WTVSequentialWithLocalPref(*this);
}

void WTVSequentialWithLocalPref::Initialize(IPAddr_t IP)
{
  WormTargetVector::Initialize(IP);
  Random *rngP = new Uniform(0,1);
  Random *rngT;

  double p = rngP->Value();
  if (p<prob) {
    rngT = new Uniform(0, localscanrange);
    IPAddr_t a,b;
    a = rngT->IntValue();
    b = localIP&(scanrange-1);
    b = b&~(localscanrange-1);
    currentIP = a|b;
  } else {
    rngT = new Uniform(0, scanrange); 
    currentIP = rngT->IntValue();
  }
  delete rngT;
}


IPAddr_t WTVSequentialWithLocalPref::Generate()
{
  WormTargetVector::Generate();

  IPAddr_t a = currentIP;

  currentIP++;
  if (currentIP>=scanrange)
    currentIP=0;

  return a;
}


