// Copyright 2004, The Boeing Company

#include <iostream>
#include <stdlib.h>

#include "application-cbr.h"
#include "application-ospf6d.h"
#include "args.h"
#include "estat.h"
#include "etrace.h"
#include "ipaddr.h"
#include "linkmatrix.h"
#include "mobility-random-waypoint.h"
#include "node.h"
#include "rng.h"
#include "simulator.h"
#include "validation.h"  
#include "wireless-grid-rectangular.h"
#include "wlan.h"

void set_global_seeds(int seed)
{
  // turn one seed into 6 repeatable seeds in the simulator using srand() 
  unsigned long seeds[6];
  srand(seed);

  seeds[0] = rand(); 
  seeds[1] = rand(); 
  seeds[2] = rand(); 
  seeds[3] = rand(); 
  seeds[4] = rand(); 
  seeds[5] = rand(); 

  Random::GlobalSeed(seeds[0],seeds[1],seeds[2],seeds[3],seeds[4],seeds[5]);
}

int main(int argc, char** argv)
{
  struct ospf6_config_args ca;
  Validation::Init(argc, argv);
  Simulator s;
  Interface *inter;
  IFVec_t vInterface;
  std::string argstring, tracefile, statfile;
  char hostname[20], logfilename[20];
  IPAddr_t first_rtrid;

  if (!set_ospf6_config_arguments(argc, argv, &ca))
    return 0;

  set_global_seeds(ca.seed);

  // If user specifies "tag=args", this means that the suffix appended
  // to the .stat or .tr filename should be the concatenation of all command
  // line arguments, e.g.:
  // random-waypoint-mobility-opt-num_nodes=20.seed=72.stat
  if (ca.tag == "args") {
    ca.tag.erase(0,4);
    for (int q = 1; q < argc; q++) {
      if (!strcmp(argv[q],"tag=args") ||
          !strcmp(argv[q],"trace") ||
          !strcmp(argv[q],"etrace") ||
          !strcmp(argv[q],"ospfv3.log"))
        continue;
      ca.tag.append(argv[q]);
      ca.tag.append(".");
    }
    ca.tag.erase(ca.tag.length()-1, 1);
  }

  // Prepare output file names
  tracefile.append(argv[0]);
  tracefile.erase(tracefile.length()-4,4); //delete "-opt" or "-dbg" from name
  statfile.append(argv[0]);
  statfile.erase(statfile.length()-4,4); // delete "-opt" or "-dbg" from name

  // Make sure that output file names are less than 256 char limit
  if ((statfile.length() + ca.tag.length() + 7) > 256) {
    int overrun =  statfile.length() + ca.tag.length() + 7 - 256;
    ca.tag.erase(ca.tag.length()- overrun, overrun);
  }

  if (ca.tag.size()) {
    tracefile.append(".");
    tracefile.append(ca.tag.c_str());
  }
  tracefile.append(".tr");

  if (ca.tag.size()) {
    statfile.append(".");
    statfile.append(ca.tag.c_str());
  }
  statfile.append(".stat");

  cout << statfile.c_str() << endl;

  // Capture command line arguments for logging purposes
  for (int q = 0; q < argc; q++) {
    argstring.append(argv[q]);
    argstring.append(" ");
  }

  // ipv4 addresses must be less than 128.0.0.0 for ospf6 because we used 
  // the first bit in the ipv4 address to indicate a linklocal address 
  IPAddr_t firstIP = IPAddr("10.0.0.1");

  //Matrix 
  LinkMatrix LM(ca.nNodes, firstIP, Mask(24)); 
  // 
  // To change the connectivity between two nodes (indexed by nodeId), 
  // use commands of the type found below, where the arguments are (in order):
  // (Time, src_node, dst_node, packet_error_rate)
  // By default, all neighbor relationships are error free.
  
///*
  //Node 0
  LM.ScheduleMatrixChange(0, 0, 4, 1);
  LM.ScheduleMatrixChange(0, 0, 5, 1);
  LM.ScheduleMatrixChange(0, 0, 6, 1);
  LM.ScheduleMatrixChange(0, 0, 7, 1);
  //Node 1
  LM.ScheduleMatrixChange(0, 1, 7, 1);
  //Node 2
  LM.ScheduleMatrixChange(0, 2, 7, 1);
  //Node 3
  LM.ScheduleMatrixChange(0, 3, 7, 1);
  //Node 4
  LM.ScheduleMatrixChange(0, 4, 0, 1);
  //Node 5
  LM.ScheduleMatrixChange(0, 5, 0, 1);
  //Node 6
  LM.ScheduleMatrixChange(0, 6, 0, 1);
  //Node 0
  LM.ScheduleMatrixChange(0, 7, 0, 1);
  LM.ScheduleMatrixChange(0, 7, 1, 1);
  LM.ScheduleMatrixChange(0, 7, 2, 1);
  LM.ScheduleMatrixChange(0, 7, 3, 1);


  LM.ScheduleMatrixChange(1900, 6, 3, 1);
  LM.ScheduleMatrixChange(1900, 3, 6, 1);

// */
  // Per-packet tracing configuration
  if (ca.trace) {
    ETrace* egs = ETrace::Instance();
    egs->Open(tracefile.c_str());
    if (ca.etrace)
      egs->SetExtended(true);
    egs->LayerOff(1); //don't etrace this layer
    egs->LayerOff(2); //don't etrace this layer
    egs->SetTraceEventLevel(2);
  }

  // Statistics collection
  EStat* estat = EStat::Instance();
  estat->CollectEStats(statfile.c_str(), ca.start_time);
  estat->LayerOff(1);
  estat->LayerOff(2);
  
  // OSPF6 initialization
  OSPF6DApplication ospf6d;
  ospf6d.CollectStats(statfile.c_str(), argstring.c_str(), ca.start_time);
  ospf6d.LogNeighbors(0.5);  // Log physical layer neighbors every 500 ms
  
  OSPF6DInstance *ospf6_inst;
  Uniform u(0,ca.start_time);
  first_rtrid = IPAddr("0.0.0.1"); 
          
  // CBR traffic initialization
  IPAddr remoteIP; 
  char buf[32];
  Rate_t rate_bps = ((Rate_t)ca.pktrate * ca.pktsize * 8) /
                    (ca.nNodes*(ca.nNodes-1));
  // start traffic generators at a random time
  Uniform cbrstart(ca.start_time-(double)(ca.pktsize*8)/rate_bps,ca.start_time); 
  
  for (Count_t i = 0; i < LM.Size(); ++i)
  {
    Node* n = LM.GetNode(i);
     
    // OSPF6 Router Configuration
    ospf6_inst = ospf6d.AddNode(n);

    sprintf(hostname, "ospf6_%d", n->Id()+1);
    sprintf(logfilename, "log_ospf6_%d.log", n->Id()+1);

    ospf6_inst->SetHostName(hostname);
    if (ca.ospfv3log)
      ospf6_inst->SetLogFileName(logfilename);
    ospf6_inst->SetRouterId(n->Id()+first_rtrid);
    ospf6_inst->SetMinLSInterval(ca.MinLSInterval);
    ospf6_inst->SetMinLSArrival(ca.MinLSArrival);

    vInterface = n->Interfaces();
    for (IFVec_t::iterator it=vInterface.begin(); it!=vInterface.end(); ++it)
    {
      ospf6_inst->SetInterfaceHelloInterval(*it, ca.HelloInterval);
      ospf6_inst->SetInterfaceDeadInterval(*it, ca.DeadInterval);
      ospf6_inst->SetInterfaceRetransmitInterval(*it, ca.RxmtInterval);
      ospf6_inst->SetInterfaceAckInterval(*it, ca.AckInterval);
      ospf6_inst->SetInterfaceArea(*it, 0);
      ospf6_inst->SetInterfaceCost(*it, ca.cost);
      // too much jitter can cause stale ospf6 state to be sent
      ospf6_inst->SetInterfaceJitter(*it, ca.interface_jitter); //msec
      ospf6_inst->SetInterfaceFloodDelay(*it, ca.interface_flood_delay); //msec

      if (strcmp((*it)->GetName(), "lo") == 0)
      {
        ospf6_inst->SetInterfaceType(*it, "loopback");
        if (rate_bps > 0 && ca.pktsize > 0)
        {
          // CBR traffic from this router to all other router's loopback addr
          for (Count_t j = 0; j < LM.Size(); ++j)
          {
            if (i == j)
              continue;
            sprintf(buf, "0.0.%d.0", j+1); //loopback addr of routers
            remoteIP = IPAddr(buf);
            CBRApplication* cbr = (CBRApplication*) n->AddApplication(
                   CBRApplication(remoteIP,1000,NO_PORT,rate_bps,ca.pktsize));
            cbr->Start(cbrstart.Value());
            cbr->Stop(ca.stop_time - 5);
          }
        }
      }
      else
      {
        if(ca.w_int_type == 1)
          ospf6_inst->SetInterfaceType(*it, "point-to-multipoint");
        else if (ca.w_int_type == 2)
        {
          ospf6_inst->SetInterfaceType(*it, "manet-reliable");
          ospf6_inst->SetInterfaceDiffHellos(*it, ca.diff_hellos);
          //set flooding type
          if (ca.w_flooding == 0)
            ospf6_inst->SetInterfaceFlooding(*it, "broadcast");
          else if (ca.w_flooding == 1)
          {
            ospf6_inst->SetInterfaceFlooding(*it, "mpr");
            ospf6_inst->SetInterfacePushBackInterval(*it, ca.PushbackInterval);
            ospf6_inst->SetInterfaceSmartPeering(*it, ca.smart_peering);
            ospf6_inst->SetInterfaceUnsynchAdj(*it, ca.unsynch_adj);
          }
          else if (ca.w_flooding == 2)
          {
            ospf6_inst->SetInterfaceFlooding(*it, "mdr");
            ospf6_inst->SetInterfaceBackupWaitInterval(*it,
                                                       ca.BackupWaitInterval);
            ospf6_inst->SetInterfaceTwoHopRefresh(*it, ca.TwoHopRefresh);
            ospf6_inst->SetInterfaceHelloRepeatCount(*it, ca.HelloRepeatCount);
            if (ca.AdjConnectivity == 0)
              ospf6_inst->SetInterfaceAdjConnectivity(*it, "fully");
            else if (ca.AdjConnectivity == 1)
              ospf6_inst->SetInterfaceAdjConnectivity(*it, "uniconnected");
            else
              ospf6_inst->SetInterfaceAdjConnectivity(*it, "biconnected");
            if (ca.LSAFullness == 0)
              ospf6_inst->SetInterfaceLSAFullness(*it, "minlsa");
            else if (ca.LSAFullness == 1)
              ospf6_inst->SetInterfaceLSAFullness(*it, "minhoplsa");
            else if (ca.LSAFullness == 2)
              ospf6_inst->SetInterfaceLSAFullness(*it, "mdrfulllsa");
            else
              ospf6_inst->SetInterfaceLSAFullness(*it, "fulllsa");
            ospf6_inst->SetInterfaceNonPersistentMDR(*it, ca.NonPersistentMDR);
          }
        }
        if (i==0)
        {
          WirelessLink *l = (WirelessLink *) (*it)->GetLink();
          l->SetAlpha(ca.alpha);
        }
      }
    }
    ospf6_inst->Start(u.Value());  // random start time of each ospf6 router
  }
  s.StopAt(ca.stop_time);
  s.Progress(40);
  s.Run();
}
