// Copyright 2004, The Boeing Company
/*
*
*               0
*             / | \
*            1- 2 -3  
*            \  |  /
*               4
*               |
*               5
* At time 1900 the link is broken between nodes 4 and 5.  
* After the neighbor relationship between 4 & 5 expires, a new router LSA is 
* originated by node 4.  This LSA is flooded through out the network in 
* an LSU.  The relay node that floods the LSU from node 4 is node
* 2.  The LSU does not reach node 0 when it is flooded from node 2.  
* Nodes 1 and 3 ack the LSA from 4.  Nodes 1 and 3 are both BDRs.
* Nodes 2, 1, and 3 each retransmit the LSA to node 0 because of 
* node 0 didn't ack in time.
* Node 0 acks.
*  
*/
      
#include <iostream>
#include <stdlib.h>

#include "application-cbr.h"
#include "application-ospf6d.h"
#include "args.h"
#include "estat.h"
#include "etrace.h"
#include "ipaddr.h"
#include "linkmatrix.h"
#include "mobility-random-waypoint.h"
#include "node.h"
#include "rng.h"
#include "simulator.h"
#include "validation.h"  
#include "wireless-grid-rectangular.h"
#include "wlan.h"

// Do not change these defaults and expect the validation output to match.

// These lines enable only a small portion of the trace file to be written out
// (saving disk space).
double default_trace_on_time = 1900;
double default_trace_off_time = 2000;

// Basic configuration
int default_seed = 1;                  // Random seed
double default_start_time = 1800;      // seconds.  All statistics discarded
                                       //   for first "start_time" seconds
double default_stop_time = 2000;       // seconds
int default_pktsize = 40;              // CBR packet size (bytes)
int default_pktrate = 2;              // CBR packets per second overall 
std::string default_tag = "";          // string tag for appending output files
                                       //   in the simulator
// Scenario configuration
int default_num_nodes = 6;            // Number of nodes in the topology
// OSPF configuration
int default_wireless_interface = 2;    // interface type:  1=ptmp, 2=manet
int default_wireless_flooding = 2;     //flooding 0=broadcast, 1=mpr, 2=mdr
int default_rxmt_int = 10;             // retransmit interval (seconds)
int default_pb_int = 4000;             // pushback interval (msec)
int default_bw_int = 1800;             // backupwait interval (msec)
int default_ack_int = 3600;            // ack interval (msec)
int default_minls_int = 5;             // min LSA Interval (seconds)
int default_minls_arrival = 1;         // min LSA Arrival (seconds)
int default_hello_int = 10;            // hello interval (seconds)
int dead_interval = 4 * default_hello_int;  // dead interval (seconds)
int cost = 1;                          // ospfv3 interface cost
int interface_jitter = 100;            // jitter of pkt sends (msec)
int interface_flood_delay = 100;       // LSA coalescing timer (msec)

void usage(int argc, char** argv) {
  cout << "Usage:  " << argv[0] << " [seed=value]"<< endl;
  cout << "         [start_time=value] [stop_time=value]"<< endl;
  cout << "         [pktsize=value] [pktrate=value]"<< endl;
  cout << "         [tag=string]"<< endl;
  cout << "         [wireless_interface=value] [wireless_flooding=value]"<<endl;
  cout << "         [rxmt_int=value] "<< endl;
  cout << "         [pushback_int=value] [ack_int=value] "<< endl;
  cout << "         [backupwait_int=value] "<< endl;
  cout << "         [minls_int=value] [minls_arrival=value]"<< endl;
  cout << "         [hello_int=value] "<< endl;
  cout << "         [diff_hellos] [tracing] [etrace] [ospfv3log]"<< endl;
  cout << " Options:"<<endl;
  cout << "     seed             : random seed, default: " 
                << default_seed << endl;
  cout << "     start_time       : statistics discarded before this time, "
                << "default: " << default_start_time << endl;
  cout << "     stop_time        : simulation stop time, default: " 
                << default_stop_time << endl;
  cout << "     pktsize          : packet size in bytes, default: " 
                << default_pktsize << endl;
  cout << "     pktrate          : number of user data pkts sent/sec (across "
                << endl;
  cout << "                        whole simulation), default: " 
                << default_pktrate << endl;
  cout << "     tag              : unique string to append to trace/statfile "
                << endl;
  cout << "                        names, default:null" << endl; 
  cout << "     wireless_interface: type of interface ptmp=1, manet=2, "
                << "default: " << default_wireless_interface <<endl;
  cout << "     wireless_flooding: flooding broadast=0, mpr=1, mdr=2, "
                << "default: " << default_wireless_flooding <<endl;
  cout << "     rxmt_int         : ospf6 retransmit interval, default: " 
                << default_rxmt_int << endl;
  cout << "     pushback_int         : ospf6 pushback interval, default: " 
                << default_pb_int << endl;
  cout << "     backupwait_int         : ospf6 backupwait interval, default: " 
                << default_bw_int << endl;
  cout << "     ack_int         : ospf6 ack interval, default: " 
                << default_ack_int << endl;
  cout << "     minls_int         : ospf6 min LSA interval, default: " 
                << default_minls_int << endl;
  cout << "     minls_arrival         : ospf6 min LSA arrival, default: " 
                << default_minls_arrival << endl;
  cout << "     hello_int         : ospf6 hello interval, default: " 
                << default_hello_int << endl;
  cout << "     diff_hellos      : enable diff hellos on wireless interface, "
                " default:off" <<  endl;
  cout << "     tracing          : enable per-packet tracing, default: off" 
                << endl;
  cout << "     etrace           : enable extended tracing, default: off" 
                << endl;
  cout << "     ospfv3log        : enable ospfv3 logging, default: off" 
                << endl;
  cout << endl;
}

void set_global_seeds(int seed)
{
  // turn one seed into 6 repeatable seeds in the simulator using srand() 
  unsigned long seeds[6];
  srand(seed);

  seeds[0] = rand(); 
  seeds[1] = rand(); 
  seeds[2] = rand(); 
  seeds[3] = rand(); 
  seeds[4] = rand(); 
  seeds[5] = rand(); 

  Random::GlobalSeed(seeds[0],seeds[1],seeds[2],seeds[3],seeds[4],seeds[5]);
}

int main(int argc, char** argv)
{
  Validation::Init(argc, argv);
  Simulator s;
  Interface *inter;
  IFVec_t vInterface;
  std::string tracefile, statfile;
  char hostname[20], logfilename[20];
  std::string tag; 
  IPAddr_t first_rtrid;

  Count_t seed, nNodes;
  Count_t w_int_type, w_flooding;
  Count_t rxmt_int,bw_int,pb_int,ack_int,minls_int,minls_arrival,hello_int;
  Count_t pktrate, pktsize;
  DCount_t start_time, stop_time; 
  double alpha;
			
  Arg("seed", seed, default_seed);
  Arg("wireless_interface", w_int_type, default_wireless_interface);
  Arg("wireless_flooding", w_flooding, default_wireless_flooding);
  Arg("rxmt_int", rxmt_int, default_rxmt_int);
  Arg("pushback_int", pb_int, default_pb_int);
  Arg("backupwait_int", bw_int, default_bw_int);
  Arg("ack_int", ack_int, default_ack_int);
  Arg("minls_int", minls_int, default_minls_int);
  Arg("minls_arrival", minls_arrival, default_minls_arrival);
  Arg("hello_int", hello_int, default_hello_int);
  Arg("pktrate", pktrate, default_pktrate);
  Arg("pktsize", pktsize, default_pktsize);
  Arg("tag", tag, default_tag);
  Arg("start_time", start_time, default_start_time);
  Arg("stop_time", stop_time, default_stop_time);
  
  Arg::ProcessArgs(argc, argv);
  nNodes = default_num_nodes;
  dead_interval = 4 * hello_int;  // dead interval (seconds)

  set_global_seeds(seed);

  if (Arg::Specified("help") || 
      Arg::Specified("-help") || 
      Arg::Specified("-h") || 
      Arg::Specified("h") || 
      Arg::Specified("?"))
  {
    usage(argc, argv);
    return 0;
  }
  cout << "Running " << nNodes << " nodes with random seed " << seed <<  endl;
  cout << "The ospf6 retransmit interval is " << rxmt_int << " secs" << endl;
  cout << "The ospf6 pushback interval is " << pb_int << " msecs" << endl;
  cout << "The ospf6 backupwait interval is " << bw_int << " msecs" << endl;
  cout << "The ospf6 ack interval is " << ack_int << " msecs" << endl;
  cout << "The ospf6 min LSA interval is " << minls_int << " secs" << endl;
  cout << "The ospf6 min LSA arrival is " << minls_arrival << " secs" << endl;
  cout << "The ospf6 hello interval is " << hello_int << " secs" << endl;
  cout << "The wireless interface type is ";
  if (w_int_type == 1)
    cout << "point-to-multipoint" << endl;
  else if (w_int_type == 2)
    cout << "manet" << endl;
  else 
  {
    cout << "bad interface type: " << w_int_type << endl;
    usage(argc, argv);
    return 0;
  } 
  cout << endl;

  cout << "The wireless flooding is ";
  if (w_flooding == 0)
    cout << "broadcast" << endl;
	else if (w_flooding == 1)
    cout << "mpr" << endl;
  else if (w_flooding == 2)
    cout << "si-cds" << endl;
  else 
  {
    cout << "bad flooding type: " << w_flooding << endl;
    usage(argc, argv);
    return 0;
  } 
  cout << endl;

  // Prepare output file names
  tracefile.append(argv[0]);
  if (tag.size()) {
    tracefile.append("-");
    tracefile.append(tag.c_str());
  }
  tracefile.append(".tr");
  statfile.append(argv[0]);
  if (tag.size()) {
    statfile.append("-");
    statfile.append(tag.c_str());
  }
  statfile.append(".stat");

  // ipv4 addresses must be less than 128.0.0.0 for ospf6 because we used 
  // the first bit in the ipv4 address to indicate a linklocal address 
  IPAddr_t firstIP = IPAddr("10.0.0.1");

  //Matrix 
  LinkMatrix LM(nNodes, firstIP, Mask(24)); 
  LM.ScheduleMatrixChange(0, 0, 4, 1);  //(Time,src_node,dst_node,error rate)
  LM.ScheduleMatrixChange(0, 0, 5, 1);

  LM.ScheduleMatrixChange(0, 1, 3, 1);
  LM.ScheduleMatrixChange(0, 1, 5, 1);

  LM.ScheduleMatrixChange(0, 2, 5, 1);

  LM.ScheduleMatrixChange(0, 3, 1, 1);
  LM.ScheduleMatrixChange(0, 3, 5, 1);

  LM.ScheduleMatrixChange(0, 4, 0, 1);

  LM.ScheduleMatrixChange(0, 5, 0, 1);
  LM.ScheduleMatrixChange(0, 5, 1, 1);
  LM.ScheduleMatrixChange(0, 5, 2, 1);
  LM.ScheduleMatrixChange(0, 5, 3, 1);

  // force flood of LSA from node 4 at time 
  LM.ScheduleMatrixChange(1900, 4, 5, 1);
  LM.ScheduleMatrixChange(1900, 5, 4, 1);

  //force loss of LSA from node 3 to node 0
  LM.ScheduleMatrixChange(1935, 2, 0, 1);
  LM.ScheduleMatrixChange(1936, 2, 0, 0);

  // Per-packet tracing configuration
  ETrace* egs = ETrace::Instance();
  egs->Open(tracefile.c_str());
  egs->SetExtended(true);
  egs->LayerOff(1); //don't etrace this layer
  egs->LayerOff(2); //don't etrace this layer
  egs->SetTraceEventLevel(2);
  egs->SetTraceOnTime(default_trace_on_time);
  egs->SetTraceOffTime(default_trace_off_time);

  // Statistics collection
  EStat* estat = EStat::Instance();
  estat->CollectEStats(statfile.c_str(), start_time);
  estat->LayerOff(1);
  estat->LayerOff(2);
  
  // OSPF6 initialization
  OSPF6DApplication ospf6d;
  ospf6d.CollectStats(statfile.c_str(), start_time);
  ospf6d.LogNeighbors(0.5);  // Log physical layer neighbors every 500 ms
  
  OSPF6DInstance *ospf6_inst;
  Uniform u(0,start_time);
  first_rtrid = IPAddr("0.0.0.1"); 
          
  // CBR traffic initialization
  IPAddr remoteIP; 
  char buf[32];
  Count_t rate = (pktrate * pktsize * 8)/(nNodes*(nNodes-1)); 
  // start traffic generators at a random time
  Uniform cbrstart(start_time-(double)(pktsize*8)/rate, start_time); 
  
  for (Count_t i = 0; i < LM.Size(); ++i)
  {
    Node* n = LM.GetNode(i);
     
    // OSPF6 Router Configuration
    ospf6_inst = ospf6d.AddNode(n);

    sprintf(hostname, "ospf6_%d", n->Id()+1);
    sprintf(logfilename, "log_ospf6_%d.log", n->Id()+1);

    ospf6_inst->SetHostName(hostname);
    if (Arg::Specified("ospfv3log"))
      ospf6_inst->SetLogFileName(logfilename);
    ospf6_inst->SetRouterId(n->Id()+first_rtrid);
    ospf6_inst->SetMinLSInterval(minls_int);
    ospf6_inst->SetMinLSArrival(minls_arrival);

    vInterface = n->Interfaces();
    for (IFVec_t::iterator it=vInterface.begin(); it!=vInterface.end(); ++it)
    {
      ospf6_inst->SetInterfaceHelloInterval(*it, hello_int);
      ospf6_inst->SetInterfaceDeadInterval(*it, dead_interval);
      ospf6_inst->SetInterfaceRetransmitInterval(*it, rxmt_int);
      ospf6_inst->SetInterfacePushBackInterval(*it, pb_int);
      ospf6_inst->SetInterfaceBackupWaitInterval(*it, bw_int);
      ospf6_inst->SetInterfaceAckInterval(*it, ack_int);
      ospf6_inst->SetInterfaceArea(*it, 0);
      ospf6_inst->SetInterfaceCost(*it, cost);
      // too much jitter can cause stale ospf6 state to be sent
      ospf6_inst->SetInterfaceJitter(*it, interface_jitter); //msec
      ospf6_inst->SetInterfaceFloodDelay(*it, interface_flood_delay); //msec

      if (strcmp((*it)->GetName(), "lo") == 0)
      {
        ospf6_inst->SetInterfaceType(*it, "loopback");
        if (rate > 0 && pktsize > 0)
        {
          // CBR traffic from this router to all other router's loopback addr
          for (Count_t j = 0; j < LM.Size(); ++j)
          {
            if (i == j)
              continue;
            sprintf(buf, "0.0.%d.0", j+1); //loopback addr of routers
            remoteIP = IPAddr(buf); 
            CBRApplication* cbr = (CBRApplication*) n->AddApplication(
                          CBRApplication(remoteIP,1000,NO_PORT,rate,pktsize));
            cbr->Start(cbrstart.Value()); 
            cbr->Stop(stop_time - 5);
          }
        } 
      }
      else
      {
        if(w_int_type == 1)
          ospf6_inst->SetInterfaceType(*it, "point-to-multipoint");
        else if (w_int_type == 2)
        {
          ospf6_inst->SetInterfaceType(*it, "manet-reliable");
          if (Arg::Specified("diff_hellos"))
            ospf6_inst->SetInterfaceDiffHellos(*it, true);
          //set flooding type
          if (w_flooding == 0)
            ospf6_inst->SetInterfaceFlooding(*it, "broadcast");
          else if (w_flooding == 1)
            ospf6_inst->SetInterfaceFlooding(*it, "mpr");
          else if (w_flooding == 2)
            ospf6_inst->SetInterfaceFlooding(*it, "mdr");
        }
        if (i==0)
        {
          WirelessLink *l = (WirelessLink *) (*it)->GetLink();
          l->SetAlpha(alpha);
        }
      }
    }
    ospf6_inst->Start(u.Value());  // random start time of each ospf6 router
  }
  s.StopAt(stop_time);
  s.Progress(40);
  s.Run();
}
